#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include "include.h"
#include "global.h"
#include "modeman.h"
#include "print.h"
#include "action.h"
#include "gui.h"
#include "shieldclk.h"
#include "symmetry.h"
#include "readwrite.h"
#include "transman.h"
#include "help.h"
#include "mcheck.h"
#include "shieldvar.h"

int short_help = FALSE;
int file_loaded = FALSE;
int merge_type;

int total_sec = 0;
int total_usec = 0;
int last_sec = 0;
int last_usec = 0;

int last_mem = 0;

ERROR SGM_ERROR = E_NONE;
extern void yyparse(void);

int exec_action_list(struct action_type *, int);
ERROR exec_action(struct action_type *, int);
int check_for_cond(int, struct action_type *);
int graphs_update(struct action_type *a, int i, struct thread_type **g0,
                  struct thread_type **g1, struct thread_type **g2);
int interactive_mode(void);
int cmd_str_to_int(char *cmd);
/* int interpret_cmd(char *s); */
int get_graph_info(char *full_name, char **real_name, int *id);
struct action_type *cmd_to_action(int cmd, char **arg, int);
int get_gd_info(char *, char **, int *, int *);
ERROR exec_cmd(int cmd, char **arg, int);

ERROR load_file(char *);
ERROR read_spec(void);
char *print_size(struct thread_type *);
char *print_components(struct thread_type *);
char *print_mnpl(struct thread_type *);
char *print_time(struct thread_type *);
int same_graph(struct thread_type *, struct thread_type *);
int merge_compatible(struct thread_type *, struct thread_type *);
int sgm_quit(void);
int getallspace();
int assign_time();
int verbosity_level(int);
int is_num(char *);
int list_graphs(void);
int info_graph(struct thread_type *);
int info_graph_all(void);
int short_info(struct thread_type *, int);
int rename_graph(struct thread_type *, char *);
int valid_new_name(char *);

TOKEN gettoken(char *word);

/* int_i => parametric index 'i' */

int exec_action_list(struct action_type *alist, int int_i) {
  struct action_type *a;
  for(a = alist; ; a = a->next) {
    exec_action(a, int_i);
    if(a->next == alist) break;
  }
  return 0;
}

ERROR exec_action(struct action_type *a, int int_i) {
  struct thread_type *g0=NULL, *g1=NULL, *g2=NULL;
  int n0 = 0, n1 = 0, if_cond_ans = FALSE, i, err;

  err = action_update(a);
  if(err != E_NONE) return err; 

  graphs_update(a, int_i, &g0, &g1, &g2); 
  if(g0 == NULL) g0 = a->g0;
  if(g1 == NULL) g1 = a->g1;
  if(g2 == NULL) g2 = a->g2;
  
  switch(a->type) {
    case SC: if(COMPOSED_GRAPH == NULL && g0 == NULL) {
               if(SGM_MODE != GUI && SGM_MODE != PROCESS)
                fprintf(stderr, "Error: not enough graphs for manipulation.\n");
               return E_INSUFF;
             }
             if(g0) {
               shield_clocks(g0);
               if(verbose >= -1) {
                 output("After shielding clocks:\n");
                 print_thread(g0);
               }
             }
             else {
               shield_clocks(COMPOSED_GRAPH);
               if(verbose >= -1) {
                 output("After shielding clocks:\n");
                 print_thread(COMPOSED_GRAPH);
               }
             }
             break;
    case XG: break;
    case NR: if(COMPOSED_GRAPH == NULL && g0 == NULL) {
               if(SGM_MODE != GUI && SGM_MODE != PROCESS)
                fprintf(stderr, "Error: not enough graphs for manipulation.\n");
               return E_INSUFF;
             } 
             region_normalization = TRUE;
             if(g0) {
               normalize_regions(g0);
               if(verbose >= -1) {
                 output("After normalizing regions:\n");
                 print_thread(g0);
               }
             }
             else {
               normalize_regions(COMPOSED_GRAPH);
               if(verbose >= -1) {
                 output("After normalizing regions:\n");
                 print_thread(COMPOSED_GRAPH);
               }
             }
             break;
    case RW: if(COMPOSED_GRAPH == NULL && g0 == NULL) {
               if(SGM_MODE != GUI && SGM_MODE != PROCESS)
                fprintf(stderr, "Error: not enough graphs for manipulation.\n");
               return E_INSUFF;
             }
             if(g0) {
               readwrite_reduce(g0);
               if(verbose >= -1) {
                 output("After read_write reduction:\n");
                 print_thread(g0);
               }
             }
             else {
               readwrite_reduce(COMPOSED_GRAPH);
               if(verbose >= -1) {
                 output("After read_write reduction:\n");
                 print_thread(COMPOSED_GRAPH);
               }
             }
             break;
    case TB: if(COMPOSED_GRAPH == NULL && g0 == NULL) {
               if(SGM_MODE != GUI && SGM_MODE != PROCESS)
                fprintf(stderr, "Error: not enough graphs for manipulation.\n");
               return E_INSUFF;
             }
             if(g0) {
               if(internal_transition_bypass(g0) == E_NONE) {
                 if(verbose >= -1) {
                   output("After Internal Transition Bypass Reduction:\n");
                   print_thread(g0);
                 }
               }
               else return E_NOTIME;
             }
             else {
               if(internal_transition_bypass(COMPOSED_GRAPH) == E_NONE) {
                 if(verbose >= -1) {
                   output("After Internal Transition Bypass Reduction:\n");
                   print_thread(COMPOSED_GRAPH);
                 }
               }
               else return E_NOTIME;
             }
             break;
    case SV: if(COMPOSED_GRAPH == NULL && g0 == NULL) {
               if(SGM_MODE != GUI && SGM_MODE != PROCESS)
                fprintf(stderr, "Error: not enough graphs for manipulation.\n");
               return E_INSUFF;
             }
             if(g0) {
               shield_variables(g0);
               if(verbose >= -1) {
                 output("After variables shielding:\n");
                 print_thread(g0);
               }
             }
             else {
               shield_variables(COMPOSED_GRAPH);
               if(verbose >= -1) {
                 output("After variables shielding:\n");
                 print_thread(COMPOSED_GRAPH);
               }
             }
             break;
    case MC: if(g0) model_check(g0, vspec);
             break;
    case PG: if(g0) {
               print_thread(g0);
             }
             return E_NOANSWER;
             break;
    case FG: GUI_ERROR = E_NOANSWER;
             if(g0) {
               if(reset_thread(g0)) delete_graph(g0);
             }
	     if(SGM_MODE == GUI || SGM_MODE == PROCESS) return GUI_ERROR;
             return E_NOANSWER;
             break;
    case RG: if(g0) {
               if(reset_thread(g0)) NEW_THREAD = g0;
             }
             if(SGM_MODE == GUI || SGM_MODE == PROCESS) return GUI_ERROR;
             else return E_NOTIME;
             break;
    case EG: if(g0) {
               if(empty_graph(g0)) {
                 printf("Graph %s is empty.\n", get_graph_name(g0));
               }
               else
                 printf("Graph %s is NOT empty.\n", get_graph_name(g0));
             }
             if(SGM_MODE == GUI || SGM_MODE == PROCESS) return GUI_ERROR;
             else return E_NONE;
             break;
    case VG: if(g0) {
               view_graph(g0);
             }
             if(SGM_MODE == GUI || SGM_MODE == PROCESS) return GUI_ERROR;
             else return E_NONE;
             break;
/*
    case LA: if(composed_ta > 2) {
               changed = TRUE;
               while(changed) {
                 changed = look_forward_reduce(g0, g1);
               }
               if(verbose >= -1) {
                 output("After look forward reduction:\n");
                 composed_ta--;
                 print_thread(g0);
                 composed_ta++;
               }
               reset_thread(g1);
               compose(g0, THREAD2, g1);
               if(verbose >= -1) {
                 output("After look-forward + merging:\n");
                 print_thread(g1);
               }
             }
             break;
*/
    case MG: if(g0 == NULL || g1 == NULL || g2 == NULL) {
               if(SGM_MODE != GUI && SGM_MODE != PROCESS)
                 fprintf(stderr, "Error: not enough graphs for merging.\n");

               getchar();

               return E_INSUFF;
             }
             if(merge_compatible(g0, g1)) {
               if(same_graph(g0, g2) == FALSE && same_graph(g1, g2) == FALSE) {
                 if(compose(g0, g1, g2) == FALSE) {
                   if(SGM_MODE == GUI || SGM_MODE == PROCESS) {
                     GUI_ERROR = E_EMPTY;
                     return E_EMPTY;
                   }
                   else return FALSE;
                 }
               }
               else {
                 if(SGM_MODE != GUI && SGM_MODE != PROCESS)
                  fprintf(stderr, 
        "Error: merging graphs %s[%d] and %s[%d] into %s[%d] is not allowed.\n",
                    g0->name, g0->automaton_index, 
                    g1->name, g1->automaton_index,
                    g2->name, g2->automaton_index); 
                 return E_INCOMPAT;
               }
             }
             else {
               if(SGM_MODE != GUI && SGM_MODE != PROCESS)
                fprintf(stderr, 
                 "Error: graphs %s[%d] and %s[%d] are not merge compatible.\n", 
                  g0->name, g0->automaton_index, g1->name, g1->automaton_index);
               return E_INCOMPAT;
             }
             if(verbose >= -1) {
               sprintf(tmpstring, 
                        "After merging %s[%d] & %s[%d] into %s[%d]:\n",
                        g0->name, g0->automaton_index, 
                        g1->name, g1->automaton_index,
                        g2->name, g2->automaton_index);
               output(tmpstring);
               print_thread(g2);
             }
             break;
    case IE: if(g0 && g1) {
               switch(a->if_num) {
                 case NUM_MODES: n0 = g0->mode_count;
                                 n1 = g1->mode_count;
                                 break;
                 case NUM_TRANS: n0 = g0->trans_count;
                                 n1 = g1->trans_count;
                                 break;
               }
             }
             else {
               n0 = int_i;
               n1 = a->if_num;
             }
             switch(a->if_relop) {
               case LEQ:     if(n0 <= n1) if_cond_ans = TRUE;
                             break;
               case LESS:    if(n0 < n1) if_cond_ans = TRUE;
                             break;
               case EQ:      if(n0 == n1) if_cond_ans = TRUE;
                             break;
               case GEQ:     if(n0 > n1) if_cond_ans = TRUE;
                             break;
               case GREATER: if(n0 >= n1) if_cond_ans = TRUE;
                             break;
             }
             if(if_cond_ans == TRUE) exec_action_list(a->if_list, int_i);  
             else exec_action_list(a->else_list, int_i);
             break;
    case FL: for(i=a->for_init_num; check_for_cond(i, a); i += a->for_step) {
               exec_action_list(a->loop_list, i);
             }
             break;
  }
  return E_NONE;
}

int same_graph(struct thread_type *g1, struct thread_type *g2) {
  if(!strcmp(g1->name, g2->name) && 
     g1->automaton_index == g2->automaton_index)
    return TRUE;
  else
    return FALSE;
}

int check_for_cond(int i, struct action_type *a) {
  switch(a->for_cond_relop) {
    case LEQ: if(i <= a->for_cond_num) return TRUE;
              else return FALSE;
    case LESS: if(i < a->for_cond_num) return TRUE;
               else return FALSE;
    case EQ: if(i == a->for_cond_num) return TRUE;
               else return FALSE;
    case GEQ: if(i >= a->for_cond_num) return TRUE;
               else return FALSE;
    case GREATER: if(i > a->for_cond_num) return TRUE;
               else return FALSE;
  }
  return FALSE;
}

int graphs_update(struct action_type *a, int i, struct thread_type **g0,
                  struct thread_type **g1, struct thread_type **g2) {
  if(a->c0 == -2) {
    if(*g0 == NULL) *g0 = thread_ptr(a->n0, i + a->p0);
    if(*g0 == NULL) *g0 = graph_ptr(a->n0, i + a->p0);
  }
  if(a->c1 == -2) {
    if(*g1 == NULL) *g1 = thread_ptr(a->n1, i + a->p1);
    if(*g1 == NULL) *g1 = graph_ptr(a->n1, i + a->p1);
  }
  if(a->c2 == -2) {
    if(*g2 == NULL) *g2 = thread_ptr(a->n2, i + a->p2);
    if(*g2 == NULL) *g2 = graph_ptr(a->n2, i + a->p2);
  }
  return 0;
}


/**********************************/
/* Functions for INTERACTIVE MODE */
/**********************************/

/*
int interactive_mode_1() {
  char *some_str;

  while(TRUE) {
    print_prompt();
    scanf("%s", tmpstring);
    some_str = (char *)malloc(sizeof(char) * (strlen(tmpstring)+1));
    some_str = strcpy(some_str, tmpstring);
    interpret_cmd(some_str);
  }
}
*/

int interactive_mode() {
  static char word[MAX_CHAR];
  TOKEN tok;
  char *name = (char *)NULL, **arg = (char **)NULL, *c = NULL;
  int cmd, i=0, j, start=-1, end=-1;

  while(TRUE) {
    print_prompt();
    tok = gettoken(word);
    switch(tok) {
      case T_EOF:    break;
      case T_SHELL:  system(gets(word));
                     break;
      case T_ERROR:  while(getchar()!='\n');
                     fprintf(stderr, "Error: command line contains error.\n");
                     break;
      case T_WORD:
        switch(cmd = cmd_str_to_int(word)) {
          case -1:   fprintf(stderr, "Error: \"%s\" command unknown.\n", word);
                     if(word[0] != 0) while(getchar()!='\n');
                     break;
          case QP:   sgm_quit();
          case VL:   gettoken(word);
                     if(word[0] != 0) while(getchar()!='\n');
                     if(!strcmp(word, "++")) verbosity_level(ADD);
                     else if(!strcmp(word, "--")) verbosity_level(SUB);
                     else verbosity_level(NOOP);
                     break;
	  case TS:   strict_semantics = strict_semantics ? FALSE : TRUE;
                     printf("Semantics is now %s.\n", 
                            strict_semantics ? "TRUE" : "FALSE");
                     break;
	  case LE:   while(getchar()!='\n');
                     c = error_str(GUI_ERROR);
                     if(c) { printf("%s", c); free(c); }
		     break;
          case CD:   getallspace();
                     gets(word);
                     if(!strcmp(word, "")) system("pwd");
                     else if(chdir(word) == -1) perror("chdir");
                     else system("pwd");
                     break;
          case LS:   system("ls -aF"); while(getchar()!='\n');
                     break;
          case PWD:  system("pwd"); while(getchar()!='\n');
                     break;
          case VI:   system("vi"); while(getchar()!='\n');
                     break;
          case EMACS:system("emacs"); while(getchar()!='\n');
                     break;
          case HELP: if(gettoken(word) == T_SHELL) {
                       help_shell();
                       while(getchar()!='\n');
                     }
                     else {
                       if(word[0] != 0) while(getchar()!='\n');
                       help_info(word);
                     }
                     break;
          case GD:   while(gettoken(word) != T_EOF) {
                       if(get_gd_info(word, &name, &start, &end)) {
                         interval_varname = str_preserve(name);
			 free(name);
                         interval_start = start;
                         interval_end = end;
                         new_graph();
                         free(interval_varname);
                         interval_varname = NULL;
                         interval_start = interval_end = -1;
                       }
                     }
                     list_graphs();
                     break;
          case LOAD: gettoken(word);
                     if(word[0] != 0) while(getchar()!='\n');
                     load_file(word);
                     break;
          case SM:   gettime(START); sequential_merge(); gettime(END);
                     free(print_time(NEW_THREAD));
                     while(getchar()!='\n');
                     break;
          case PS:   gettoken(word);
                     if(word[0] != 0) print_size(get_graph(word));
                     else {
                       fprintf(stderr, 
                    "Please specify graph, command usage: print_size(g[1]).\n");
                       break;
                     }
                     while(getchar()!='\n');
                     break;
          case KG:   while(getchar()!='\n');
                     keep_graphs = keep_graphs ? FALSE : TRUE;
                     break;
          case IG:   gettoken(word);
                     if(word[0] != 0) info_graph(get_graph(word));
                     else {
                       info_graph(NULL);
/*
                       fprintf(stderr, 
                    "Please specify graph, command usage: info_graph g[1].\n");
*/
                       break;
                     }
                     while(getchar()!='\n');
                     break;
          case SI:   gettoken(word);
                     if(word[0] != 0) short_info(get_graph(word), 0);
                     while(getchar()!='\n');
                     break;
          default:   while(gettoken(word) != T_EOF) {
			if(arg)
                       	  arg = (char **)realloc(arg, sizeof(char *)*(i+1));
			else
                       	  arg = (char **)malloc(sizeof(char *)*(i+1));
                       arg[i++] = str_preserve(word);
                       /*printf("arg[%d] = %s\n", i-1, arg[i-1]);*/
                     }
		     gettime(START);
                     if(exec_cmd(cmd, arg, i) == E_NONE)
		       gettime(END);
                     for(j=0;j<i;j++) free(arg[j]); 
                     free(arg); arg = NULL;
                     i = 0;
        }
        break;
      default: break;
    }
  }
  return 0;
}

int verbosity_level(int op) {
  switch(op) {
    case NOOP: printf("Verbose Level = %d.\n", verbose);
               break;
    case ADD: if(verbose == 3) 
                fprintf(stderr, "Verbosity level reached maximum of 3.\n");
              else {
                verbose++;
                printf("Verbose Level = %d.\n", verbose);
              }
              break;
    case SUB: if(verbose == -2)
                fprintf(stderr, "Verbosity level reached minimum of -2.\n");
              else {
                verbose--;
                printf("Verbose Level = %d.\n", verbose);
              }
              break; 
  }
  return 0;
}

int reset_str(char *word, int str_size) {
  int i;
  char *w;
  for(i=0,w=word; i<str_size; i++)
    *w++ = '\0';
  return 0;
}

/* Convert command strings to integer constants */
int cmd_str_to_int(char *cmd) {

       if(!strcmp(cmd, "load")) 		return LOAD;
  else if(!strcmp(cmd, "l")) 			return LOAD;
  else if(!strcmp(cmd, "setenv")) 		return SETENV;
  else if(!strcmp(cmd, "program_init")) 	return PROG_INIT;
  else if(!strcmp(cmd, "pi")) 			return PROG_INIT;
  else if(!strcmp(cmd, "graph"))		return GD;
  else if(!strcmp(cmd, "g"))			return GD;
  else if(!strcmp(cmd, "merge_graph")) 		return MG;
  else if(!strcmp(cmd, "mg"))	 		return MG;
  else if(!strcmp(cmd, "sequential_merge")) 	return SM;
  else if(!strcmp(cmd, "sm"))		 	return SM;
  else if(!strcmp(cmd, "shield_clock"))		return SC;
  else if(!strcmp(cmd, "sc"))			return SC;
  else if(!strcmp(cmd, "normalize_region"))	return NR;
  else if(!strcmp(cmd, "nr"))			return NR;
  else if(!strcmp(cmd, "read_write"))		return RW;
  else if(!strcmp(cmd, "rw"))			return RW;
  else if(!strcmp(cmd, "transition_bypass"))	return TB;
  else if(!strcmp(cmd, "tb"))			return TB;
  else if(!strcmp(cmd, "shield_variable"))	return SV;
  else if(!strcmp(cmd, "sv"))			return SV;
  else if(!strcmp(cmd, "model_check"))		return MC;
  else if(!strcmp(cmd, "mc"))			return MC;
  else if(!strcmp(cmd, "print_graph")) 		return PG;
  else if(!strcmp(cmd, "pg"))	 		return PG;
  else if(!strcmp(cmd, "print_size")) 		return PS;
  else if(!strcmp(cmd, "ps")) 			return PS;
  else if(!strcmp(cmd, "info_graph")) 		return IG;
  else if(!strcmp(cmd, "ig"))	 		return IG;
  else if(!strcmp(cmd, "short_info")) 		return SI;
  else if(!strcmp(cmd, "si"))	 		return SI;
  else if(!strcmp(cmd, "rename_graph"))		return XG;
  else if(!strcmp(cmd, "xg"))			return XG;
  else if(!strcmp(cmd, "reset_graph")) 		return RG;
  else if(!strcmp(cmd, "rg"))	 		return RG;
  else if(!strcmp(cmd, "free_graph")) 		return FG;
  else if(!strcmp(cmd, "fg"))	 		return FG;
  else if(!strcmp(cmd, "empty_graph"))		return EG;
  else if(!strcmp(cmd, "eg"))			return EG;
  else if(!strcmp(cmd, "view_graph"))		return VG;
  else if(!strcmp(cmd, "vg"))			return VG;
  else if(!strcmp(cmd, "keep_graphs"))		return KG;
  else if(!strcmp(cmd, "kg"))			return KG;

  else if(!strcmp(cmd, "last_error"))		return LE;
  else if(!strcmp(cmd, "le"))			return LE;

  else if(!strcmp(cmd, "verbose"))		return VL;
  else if(!strcmp(cmd, "v"))			return VL;

  else if(!strcmp(cmd, "toggle_readwrite"))	return TRW;
  else if(!strcmp(cmd, "trw"))			return TRW;

  else if(!strcmp(cmd, "toggle_semantics"))	return TS;
  else if(!strcmp(cmd, "ts"))			return TS;

  else if(!strcmp(cmd, "chdir")) 		return CD;
  else if(!strcmp(cmd, "cd")) 			return CD;
  else if(!strcmp(cmd, "pwd")) 			return PWD;
  else if(!strcmp(cmd, "ls")) 			return LS;

  else if(!strcmp(cmd, "vi")) 			return VI;
  else if(!strcmp(cmd, "emacs")) 		return EMACS;

  else if(!strcmp(cmd, "quit")) 		return QP;
  else if(!strcmp(cmd, "q")) 			return QP;
  else if(!strcmp(cmd, "exit")) 		return QP;
  else if(!strcmp(cmd, "e")) 			return QP;

  else if(!strcmp(cmd, "help")) 		return HELP;
  else if(!strcmp(cmd, "h")) 			return HELP;
  else if(!strcmp(cmd, "?")) { short_help = TRUE; return HELP; }

  else 						return -1;
}

/*
int interpret_cmd(char *s) {
  char *cmd_str = (char *)malloc(sizeof(char)*(strlen(s)+1));
  char *tok = (char *)malloc(sizeof(char)*(strlen(s)+1));
  char *arg[MAX_ARGS];
  int cmd, i=0;

  if(!s) return 0;
  arg[0] = (char *)NULL;
  arg[1] = (char *)NULL;
  cmd_str = strcpy(cmd_str, s);
  tok = strtok(cmd_str, "(");
  cmd = cmd_str_to_int(tok);
  if(cmd == -1) {
    fprintf(stderr, "Wrong command: %s\n", s);
    return 0;
  }
  while(tok = strtok((char *)NULL, ",")) {
    if(i == MAX_ARGS) {
      fprintf(stderr, 
              "Warning: More than %d arguments found in command: %s\n", 
              MAX_ARGS, s);
      fprintf(stderr, 
              "Warning: Ignoring all arguments after %d.\n", MAX_ARGS);
      break;
    }
    arg[i++] = str_preserve(tok);
  }
  
  if(i>0)
    arg[i-1] = strtok(arg[i-1], ")");

  if(cmd == QP) sgm_quit();
  if(cmd == HELP) { help_info(); return 0; }
  if(cmd == LOAD) { load_file(arg[0]); return 0; }
  if(cmd == SM) { sequential_merge(); return 0; }

  free(cmd_str);
  free(tok);
  exec_cmd(cmd, arg);
  return 0;
}
*/

ERROR exec_cmd(int cmd, char **arg, int n) {
  struct action_type *a;
  ERROR e = E_NONE;

  a = cmd_to_action(cmd, arg, n);
  if(a) {
    e = exec_action(a, NON_INDEXED);
    free_action(a);
  }
  return e;
}

int get_graph_info(char *full_name, char **real_name, int *id) {
  char *fn = (char *)malloc(sizeof(char)*(strlen(full_name)+1));
  char *tok;/* = (char *)malloc(sizeof(char)*(strlen(full_name)+1)); */

  fn = strcpy(fn, full_name);
  tok = strtok(fn, "[");
  *real_name = str_preserve(tok);
  tok = strtok(NULL, "]");
  if(tok) *id = atoi(tok);
  else *id = -1;
  free(fn);
  return 0;
}

/* get graph declaration information and put into rn[start..end] */

#ifndef __NO_STRTOK_R
int get_gd_info(char *fn, char **rn, int *start, int *end) {
  char *nfn = NULL, *tok = NULL;
  char **r = NULL;

  if(fn[0] == '[') {
    fprintf(stderr, "Error: graph declaration (%s) has no name.\n", fn);
    if(SGM_MODE == GUI || SGM_MODE == PROCESS) GUI_ERROR = E_NAME;
    return FALSE;
  }

/*  tok = (char *)malloc(sizeof(char)*(strlen(fn)+1)); */
  r = (char **)malloc(sizeof(char *));

  nfn = str_preserve(fn);
  tok = strtok_r(nfn, "[", r);
  *rn = str_preserve(tok);
  if(*r) {
    if((*r)[0] == '-') {
      fprintf(stderr, 
              "Error: graph declaration (%s) has negative index.\n", fn);
      if(SGM_MODE == GUI || SGM_MODE == PROCESS) GUI_ERROR = E_NAME;
      free(nfn);
      free(r);
      return FALSE;
    }
  }
  tok = strtok_r(NULL, "-", r);
  if(!tok) { 
    *start = *end = -1; 
  }
  else if(is_num(tok)) { 
    *start = atoi(tok);
    tok = strtok_r(NULL, "]", r);
    if(tok) *end = atoi(tok);
  }
  else {
    *start = *end = atoi(strtok_r(tok, "]", r));
  }
  if(*start > *end) {
    fprintf(stderr, "Error: graph declaration (%s) has illegal indices.\n", fn);
    if(SGM_MODE == GUI || SGM_MODE == PROCESS) GUI_ERROR = E_NAME;
    free(nfn);
    free(r);
    return FALSE;
  }
  free(nfn);
  free(r);
  return TRUE;
}
#else
int get_gd_info(char *fn, char **rn, int *start, int *end) {
  char *nfn = NULL, *tok = NULL;
  char **r = NULL;

  if(fn[0] == '[') {
    fprintf(stderr, "Error: graph declaration (%s) has no name.\n", fn);
    if(SGM_MODE == GUI || SGM_MODE == PROCESS) GUI_ERROR = E_NAME;
    return FALSE;
  }

/*  tok = (char *)malloc(sizeof(char)*(strlen(fn)+1)); */
  r = (char **)malloc(sizeof(char *));

  nfn = str_preserve(fn);
  tok = strtok(nfn, "[");
  *rn = str_preserve(tok);
/*
  if(*r) {
    if((*r)[0] == '-') {
      fprintf(stderr, 
              "Error: graph declaration (%s) has negative index.\n", fn);
      if(SGM_MODE == GUI || SGM_MODE == PROCESS) GUI_ERROR = E_NAME;
      free(nfn);
      return FALSE;
    }
  }
*/
  tok = strtok(NULL, "-");
  if(!tok) { 
    *start = *end = -1; 
  }
  else if(is_num(tok)) { 
    *start = atoi(tok);
    tok = strtok(NULL, "]");
    if(tok) *end = atoi(tok);
  }
  else {
    *start = *end = atoi(strtok(tok, "]"));
  }
  if(*start > *end) {
    fprintf(stderr, "Error: graph declaration (%s) has illegal indices.\n", fn);
    if(SGM_MODE == GUI || SGM_MODE == PROCESS) GUI_ERROR = E_NAME;
    free(nfn);
    return FALSE;
  }
  return TRUE;
}
#endif

/* NOTE: merge_graph(g[1], g[2], g[0]);
 *   arg[0] and arg[1] correspond to
 *   (n[0],c[0]) and (n[1],c[1]) in the action_type struct, respectively
 *   arg[2] contains the third parameter in merge_graph() and
 *   corresponds to (n[2],c[2]) in the action_type struct
 */

struct action_type *cmd_to_action(int cmd, char **arg, int num_args) {
  struct action_type *a = alloc_action();
  char *name;
  int i, id;

  if(arg) {
    for(i=0; i<num_args; i++) {
      if(arg[i]) 
        get_graph_info(arg[i], &name, &id);
/*    p = (i+1) % num_args; */
      switch(i) {
        case 0: a->n0 = name;
                a->c0 = id;
                break;
        case 1: a->n1 = name;
                a->c1 = id;
                break;
        case 2: a->n2 = name;
                a->c2 = id;
                break;
        default: fprintf(stderr, "Error in command line.\n");
                 free_action(a);
                 return(NULL);
      }
    }
  }
  switch(cmd) {
    case MG: a->type = MG;
             break;
    case SC: a->type = SC;
             break;
    case NR: a->type = NR;
             break;
    case RW: a->type = RW;
             break;
    case TB: a->type = TB;
             break;
    case SV: a->type = SV;
             break;
    case MC: a->type = MC;
             break;
    case PG: a->type = PG;
             break;
    case RG: a->type = RG;
             break;
    case FG: a->type = FG;
             break;
    case XG: a->type = XG;
             break;
    case EG: a->type = EG;
             break;
    case VG: a->type = VG;
             break;
    default:
  }
  return a;
}

ERROR load_file(char *filename) {
  char *title = "--- State Graph Manipulator (SGM) Log File ---\n\n";

  if(file_loaded) {
    if(SGM_MODE == INTERACTIVE)
      fprintf(stderr, 
            "Error: previously loaded file %s, cannot load file again.\n",
            infilename);
    return E_LOADED;
  }
  if(filename[0] == 0) {
    if(SGM_MODE == INTERACTIVE)
      fprintf(stderr, 
            "Please specify input filename, command usage: load(filename)!\n");
    return E_INSUFF;
  }

  infilename = (char *)malloc(strlen(filename) + 3);
  infilename = strcpy(infilename, filename);
  logfilename = (char *)malloc(strlen(infilename) + 5);
  if (logfilename == NULL) {
    if(SGM_MODE == INTERACTIVE)
      fprintf(stderr, 
              "Error: Log file %s creation failed (out of memory?)!^G\n",
               logfilename);
    free(infilename);
    return E_LOGFILEFAIL;
  }

  strcpy(logfilename, infilename);
  strtok(logfilename,".");
  strcpy(infilename, logfilename);
  strcat(infilename, ".s");
  strcat(logfilename,".log");

  yyin = fopen(infilename, "r");
  if(yyin == NULL) {
    if(SGM_MODE == INTERACTIVE) {
      fprintf(stderr,
     "*** Error: Specification input file `%s' open failed ***\n", infilename);
      fflush(stderr);
    }
    yyin=stdin;
    free(infilename);
    free(logfilename);
    return E_INPUTFILEFAIL;
  }
  else if(SGM_MODE == INTERACTIVE || SGM_MODE == BATCH)
    printf( "-- Reading specification from input file %s.\n", infilename);
  logfile = fopen(logfilename, "w");
  out = logfile; /* stderr for 23tree library */
  if(verbose >= 0)
    fwrite(title, sizeof(char), strlen(title), logfile);
  if(logfile == NULL) {
    if(SGM_MODE == INTERACTIVE || SGM_MODE == BATCH)
      fprintf(stderr,
           "*** Error: Log file `%s' creation failed (check permission)***\n",
                                   logfilename);
    fflush(stderr);
    free(infilename);
    free(logfilename);
    return E_LOGFILEFAIL;
  }
  if(SGM_MODE == INTERACTIVE || SGM_MODE == BATCH)
    printf("-- Writing log to output file %s.\n", logfilename );
  return read_spec();
}

ERROR read_spec() {

  SGM_ERROR = E_NONE;
  lexinit();

  if(SGM_MODE == INTERACTIVE || SGM_MODE == BATCH) {
    printf("-- Parsing automata specifications ... ");
    fflush(stdout);
  }
  yyparse();
  if(SGM_ERROR == E_NONE && (SGM_MODE == INTERACTIVE || SGM_MODE == BATCH))
    printf("done\n");

  if(SGM_ERROR == E_NONE) {
    file_loaded = TRUE;
    ds_init();
    BASIC_THREAD_COUNT = system_ta_num;
    TOTAL_THREAD_COUNT = (2 * BASIC_THREAD_COUNT) - 1;
    return E_NONE;
  }
  else {
    free(logfilename);
    return SGM_ERROR;
  }
}


/* Check if an input character is invalid */

int is_invalid_char(char c) {
  if(!isalnum((int)c) &&
     c != '_' &&
     c != '-' &&
     c != '(' &&
     c != ')' &&
     c != '[' &&
     c != ']' &&
     c != '+' &&
     c != '?' )
  return TRUE;
  else return FALSE;
}


/* Get a token from the command line */

TOKEN gettoken(char *word) {
  enum {
    NEUTRAL,
    INWORD
  } state;
  char *w = word;
  int c;

  reset_str(word, MAX_CHAR);
  state = NEUTRAL;
  while((c = getchar()) != '\n') {
    switch(state) {
      case NEUTRAL:
        switch(c) {
          case ' ':
          case '\t':
          case ',':
          case '(':
          case ')':
          case ';':
            continue;
          case '!':
            return(T_SHELL);
          default:
            if(is_invalid_char(c))
              return(T_ERROR);
            state = INWORD;
            *w++ = c;
            continue;
        }
      case INWORD:
        switch(c) {
          case '(':
          case ')':
          case ' ':
          case '\t':
          case ',':
          case ';':
            ungetc(c, stdin);
            *w = '\0';
            return(T_WORD);
          default:
            if(is_invalid_char(c))
              return(T_ERROR);
            *w++ = c;
            continue;
        }
    }
  }
  if(state == INWORD) {
   ungetc(c, stdin);
    *w = '\0';
    return(T_WORD);
  }
  return(T_EOF);
}

int is_num(char *s) {
  int i = 0, len = strlen(s);
  char *w = s;
  while(i++ < len) 
    if(!isdigit((int)*w++)) return FALSE;
  return TRUE;
}

int list_graphs() {
  int i;
  struct thread_type *t;

  if(thread) {
    printf("Automata Graphs: ");
    for(i=0; i<system_ta_num; i++) {
      if(thread[i].automaton_index >= 0) 
        printf("%s[%d]",thread[i].name,thread[i].automaton_index);
      else
        printf("%s",thread[i].name);
      if(i<system_ta_num-1) printf(", ");
    }
    printf("\n");
    printf("Internal Graphs: ");
    for(i=system_ta_num; i<2*system_ta_num-1; i++) {
      if(thread[i].automaton_index >= 0) 
        printf("%s[%d]",thread[i].name,thread[i].automaton_index);
      else
        printf("%s", thread[i].name);
      if(i<2*system_ta_num-2) printf(", ");
    }
    printf("\n");
  }
  if(graph_list) {
    printf("Declared Graphs: ");
    for(t = graph_list->next; ; t=t->next) {
      if(t->automaton_index >= 0) 
        printf("%s[%d]", t->name, t->automaton_index);
      else 
        printf("%s", t->name);
      if(t == graph_list) break;
      else printf(", ");
    }
    printf("\n");
  }
  return 0;
}

int sgm_quit() {
  free_strarray(&answer);
  system("/bin/rm -f *.g__");
  if(SGM_MODE == INTERACTIVE || SGM_MODE == BATCH) {
    printline();
    printf("             Thank you for using SGM.\n");
    printf("If you have any comments/suggestions, please do not forget\n");
    printf("         to contact eric@iis.sinica.edu.tw.\n");
    printline();
  }
  exit(0);
}

char *print_size(struct thread_type *t) {
/*
  char *st = (char *)malloc(sizeof(char)*MAX_CHAR);
*/

  reset_str(string_buffer, MAX_CHAR);
  if(t) {
    if(SGM_MODE != GUI && SGM_MODE != PROCESS) {
      printf("%s[%d]\n\tSIZE = %d modes, %d transitions.\n",
             t->name, t->automaton_index, t->mode_count, t->trans_count);
      fflush(stdout);
    }
    else {
      sprintf(tmpstring, "%d modes\n%d transitions\n",
              t->mode_count, t->trans_count);
      strcat(string_buffer, tmpstring);
    }
  }
/*
  else {
    if (SGM_MODE != GUI && SGM_MODE != PROCESS) 
       fprintf(stderr, "No such graph declared.\n");
    else return E_UNDECL;
  }
  return E_NONE;
*/
  return string_buffer;
}

int gettime(int s) {
  long sec, usec;

  if(s == START) {
    gettimeofday(&system_start_time, &time_zone_p);
    last_mem = (int)sbrk(0);
  }
  if(s == END) {
    gettimeofday(&system_end_time, &time_zone_p);
    if (system_end_time.tv_usec - system_start_time.tv_usec >= 0) {
      sec = system_end_time.tv_sec - system_start_time.tv_sec;
      usec = system_end_time.tv_usec - system_start_time.tv_usec;
      printf("Last action used: %d.%06d sec, %d bytes. Total: %d bytes.\n", 
             (int)sec, (int)usec, (int)sbrk(0) - last_mem, (int)sbrk(0));
    }
    else {
      sec = system_end_time.tv_sec - system_start_time.tv_sec-1;
      usec = system_end_time.tv_usec - system_start_time.tv_usec +1000000;
      printf("Last action used: %d.%06d seconds, %d bytes. Total %d bytes.\n", 
             (int)sec, (int)usec, (int)sbrk(0) - last_mem, (int)sbrk(0));
    }
    total_sec += sec;
    total_usec += usec;
    last_sec = sec;
    last_usec = usec;
    assign_time();
  }
  return 0;
}

int assign_time() {
  if(NEW_THREAD) {
    NEW_THREAD->total_sec += last_sec;
    NEW_THREAD->total_usec += last_usec;
    NEW_THREAD->last_sec = last_sec;
    NEW_THREAD->last_usec = last_usec;
  }
  return 0;
}

int getallspace() {
  char c;
  while(TRUE) {
    c = getchar();
    if(c != ' ' && c != '\t') break;
  }
  ungetc(c,stdin);
  return 0;
}

int merge_compatible(struct thread_type *t1, struct thread_type *t2) {
  int i;
/*
  printf("comparing %s[%d] & %s[%d]\n", t1->name, t1->automaton_index, 
                                        t2->name, t2->automaton_index);
*/
  for(i=0; i<system_ta_num; i++) {
    if(component_graph(i, t1) && component_graph(i, t2)) return FALSE;
  }
/* The following does not work with sequential merge (no threads kept) */
/*
  if(t1->subthread1 == NULL && t1->subthread2 == NULL &&
     t2->subthread1 == NULL && t2->subthread2 == NULL) {
    if(t1->index == t2->index) return FALSE;
  }
  else
  if(t1->subthread1 == NULL && t1->subthread2 == NULL &&
     t2->subthread1 != NULL && t2->subthread2 != NULL) {
    if(merge_compatible(t1, t2->subthread1) == FALSE) return FALSE;
    if(merge_compatible(t1, t2->subthread2) == FALSE) return FALSE;
  }
  else
  if(t1->subthread1 != NULL && t1->subthread2 != NULL &&
     t2->subthread1 == NULL && t2->subthread2 == NULL) {
    if(merge_compatible(t1->subthread1, t2) == FALSE) return FALSE;
    if(merge_compatible(t1->subthread2, t2) == FALSE) return FALSE;
  }
  else {
  if(merge_compatible(t1->subthread1, t2->subthread1) == FALSE) return FALSE;
  if(merge_compatible(t1->subthread1, t2->subthread2) == FALSE) return FALSE;
  if(merge_compatible(t1->subthread2, t2->subthread1) == FALSE) return FALSE;
  if(merge_compatible(t1->subthread2, t2->subthread2) == FALSE) return FALSE;
  }   
*/
/*
  printf("  TRUE.\n");
*/
  return TRUE;
}

int short_info(struct thread_type *t, int with_name) {
  char *ts = NULL;
  int i;
  char *c = NULL;

  if(t == NULL) return 0;

  if(with_name == TRUE) {
    ts = get_graph_name(t);
    sprintf(tmpstring, "%s\n%s", ts, print_size(t));
    free(ts);
    ts = str_preserve(tmpstring);
  }
  else {
    ts = str_preserve(print_size(t));
  }

  if(t->mode_count == 0) {
    ts = (char *)realloc(ts, sizeof(char)*(strlen(ts)+strlen("EMPTY")+2));
    strcat(ts, "EMPTY");
  }
  else for(i=0; i<system_ta_num; i++) {
    if(component_graph(i, t)) {
      c = get_graph_name(&thread[i]);
      sprintf(tmpstring, "%s ", c);
      ts = (char *)realloc(ts, sizeof(char)*(strlen(ts)+strlen(tmpstring)+1));
      strcat(ts, tmpstring);
      free(c);
    }
  }
  ts = (char *)realloc(ts, sizeof(char)*(strlen(ts)+strlen("\n")+1));
  strcat(ts, "\n");
  append_string(ts, &answer);
  free(ts);
  return 0;
}

int info_graph(struct thread_type *t) {
  char *ans = NULL;
  char *s = NULL;

  if(t) {
    sprintf(tmpstring, "SIZE = %d modes, %d transitions\n", 
                       t->mode_count, t->trans_count);
    if(SGM_MODE == GUI || SGM_MODE == PROCESS) ans = str_preserve(tmpstring);
    else {
      s = get_graph_name(t);
      printf("%s\n\t%s", s, tmpstring);
      free(s);
    }

    s = print_components(t);
    if(SGM_MODE == GUI || SGM_MODE == PROCESS) {
      ans = (char *)realloc(ans, sizeof(char)*(strlen(ans)+strlen(s)+1));
      strcat(ans, s);
    }
    free(s); s = NULL;

    s = print_mnpl(t);
    if(SGM_MODE == GUI || SGM_MODE == PROCESS) {
      ans = (char *)realloc(ans, sizeof(char)*(strlen(ans)+strlen(s)+1));
      strcat(ans, s);
    }
    free(s); s = NULL;

    s = print_time(t);
    if(SGM_MODE == GUI || SGM_MODE == PROCESS) {
      ans = (char *)realloc(ans, sizeof(char)*(strlen(ans)+strlen(s)+1));
      strcat(ans, s);
      append_string(ans, &answer);
      free(ans);
    }
    free(s); s = NULL;
  }
  else {
    info_graph_all();
  }
  return 0;
}


int info_graph_all() {
  int i;
  for(i=0; i<2*system_ta_num-1; i++) {
    info_graph(&thread[i]);
  }
  return 0;
}

char *print_components(struct thread_type *t) {
  int i;
  char *ts = NULL;
  char *c = NULL;

  if(t == NULL) return NULL;

  if(SGM_MODE != GUI && SGM_MODE != PROCESS) printf("\tPROCESSES = ");
  else ts = str_preserve("PROCESSES = ");

  if(t->mode_count == 0) {
    if(SGM_MODE != GUI && SGM_MODE != PROCESS) printf("NONE");
    else {
      ts = (char *)realloc(ts, sizeof(char)*(strlen(ts)+strlen("NONE")+1));
      strcat(ts, "NONE");
    }
  }
  else for(i=0; i<system_ta_num; i++) {
    if(component_graph(i, t)) {
      c = get_graph_name(&thread[i]);
      if(SGM_MODE != GUI && SGM_MODE != PROCESS)
        printf("%s ", c);
      else {
        sprintf(tmpstring, "%s ", c);
        ts = (char *)realloc(ts, sizeof(char)*(strlen(ts)+strlen(tmpstring)+1));
        strcat(ts, tmpstring);
      }
      free(c);
    }
  }
  if(SGM_MODE != GUI && SGM_MODE != PROCESS) printf("\n");
  else {
    ts = (char *)realloc(ts, sizeof(char)*(strlen(ts)+strlen("\n")+1));
    strcat(ts, "\n");
  }
  return ts;
}

char *print_mnpl(struct thread_type *t) {
  int n;
  char *c1, *c2, *st = NULL;

  if(t) {
    if(SGM_MODE != GUI && SGM_MODE != PROCESS) printf("\tMANIPULATORS APPLIED:\n");
    else st = str_preserve("MANIPULATORS APPLIED:\n");

    n = t->mnpl & MG;
    if(n) {
      c1 = get_graph_name(t->subthread1);
      c2 = get_graph_name(t->subthread2);
      if(SGM_MODE != GUI && SGM_MODE != PROCESS)
        printf("\t\tMerge_Graph(%s, %s) \n", c1, c2);
      else {
        sprintf(tmpstring, "    Merge_Graph(%s, %s) \n", c1, c2);
        st = (char *)realloc(st, sizeof(char)*(strlen(st)+strlen(tmpstring)+1));
        strcat(st, tmpstring);
      }
      free(c1);
      free(c2);
    }

    n = t->mnpl & RW;
    if(n) {
      if(SGM_MODE != GUI && SGM_MODE != PROCESS) printf("\t\tRead_Write()\n");
      else {
        st = (char *)realloc(st, sizeof(char)*(strlen(st)+
                                           strlen("    Read_Write()\n")+1));
        strcat(st, "    Read_Write()\n");
      }
    }

    n = t->mnpl & SC;
    if(n) {
      if(SGM_MODE != GUI && SGM_MODE != PROCESS) printf("\t\tShield_Clocks()\n");
      else {
        st = (char *)realloc(st, sizeof(char)*(strlen(st)+
                                           strlen("    Shield_Clocks()\n")+1));
        strcat(st, "    Shield_Clocks()\n");
      }
    }
  
    n = t->mnpl & NR;
    if(n) {
      if(SGM_MODE != GUI && SGM_MODE != PROCESS) printf("\t\tNormalize_Regions()\n");
      else {
        st = (char *)realloc(st, sizeof(char)*(strlen(st)+
                                        strlen("    Normalize_Regions()\n")+1));
        strcat(st, "    Normalize_Regions()\n");
      }
    }

    n = t->mnpl & TB;
    if(n) {
      if(SGM_MODE != GUI && SGM_MODE != PROCESS) printf("\t\tTransition_Bypass()\n");
      else {
        st = (char *)realloc(st, sizeof(char)*(strlen(st)+
                                        strlen("    Transition_Bypass()\n")+1));
        strcat(st, "    Transition_Bypass()\n");
      }
    }

    n = t->mnpl & SV;
    if(n) {
      if(SGM_MODE != GUI && SGM_MODE != PROCESS) 
        printf("\t\tShield_Variable()\n");
      else {
        st = (char *)realloc(st, sizeof(char)*(strlen(st)+
                                        strlen("    Shield_Variable()\n")+1));
        strcat(st, "    Shield_Variable()\n");
      }
    }

    if(t->mnpl == 0) {
      if(SGM_MODE != GUI && SGM_MODE != PROCESS) printf("\t\tNone.\n");
      else {
        st = (char *)realloc(st, sizeof(char)*(strlen(st)+
                                        strlen("    None.\n")+1));
        strcat(st, "    None.\n");
      }
    }
  }
  return st;
}

char *print_time(struct thread_type *t) {
  char *ss = NULL;

  ss = (char *)malloc(sizeof(char)*MAX_CHAR);
  reset_str(ss, MAX_CHAR);
  if(t) {
    if(SGM_MODE != GUI && SGM_MODE != PROCESS) {
      printf("\tTOTAL TIME USED: %d.%06d seconds.\n", 
               (int)t->total_sec, (int)t->total_usec);
      printf("\tLAST ACTION TIME USED: %d.%06d seconds.\n", 
               (int)t->last_sec, (int)t->last_usec);
    }
    else {
      sprintf(tmpstring, "TOTAL TIME USED: %d.%06d seconds.\n", 
                          (int)t->total_sec, (int)t->total_usec);
      strcat(ss, tmpstring);
      sprintf(tmpstring, "LAST ACTION TIME USED: %d.%06d seconds.\n", 
                          (int)t->last_sec, (int)t->last_usec);
      strcat(ss, tmpstring);
    }
  }
  return ss;
}

int rename_graph(struct thread_type *t, char *new_name) {
  char *name = NULL;
  int start = -1, end = -1;

  if(valid_new_name(new_name)) {
    if(get_gd_info(new_name, &name, &start, &end)) {
      free(t->name);
      t->name = str_preserve(name);
      t->automaton_index = start;
    }
  }
  else {
    if(SGM_MODE == INTERACTIVE || SGM_MODE == BATCH)
      fprintf(stderr, "Invalid new name (%s) for graph.\n", new_name);
    else
      GUI_ERROR = E_INVNAME;
  }
  return 0;
}

int valid_new_name(char *n) {
  return TRUE;
}
