#include "include.h"
#include "global.h"
#include "bddman.h"
#include "modeman.h"
#include "readwrite.h"
#include "print.h"
#include "transman.h"



/***** GLOBAL VARIABLES *****/

bdd_manager bddm;
struct a23tree_type *bdd_tree;



/***** FUNCTION DECLARATIONS *****/

void init_bdd(void);
bdd new_bdd(void);
struct proposition_type *new_prop(struct boolconj_type *);
int create_props(struct variable_type *);
int create_props2();
void install_bdd(struct boolconj_type *);
void install_bdd2(struct boolconj_type *);
int install_boolexp_bdds(void);
int bdd_comp(bdd, bdd);
/*
int bdd_comp_internal(bdd, bdd);
*/
bdd BDD_AND(bdd, bdd);

bdd prop_bdd_and(struct proposition_type *, struct proposition_type *);


/***** FUNCTION DEFINITIONS *****/

void init_bdd() {
  struct variable_type *v;
  bddm = bdd_init();
  if(var_tail) {
    for(v = var_tail->next; ; v=v->next) {
      if(v->type == V_REGISTER) 
        create_props(v);
      if(v == var_tail) break;
    }
/*
    create_props2();
*/
  }
}

int create_props(struct variable_type *v) {
  struct boolconj_type *c, *bb;
  struct automaton *a, *b;
  int i, j, k, vid, written = FALSE;
  struct mode_type *m;
  struct transition_type *tr;
  struct boolexp_type *x;

  c = alloc_boolconj(); 
  c->var1 = v->index;
  c->op = EQ;
  if(automaton_tail) {
    for(i=0, a=automaton_tail->next; i<system_ta_num; i++, a=a->next) {
      vid = get_dvar_id(v->index, a);
      if(vid > -1) {
        for(j=1; j<=val_count[i][vid]; j++) {
          c->constant = varval[i][vid][j];
          for(k=0, b=automaton_tail->next; k<i; k++, b=b->next) {
            if(varval_written_auto(c->var1, c->constant, b)) {
              written = TRUE;
              break;
            }
          }
          if(!written) new_prop(c);	/* var=val is not written before */
          else written = FALSE;
        }
      }
      else {
        fprintf(stderr, "Fatal Error in create_props of bddman.c: ...\n");
        fprintf(stderr, "variable %s[%d] not found in varval data structure.\n",
                         v->name, v->id);
        exit(1);
      }
    }
/*
    if(read_only_variable(v)) {
*/
      for(i=0, a=automaton_tail->next; i<system_ta_num; i++, a=a->next) {
        if(a->init_boolexp) {
          for(x=a->init_boolexp->next; ; x=x->next) {
            if(x->boolconj) {
              for(bb=x->boolconj->next; ; bb=bb->next) {
                if(bb->var2 > 0) { /* ignore boolconj with 2 discrete vars */
                  if(bb == x->boolconj) break;
                  else continue;
                }
                if(bb->var1 == v->index) {
                  c->constant = bb->constant;
                  new_prop(c);
                }
                if(bb == x->boolconj) break;
              }
            }
            if(x == a->init_boolexp) break;
          }
        }
        if(a->mode_cycle) {
          for(m=a->mode_cycle->next; ; m=m->next) {
            if(m->out_trans_cycle) {
              for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
                if(tr->trn->prp_trigger_cycle) {
                  for(bb=tr->trn->prp_trigger_cycle->next; ; bb=bb->next) {
                    if(bb->var2 > 0) { /*ignore boolconj with 2 discrete vars */
                      if(bb == tr->trn->prp_trigger_cycle) break;
                      else continue;
                    }
                    if(bb->var1 == v->index) {
                      c->constant = bb->constant;
                      new_prop(c);
                    }
                    if(bb == tr->trn->prp_trigger_cycle) break;
                  }
                }
                if(tr == m->out_trans_cycle) break;
              }
            }
            if(m == a->mode_cycle) break;
          }
        } /* END if(a->mode_cycle) */
      }
/*
    }
*/
  }
  free(c);
  return 0;
}

/* Create propositions for boolean conjuncts with two variables */
int create_props2() {
  int i; 
  struct automaton *a;
  struct mode_type *m;
  struct transition_type *tr;
  struct boolexp_type *x;
  struct boolconj_type *c, *bb;

  c = alloc_boolconj(); 
  c->op = EQ;
  if(automaton_tail) {
      for(i=0, a=automaton_tail->next; i<system_ta_num; i++, a=a->next) {
        if(a->init_boolexp) {
          for(x=a->init_boolexp->next; ; x=x->next) {
            if(x->boolconj) {
              for(bb=x->boolconj->next; ; bb=bb->next) {
                c->var1 = bb->var1;
                c->var2 = bb->var2;
                c->constant = bb->constant;
                new_prop(c);
                if(bb == x->boolconj) break;
              }
            }
            if(x == a->init_boolexp) break;
          }
        }
        if(a->mode_cycle) {
          for(m=a->mode_cycle->next; ; m=m->next) {
            if(m->out_trans_cycle) {
              for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
                if(tr->trn->prp_trigger_cycle) {
                  for(bb=tr->trn->prp_trigger_cycle->next; ; bb=bb->next) {
                    c->var1 = bb->var1;
                    c->var2 = bb->var2;
                    c->constant = bb->constant;
                    new_prop(c);
                    if(bb == tr->trn->prp_trigger_cycle) break;
                  }
                }
                if(tr == m->out_trans_cycle) break;
              }
            }
            if(m == a->mode_cycle) break;
          }
        } /* END if(a->mode_cycle) */
      }
  }
  free(c);
  return 0;
}

bdd new_bdd() {
  bdd f;

  if(prop_tail) {
    f = bdd_new_var_after(bddm, prop_tail->prop_bdd);
  }
  else
    f = bdd_new_var_first(bddm);
  return f;
}
  
struct proposition_type *new_prop(struct boolconj_type *c) {
  struct proposition_type * p; 
  bdd f;
 
  if(prop_tail) {
    for(p=prop_tail->next; ; p=p->next) {
      if(p->var1 == c->var1 && p->var2 == c->var2 && p->op == c->op &&
         p->value == c->constant) return p;
      if(p == prop_tail) break;
    }
  }
  f = new_bdd();
  p = (struct proposition_type *)malloc(sizeof(struct proposition_type));
  p->prop_bdd = f;
  p->var1 = c->var1;
  p->var2 = c->var2;
  p->op = c->op;
  p->value = c->constant;
  if(prop_tail) {
    p->next = prop_tail->next;
    prop_tail->next = p;
  }
  else p->next = p;
  prop_tail = p;
  if(verbose >= 3) {
    output("Created new proposition/bdd: ");
    print_bdd(p->prop_bdd);
  }
  return p;
}

void install_bdd(struct boolconj_type *c) {
  struct proposition_type *p;

  if(c->op == TRUE) {
    c->prop_bdd = BDD_TRUE;
    return;
  }
  c->prop_bdd = BDD_FALSE;
  if(prop_tail) {
    p = prop_tail->next;
    switch(c->op) {
      case LESS:
        for(; ; p = p->next) {
          if(p->var1 == c->var1 && p->var2 == c->var2) {
            if (p->value < c->constant) 
              c->prop_bdd = bdd_or(bddm, c->prop_bdd, p->prop_bdd);
            else {
             if(c->prop_bdd == BDD_FALSE) 
               c->prop_bdd = bdd_not(bddm, p->prop_bdd);
             else 
             c->prop_bdd=bdd_and(bddm, c->prop_bdd, bdd_not(bddm, p->prop_bdd));
            }
          }
          if(p == prop_tail) break;
        } 
        break;
      case EQ:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && p->var2 == c->var2 && 
             c->constant == p->value) 
            c->prop_bdd = p->prop_bdd;
          if(p == prop_tail) break;
        }
        if(c->prop_bdd == BDD_FALSE) break;
        p = prop_tail->next;
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->var2 == p->var2) { 
            if(c->constant != p->value) 
             c->prop_bdd=bdd_and(bddm, c->prop_bdd, bdd_not(bddm, p->prop_bdd));
          }
          if(p == prop_tail) break;
        }
        break;
      case GEQ:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->var2 == p->var2) {
            if(p->value >= c->constant) 
              c->prop_bdd = bdd_or(bddm, c->prop_bdd, p->prop_bdd);
          }
          if(p == prop_tail) break;
        }
        if(c->prop_bdd == BDD_FALSE) break;
        p = prop_tail->next;
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->var2 == p->var2) { 
            if(p->value < c->constant) 
             c->prop_bdd=bdd_and(bddm, c->prop_bdd, bdd_not(bddm, p->prop_bdd));
          }
          if(p == prop_tail) break;
        }
        break;
      case NEQ:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->var2 == p->var2 && c->constant!=p->value) 
              c->prop_bdd = bdd_or(bddm, c->prop_bdd, p->prop_bdd);
          if(p == prop_tail) break;
        }
        p = prop_tail->next;
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->var2 == p->var2 && p->value==c->constant)
             c->prop_bdd=bdd_and(bddm, c->prop_bdd, bdd_not(bddm, p->prop_bdd));
          if(p == prop_tail) break;
        }
        break;
      case LEQ:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->var2 == p->var2 && p->value<= c->constant)
              c->prop_bdd = bdd_or(bddm, c->prop_bdd, p->prop_bdd);
          if(p == prop_tail) break;
        }
        if(c->prop_bdd == BDD_FALSE) break;
        p = prop_tail->next;
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->var2 == p->var2 && p->value > c->constant)
             c->prop_bdd=bdd_and(bddm, c->prop_bdd, bdd_not(bddm, p->prop_bdd));
          if(p == prop_tail) break;
        }
        break;
      case GREATER:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->var2 == p->var2 && p->value > c->constant)
              c->prop_bdd = bdd_or(bddm, c->prop_bdd, p->prop_bdd);
          if(p == prop_tail) break;
        }
        p = prop_tail->next;
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->var2 == p->var2 && 
             p->value <= c->constant) {
            if(c->prop_bdd == BDD_FALSE)
              c->prop_bdd = bdd_not(bddm, p->prop_bdd);
            else
             c->prop_bdd=bdd_and(bddm, c->prop_bdd, bdd_not(bddm, p->prop_bdd));
          }
          if(p == prop_tail) break;
        }
        break;
    }
  }
}

void install_bdd2(struct boolconj_type *c) {
  struct proposition_type *p, *q;
  int val;
  
  c->prop_bdd = BDD_FALSE;

  if(prop_tail) {
    for(p = prop_tail->next; ; p = p->next) {
      if(c->var1 == p->var1) {
        for(q = prop_tail->next; ; q = q->next) {
          if(c->var2 == q->var1) {
            if(c->mod_num) 
              val = (int)fmod(p->value - q->value, c->mod_num);
            else if(c->mod_var) {
              fprintf(stderr, 
                      "Error: MOD var in boolean conjunct not implemented!\n");
              exit(1);
            }
            else
              val = p->value - q->value;
            switch(c->op) {
              case LESS: 
                if(val < c->constant) {
                  c->prop_bdd = bdd_or(bddm, c->prop_bdd, prop_bdd_and(p, q));
                }
                else {
                  if(c->prop_bdd == BDD_FALSE)
                    c->prop_bdd = bdd_not(bddm, prop_bdd_and(p, q));
                  else
                    c->prop_bdd=bdd_and(bddm, c->prop_bdd, 
                                        bdd_not(bddm, prop_bdd_and(p, q)));
                }
                break;
              case LEQ:
                if(val <= c->constant) {
                  c->prop_bdd = bdd_or(bddm, c->prop_bdd, prop_bdd_and(p, q));
                }
                else {
                  if(c->prop_bdd == BDD_FALSE)
                    c->prop_bdd = bdd_not(bddm, prop_bdd_and(p, q));
                  else
                    c->prop_bdd=bdd_and(bddm, c->prop_bdd, 
                                        bdd_not(bddm, prop_bdd_and(p, q)));
                }
                break;
              case EQ:
                if(val == c->constant) {
                  c->prop_bdd = bdd_or(bddm, c->prop_bdd, prop_bdd_and(p, q));
                }
                else {
                  if(c->prop_bdd == BDD_FALSE)
                    c->prop_bdd = bdd_not(bddm, prop_bdd_and(p, q));
                  else
                    c->prop_bdd=bdd_and(bddm, c->prop_bdd, 
                                        bdd_not(bddm, prop_bdd_and(p, q)));
                }
                break;
              case NEQ:
                if(val != c->constant) {
                  c->prop_bdd = bdd_or(bddm, c->prop_bdd, prop_bdd_and(p, q));
                }
                else {
                  if(c->prop_bdd == BDD_FALSE)
                    c->prop_bdd = bdd_not(bddm, prop_bdd_and(p, q));
                  else
                    c->prop_bdd=bdd_and(bddm, c->prop_bdd, 
                                        bdd_not(bddm, prop_bdd_and(p, q)));
                }
                break;
              case GEQ:
                if(val >= c->constant) {
                  c->prop_bdd = bdd_or(bddm, c->prop_bdd, prop_bdd_and(p, q));
                }
                else {
                  if(c->prop_bdd == BDD_FALSE)
                    c->prop_bdd = bdd_not(bddm, prop_bdd_and(p, q));
                  else
                    c->prop_bdd=bdd_and(bddm, c->prop_bdd, 
                                        bdd_not(bddm, prop_bdd_and(p, q)));
                }
                break;
              case GREATER:
                if(val > c->constant) {
                  c->prop_bdd = bdd_or(bddm, c->prop_bdd, prop_bdd_and(p, q));
                }
                else {
                  if(c->prop_bdd == BDD_FALSE)
                    c->prop_bdd = bdd_not(bddm, prop_bdd_and(p, q));
                  else
                    c->prop_bdd=bdd_and(bddm, c->prop_bdd, 
                                        bdd_not(bddm, prop_bdd_and(p, q)));
                }
                break;
            }
          }
          if(q == prop_tail) break;
        }
      }
      if(p == prop_tail) break;
    }
  }
}

bdd prop_bdd_and(struct proposition_type *p, struct proposition_type *q) {
  struct proposition_type *x;
  bdd f = BDD_TRUE;

  if(prop_tail) {
    for(x = prop_tail->next; ; x = x->next) {
      if(x->var1 == p->var1) {
        if(x->value == p->value) f = BDD_AND(f, x->prop_bdd);
        else f = BDD_AND(f, bdd_not(bddm, x->prop_bdd));
      }
      else if(x->var1 == q->var1) {
        if(x->value == q->value) f = BDD_AND(f, x->prop_bdd);
        else f = BDD_AND(f, bdd_not(bddm, x->prop_bdd));
      }
      if(x == prop_tail) break;
    }
  }
  return f;
}

/*
void install_bdd(struct boolconj_type *c) {
  struct proposition_type *p, *a, *b;

  if(prop_tail) {
    p = prop_tail->next;
    switch(c->op) {
      case LESS:
      case EQ:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->op == p->op && c->constant == p->value) {
            c->prop_bdd = p->prop_bdd;
            return;
          }
          if(p == prop_tail) break;
        }
        break;
      case GEQ:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && p->op == LESS && c->constant == p->value) {
            c->prop_bdd = bdd_not(bddm, p->prop_bdd);
            return;
          }
          if(p == prop_tail) break;
        }
        break;
      case NEQ:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && p->op == EQ && c->constant == p->value) {
            c->prop_bdd = bdd_not(bddm, p->prop_bdd);
            return;
          }
          if(p == prop_tail) break;
        }
        break;
      case LEQ:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->constant == p->value) {
            if(p->op == LESS) a = p;
            if(p->op == EQ) b = p;
          }
          if(a && b) {
            c->prop_bdd = bdd_or(bddm, a->prop_bdd, b->prop_bdd);
            return ;
          }
          if(p == prop_tail) break;
        }
        break;
      case GREATER:
        for(; ; p = p->next) {
          if(c->var1 == p->var1 && c->constant == p->value) {
            if(p->op == LESS) a = p;
            if(p->op == EQ) b = p;
          }
          if(a && b) {
            c->prop_bdd = bdd_nor(bddm, a->prop_bdd, b->prop_bdd);
            return ;
          }
          if(p == prop_tail) break;
        }
        break;
    }
  }
  if(!prop_tail || !(c->prop_bdd)) { 
    switch(c->op) {
      case LESS:
      case EQ:
            c->prop_bdd = (new_prop(c))->prop_bdd;
            return;
      case GEQ:
            c->op = LESS;
            c->prop_bdd = bdd_not(bddm, (new_prop(c))->prop_bdd);
            c->op = GEQ;
            return;
      case NEQ:
            c->op = EQ;
            c->prop_bdd = bdd_not(bddm, (new_prop(c))->prop_bdd);
            c->op = NEQ;
            return;
      case LEQ:
            if(!a) {
              c->op = LESS;
              a = new_prop(c);
              c->op = GEQ;
            }
            if(!b) {
              c->op = EQ;
              b = new_prop(c);
              c->op = GEQ;
            }
            c->prop_bdd = bdd_or(bddm, a->prop_bdd, b->prop_bdd);
            return;
      case GREATER:
            if(!a) {
              c->op = LESS;
              a = new_prop(c);
              c->op = GEQ;
            }
            if(!b) {
              c->op = EQ;
              b = new_prop(c);
              c->op = GEQ;
            }
            c->prop_bdd = bdd_nor(bddm, a->prop_bdd, b->prop_bdd);
            return;
    }
  }
}
*/

int install_init_boolexp_bdds() {
  struct automaton *a;
  struct boolexp_type *b;
  struct boolconj_type *c;
  bdd f;
  bdd g = bdd_one(bddm);
  bdd h = bdd_zero(bddm);

 if(automaton_tail)
  for(a = automaton_tail->next; ; a = a->next) {
    if(a->init_boolexp) {
      for(b = a->init_boolexp->next; ; b = b->next) {
        if(b->boolconj) {
          for(c = b->boolconj->next; ; c = c->next) {
            install_bdd(c);
            f = c->prop_bdd;
            g = bdd_and(bddm, f, g);
            if(c == b->boolconj) break;
          }  
          h = bdd_or(bddm, g, h); 
          g = bdd_one(bddm); 
        }
        if(b == a->init_boolexp) break;
      }
      a->init_boolexp->boolexp_bdd = h;
      if(verbose >= 2) {
        if(a->index >= 0)
          sprintf(tmpstring,
                "Installed bdd for init_boolexp of automaton %s[%d]:\n",
                                               a->name, a->index); 
        else if(a->index == NON_PARAM)
          sprintf(tmpstring,
                "Installed bdd for init_boolexp of automaton %s:\n", a->name); 
        output(tmpstring);
        print_bdd(a->init_boolexp->boolexp_bdd);
        output("\n");
      }
      h = bdd_zero(bddm);
      g = bdd_one(bddm);
    }
    else {
      a->init_boolexp = alloc_boolexp();
      a->init_boolexp->boolexp_bdd = BDD_TRUE;
      a->init_boolexp->boolconj = alloc_boolconj();
      a->init_boolexp->boolconj->op = TRUE;
    }
    if(a == automaton_tail) break;
  }
  return 0;
}
      
int install_boolexp_bdds() {
  struct automaton *a;
  struct mode_type *m;
  struct boolexp_type *b;
  struct boolconj_type *c;
  bdd f;
  bdd g = bdd_one(bddm);
  bdd h = bdd_zero(bddm);

 if(automaton_tail)
  for(a = automaton_tail->next; ; a = a->next) {
    if(a->mode_cycle)
    for(m = a->mode_cycle->next; ; m = m->next) {
      if(m->boolexp) {
        for(b = m->boolexp->next; ; b = b->next) {
          if(b->boolconj) {
            for(c = b->boolconj->next; ; c = c->next) {
              install_bdd(c);
              f = c->prop_bdd;
              g = bdd_and(bddm, f, g);
              if(c == b->boolconj) break;
            }  
            h = bdd_or(bddm, g, h); 
            g = bdd_one(bddm); 
          }
          if(b == m->boolexp) break;
        }
        m->boolexp->boolexp_bdd = h;
        if(verbose >= 2) {
          sprintf(tmpstring, 
                  "Installed bdd for boolexp of mode %s: \n", m->name); 
          output(tmpstring);
          print_bdd(m->boolexp->boolexp_bdd);
          output("\n");
        }
      }
      else {
        m->boolexp = alloc_boolexp();
        m->boolexp->boolexp_bdd = BDD_TRUE;
/*
        m->boolexp->boolconj->op = TRUE;
*/
      }
      h = bdd_zero(bddm);
      g = bdd_one(bddm);
      if(m == a->mode_cycle) break;
    }
    if(a == automaton_tail) break;
  }
  return 0;
}

void install_trigger_bdds() {
  struct automaton *a;
  struct mode_type *m;
  struct transition_type *t;
  struct boolconj_type *c;
  bdd f;
  bdd g = bdd_one(bddm);

 if(automaton_tail)
  for(a = automaton_tail->next; ; a = a->next) {
    if(a->mode_cycle)
    for(m = a->mode_cycle->next; ; m = m->next) {
      if(m->out_trans_cycle) {
        for(t = m->out_trans_cycle->next; ; t = t->next) {
          if(t->trn->prp_trigger_cycle) {
            for(c = t->trn->prp_trigger_cycle->next; ; c = c->next) {
              if(c->var2) install_bdd2(c);
              else install_bdd(c);
              f = c->prop_bdd;
              g = bdd_and(bddm, f, g);
              if(c == t->trn->prp_trigger_cycle) { /*end of cycle, install bdd*/
                t->trn->trigger_bdd = g;
                if(verbose >= 2) {
                  sprintf(tmpstring, 
                   "Installed bdd for trigger of transition #%d (mode %s): \n", 
                         t->index, m->name); 
                  output(tmpstring);
                  print_bdd(t->trn->trigger_bdd);
                  output("\n");
                }
                break;
              }
            } /* END for c */  
            g = bdd_one(bddm); 
          }
          else {
            t->trn->prp_trigger_cycle = alloc_boolconj();
            t->trn->prp_trigger_cycle->op = TRUE;
            t->trn->prp_trigger_cycle->next = t->trn->prp_trigger_cycle;
            t->trn->trigger_bdd = BDD_TRUE;
          }
/*
          install_in_trans_trigger_bdd(m, t);
*/
          if(t == m->out_trans_cycle) break;
        }  /* END for t */
      }
      if(m == a->mode_cycle) break;
    }  /* END for m */
    if(a == automaton_tail) break;
  } /* END for a */
}

/*
int install_in_trans_trigger_bdd(struct mode_type *m, 
                                 struct transition_type *t) {
  struct transition_type *ptr;

  if(t->trn->trigger_bdd) {
    for(ptr = t->mode->in_trans_cycle->next; ; ptr=ptr->next) {
      if(ptr->index == t->index) {
        ptr->trn->trigger_bdd = t->trn->trigger_bdd;
      }
      if(ptr == t->mode->in_trans_cycle) break;
    }
  }
  return;
}
*/

int bdd_comp(bdd bdd1, bdd bdd2) {
  int comp_result;
  if(prop_tail) 
    comp_result = bdd_compare(bddm, bdd1, bdd2, prop_tail->prop_bdd);
  else return EQUAL;
  if(comp_result > 0) return EARLIER;
  else if(comp_result < 0) return LATER;
  else return EQUAL;
}

bdd BDD_AND(bdd bdd1, bdd bdd2) {
  if(!bdd1 && !bdd2) {
    return BDD_TRUE;
  }
  else if(!bdd1 && bdd2) {
    return bdd2;
  }
  else if(bdd1 && !bdd2) {
    return bdd1;
  }
  else {   /* if(bdd1 && bdd2) { */
    return bdd_and(bddm, bdd1, bdd2);
  }
}

bdd BDD_OR(bdd bdd1, bdd bdd2) {
  if(!bdd1 && !bdd2) {
    return BDD_TRUE;
  }
  else if(!bdd1 && bdd2) {
    return bdd2;
  }
  else if(bdd1 && !bdd2) {
    return bdd1;
  }
  else { /* if(bdd1 && bdd2) { */
    return bdd_or(bddm, bdd1, bdd2);
  }
}
