/*
 * Wireless Transport Protocol 
 * (Class 1 transactions) GSM SMS Version
 */

define(R,40)
define(A,5)
define(W,300)
define(MAX_RCR,4)
define(AEC_MAX,4)

define(STATUS_RUNNING,0)
define(STATUS_COMPLETE,1)
define(STATUS_ABORT,2)

automata Network, Initiator, Responder;

-- local variables
clock 		xi,  		-- local clock of Initiator
      		xr; 		-- local clock of Responder
register 	statusI,  	-- status of Initiator
         	statusR;  	-- status of Responder

-- global variables
register 	RCR, 		-- Retransmission Counter
		AEC;		-- Acknowledgment Expiration Counter
register 	TIDOK, 		-- Transaction ID Ok
		tidok_sent; 	-- Transaction ID Ok SENT

automaton Network {

synclabs i_send_invoke, i_send_ack, i_send_abort, i_rcv_ack, i_rcv_abort,
	 r_rcv_invoke, r_rcv_ack, r_rcv_abort, r_send_ack, r_send_abort;

initially network_initial;

  mode network_initial {
    invariant True;
    when True sync i_send_invoke may goto network_i_invoke; 	/* 1 */
    when True sync i_send_ack    may goto network_i_ack;	/* 2 */ 
    when True sync i_send_abort  may goto network_i_abort;	/* 3 */ 
    when True sync r_send_ack    may goto network_r_ack;	/* 4 */ 
    when True sync r_send_abort  may goto network_r_abort;	/* 5 */ 
  }

  /* we will never lose any packet, though it might be delayed */

  mode network_i_invoke {
    invariant True;
    when True sync r_rcv_invoke  may goto network_initial;	/* 6 */ 
  }

  mode network_i_ack {
    invariant True;
    when True sync r_rcv_ack     may goto network_initial;	/* 7 */ 
  }

  mode network_i_abort { 
    invariant True;
    when True sync r_rcv_abort   may goto network_initial;	/* 8 */ 
  }

  mode network_r_ack { 
    invariant True;
    when True sync i_rcv_ack     may goto network_initial;	/* 9 */ 
  }

  mode network_r_abort { 
    invariant True;
    when True sync i_rcv_abort   may goto network_initial;	/* 10 */ 
  }
}


automaton Initiator {

synclabs i_send_invoke, i_send_ack, i_send_abort, i_rcv_ack, i_rcv_abort;

initially init_null and xi = 0 and RCR = 0 & tidok_sent = 0 & 
                                   statusI = STATUS_RUNNING;

mode init_null {
  invariant True;
  /* ignore event TR-Invoke.req */
  when True sync i_send_invoke 	may goto init_result_wait;
            RCR:= 0; xi:= 0; TIDOK:=0; tidok_sent:=0;
	    statusI := STATUS_RUNNING; 				/* 11 */ 
}

mode init_result_wait {
  invariant True;
  /* ignore event TR-Abort.req */
  when xi < R sync i_send_abort may goto init_null;
	    statusI:= STATUS_ABORT; 				/* 12 */ 

  /* goto sub-mode while event RcvAck occurs */
  when xi < R sync i_rcv_ack may goto init_result_wait_ack; 	/* 13 */ 

  /* ignore action TR-Abort.ind */
  when xi < R sync i_rcv_abort may goto init_null; 
            statusI:= STATUS_ABORT; 				/* 14 */ 

  /* TimerTo_R */
  when xi = R and RCR < MAX_RCR & tidok_sent = 0 sync i_send_invoke
	    may goto init_result_wait; RCR:=RCR+1; xi:= 0; 	/* 15 */

  /* ignore TIDok sent with Ack */
  when xi = R and RCR < MAX_RCR & tidok_sent = 1 sync i_send_ack  
            may goto init_result_wait; RCR:=RCR+1; xi:= 0;	/* 16 */ 

  /* ignore indication */
  when xi = R and RCR = MAX_RCR may goto init_null; 
	    statusI:= STATUS_ABORT; 				/* 17 */ 
}

mode init_result_wait_ack {
  /* ignore action Stop Timer and TR-Abort.cnf */
  invariant True;
  when TIDOK = 0 may goto init_null; statusI:= STATUS_COMPLETE; /* 18 */ 

  /* doing TIDve */
/* ERIC: Added RCR<MAX_RCR */
  when TIDOK = 1 & RCR < MAX_RCR sync i_send_ack 
            may goto init_result_wait;
	    tidok_sent:= 1; RCR:=RCR+1; xi:= 0; 		/* 19 */ 
} 

}


automaton Responder {

synclabs r_rcv_invoke, r_rcv_ack, r_rcv_abort, r_send_ack, r_send_abort;

initially resp_listen and xr = 0 and AEC = 0 & statusR = STATUS_RUNNING;

mode resp_listen {
  invariant True;
  /* simplification of Invalid-TID-test */
  when True sync r_rcv_invoke may goto resp_listen_invoke; 
	    statusR:= STATUS_RUNNING; 				/* 20 */ 

  /* ignore TR-Invoke.ind */
  when True sync r_rcv_invoke may goto resp_invoke_resp_wait; 
            AEC:=0; xr:=0; statusR:= STATUS_RUNNING; 		/* 21 */ 
}

mode resp_listen_invoke {
  invariant True;
  when True sync r_send_ack may goto resp_tidok_wait;	
	    TIDOK:=1; 						/* 22 */
}

mode resp_tidok_wait {
  invariant True;

  /* ignore TR-Invoke.ind */
  when True sync r_rcv_ack may goto resp_invoke_resp_wait; 
	  TIDOK:=0; AEC:=0; xr:=0; 				/* 23 */ 

  when True sync r_rcv_abort may goto resp_listen;	
	  statusR:= STATUS_ABORT; 				/* 24 */ 

  /* goto sub-mode while event RcvInvoke occurs */
  when True sync r_rcv_invoke may goto resp_tidok_wait_invoke;	/* 25 */ 
}

mode resp_tidok_wait_invoke { 
  invariant True;
  /* ignore RID problem */
  when True sync r_send_ack may goto resp_tidok_wait; 		/* 26 */ 
} 

mode resp_invoke_resp_wait {
  invariant True;

when xr<=A sync r_send_ack may goto resp_wait_timeout; xr:= 0; 	/* 27 */ 

/* ignore event TR-Abort.req */
when xr < A sync r_rcv_abort may goto resp_listen;
	  statusR:= STATUS_ABORT; 				

when xr < A sync r_send_abort may goto resp_listen; 
	  statusR:= STATUS_ABORT; 				/* 28 */ 

/* ignore event RcvInvoke */
when xr < A sync r_rcv_invoke may goto resp_invoke_resp_wait; 	/* 29 */ 

/* TimerTO_A */
when xr = A and AEC < AEC_MAX may goto resp_invoke_resp_wait; 
	  AEC:=AEC+1; xr := 0;  				/* 30 */ 

/* ERIC: Removed comments */
when xr = A and AEC = AEC_MAX sync r_send_abort
	  may goto resp_listen; statusR:= STATUS_ABORT; 

/* ERIC: Removed AEC = AEC_MAX , Added xr := 0;*/
/* ERIC: Commented it out */
/*
when xr = A sync r_send_ack 
	  may goto resp_wait_timeout; xr := 0;			*/

}

mode resp_wait_timeout {
  invariant True;

/* goto sub-mode while event RcvInvoke occurs */
when xr < W sync r_rcv_invoke may goto resp_wait_timeout_invoke; /* 32 */

/* ignore T-TRAbort.ind */
when xr < W sync r_rcv_abort may goto resp_listen; 
	  statusR:= STATUS_ABORT; 				/* 33 */ 

/* Transaction succeed! Clear Transaction */
when xr = W may goto resp_listen; statusR:= STATUS_COMPLETE; 	/* 34 */ 

/* ignore event TR-Abort.req */
when xr < W sync r_send_abort may goto resp_listen;
	  statusR:= STATUS_ABORT; 				/* 35 */ 
}

mode resp_wait_timeout_invoke {
  invariant True;
when True sync r_send_ack may goto resp_wait_timeout;		/* 36  */ 
}

}

manipulation
  sequential_merge();
  read_write();
  shield_clock();
  transition_bypass();
--  normalize_region();

/*
risk
status[2] = STATUS_COMPLETE and status[3] = STATUS_ABORT; 
*/
