/* GLOBAL VARIABLES AND FUNCTIONS */

#ifndef __GLOBAL_H
#define __GLOBAL_H

#include "include.h"
#include "action.h"

extern struct thread_type *CURRENT_GRAPH, *COMPOSED_GRAPH, *REDUCED_GRAPH;
extern struct thread_type *NEW_THREAD, *THREAD1, *THREAD2;
extern int BASIC_THREAD_COUNT;
extern int TOTAL_THREAD_COUNT; 
extern char *tmpstring;
extern int composed_ta;
extern char *prompt;
extern int verbose;
extern int stdoutput;
extern int lastcomp;
extern int clock_shielding;
extern int on_the_fly_normalization;
extern int decl_graph_num;        /* total number of graphs declared
                                  (including system TA graphs) */

/*
extern int zone_reduction;
extern int partial_order_reduction;
extern int partial_order_with_symmetry_reduction;
*/
extern int region_normalization;
extern int model_checking;
extern int no_permutation;
extern int strict_semantics;
extern int changed;
extern int lazy_bdd_merging;
extern int look_forward_reduction;
extern int sibling_trans_multi_reduction;
extern int EIV;
extern int EOI;
extern int EDV;
extern int EOT;

extern int yylineno;
#ifdef __LINUX
extern char *yytext;
#else
extern char yytext[];
#endif
extern int yyleng;
extern FILE *logfile;

extern FILE *yyin;
extern void lexinit(void);
extern FILE *out;
extern char *infilename;
extern char *logfilename;

extern int *automata_num;      		/* automata counts */
extern char **automata_name;   		/* automata names */
extern char *parameter;        		/* current parametric index */
extern struct automaton *current_ta;  	/* points to the currently parsed TA */
extern int current_ta_num;            	/* number of currently parsed TA */
extern int system_ta_num;             	/* total number of system TA */

extern struct thread_type *thread;/* a thread is a process */

extern int automaton_count;       /* total number of automata */
extern int total_exp;             /* total number of expressions */

extern int *automata_num;    	  /* automata counts */
extern char **automata_name; 	  /* automata names */
extern char *parameter;      	  /* current parametric index */

extern int var_count;             /* total number of variables */
extern int tmp_var_count;         /* number of variables of the same type */
extern int var_type;              /* current variable type */
extern int interval_start;        /* variable index interval starting index */
extern int interval_end;          /* variable index interval ending index */
extern char *interval_varname;    /* variable name of an interval */

extern int index_relop;         /* indicative of the index range: <i, >i, !=i */
extern int index_quantifier;    /* quantifier of the index range: 
                                     FORALL, EXISTS */

extern int curr_exp_type;         /* current expression type: CLK_INEQ or BOOL_EXP */
extern int boolconj_count;        /* current count of boolean conjuncts */
extern int boolexp_count;         /* current count of boolean expressions */
extern int clkconj_count;         /* current count of clock conjuncts */
extern int inside_exp;            /* whether still parsing inside a conjunct */
extern int neg_conj;              /* whether a '~' expression */
extern int sign;                  /* LHS = +1, RHS = -1 */
extern int need_transform;        /* if the clkconj needs transforming ... */
                                  /* this occurs when: (a) x ~ y - d       */
                                  /*                   (b) x + d ~ y       */
                                  /* i.e., sign==1 during install_const()  */


extern int mode_count;            /* total number of modes */
extern int trans_count;           /* currently parsed transition count */
extern int prop_count;            /* current proposition count */

extern int zone_count;		  /* zone count */

extern int last_int;              /* last parsed integer */
extern char *string_buffer;       /* last parsed string */
  
extern int max_const;             /* maximum constant in specification */

extern int neg_spec;              /* whether a '~' spec */
extern struct spec_type *vspec, *cspec;  /* vspec: head of specification
                                          * cspec: current specification */
extern int blevel;		  /* bracket level */
extern int p_num;		  /* parametric index increment/decrement */
extern int merge_type;		  /* type of merging: SM, PM, ...         *
                                   *  declared in action.c                */

extern struct action_type *action_list, *curr_action, *tmp_action;
extern struct thread_type *graph_list;
extern struct manipulator_type *manipulator_list;

extern struct variable_type *var_tail;
extern struct automaton *automaton_tail;
extern struct mode_type **mode_tail;
extern struct spec_trans_type **out_trans_tail;

extern struct boolconj_type **boolconj_tail;
extern struct boolexp_type **boolexp_tail;
extern struct clkconj_type **clkconj_tail;
  
extern struct proposition_type *prop_tail;
extern struct spec_type *spec_tail;

extern void printline(void);
/*************************************************************
 * functions defined in include.c
 */

/***** variable processing *****/
extern struct variable_type *var_ptr(char *, int); 
                                        /* get pointer to variable from name */
extern struct variable_type *var_ptr_from_index(int); 
                                        /* get pointer to variable from index */
extern int var_index(char *, int);       /* get index of variable from name */
extern char *var_name(int);              /* get name of variable from index */
extern int var_id(int);			 /* get id of variable from index */
int get_var_type(int);            	 /* get variable type from index */
extern int insert_var(char *);          /* insert a parsed variable */
extern int add_attribute(int);          /* add variable attribute */

extern int init_tails(void);
/***** automaton processing *****/
extern int total_ta(char *);		/* get number of some ta from name */
extern int update_automata(void);	/* update automata name & count */
extern int new_automata(void);              /* new automata */
extern struct automaton *get_ta(char *, int);
extern int check_param(char *);

extern int install_init_mode(char *);
extern int install_init_cond(void);

/***** expression processing *****/
extern struct boolexp_type *alloc_boolexp(void);  /* allocate a new boolexp */
extern struct boolconj_type *alloc_boolconj(void); /* allocate a new boolconj */
extern int new_boolexp(void);               /* new boolean expression */
extern int new_boolconj(int);           /* new boolean conjunct */
extern int new_boolconj_array(void);        /* new boolean conjunct array */

extern int new_clkconj(void);               /* new clock conjunct */
extern struct clkconj_type *alloc_clkconj(void);  /* allocate a new clkconj */
struct clkconj_type *clkconj_copy1(struct clkconj_type *);
                                  /* copy a single clkconj */
struct clkconj_type *clkconj_copy(struct clkconj_type *);
                                  /* copy a clkconj cycle */
struct boolconj_type *boolconj_copy(struct boolconj_type *);
struct boolconj_type *boolconj_copy1(struct boolconj_type *);
struct assign_type *assign_copy(struct assign_type *);

extern int new_exp(char *, char *);     /* new expression: boolean or clock */
extern int reset_exp_tails(void);	 	/* reset expression tails */

extern int free_clkconj(struct clkconj_type **);/* free clock conjunct */
extern int free_boolexp(struct boolexp_type *); /* free a cycle of boolexp */
extern int free_one_boolexp(struct boolexp_type *); /* free a single boolexp */
extern int free_boolconj(struct boolconj_type *); /* free a single boolconj */


extern int install_const(void);             /* install expression terms */
extern int install_term(char *, char *);/*         and operators    */
extern int install_param(void);             /* install a parameter */
extern int install_relop(int);

extern int transform_clkconj(void);       	/* transform clkconj */
extern int reverse_op(int);           	/* reverse operator */

extern int compatible(int, int, int, int, int*, int*);

extern struct clkconj_type *clkconj_and(struct clkconj_type *, 
                                        struct clkconj_type *);

/***** mode and transition processing *****/

extern int new_mode(char *);             /* new mode */
extern int install_inv(void);                /* install mode invariant */
extern int new_transition(void);             /* new transition */
extern int install_trigger(void);           /* install transition trigger */
extern int assign_goto_mode(char *);    /* install target mode of transition */
extern int new_assign(void);
extern struct assign_type *alloc_assign(void);
extern int install_assign_value(void);
extern int install_assign_param(void);
extern int install_assign_var(char *, char *);
extern int install_assignment(void);
extern int install_out_trans(void);      /*install out_trans_tail to mode_tail*/
extern int install_mode(void);           /*install mode_tail[i] to automaton*/
extern struct transition_type *copy_transition(struct transition_type *);
                                         /* copy transition */
extern struct mode_type *source_mode(struct transition_type *);


/***** BDD printing ******/
extern char *naming_fn(bdd_manager, bdd, pointer);     
                                         /* naming function for bdds */
extern int print_bdd(bdd);              /* print a bdd */
extern int print_prop_bdds(void);           /* print proposition BDDs */



/***** utility functions ******/
extern char *str_preserve(char *);       /* make a copy of string */
extern char *itoa(int);                  /* integer to string transformation */
extern int pos(char *, char **, int);    /* position of string in array */
extern char *relop_string(int);		 /* return relation operator string */
extern void yywarningmessage(char *);    /* yacc warning message */

extern int get_max_const(void);		/* get the maximum constant */

extern char *bool_string(int);		/* get the string for a boolean 
                                           connective */
extern struct spec_type *set_cspec(void);   /* set the cspec pointer */
extern char *automaton_fullname(struct automaton *);
extern struct thread_type *thread_ptr(char *, int);
extern struct thread_type *graph_ptr(char *, int);

extern struct action_type *alloc_action(void);

extern struct thread_type *get_graph(char *);

extern int remove_duplicate_trans();
extern int init_action_management();
extern ERROR action_update(struct action_type *);
extern int delete_graph(struct thread_type *);
extern int new_graph(void);
extern int sequential_merge(void);
extern int free_action(struct action_type *a);
extern int ds_init(void);
extern int clkconj_compare(struct clkconj_type *, struct clkconj_type *);
extern int install_in_trans(void);
extern int install_trans_mode(void);
extern int free_tails(void);
extern int warn_constant(void);
extern int is_param(char *, struct automaton *);
extern int warn_param(void);
extern int install_assign_op(int);
extern int init_specification(void);
extern int new_spec(int);
extern int init_spec_atom(void);
extern int new_mode_spec(void);
extern int install_exp(void);
extern int install_auto_name(int);
extern int init_action_list(void);
extern int init_graph_list(void);
extern int new_action(void);
extern int install_graph(int);
extern void program_init(void);

extern int free_assign_cycle(struct assign_type **);

extern struct timeval        system_start_time;
extern struct timeval        system_end_time;
extern struct timezone       time_zone_p;
extern struct rusage         usage;

extern int keep_graphs;
extern MODE SGM_MODE;
extern int stats;
extern int internal_trans_bypass;
#endif
