#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include "include.h"
#include "global.h"
#include "print.h"
#include "action.h"
#include "help.h"

char **answer = NULL;
const char *nextchar = (char *)NULL;
char word[MAX_CHAR];
ERROR GUI_ERROR = E_NONE;

int gui_mode(void);
int reset_answer(void);
int free_strarray(char ***);
char **RequestFromGUI(const char *, int);
TOKEN strtoken(const char *);
char strchar(void);
char *getallchar(void);
int getallstrspace(void);
int make_answer(int);
char *error_str(ERROR);
int append_string(char *, char ***);
int print_answer(void);
int empty_graph(struct thread_type *);


/******************************************************/

int gui_mode() {
  int param=0;  
  char request[MAX_CHAR];
  char *w = request;
  int c;

  while(TRUE) {
    print_prompt();
    while((c = getchar())!='\n') {
      *w++ = c;
    }
    *w = '\0';
/*
    printf("Got command line: %s.\n", request);
*/
    RequestFromGUI(request, param);
    w = request;
    reset_str(request, MAX_CHAR);
  }
  return 0;
}

int reset_answer() {
  free_strarray(&answer);
  answer = (char **)malloc(sizeof(char *));
  answer[0] = (char *)NULL;
  return 0;
}

int free_strarray(char ***sa) {
  int i=0;
  char *s;

  if((*sa) == NULL) return 0;
  for(s=(*sa)[i]; s; s=(*sa)[++i]) {
    free(s);
    s = NULL;
  }
  free(*sa);
  *sa = NULL;
  return 0;
}

char **RequestFromGUI(const char *request, int param) {
  TOKEN tok;
  char *name = (char *)NULL, **arg = (char **)NULL;
  int cmd = -2, i=0, j, start=-1, end=-1;
  struct thread_type *g;

  reset_answer();
/*
  printf("Got request from GUI:\n%s\nEND of request.\n", request);
*/
  if(strcmp(request, "last_error")) GUI_ERROR = E_NONE;
  tok = strtoken(request);
  switch(tok) {
      case T_EOF:    break;
      case T_SHELL:  system(getallchar());
                     break;
      case T_ERROR:  nextchar = NULL;
                     fprintf(stderr, "Error: command line contains error.\n");
                     break;
      case T_WORD:
        switch(cmd = cmd_str_to_int(word)) {
	  case -2:   break;
          case -1:   fprintf(stderr, "Error: \"%s\" command unknown.\n", word);
                     if(word[0] != 0)
                       nextchar = NULL;
                     GUI_ERROR = E_CMD;
                     break;
          case QP:   sgm_quit();
	  case LE:   
		     break;
          case VL:   strtoken(NULL);
                     if(word[0] != 0)
                       nextchar = NULL;
                     if(!strcmp(word, "++")) verbosity_level(ADD);
                     else if(!strcmp(word, "--")) verbosity_level(SUB);
                     else verbosity_level(NOOP);
                     break;
          case TRW:  lazy_bdd_merging = lazy_bdd_merging ? FALSE : TRUE;
                     break;
          case TS:   strict_semantics = strict_semantics ? FALSE : TRUE;
                     break;
          case CD:   getallstrspace();
                     getallchar(); 
                     if(!strcmp(word, "")) system("pwd");
                     else if(chdir(word) == -1) perror("chdir");
                     else system("pwd");
                     break;
          case LS:   system("ls -aF");
                     nextchar = NULL;
                     break;
          case PWD:  system("pwd");
                     nextchar = NULL;
                     break;
          case VI:   system("vi");
                     nextchar = NULL;
                     break;
          case EMACS:system("emacs");
                     nextchar = NULL;
                     break;
          case HELP: if(strtoken(NULL) == T_SHELL) {
                       help_shell();
                       nextchar = NULL;
                     }
                     else {
                       if(word[0] != 0)
                         nextchar = NULL;
                       help_info(word);
                     }
                     break;
          case GD:   reset_answer();
                     while(strtoken(NULL) != T_EOF) {
                       if(get_gd_info(word, &name, &start, &end)) {
                         interval_varname = str_preserve(name);
                         interval_start = start;
                         interval_end = end;
                         new_graph();
                         free(interval_varname);
                         interval_varname = NULL;
                         interval_start = interval_end = -1;
                       }
                     }
/*
                     list_graphs();
*/
                     break;
	  case PROG_INIT: program_init();
                     break;
	  case SETENV: getallstrspace();
                       getallchar(); 
		       putenv(str_preserve(word));
                       printf("SGM_DIR = %s\n", getenv("SGM_DIR"));
		       break;  
          case LOAD: strtoken(NULL);
                     if(word[0] != 0)
                       nextchar = NULL;
                     GUI_ERROR = load_file(word);
                     break;
          case SM:   gettime(START); sequential_merge(); gettime(END);
                     free(print_time(NEW_THREAD));
                       nextchar = NULL;
                     break;
          case PS:   strtoken(NULL);
                     reset_answer();
                     if(word[0] != 0) {
                       g = get_graph(word);
                       if(g) append_string(print_size(g), &answer);
                       else GUI_ERROR = E_UNDECL;
                     }
                     else {
                       GUI_ERROR = E_CMD;
                       break;
                     }
                     nextchar = NULL;
                     break;
          case KG:   keep_graphs = keep_graphs ? FALSE : TRUE;
                     break;
          case IG:   strtoken(NULL);
                     reset_answer();
                     if(word[0] != 0) info_graph(get_graph(word));
                     else {
                       info_graph(NULL);
                       break;
                     }
                     nextchar = NULL;
                     break;
          case SI:   strtoken(NULL);
                     reset_answer();
                     if(word[0] != 0) short_info(get_graph(word), 0);
                     nextchar = NULL;
                     break;
	  case EG:   strtoken(NULL);
                     reset_answer();
                     if(word[0] != 0) {
                       if(empty_graph(get_graph(word))) GUI_ERROR = E_EMPTY;
                       else GUI_ERROR = E_NONE;
                     }
                     else {
                       GUI_ERROR = E_INSUFF;
                       break;
                     }
                     break;
          case VG:   strtoken(NULL);
                     reset_answer();
                     if(word[0] != 0) {
                       view_graph(get_graph(word));
                     }
                     else {
                       GUI_ERROR = E_INSUFF;
                     }
                     break;
	  case XG:   strtoken(NULL);
		     if(word[0] != 0) {
			g = get_graph(word);
			if(g) {
	  		  strtoken(NULL);
			  if(word[0] != 0) {
			    rename_graph(g, word);
                          }
			  else GUI_ERROR = E_INSUFF;
                        }
                        else GUI_ERROR = E_UNDECL;
		     }
                     else {
			GUI_ERROR = E_INSUFF;
		     }
		     reset_answer();
                     break;
          default:   while(strtoken(NULL) != T_EOF) {
		       if(arg)
                         arg = (char **)realloc(arg, sizeof(char *)*(i+1));
		       else 
                         arg = (char **)malloc(sizeof(char *)*(i+1));
                       arg[i++] = str_preserve(word);
                       /*printf("arg[%d] = %s\n", i-1, arg[i-1]);*/
                     }
		     gettime(START);
                     GUI_ERROR = exec_cmd(cmd, arg, i);
                     if(GUI_ERROR == E_NONE)
		       gettime(END);
                     for(j=0;j<i;j++) free(arg[j]); 
                     free(arg); arg = NULL;
                     i = 0;
        }
      break;
      default: break;
  }
  reset_str(word, MAX_CHAR);
  make_answer(cmd);
  /* print_answer(); */
  return answer;
}

/* Get a token from the GUI request */

TOKEN strtoken(const char *request) {
  enum {
    NEUTRAL,
    INWORD
  } state;
  char *w;
  char c;

  reset_str(word, MAX_CHAR);
  w = word;
  state = NEUTRAL;
  if(request)
    nextchar = request;
  while((c = strchar()) != '\0') {
    switch(state) {
      case NEUTRAL:
        switch(c) {
          case ' ':
          case '\t':
          case ',':
          case '(':
          case ')':
          case ';':
            continue;
          case '!':
            return(T_SHELL);
          default:
            if(is_invalid_char(c))
              return(T_ERROR);
            state = INWORD;
            *w++ = c;
            continue;
        }
      case INWORD:
        switch(c) {
          case '(':
          case ')':
          case ' ':
          case '\t':
          case ',':
          case ';':
/*          ungetc(c, stdin); */
            *w = '\0';
            return(T_WORD);
          default:
            if(is_invalid_char(c))
              return(T_ERROR);
            *w++ = c;
            continue;
        }
    }
  }
  if(state == INWORD) {
/* ungetc(c, stdin); */
    *w = '\0';
    return(T_WORD);
  }
  return(T_EOF);
}

char strchar() {
  char *n = NULL;

  if(nextchar && strcmp(nextchar,"")) {
    n = (char *)nextchar;
    nextchar++;
    return n[0];
  }
  return '\0';
}

char *getallchar() {
  return strcpy(word, nextchar);
}

int getallstrspace() {
  while(nextchar[0] == ' ') {
    nextchar++;
  }
  return 0;
}

int make_answer(int cmd) {
  int i;
  char *ans = NULL;

  switch(cmd) {
    case -1:   reset_answer();
/*
               ans = error_str(GUI_ERROR);
               append_string(ans, &answer);
*/
               break;
    case QP:
    case CD:
    case TS:
    case VL:   reset_answer();
               break;
    case LS:   
    case PWD:  
    case VI:   
    case EMACS:
    case HELP: 
    case GD:   
               break;
    case LOAD: reset_answer();
      switch(GUI_ERROR) {
        case E_NONE: 
                     for(i = 0; i < 2*system_ta_num-1; i++) {
/*
                       sprintf(tmpstring, "%s[%d]\n%s",
                          thread[i].name, thread[i].automaton_index, 
                          print_size(&thread[i]));
                       append_string(tmpstring, &answer);
*/
                       short_info(&thread[i], 1);
                     }
/*
                   for(i=0, g=&thread[0]; i<2*system_ta_num-1; g=&thread[++i]) {
                       append_string(print_size(g), &answer);
                   }
*/
                   break;
        case E_INSUFF:
        case E_LOADED:
        case E_LOGFILEFAIL:
        case E_INPUTFILEFAIL:
			reset_answer();
/*
                     ans = error_str(GUI_ERROR);
                     append_string(ans, &answer);
*/
                     break;
        default: break;
      }
      break;
    case SM:   reset_answer();
               break;
    case LE:   reset_answer();
               ans = error_str(GUI_ERROR);
               if(ans) {
                 append_string(ans, &answer);
               }
               break;
    case PS:   if(answer[0]) break;
/*
               ans = error_str(GUI_ERROR);
               if(ans) {
                 append_string(ans, &answer);
               }
*/
               break;
    case IG:   if(answer[0]) break;
               break;
    case SI:   if(answer[0]) break;
               break;
    default:   reset_answer();
/*
               ans = error_str(GUI_ERROR);
               if(ans) {
                 append_string(ans, &answer);
               }
               else 
*/
               if(GUI_ERROR != E_NOANSWER && GUI_ERROR != E_NONE && 
                  GUI_ERROR != E_NOTIME) info_graph(NEW_THREAD);
               break;
  }
  return 0;
}

char *error_str(ERROR e) {
  switch(e) {
    case E_NONE: 
    case E_NOANSWER: 
    case E_NOTIME:
      return NULL;
      break;
    case E_INSUFF: 
      sprintf(tmpstring, "Error: insufficient arguments.\n");
      break;
    case E_INVNAME:
      sprintf(tmpstring, "Error: invalid new graph name.\n");
      break;
    case E_LOADED: 
      sprintf(tmpstring, "Error: only one file can be loaded per session.\n");
      break;
    case E_INPUTFILEFAIL:
      sprintf(tmpstring, 
              "Error: input file read failed (check permission/existence).\n");
      break;
    case E_LOGFILEFAIL:
      sprintf(tmpstring, "Error: log file creation failed.\n");
      break;
    case E_CMD:
      sprintf(tmpstring, "Error: command line contains error.\n");
      break;
    case E_INCOMPAT:
      sprintf(tmpstring, "Error: incompatible graphs for merging.\n");
      break;
    case E_EMPTY:
      sprintf(tmpstring, "Error: empty graph(s).\n");
      break;
    case E_UNDECL:
      sprintf(tmpstring, "Error: argument graph(s) undeclared.\n");
      break;
    case E_DECL:
      sprintf(tmpstring, "Error: argument graph(s) already declared.\n");
      break;
    case E_RESET_FAIL:
      sprintf(tmpstring, "Error: a system graph cannot be reset or freed.\n");
      break;
    case E_PARSE_FAIL:
      sprintf(tmpstring, "Error: syntax error in SGM Input File %s.\n"
                         "Check standard console for errors.\n", infilename);
      free(infilename);
      break;
    case E_NAME:
      sprintf(tmpstring, "Error: illegal name.\n");
      break;
    default:
      sprintf(tmpstring, "Error: unknown error code %d.\n", e);
      break;
  }
  return tmpstring;
}

/* Append string "st" to string-array "a"
 * Note: "a" is an array of strings with the last string being a NULL string 
 */

int append_string(char *st, char ***a) {
  int i = 0;

  if((*a) == NULL) reset_answer();
  else for(; (*a)[i]; ++i) ;

  (*a)[i] = str_preserve(st);
  (*a) = (char **)realloc((*a), sizeof(char *)*(i+2));
  (*a)[i+1] = (char *)NULL;
  return 0;
}

int print_answer() {
  int i = 0;
  char *s;

  if(answer == NULL) {
    fprintf(stderr, "ERROR: No answer from SGM for GUI!\n");
    return 0;
  }
  fprintf(stderr, "-------TO GUI-----------\n");
  for(s=answer[0]; s; s=answer[++i]) {
    fprintf(stderr, "ANSWER[%d]:\n", i);
    fprintf(stderr, "%s", s);
  }
  fprintf(stderr, "------------------------\n");
  return 0;
}

int empty_graph(struct thread_type *t) {
  if(t == NULL) return TRUE;
  if(t->num_ta == 0) return TRUE;
  else return FALSE;
}
