#include "include.h"
#include "global.h"
#include "print.h"
#include "action.h"
#include "gui.h"
#include "help.h"
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>

int read_file(char *);
int help_printf(char *) ;
int gui_help(char *);
int help_info(char *);
int help_short();
int help_main();
int help_cmd();
int help_l();
int help_gd();
int help_mg();
int help_v();
int help_sm();
int help_sc();
int help_nr();
int help_rw();
int help_pg();
int help_ps();
int help_ig();
int help_rg();
int help_fg();
int help_mc();
int help_q();
int help_shell();
int help_error(char *);

int help_printf(char *s) {
  if(SGM_MODE == GUI || SGM_MODE == PROCESS) append_string(s, &answer);
  else printf("%s", s);
  return 0;
}

int help_info(char *s) {
  int cmd;
  if(SGM_MODE == GUI || SGM_MODE == PROCESS)
    reset_answer();
  if(!strcmp(s, "")) {
    if(short_help) help_short();
    else help_main();
    if(short_help) short_help = FALSE;
    return 0;
  }
  if(!strcmp(s, "commands") || !strcmp(s, "command") || !strcmp(s, "c")) {
    help_cmd();
    if(short_help) short_help = FALSE;
    return 0;
  }
  cmd = cmd_str_to_int(s);
  switch(cmd) {
    case LOAD: if(SGM_MODE == INTERACTIVE) read_file("load.txt");
	       else read_file("load1.txt");
               break;
    case GD:   if(SGM_MODE == INTERACTIVE) read_file("gd.txt");
	       else read_file("gd1.txt");
               break;

    case MG:   if(SGM_MODE == INTERACTIVE) read_file("mg.txt");
	       else read_file("mg1.txt");
               break;
    case SM:   if(SGM_MODE == INTERACTIVE) read_file("sm.txt");
	       else read_file("sm1.txt");
               break;

    case RW:   if(SGM_MODE == INTERACTIVE) read_file("rw.txt");
	       else read_file("rw.txt");
               break;
    case SC:   if(SGM_MODE == INTERACTIVE) read_file("sc.txt");
	       else read_file("sc1.txt");
               break;
    case NR:   if(SGM_MODE == INTERACTIVE) read_file("nr.txt");
	       else read_file("nr1.txt");
               break;

    case MC:   if(SGM_MODE == INTERACTIVE) read_file("mc.txt");
	       else read_file("mc1.txt");
               break;
    case VL:   if(SGM_MODE == INTERACTIVE) read_file("vl.txt");
	       else read_file("vl1.txt");
               break;
    case IG:   help_ig();
               break;
    case PG:   help_pg();
               break;
    case PS:   help_ps();
               break;
    case RG:   help_rg();
               break;
    case FG:   help_fg();
               break;
    case CD:   system("man chdir");
               break;
    case PWD:  system("man pwd");
               break;
    case LS:   system("man ls");
               break;
    case VI:   system("man vi");
               break;
    case EMACS:system("man emacs");
               break;
    case QP:   help_q();
               break;
    case HELP: help_cmd();
               break;
    default:   switch(gui_help(s)) {
		case INTRODUCTION: read_file("introduction.txt");
		                   break;
		case FILE_NEW_EDIT: read_file("file_new_edit.txt");
		                   break;
		case FILE_LOAD_UNLOAD: read_file("file_load_unload.txt");
		                   break;
		case GRAPH_OPEN_CLOSE: read_file("graph_open_close.txt");
		                   break;
		case GRAPH_VIEW: read_file("graph_view.txt");
		                   break;
		case GRAPH_RESET_FREE: read_file("graph_reset_free.txt");
		                   break;
		case GRAPH_RENAME: read_file("graph_rename.txt");
		                   break;
		case GRAPH_PRINT_SAVE: read_file("graph_print_save.txt");
		                   break;
		case OPTIONS_SEMANTICS: read_file("options_semantics.txt");
		                   break;
		case SPECIFY_TCTL: read_file("specify_tctl.txt");
		                   break;
		default: help_error(s);
               }
  }
  if(short_help) short_help = FALSE;
  return 0;
}

int gui_help(char *s) {
  if(!strcmp(s, "introduction")) return INTRODUCTION;
  if(!strcmp(s, "file_new_edit")) return FILE_NEW_EDIT;
  if(!strcmp(s, "file_load_unload")) return FILE_LOAD_UNLOAD;
  if(!strcmp(s, "graph_open_close")) return GRAPH_OPEN_CLOSE;
  if(!strcmp(s, "graph_view")) return GRAPH_VIEW;
  if(!strcmp(s, "graph_reset_free")) return GRAPH_RESET_FREE;
  if(!strcmp(s, "graph_rename")) return GRAPH_RENAME;
  if(!strcmp(s, "graph_print_save")) return GRAPH_PRINT_SAVE;
  if(!strcmp(s, "options_semantics")) return OPTIONS_SEMANTICS;
  if(!strcmp(s, "specify_tctl")) return SPECIFY_TCTL;
  return -1;
}

int read_file(char *c) {
  FILE *f;
  char *s;
  
  if( !(s = getenv("SGM_DIR")) ) {
    fprintf(stderr, "Warning: environment variable SGM_DIR not set,"
                    " awkward behavior may result.\nPlease set it to"
                    " the SGM directory.\n");
  }
  if(s) sprintf(tmpstring, "%s/help", s);
  else sprintf(tmpstring, "help");
  if( chdir(tmpstring) < 0) {
    fprintf(stderr, "Error: Help directory (%s) not found.\n", tmpstring); 
  }
  else if( !(f = fopen(c, "r")) ) {
    fprintf(stderr, "Error: Help file (%s) not found.\n", c); 
  }
  else {
    while(fgets(tmpstring, MAX_CHAR, f)) {
      help_printf(tmpstring);
    }
    fclose(f);
    return 0;
  }
  return 1;
}

/* SHORT HELP */
int help_short() {
  help_printf("Graph: l, g, sm, mg, rw, sc, nr, mc, pg, ps, ig, rg, fg\n");
  help_printf("Misc : v, q, e\n");
  help_printf("Shell: !, cd, pwd, ls, vi, emacs\n");   
  return 0;
}

/* MAIN HELP */
int help_main() {
  printline();
  help_printf("  SGM commands are classified into 3 classes:\n\n");
  help_printf("  (A) STATE-GRAPH COMMANDS  :\n");
  help_printf("       l  (load),             g (graph),       sm (sequential_merge),\n");
  help_printf("      mg (merge_graph),      rw (read_write),  sc (shield_clock),\n");
  help_printf("      nr (normalize_region), mc (model_check), pg (print_graph),\n");
  help_printf("      ps (print_size),       ig (info_graph),  rg (reset_graph),\n");
  help_printf("      fg (free_graph).\n\n");
  help_printf("  (B) MISCELLANEOUS COMMANDS:\n");
  help_printf("      v (verbose), q (quit), e (exit).\n\n");
  help_printf("  (C) SHELL COMMANDS        :\n");
  help_printf("      !, cd (chdir), pwd, ls, vi, emacs.\n\n");

  help_printf("  For further help, PLEASE TYPE:\n");
  help_printf("    * help commands (or h c) => detailed list of all commands\n");
  help_printf("    * ?                      => list of commands in short forms\n");
  help_printf("    * help <cmd> (or h <cmd>)=> help on <cmd>, any known command\n");
  printline();
  return 0;
}

/* DETAILED LIST OF ALL COMMANDS */
int help_cmd() {
  printline();
  help_printf("     SGM commands\t\tShort Forms\tManipulators/Explanations\n");
  help_printf("     GRAPH COMMANDS                 \n");
  help_printf(" (0) load filename\t\t    l ...\tLoad specification file\n");
  help_printf(" (1) graph g[0-5], h[4];\t    g ...\tGraph declarations\n");
  help_printf("     graph;\t\t\t    g\t\tList all graphs\n");
  help_printf(" (2) sequential_merge\t\t    sm ...\tSequentially Merge all Graphs\n");
  help_printf(" (3) merge_graph g[1], g[2], g[3]   mg ...\tMerge g[1], g[2] into g[3]\n");
  help_printf(" (4) read_write g[1]\t\t    rw ...\tRead-Write Reduction\n");
  help_printf(" (5) shield_clock g[1]\t\t    sc ...\tShield-Clock Reduction\n");
  help_printf(" (6) normalize_region g[1]\t    nr ...\tSymmetry-Based Reduction\n");
  help_printf(" (7) model_check g[1]\t\t    mc ...\tModel-Check\n");
  help_printf(" (8) print_graph g[1]\t\t    pg ...\tPrint Complete State-Graph\n");
  help_printf(" (9) print_size g[1]\t\t    ps ...\tPrint Size of State-Graph\n");
  help_printf("(10) info_graph g[1]\t\t    ig ...\tInformation on State-Graph\n");
  help_printf("(11) reset_graph g[1]\t\t    rg ...\tReset State-Graph\n");
  help_printf("(12) free_graph g[1]\t\t    fg ...\tFree State-Graph\n");

  help_printf("     MISCELLANEOUS COMMANDS                 \n");
  help_printf("(13) verbose\t\t\t    v\t\tShow Log File Verbosity Level\n");
  help_printf("     verbose ++\t\t\t    v ++\tIncrement Verbosity Level\n");
  help_printf("     verbose --\t\t\t    v --\tDecrement Verbosity Level\n");
  help_printf("(14) quit, quit(), exit, exit()     q      \tExit SGM\n");

  help_printf("     SHELL COMMANDS \n");
  help_printf("(15) chdir somedir\t\t    cd ...\tChange working directory\n");
  help_printf("(16) pwd\t\t\t    pwd\t\tPresent working directory\n");
  help_printf("(17) ls\t\t\t\t    ls\t\tList directory files\n");
  printline();
  return 0;
}

/* LOAD */
int help_l() {
  printline();
  help_printf("  HELP on  : 'load' or l':\n");
  help_printf("  Syntax   : load <filename>, where <filename> is the name of\n");
  help_printf("                              an input file with a '.s' extension\n");
  help_printf("  Semantics: This command basically loads the system description.\n");
  help_printf("             The description must be written in a file with a '.s'\n");
  help_printf("             extension. The input language grammar is given in the\n");
  help_printf("             SGM User Manual along with examples.\n");
  help_printf("  Examples : l example.s\n");
  printline();
  return 0;
}

int help_gd() {
  printline();
  help_printf("  HELP on  : graph or g:\n");
  help_printf("  Syntax   : graph {name, ...}, where name is a new graph name\n");
  help_printf("  Semantics: This command executed without any parameters will \n");
  help_printf("             display all existing graphs. Any list of comma-\n");
  help_printf("             separated parameters will be taken as new graph\n");
  help_printf("             declarations.\n");
  help_printf("  Examples : graph              -- all graphs will be displayed\n");
  help_printf("             g A[1] Sender[3]   -- two new graphs will be created\n");
  printline();
  return 0;
}

/* MERGE_GRAPH */
int help_mg() {
  printline();
  help_printf("  HELP on  : 'merge_graph' or 'mg':\n");
  help_printf("  Syntax   : merge_graph <src_graph1> <src_graph2> <dest_graph>\n");
  help_printf("             where <src_graph1>, <src_graph2>, and <dest_graph>\n");
  help_printf("             are all existing state-graphs.\n");
  help_printf("  Semantics: This command merges two state-graphs <src_graph1>,\n");
  help_printf("             <src_graph2> into a merged state-graph <dest_graph>.\n");
  help_printf("  Examples : mg A[1] A[2] h[3]\n");
  printline();
  return 0;
}

/* VERBOSE */
int help_v() {
  printline();
  help_printf("  HELP on  : 'verbose' or 'v':\n");
  help_printf("  Syntax   : verbose | verbose ++ | verbose --\n");
  help_printf("  Semantics: This command changes the verbosity level of the log file\n");
  help_printf("             Default level is '-2'. 'verbose' will show current level.\n");
  help_printf("             'verbose ++' will increment the level to a maximum of 3.\n");
  help_printf("             'verbose --' will decrement the level to a minimum of -2.\n");
  help_printf("  Examples : verbose\n");
  help_printf("             v ++\n");
  printline();
  return 0;
}

int help_sm() {
  printline();
  help_printf("  HELP on  : 'sequential_merge' or 'sm':\n");
  help_printf("  Syntax   : sequential_merge\n");
  help_printf("  Semantics: This command merges all the declared system process\n");
  help_printf("             graphs into a global graph and performs all actions\n");
  help_printf("             declared in the loaded input file.\n");
  help_printf("  Examples : sequential_merge\n");
  printline();
  return 0;
}

int help_sc() {
  printline();
  help_printf("  HELP on  : 'shield_clock' or 'sc':\n");
  help_printf("  Syntax   : shield_clock <g>, where <g> some graph name.\n");
  help_printf("  Semantics: This command reduces size of the graph <g> by \n");
  help_printf("             trying to shield up all clocks which are not used\n");
  help_printf("             in the current mode.\n");
  help_printf("  Examples : shield_clock g[1]\n");
  printline();
  return 0;
}

int help_nr() {
  printline();
  help_printf("  HELP on  : 'normalize_region' or 'nr':\n");
  help_printf("  Syntax   : normalize_region <g>, where <g> some graph name.\n");
  help_printf("  Semantics: This command reduces size of the graph <g> by \n");
  help_printf("             trying to normalize all regions through symmetry-based\n");
  help_printf("             reduction.\n");
  help_printf("  Examples : normalize_region g[1]\n");
  printline();
  return 0;
}

int help_rw() {
  printline();
  help_printf("  HELP on  : 'read_write' or 'rw':\n");
  help_printf("  Syntax   : read_write <g>, where <g> some graph name.\n");
  help_printf("  Semantics: This command reduces size of the graph <g> by \n");
  help_printf("             trying to eliminate impossible computations through\n");
  help_printf("             a static analysis of the reads/writes on transitions\n");
  help_printf("             reduction.\n");
  help_printf("  Examples : read_write g[1]\n");
  printline();
  return 0;
}

int help_pg() {
  printline();
  help_printf("  HELP on  : 'print_graph' or 'pg':\n");
  help_printf("  Syntax   : print_graph <g>, where <g> is a graph name.\n");
  help_printf("  Semantics: The details of the graph is printed into the log file.\n");
  help_printf("  Examples : print_graph g[1]\n");
  printline();
  return 0;
}

int help_ps() {
  printline();
  help_printf("  HELP on  : 'print_size' or 'ps':\n");
  help_printf("  Syntax   : print_size <g>, where <g> is a graph name.\n");
  help_printf("  Semantics: The size of the graph in terms of the number of regions\n");
  help_printf("             and the number of transitions is displayed.\n");
  help_printf("  Examples : print_size g[1]\n");
  printline();
  return 0;
}

int help_ig() {
  printline();
  help_printf("  HELP on  : 'info_graph' or 'ig':\n");
  help_printf("  Syntax   : info_graph | info_graph <g>, where <g> is a graph name.\n");
  help_printf("  Semantics: Information of a particular graph or all graphs can be obtained.\n");
  help_printf("             Information include: SIZE, PROCESSES, MANIPULATORS APPLIED.\n");
  help_printf("  Examples : info_graph g[1]\n");
  printline();
  return 0;
}

int help_rg() {
  printline();
  help_printf("  HELP on  : 'reset_graph' or 'rg':\n");
  help_printf("  Syntax   : reset_graph <g>, where <g> is a graph name\n");
  help_printf("  Semantics: All contents of the graph are freed and counters reset.\n");
  help_printf("             The graph can be reused as a new declaration.\n");
  help_printf("  Examples : reset_graph g[1]\n");
  printline();
  return 0;
}

int help_fg() {
  printline();
  help_printf("  HELP on  : 'free_graph' or 'fg':\n");
  help_printf("  Syntax   : free_graph <g>, where <g> is a graph name\n");
  help_printf("  Semantics: The graph is first reset and then freed.\n");
  help_printf("             A freed graph cannot be used again.\n");
  help_printf("  Examples : free_graph g[1]\n");
  printline();
  return 0;
}

int help_mc() {
  printline();
  help_printf("  HELP on  : 'model_check' or 'mc':\n");
  help_printf("  Syntax   : model_check <g>, where <g> is a graph name.\n");
  help_printf("  Semantics: The graph <g> is model-checked against the specification\n");
  help_printf("             input through the load file command.\n");
  help_printf("  Examples : model_check g[5]\n");
  printline();
  return 0;
}

int help_q() {
  printline();
  help_printf("  HELP on  : 'quit' or 'exit' or 'q' or 'e':\n");
  help_printf("  Syntax   : quit\n");
  help_printf("  Semantics: SGM program quits immediately.\n");
  help_printf("  Examples : quit\n");
  printline();
  return 0;
}

int help_shell() {
  printline();
  help_printf("  HELP on  : '!':\n");
  help_printf("  Syntax   : !<shell_cmd>, where <shell_cmd> is any shell command.\n");
  help_printf("  Semantics: All shell (external) commands such as ls, cd, ... can be\n"); 
  help_printf("             executed directly through the shell escape character '!'.\n");
  help_printf("  Examples : !ls -al\n");
  printline();
  return 0;
}

int help_error(char *s) {
  if(SGM_MODE == GUI || SGM_MODE == PROCESS) {
    GUI_ERROR = E_CMD;
    sprintf(tmpstring, "Help Error: '%s' no such command.\n", s);
    append_string(tmpstring, &answer);
  }
  else fprintf(stderr, "Help Error: '%s' no such command.\n", s);
  return 0;
}
