#include "include.h"
#include "modeman.h"
#include "bddman.h"
#include "symmetry.h"
#include "global.h"
#include "action.h"
#include "gui.h"
#include "print.h"
#include "transman.h"

/***********************************************************
 * EXTERNAL VARIABLES 
 ***********************************************************/

extern int yylineno;
#ifdef __LINUX
extern char *yytext;
#else
extern char yytext[];
#endif
extern int yyleng;
extern FILE *logfile;
extern char *tmpstring;

/***********************************************************
 * GLOBAL VARIABLES   
 ***********************************************************/


int *automata_num = NULL;      /* automata counts */
char **automata_name = NULL;   /* automata names */
char *parameter = NULL;        /* current parametric index */
struct automaton *current_ta;  /* points to the currently parsed TA */ 
int current_ta_num;	       /* number of currently parsed TA */
int system_ta_num;	       /* total number of system TA */
int decl_graph_num;	       /* total number of graphs declared  
                                  (including system TA graphs) */

struct thread_type *thread;    /* a thread, a process, a graph are all same */

int total_exp;        	   /* total number of expressions */
int automaton_count;       /* total number of automata */

int var_count;             /* total number of variables */
int tmp_var_count;         /* number of variables of the same type */
int var_type;              /* current variable type */
int interval_start = -1;   /* variable index interval starting index */
int interval_end = -1;;    /* variable index interval ending index */
char *interval_varname;    /* variable name of an interval */

int index_relop = -1;      /* indicating index range: <i, >i, !=i */
int index_quantifier = -1; /* quantifier of index range: FORALL, EXISTS */

int curr_exp_type;         /* current expression type: CLK_INEQ or BOOL_EXP */
/*
int boolconj_count; */     /* current count of boolean conjuncts */
/*
int boolexp_count; */      /* current count of boolean expressions */
/*
int clkconj_count; */        /* current count of clock conjuncts */

int inside_exp;            /* whether still parsing inside a conjunct */
int neg_conj;              /* whether a '~' conjunct */
int sign;                  /* LHS = +1, RHS = -1 */
int need_transform;        /* if the clkconj needs transforming ... */
                           /* this occurs when: (a) x ~ y - d */
                           /*                   (b) x + d ~ y */
                           /* i.e., sign==1 during install_const() */

int prop_count;		   /* total number of propositions */
int mode_count;            /* total number of modes */
int trans_count;           /* currently parsed transition count */
int assign_count;          /* currently parsed assignment count */

int last_int;              /* last parsed integer */
char *string_buffer;       /* last parsed string */

int max_const;             /* maximum constant in specification */

int sync_count;             /* total number of synchronization labels */

struct variable_type *var_tail = (struct variable_type *)NULL;
struct proposition_type *prop_tail = (struct proposition_type *)NULL;
struct automaton *automaton_tail = (struct automaton *)NULL; 

struct variable_type *sync_tail = (struct variable_type *)NULL;
struct var_link_type **sync_cycle = (struct var_link_type **)NULL;
struct var_link_type **tr_sync_cycle = (struct var_link_type **)NULL;

struct mode_type **mode_tail; 
struct spec_trans_type **out_trans_tail;
struct boolconj_type **boolconj_tail;
struct boolexp_type **boolexp_tail;
struct clkconj_type **clkconj_tail;
struct assign_type **assign_tail;
struct spec_type *vspec, *cspec;
struct action_type *action_list, *curr_action, *tmp_action;
struct thread_type *graph_list;

struct boolexp_type **boolexp_expand_start;
struct boolconj_type **boolconj_expand_start;
struct clkconj_type **clkconj_expand_start;

int neg_spec = FALSE;   /* whether a '~' tctl_spec */
int blevel = 0;		/* current bracket level for nested brackets*/
int p_num = 0;		/* parametric index increment/decrement */

/***********************************************************
 * FUNCTION DECLARATIONS 
 ***********************************************************/

/***** variable processing *****/
struct variable_type *var_ptr(char *, int); 
                                 /* get pointer to variable from name & id */
struct variable_type *var_ptr_from_index(int);
                                 /* get pointer to variable from index */
int var_index(char *, int);      /* get index of variable from name & id */
char *var_name(int);             /* get name of variable from index */
int var_id(int);		 /* get id of variable from index */
int get_var_type(int);		 /* get variable type from index */
int insert_var(char *);          /* insert a parsed variable */
int add_attribute(int);          /* add variable attribute */

int init_tails(void);		 /* Initial all data structure tail pointers */

/***** automaton processing *****/

int total_ta(char *);		/* get the total number of some ta from name */
int update_automata(void);	/* update automata names & counts */
int new_automata(void);         /* new automaton */
struct automaton *get_ta(char *n, int k); /* get k'th automaton with name n */
int check_param(char *);        /* check if parameter is current, exit o/w */
int is_param(char *, struct automaton *); /* check if it is the parameter of 
                                             that automaton */

int install_init_mode(char *);	/* install initial mode name */
int install_init_cond(void);	/* install init condition: boolexp and xconj*/

/***** expression processing *****/
struct boolexp_type *alloc_boolexp(void); 	/* allocate a new boolexp */
struct boolconj_type *alloc_boolconj(void); 	/* allocate a new boolconj */
int new_boolexp(void);               		/* new boolean expression */
int new_boolconj(int);           		/* new boolean conjunct */
int new_boolconj_array(void);        		/* new boolean conjunct array */

int new_clkconj(void);               		/* new clock conjunct */
struct clkconj_type *alloc_clkconj(void); 	/* allocate a new clkconj */
struct clkconj_type *clkconj_copy1(struct clkconj_type *c);
                                   /* copy a single clkconj */
struct clkconj_type *clkconj_copy(struct clkconj_type *);
                                   /* copy a clkconj cycle */
struct boolconj_type *boolconj_copy(struct boolconj_type *x);
struct boolconj_type *boolconj_copy1(struct boolconj_type *x);
struct assign_type *assign_copy(struct assign_type *);

int new_exp(char *, char *);       /* new expression: boolean or clock */
int reset_exp_tails(void);	   /* reset to NULL all expression tails */

int free_clkconj(struct clkconj_type **);
int free_boolexp(struct boolexp_type *);
int free_one_boolexp(struct boolexp_type *);
int free_boolconj(struct boolconj_type *);

int install_const(void);              /* install expression terms */
int install_term(char *, char *);     /*         and operators    */
int install_param(void);              /* install a parameter      */
int install_relop(int);

int transform_clkconj(void);	      /* transform clkconj */
int reverse_op(int);		      /* reverse operator */

int compatible(int, int, int, int, int*, int*);

struct clkconj_type *clkconj_and(struct clkconj_type *, struct clkconj_type *);
int clkconj_compare(struct clkconj_type *, struct clkconj_type *);
int clkconj_contained(struct clkconj_type *, struct clkconj_type *);
int clkconj_compare_one(struct clkconj_type *, struct clkconj_type *);

int expand_boolconj(char *, int);
int expand_boolexp(char *, int);
int delete_boolconj_tail(struct boolconj_type **, int);

/***** mode and transition processing *****/
int new_mode(char *);             /* new mode */
int install_inv(void);                /* install mode invariant */
int new_transition(void);            /* new transition */
int install_trigger(void);           /* install transition trigger */
int assign_goto_mode(char *);    /* install target mode of transition */
int new_assign(void);		  /* new assignment */
struct assign_type *alloc_assign();	/* alloc an assign_type struct */
int install_assign_value(void);	  /* install value in assignment */
int install_assign_var(char *, char *); /* install variable name in assignment */
int install_assignment(void);        /* install assign_tail to out_trans_tail */
int install_out_trans(void);         /* install out_trans_tail to mode_tail */
int install_in_trans(void);          /* install in_trans_cycle of all modes */

int install_mode(void);              /* install modes to automata */

struct transition_type *
copy_transition(struct transition_type *); /* copy transition */
int install_trans_mode(void);        /* install mode pointers of all transitions */

/***** utility functions ******/
char *str_preserve(char *);       /* make a copy of string */
char *itoa(int);                  /* integer to string transformation */
char *relop_string(int);	  /* return relation operator string from int */
void yywarningmessage(char *);    /* yacc warning message */

/***** specification functions *****/
int init_specification(void);
struct spec_type *alloc_spec(void);
int new_spec(int type);
int init_spec_atom(void);
struct spec_type *set_cspec(void);
int install_exp(void);
int new_mode_spec(void);
int install_auto_name(int n);
char *bool_string(int b);
char *automaton_fullname(struct automaton *a);

/***** manipulation functions *****/
int init_action_list(void);
int init_graph_list(void);
int new_graph(void);
struct thread_type *alloc_thread(void);
struct action_type *alloc_action(void);
int new_action(void);
int install_graph(int n);
struct thread_type *thread_ptr(char *, int);
struct thread_type *graph_ptr(char *, int);

int check_relop(int, int);
int init_action_management(void);
int list_update(struct action_type *);
ERROR action_update(struct action_type *);
struct thread_type *get_graph(char *);
int remove_duplicate_trans(void);

/* synchronization functions */
int init_sync();
int sync_index(char *, int);
struct variable_type *sync_ptr(char *, int);
int insert_sync(char *);
int install_sync();
int init_tr_sync();
int insert_trans_sync(char *);
int install_trans_sync();


/***********************************************************
 * FUNCTION DEFINITIONS 
 ***********************************************************/

/* Get the total number of Timed Automata with name "n" */

int total_ta(char *n) {
  int i;
  if(automata_num) {
    for(i=1; i<=automata_num[0]; i++) { 
      if(!strcmp(automata_name[i], n)) return automata_num[i];
    }
    fprintf(stderr, "Error: timed automata %s not declared.\n", n);
  }
  else 
    fprintf(stderr, "Error: No timed automaton declared.\n");
  return -1;
}


/* automata_num : [0] => number of different types of automata
 *                [i] => number of the ith type of automata
 * automata_name: [0] => NULL
 *                [i] => name of the ith type of automata
 */

int update_automata(void) {
  if(automata_num) {
    automata_num = (int *)realloc(automata_num, 
                                  sizeof(int)*(automata_num[0]+2));
    automata_num[0]++;
    automata_num[automata_num[0]] = interval_end-interval_start+1;
    automata_name = (char **)realloc(automata_name,
                                     sizeof(char *)*(automata_num[0]+2));
    automata_name[automata_num[0]] = str_preserve(interval_varname);
  }
  else {
    automata_num = (int *)malloc(sizeof(int) * 2);
    automata_num[0] = 1;
    automata_num[1] = interval_end-interval_start+1;
    automata_name = (char **)malloc(sizeof(char *) * 2);
    automata_name[0] = NULL;
    automata_name[1] = str_preserve(interval_varname);
  }
  return 0;
}


/* Allocate a new automata: actually (interval_end-interval_start+1) number
 * of automata with index interval_start ... interval_end and same name
 */

int new_automata(void) {
  struct automaton *x;
  int i, j;

  update_automata();

  for(i=interval_start, j=0; i<=interval_end; i++, j++) {
    x = (struct automaton *)malloc(sizeof(struct automaton));
    x->name=(char *)malloc(1 + strlen(interval_varname));
    strcpy(x->name, interval_varname);
    x->par = str_preserve(parameter);
    free(parameter);
    parameter = NULL;
    x->u_index = system_ta_num + j;
    x->index = i;
    x->init_mode = (char *)NULL;
    x->init_clkconj = (struct clkconj_type *)NULL;
    x->init_boolexp = (struct boolexp_type *)NULL;
    x->mode_cycle = (struct mode_type *)NULL;
    if(automaton_tail) {
      x->next = automaton_tail->next;
      automaton_tail->next = x;
    }
    else x->next = x;
    automaton_tail = x;
  }
  return 0;
}


/* Get the k'th automaton pointer with name "n" */

struct automaton *get_ta(char *n, int k) {
  int i;
  struct automaton *a;

  if(automaton_tail) {
    for(a=automaton_tail->next, i=1; ; a=a->next) {
      if(!strcmp(a->name, n)) {
        if(i == k) return a;
        else i++;
      }
      if(a == automaton_tail) break;
    }
/*
    fprintf(stderr, "Error (line #%d): timed automaton %s not declared.\n", 
                    yylineno, n);
    exit(1);
*/
  }
/*
  fprintf(stderr, "Syntax Error: no timed automaton declared.\n");
  exit(1);
*/
  return NULL;
}

/* Check whether the parameter "n" in the definition is the same as that
 * in the declaration "current_ta->par"
 */

int check_param(char *n) {
  if(!strcmp(n, current_ta->par)) {
    return TRUE;
  }
  else {
    fprintf(stderr, 
     "Error (line #%d): parameter %s does not match that in the declaration.\n",
        yylineno, n);
    exit(1);
  }
}

int is_param(char *n, struct automaton *a) {
  if(a->par == NULL) return FALSE;
  if(!strcmp(n, a->par)) return TRUE;
  else return FALSE;
}

/* Install the initial mode name to each of the current_ta_num TA 
 * starting from current_ta in the cycle of automaton_tail
 */

int install_init_mode(char *mode) {
  struct automaton *x;
  int i;

  for(i=1, x=current_ta; i<=current_ta_num; i++, x=x->next) {
    x->init_mode = (char *)malloc(1 + strlen(mode));
    strcpy(x->init_mode, mode);
  }
  return 0;
}

/* Install the initial condition to each of the current_ta_num (total_exp) TA 
 * starting from current_ta in the cycle of automaton_tail
 */

int install_init_cond() {
  struct automaton *x;
  int i;

  for(i = 1, x = current_ta; i <= total_exp; i++, x = x->next) {
    x->init_clkconj = clkconj_tail[i];
    x->init_boolexp = boolexp_tail[i];
  }
  return 0;
}


/* Initialize all tail pointers to different data structures 
 * Totally current_ta_num+1 pointers in each data structure:
 * [0]=NULL, [1]...[current_ta_num] will point to data structures
 */

int init_tails() {
  int i;

  mode_tail = (struct mode_type **)
              malloc((1+current_ta_num) * sizeof(struct mode_type *));
  out_trans_tail = (struct spec_trans_type **)
              malloc((1+current_ta_num) * sizeof(struct spec_trans_type *));
  assign_tail = (struct assign_type **)
              malloc((1+current_ta_num) * sizeof(struct assign_type *));
  clkconj_tail = (struct clkconj_type **)
              malloc((1+current_ta_num) * sizeof(struct clkconj_type *));
  boolconj_tail = (struct boolconj_type **)
              malloc((1+current_ta_num) * sizeof(struct boolconj_type *));
  boolexp_tail = (struct boolexp_type **)
              malloc((1+current_ta_num) * sizeof(struct boolexp_type *));
  clkconj_expand_start = (struct clkconj_type **)
              malloc((1+current_ta_num) * sizeof(struct clkconj_type *));
  boolconj_expand_start = (struct boolconj_type **)
              malloc((1+current_ta_num) * sizeof(struct boolconj_type *));
  boolexp_expand_start = (struct boolexp_type **)
              malloc((1+current_ta_num) * sizeof(struct boolexp_type *));

  for(i=0; i<=current_ta_num; i++) {
    mode_tail[i] = (struct mode_type *)NULL;
    out_trans_tail[i] = (struct spec_trans_type *)NULL;
    assign_tail[i] = (struct assign_type *)NULL;
  }
  reset_exp_tails();
  return 0;
}

int free_tails() {
  free(mode_tail);
  mode_tail = NULL;

  free(out_trans_tail);
  out_trans_tail = NULL;

  free(assign_tail);
  assign_tail = NULL;

  free(boolexp_tail);
  boolexp_tail = NULL;

  free(boolconj_tail);
  boolconj_tail = NULL;

  free(clkconj_tail);
  clkconj_tail = NULL;

  free(clkconj_expand_start);
  clkconj_expand_start = NULL;

  free(boolconj_expand_start);
  boolconj_expand_start = NULL;

  free(boolexp_expand_start);
  boolexp_expand_start = NULL;

  return 0;
}


/* Reset all expression tails 
 */

int reset_exp_tails() {
  int i;
  for(i=0; i<=current_ta_num; i++) {
    if(clkconj_tail) clkconj_tail[i] = (struct clkconj_type *)NULL;
    if(boolconj_tail) boolconj_tail[i] = (struct boolconj_type *)NULL;
    if(boolexp_tail) boolexp_tail[i] = (struct boolexp_type *)NULL;
    if(clkconj_expand_start) 
      clkconj_expand_start[i] = (struct clkconj_type *)NULL;
    if(boolconj_expand_start)
      boolconj_expand_start[i] = (struct boolconj_type *)NULL;
    if(boolexp_expand_start)
      boolexp_expand_start[i] = (struct boolexp_type *)NULL;
  }
  return 0;
}


/**********************************************************
 **            Variables Related Functions               **
 **********************************************************/

/* Get variable pointer from name[id], or
 * id = NON_PARAM  (-1): the variable has no id
 * id = EXISTS_VAR (-2): any variable with name "name"
 */

struct variable_type *var_ptr(char *name, int id) {
  struct variable_type *x;
  if(var_tail) {
    for(x=var_tail->next; ; x=x->next) {
      if(!strcmp(x->name, name) && 
         (x->id == id || id==EXISTS_VAR)
        )
        return x;
      if(x == var_tail) break;
    }
  }
  return (struct variable_type *)NULL; /* not found */
}


/* Get variable pointer from its unique index */

struct variable_type *var_ptr_from_index(int i) {
  struct variable_type *x;
  if(var_tail) {
    for(x=var_tail->next; ; x=x->next) {
      if(x->index == i) return x;
      if(x == var_tail) break;
    }
  }
  return (struct variable_type *)NULL;
}


/* Get variable unique index from name[id] */

int var_index(char *name, int id){
  struct variable_type *x;
  if(var_tail) {
    for(x=var_tail->next; ; x=x->next) {
      if(!strcmp(x->name, name) && x->id == id) 
        return x->index;
      if(x == var_tail) break;
    }
  }
  return -1;  /* not found */
}


/* Get variable name from its unique index */

char *var_name(int index){
  struct variable_type *x;

  if(index == DELTA) {
    return(str_preserve("DELTA"));
  }
  if(var_tail) {
    for(x=var_tail->next; ; x=x->next) {
      if (x->index == index) {
        return(x->name);
      }
      if(x == var_tail) break;
    }
  }
  return (char *)NULL;  /* not found */
}


/* Get variable [id] from its unique index */

int var_id(int index){
  struct variable_type *x;

  if(var_tail) {
    for(x=var_tail->next; ; x=x->next) {
      if(x->index == index) return x->id;
      if(x == var_tail) break;
    }
  }
  return -2; /* not found */ 
}

/* Get variable type from index */

int get_var_type(int index){
  struct variable_type *x = var_tail;
  if(x) {
    x=x->next;
    for(; x; x=x->next) {
      if(x->index == index) {
        return x->type;
      }
      if(x==var_tail) break;
    }
  }
  return -1;
}

/* Allocate a new variable */

int insert_var(char *name){
  int i, tmp_count;
  struct variable_type *x;

  if(interval_start == -1) {
    if(-1 != var_index(name, -1)) {
      fprintf(stderr, "Warning: Ignoring duplicate variable declaration: %s.\n",
                      name); 
      var_count--; return 0;
    }
    x = (struct variable_type *)malloc(sizeof(struct variable_type));
    x->index = var_count;
    x->id = NON_PARAM;              /* Not a parametric variable like x[i] */
    x->par_val = FALSE;  /* Does not have a parametric value */
    x->name = (char *)malloc(1 + strlen(name));
    strcpy(x->name, name);

    if(var_tail) {
      x->next = var_tail->next;
      var_tail->next = x;
    }
    else x->next = x;
    var_tail = x;
  }
  else if(interval_start >= 0) {
    for(i=interval_start; i<=interval_end; i++) {
      if(var_index(name, i) != -1) {
        fprintf(stderr, 
                "Warning: Duplicate variable declaration: %s[%d], ignoring all variables in the interval [%d..%d].\n", 
                name, i, interval_start, interval_end); 
        var_count -= (interval_end - interval_start + 1);
        return 0;
      }
    }
    tmp_count = var_count - (interval_end - interval_start);
    for(i=interval_start; i<=interval_end; i++) {
      x = (struct variable_type *)malloc(sizeof(struct variable_type));
      x->index = tmp_count++;
      x->id = i;
      x->name = (char *)malloc(1 + strlen(name));
      strcpy(x->name, name);

      if(var_tail) {
        x->next = var_tail->next;
        var_tail->next = x; 
      }
      else x->next = x;
      var_tail = x;
    }
  }
  else if(interval_start != interval_end) {
   fprintf(stderr, "Error: illegal variable interval index values: [%d..%d].\n",
                  interval_start, interval_end);
   fprintf(stderr, "Error: ignoring all variable declarations in %s[%d..%d].\n",
                  interval_varname, interval_start, interval_end);
   return 1;
  }
  else {
    fprintf(stderr, "Error: illegal variable index value: [%d].\n",
                    interval_start);
    fprintf(stderr, "Error: ignoring variable declaration %s[%d].\n",
                    interval_varname, interval_start);
    return 1;
  }
  return 0;
}


/* Add the variable attribute */

int add_attribute(int type){
  int i = var_count - tmp_var_count;
  struct variable_type *x;

  for(x=var_tail->next; i; i--)      /* move to the last group of variables */
    x = x->next;
    
  for(; tmp_var_count; tmp_var_count--) { 
    x->type = type; 
    x = x->next; 
  }
  return 0;
}



/*******************************************************
 **          Expressions Related Functions            **
 *******************************************************/

/* Create a new boolean expression */

int new_boolexp() {
  struct boolexp_type *x;
  int i;

  for(i=1; i<=total_exp; i++) { 
    x = alloc_boolexp();
    if(boolexp_tail[i]) {
      x->next = boolexp_tail[i]->next;
      boolexp_tail[i]->next = x;
    }
    else x->next = x;
    boolexp_tail[i] = x;

    boolconj_tail[i] = (struct boolconj_type *)NULL;
    new_boolconj(i);
  }
  return 0;
}

/* create and append one single boolexp to boolexp_tail[i] 
 * Used in expand_boolexp()  ...
 */
int new_boolexp1(int i) {
  struct boolexp_type *x;

  x = alloc_boolexp();
  if(boolexp_tail[i]) {
    x->next = boolexp_tail[i]->next;
    boolexp_tail[i]->next = x;
  }
  else x->next = x;
  x->boolconj = alloc_boolconj();
  boolexp_tail[i] = x;
  return 0;
}

/* Allocate a boolean expression */

struct boolexp_type *alloc_boolexp() {
  struct boolexp_type *x = (struct boolexp_type *)
                           malloc(sizeof(struct boolexp_type));
  x->boolexp_bdd = (bdd)NULL;
  x->boolconj = (struct boolconj_type *)NULL;
  x->next = x;
  return x;
}

/* Allocate a boolean conjunct */

struct boolconj_type *alloc_boolconj() {
  struct boolconj_type *x = (struct boolconj_type *)
                            malloc(sizeof(struct boolconj_type));
  x->var1 = 0;
  x->var2 = 0;
  x->op = -1;
  x->constant = 0;
  x->mod_num = 0;
  x->mod_var = 0;
  x->prop_bdd = (bdd)NULL;
  x->next = x;
  return x;
}

/* Create a new boolean conjunct for booexp_tail[i] */

int new_boolconj(int i) {
  struct boolconj_type *x;

  x = alloc_boolconj();
  if(boolconj_tail[i]) {
    x->next = boolconj_tail[i]->next;
    boolconj_tail[i]->next = x;
  }
  else x->next = x;
  boolconj_tail[i] = x;
  boolexp_tail[i]->boolconj = boolconj_tail[i];
  return 0;
}

/* Create an array of total_exp new boolean conjuncts */

int new_boolconj_array() {
  int i;
  for(i=1; i<=total_exp; i++) 
    new_boolconj(i);
  return 0;
}

/* Allocate a clock conjunct */

struct clkconj_type *alloc_clkconj() {
  struct clkconj_type *x =
              (struct clkconj_type *)malloc(sizeof(struct clkconj_type));
  x->clk1 = x->clk2 = x->diff = 0;
  x->op = -1;
  x->next = x;
  return x;
}

/* Create a new clock conjunct */

int new_clkconj() {
  struct clkconj_type *x;
  int i;

  for(i=1; i<=total_exp; i++) {
    x = alloc_clkconj();
    if(clkconj_tail[i]) {		  /* clkconj_tail points to the tail */
      x->next = clkconj_tail[i]->next;    /* of a cycle of conjuncts         */
      clkconj_tail[i]->next = x;
    }
    else x->next = x;                  	  /* clkconj_tail is *NULL */
    clkconj_tail[i] = x;  
  }
  return 0;
}

/* Copy a *single* clock conjunct */

struct clkconj_type *clkconj_copy1(struct clkconj_type *c){
  struct clkconj_type *ans = alloc_clkconj();
  ans->clk1 = c->clk1;
  ans->clk2 = c->clk2;
  ans->op = c->op;
  ans->diff = c->diff;
  ans->next = (struct clkconj_type *)NULL;
  return ans;
}

/* Copy a cycle of clock conjuncts */

struct clkconj_type *clkconj_copy(struct clkconj_type *x) {
  struct clkconj_type *c=(struct clkconj_type *)NULL;
  struct clkconj_type *copy = (struct clkconj_type *)NULL;
  struct clkconj_type *ans = (struct clkconj_type *)NULL;

  if(x) {
    for(c=x->next; ; c=c->next) {
      copy = clkconj_copy1(c);
      if(ans) {
        copy->next = ans->next;
        ans->next = copy;
      }
      else copy->next = copy;
      ans = copy;
      if(c == x) break;
    }
  }
  c = NULL;
  copy = NULL;
  return ans;
}

/* Copy a cycle of boolean conjuncts */

struct boolconj_type *boolconj_copy(struct boolconj_type *x) {
  struct boolconj_type *c=(struct boolconj_type *)NULL;
  struct boolconj_type *copy = (struct boolconj_type *)NULL;
  struct boolconj_type *ans = (struct boolconj_type *)NULL;

  if(x) {
    for(c=x->next; ; c=c->next) {
      copy = (struct boolconj_type *)malloc(sizeof(struct boolconj_type));
      copy->var1 = c->var1;
      copy->var2 = c->var2;
      copy->op = c->op;
      copy->constant  = c->constant;
      copy->mod_num = c->mod_num;
      copy->mod_var = c->mod_var;
      copy->prop_bdd = c->prop_bdd;
      bdd_unfree(bddm, c->prop_bdd);
      if(ans) {
        copy->next = ans->next;
        ans->next = copy;
      }
      else copy->next = copy;
      ans = copy;
      if(c == x) break;
    }
  }
  return ans;
}

/* Copy a single boolean conjunct */

struct boolconj_type *boolconj_copy1(struct boolconj_type *x) {
  struct boolconj_type *copy = (struct boolconj_type *)NULL;

  if(x) {
    copy = (struct boolconj_type *)malloc(sizeof(struct boolconj_type));
    copy->var1 = x->var1;
    copy->var2 = x->var2;
    copy->op = x->op;
    copy->constant  = x->constant;
    copy->mod_num = x->mod_num;
    copy->mod_var = x->mod_var;
    copy->prop_bdd = x->prop_bdd;
    bdd_unfree(bddm, x->prop_bdd);
    copy->next = copy;
  }
  return copy;
}

/* Copy a cycle of assignments */

struct assign_type *assign_copy(struct assign_type *x) {
  struct assign_type *c=(struct assign_type *)NULL;
  struct assign_type *copy = (struct assign_type *)NULL;
  struct assign_type *ans = (struct assign_type *)NULL;

  if(x) {
    for(c = x->next; ; c = c->next) {
      copy = alloc_assign();
      copy->var = c->var;
      copy->use_var = c->use_var;
      copy->value  = c->value;
      copy->mod_num = c->mod_num;
      copy->mod_var = c->mod_var;
      copy->op = c->op;
      if(ans) {
        copy->next = ans->next;
        ans->next = copy;
      }
      else copy->next = copy;
      ans = copy;
      if(c == x) break;
    }
  }
  return ans;
}


/* Check whether name[par] is a clock or a discrete variable,
 * par is a parametric string "i", "j", ... or NULL (non-parametric variable)
 * create a new expression accordingly.
 */

int new_exp(char *name, char *par) {
  int t;

  if(par) {
    if(var_ptr(name, EXISTS_VAR) == NULL) {
     fprintf(stderr, "\nError: %s[%s] undeclared in line #%d.\n", 
                     name, par, yylineno);
     exit(1);
    }
  }
  else {
    if(var_ptr(name, NON_PARAM) == NULL) {
     fprintf(stderr, "\nError: %s undeclared in line #%d.\n", 
                     name, yylineno);
     exit(1);
    } 
  }
  t = (var_ptr(name, EXISTS_VAR))->type;
  if(t == V_CLOCK) {
    curr_exp_type = CLK_INEQ;
    new_clkconj();
  }
  else if(t == V_REGISTER) {
    curr_exp_type = BOOL_EXP;
    new_boolexp();
  }
  return 0;
}

/* Warn about an expression beginning with a constant.
 * An expression MUST begin with a variable.
 */
int warn_constant(void) {
  fprintf(stderr, 
"\nError: Illegal to begin an expression with a constant number (%d) on line #%d.\n", last_int, yylineno); 
  exit(1);
}

/* Warn about an expression beginning with a parameter.
 * An expression MUST begin with a variable.
 */
int warn_param(void) {
  fprintf(stderr, 
    "\nError: Illegal to begin an expression with a parameter on line #%d.\n", 
              yylineno); 
  exit(1);
}

/* Free a cycle of clock conjuncts */

int free_clkconj(struct clkconj_type **xconj_cycle) {
  struct clkconj_type *x = (struct clkconj_type *)NULL;

  /* break the clock conjunct cycle first */
  if(*xconj_cycle) {                
    x = (*xconj_cycle)->next;
    (*xconj_cycle)->next = (struct clkconj_type *)NULL;
  }
  while(x) {
    (*xconj_cycle) = x->next;
    free(x);
    x = (*xconj_cycle);
  }
  *xconj_cycle = NULL;
  return 0;
}

/* Free a cycle of boolean expressions */

int free_boolexp(struct boolexp_type *b) {
  struct boolexp_type *x = (struct boolexp_type *)NULL;
  if(b == NULL) return TRUE;
  x = b->next;
  b->next = (struct boolexp_type *)NULL;
  while(x) {
    b = x->next;
    bdd_free(bddm, x->boolexp_bdd);
    if(x->boolconj) 
      free_boolconj(x->boolconj);
    free(x);
    x = b;
  }
  b = NULL;
  return 0;
}

/*
int free_one_boolexp(struct boolexp_type *b) {
  struct boolconj_type *x = (struct boolconj_type *)NULL;
  printf("  freeing one boolexp: "); print_boolexp(b); printf("\n");
  bdd_free(bddm, b->boolexp_bdd);
  if(b->boolconj) {
    x = b->boolconj->next;
    b->boolconj->next = (struct boolconj_type *)NULL;
    while(x) {
      printf("In here 0 ...\n");
      b->boolconj = x->next;
      printf("In here 1 ...\n");
      free_boolconj(x);
      printf("In here 2 ...\n");
      x = b->boolconj;
    }
  }
  free(b);
  b = NULL;
  return 0;
}
*/

/* Free a cycle of boolean conjuncts */

int free_boolconj(struct boolconj_type *b) {
  struct boolconj_type *x = (struct boolconj_type *)NULL;

  if(b && b->next == NULL) 
    x = b;
  else if(b) {
    x = b->next;
    b->next = (struct boolconj_type *)NULL;
  }
  while(x) {
    b = x->next;
    /*bdd_free(bddm, x->prop_bdd); */
    free(x); 
    x = b;
  }
  b = NULL;
  return 0;
}

/* Install a parsed constant into clkconj_tail or boolconj_tail cycles
 */
int install_const() {
  int i;
  struct clkconj_type *c;
  struct boolconj_type *b;
  struct boolexp_type *e;

  if(index_relop == -1) {
    if(curr_exp_type == CLK_INEQ) {
      if(sign == 1) need_transform = 1;
      for(i=1; i<=total_exp; i++) 
        clkconj_tail[i]->diff = last_int;
    }
    else if(curr_exp_type == BOOL_EXP) {
      for(i=1; i<=total_exp; i++) 
        boolconj_tail[i]->constant = last_int;
    }
  }
  else {
    if(curr_exp_type == CLK_INEQ) {
      if(sign == 1) need_transform = 1;
      for(i=1; i<=total_exp; i++) {
        for(c=clkconj_expand_start[i]; ; c=c->next) {
          if(c) c->diff = last_int;
          if(c == clkconj_tail[i]) break;
        }
      }
    }
    else if(curr_exp_type == BOOL_EXP) {
      if(index_quantifier == FORALL) 
        for(i=1; i<=total_exp; i++) {
          if(boolconj_expand_start[i]) {
            for(b=boolconj_expand_start[i]; ; b=b->next) {
              if(b) b->constant = last_int;
              if(b == boolconj_tail[i]) break;
            }
          }
        }
      else if(index_quantifier == EXISTS)
        for(i=1; i<=total_exp; i++) {
          if(boolexp_expand_start[i]) {
            for(e=boolexp_expand_start[i]; ; e=e->next) {
              if(e) e->boolconj->constant = last_int;
              if(e == boolexp_tail[i]) break;
            }
          }
        }
    }
  }
  return 0;
}

/* Transform clkconj:  if the clkconj needs transforming (need_transform == 1)
 *                     this occurs when: (a) x ~ y - d
 *                                       (b) x + d ~ y
 *                     i.e., sign == 1 during install_const() 
 */
int transform_clkconj() {
  int i, tmp;

  if(need_transform) {
    for(i=1; i<=current_ta_num; i++) {
                           /* swap clk1 and clk2 of clkconj_tail[i] */
      tmp = clkconj_tail[i]->clk1;
      clkconj_tail[i]->clk1 = clkconj_tail[i]->clk2;
      clkconj_tail[i]->clk2 = tmp;
                          /* reverse operator: < into >, >= into <=, ... */
      clkconj_tail[i]->op = reverse_op(clkconj_tail[i]->op);
    }
  }
  return 0;
}


/* Get the reverse of operator op */

int reverse_op(int op) {
  switch(op) {
    case LESS: 		return GREATER;
    case LEQ:  		return GEQ;
    case EQ:   		return EQ;
    case GEQ:  		return LEQ;
    case GREATER: 	return LESS;
    case NEQ:     	return NEQ;
    default: 		fprintf(stderr, 
                                "\nFatal error in reverse_op(%d)\n", op);
             		exit(1);
  }
}


/* Install a parameter into clkconj_tail or boolconj_tail
 */
int install_param() {
  int i;
  struct variable_type *v = NULL;
  struct clkconj_type *c;
  struct boolconj_type *b;
  struct boolexp_type *e;

  if(index_relop == -1) {
    if(curr_exp_type == CLK_INEQ) 
      for(i=1; i<=total_exp; i++) {
        clkconj_tail[i]->diff = i;
        v = var_ptr_from_index(clkconj_tail[i]->clk1);
        v->par_val = TRUE;
      }
    else if(curr_exp_type == BOOL_EXP)
      for(i=1; i<=total_exp; i++) {
        boolconj_tail[i]->constant = i;
        v = var_ptr_from_index(boolconj_tail[i]->var1);
        if(v) v->par_val = TRUE;
        v = var_ptr_from_index(boolconj_tail[i]->var2);
        if(v) v->par_val = TRUE;
      }
  }
  else {
    if(curr_exp_type == CLK_INEQ)
      for(i=1; i<=total_exp; i++) {
        for(c=clkconj_expand_start[i]; ; c=c->next) {
          if(c) c->diff = i;
          v = var_ptr_from_index(c->clk1);
          v->par_val = TRUE;
          if(c==clkconj_tail[i]) break;
        }
      }
    else if(curr_exp_type == BOOL_EXP) {
      if(index_quantifier == FORALL)
        for(i=1; i<=total_exp; i++) {
          for(b=boolconj_expand_start[i]; ; b=b->next) {
            if(b) b->constant = i;
            v = var_ptr_from_index(b->var1);
            if(v) v->par_val = TRUE;
            v = var_ptr_from_index(b->var2);
            if(v) v->par_val = TRUE;
            if(b==boolconj_tail[i]) break;
          }
        }
      else if(index_quantifier == EXISTS)
        for(i=1; i<=total_exp; i++) {
          for(e=boolexp_expand_start[i]; ; e=e->next) {
            if(e) e->boolconj->constant = i;
            v = var_ptr_from_index(e->boolconj->var1);
            if(v) v->par_val = TRUE;
            v = var_ptr_from_index(e->boolconj->var2);
            if(v) v->par_val = TRUE;
            if(e==boolexp_tail[i]) break;
          }
        }
    }
  }
  v = NULL;
  return 0;
}

/* Install a term into clkconj_tail or boolconj_tail
 */
int install_term(char *name, char *par) {
  int i, v, n=1;

  if(par) {
    if(var_ptr(name, EXISTS_VAR) == NULL) {
     fprintf(stderr, "\nError: %s[%s] undeclared in line #%d.\n",
                     name, par, yylineno);
     exit(1);
    }
  }
  else {
    if(var_ptr(name, NON_PARAM) == NULL) {
     fprintf(stderr, "\nError: %s undeclared in line #%d.\n",
                     name, yylineno);
     exit(1);
    }
  }

  if(curr_exp_type == CLK_INEQ) {
    for(i=1; i<=total_exp; i++) {
      if(par) v=var_index(name, i);
      else v=var_index(name, NON_PARAM);
      if(v == -1) {
        if(par)
          fprintf(stderr, "\nUndeclared variable %s[%d] on line # %d.\n", 
                          string_buffer, i, yylineno);
        else
          fprintf(stderr, "\nUndeclared variable %s on line # %d.\n", 
                          string_buffer, yylineno);
        exit(1);
      }
      if(clkconj_tail[i]->clk1 == 0)
        clkconj_tail[i]->clk1 = v;
      else if(((clkconj_tail[i])->clk2) == 0)
        clkconj_tail[i]->clk2 = v;
      else {
        fprintf(stderr,
            "\nError: more than 2 clock variables in a single conjunct.\n");
        fprintf(stderr, "\nExiting ... on parsing line #%d.\n", yylineno);
        exit(1);
      }
    }
  }
  else { /* BOOL_EXP */
   if(index_relop == -1) {
    for(i=1; i<=total_exp; i++) {
      if(par) v=var_index(name, i);
      else v=var_index(name, NON_PARAM);
      if(v == -1) {
        if(par)
          fprintf(stderr, "\nUndeclared variable %s[%d] on line # %d.\n", 
                          string_buffer, i, yylineno);
        else
          fprintf(stderr, "\nUndeclared variable %s on line # %d.\n", 
                          string_buffer, yylineno);
        exit(1);
      }
      if(boolconj_tail[i]->var1 == 0) {
        boolconj_tail[i]->var1 = v;
        n=1;
      }
      else if(boolconj_tail[i]->var2 == 0) {
        boolconj_tail[i]->var2 = v;
        n=2;
      }
      else {
        fprintf(stderr, 
         "\nError: more than two proposition variables in a single conjunct\n");
        fprintf(stderr, "\nExiting ... on parsing line #%d\n", yylineno);
        exit(1);
      }
    }
   }
   else if(index_quantifier == FORALL) expand_boolconj(name, n);
   else if(index_quantifier == EXISTS) expand_boolexp(name, n);
  }
  return 0;
}

int expand_boolconj(char *name, int n) {
  int i, j, v;

  switch(index_relop) {
    case LEQ: for(i=1; i<=total_exp; i++) {
                boolconj_expand_start[i] = boolconj_tail[i];
                for(j=1; j<=i; j++) {
                  v = var_index(name, j);
                  if(n==1) boolconj_tail[i]->var1 = v;
                  if(n==2) boolconj_tail[i]->var2 = v;
                  if(j<i) new_boolconj(i);
                }
              }
              break;
    case LESS: for(i=1; i<=total_exp; i++) {
                 if(i==1) {
                   if(boolconj_tail[i] != NULL) {
                     delete_boolconj_tail(&(boolconj_tail[i]), i);
                   }
                 }
                 else {
                   boolconj_expand_start[i] = boolconj_tail[i];
                   for(j=1; j<i; j++) {
                     v = var_index(name, j);
                     if(n==1) boolconj_tail[i]->var1 = v;
                     if(n==2) boolconj_tail[i]->var2 = v;
                     if(j<i-1) new_boolconj(i);
                   }
                 }
               }
               break;
    case NEQ: for(i=1; i<=total_exp; i++) {
                boolconj_expand_start[i] = boolconj_tail[i];
                for(j=1; j<=total_exp; j++) {
                  if(j != i) {
                    v = var_index(name, j);
                    if(n == 1) boolconj_tail[i]->var1 = v;
                    if(n == 2) boolconj_tail[i]->var2 = v;
                    if(j<total_exp) new_boolconj(i);
                  }
                }
              }
              break;
    case GREATER: for(i=1; i<=total_exp; i++) {
                    if(i==total_exp) {
                      if(boolconj_tail[i] != NULL)
                        delete_boolconj_tail(&(boolconj_tail[i]), i);
                    }
                    else {
                      boolconj_expand_start[i] = boolconj_tail[i];
                      for(j=i+1; j<=total_exp; j++) {
                        v = var_index(name, j);
                        if(n==1) boolconj_tail[i]->var1 = v;
                        if(n==2) boolconj_tail[i]->var2 = v;
                        if(j<total_exp) new_boolconj(i);
                      }
                    }
                  }
                  break;
    case GEQ: for(i=1; i<=total_exp; i++) {
                boolconj_expand_start[i] = boolconj_tail[i];
                for(j=i; j<=total_exp; j++) {
                  v = var_index(name, j);
                  if(n==1) boolconj_tail[i]->var1 = v;
                  if(n==2) boolconj_tail[i]->var2 = v;
                  if(j<total_exp) new_boolconj(i);
                }
              }
              break;
    default: fprintf(stderr,
                     "Error: illegal relop (%d) in expand_boolconj().\n",
                     index_relop);
             exit(1);
             break;
  }
  return 0;
}

int delete_boolconj_tail(struct boolconj_type **t, int i) {
  struct boolconj_type *b;

  if((*t) == NULL) return TRUE;

  if((*t)->next == *t) {
    free_boolconj(*t);
    *t = NULL;
    boolconj_tail[i] = NULL;
    boolexp_tail[i]->boolconj = NULL;
    free_boolexp(boolexp_tail[i]);
    boolexp_tail[i] = NULL;
/*
    delete_boolexp_tail(&(boolexp_tail[i]));
*/
  }
  else for(b=(*t)->next; ; b=b->next) {
    if(b->next == *t) {
      b->next = (*t)->next;
      (*t)->next = *t;
      free_boolconj(*t);
      boolexp_tail[i]->boolconj = *t = b;
      break;
    }
  }
  return 0;
}

int delete_boolexp_tail(struct boolexp_type **t) {
  struct boolexp_type *b;
  if(*t) {
    if((*t)->next == *t) {
      free_boolexp(*t);
      *t = NULL;
    }
    else for(b=(*t)->next; ; b=b->next) {
      if(b->next == *t) {
        b->next = (*t)->next;
        (*t)->next = *t;
        free_boolexp(*t);
        *t = b;
        break;
      }
    }
  }
  return 0;
}

int expand_boolexp(char *name, int n) {
  int i, j, v;

  switch(index_relop) {
    case LEQ: for(i=1; i<=total_exp; i++) {
                boolexp_expand_start[i] = boolexp_tail[i];
                for(j=1; j<=i; j++) {
                  v = var_index(name, j);
                  if(n==1) boolexp_tail[i]->boolconj->var1 = v;
                  if(n==2) boolexp_tail[i]->boolconj->var2 = v;
                  if(j<i) new_boolexp1(i);
                }
              }
              break;
    case LESS: for(i=1; i<=total_exp; i++) {
                 if(i==1) {
                   if(boolexp_tail[i] != NULL) {
                     delete_boolexp_tail(&(boolexp_tail[i]));
                   }
                 }
                 else {
                   boolexp_expand_start[i] = boolexp_tail[i];
                   for(j=1; j<i; j++) {
                     v = var_index(name, j);
                     if(n==1) boolexp_tail[i]->boolconj->var1 = v;
                     if(n==1) boolexp_tail[i]->boolconj->var2 = v;
                     if(j<i-1) new_boolexp1(i);
                   }
                 }
               }
               break;
    case NEQ: for(i=1; i<=total_exp; i++) {
                boolexp_expand_start[i] = boolexp_tail[i];
                for(j=1; j<=total_exp; j++) {
                  if(j != i) {
                    v = var_index(name, j);
                    if(n==1) boolexp_tail[i]->boolconj->var1 = v;
                    if(n==2) boolexp_tail[i]->boolconj->var2 = v;
                    if(j<total_exp) new_boolexp1(i);
                  }
                }
              }
              break;
    case GREATER: for(i=1; i<=total_exp; i++) {
                    if(i==total_exp) {
                      if(boolexp_tail[i] != NULL)
                        delete_boolexp_tail(&(boolexp_tail[i]));
                    }
                    else {
                      boolexp_expand_start[i] = boolexp_tail[i];
                      for(j=i+1; j<=total_exp; j++) {
                        v = var_index(name, j);
                        if(n==1) boolexp_tail[i]->boolconj->var1 = v;
                        if(n==2) boolexp_tail[i]->boolconj->var2 = v;
                        if(j<total_exp) new_boolexp1(i);
                      }
                    }
                  }
                  break;
    case GEQ: for(i=1; i<=total_exp; i++) {
                boolexp_expand_start[i] = boolexp_tail[i];
                for(j=i; j<=total_exp; j++) {
                  v = var_index(name, j);
                  if(n==1) boolexp_tail[i]->boolconj->var1 = v;
                  if(n==2) boolexp_tail[i]->boolconj->var2 = v;
                  if(j<total_exp) new_boolexp1(i);
                }
              }
              break;
    default: fprintf(stderr,
                     "Error: illegal relop (%d) in expand_boolconj().\n",
                     index_relop);
             exit(1);
             break;
  }
  return 0;
}

/* Install the relation operator optype */

int install_relop(int optype){
  struct clkconj_type *c;
  struct boolconj_type *b;
  struct boolexp_type *e;
  int i;
  int new_op = optype;
  if(neg_conj) {
    switch(optype) {
      case LESS: new_op = GEQ;
                 break;
      case LEQ: new_op = GREATER;
                break;
      case EQ: new_op = NEQ;
               break;
      case GEQ: new_op = LESS;
                break;
      case GREATER: new_op = LEQ;
                    break;
      case NEQ: new_op = EQ;
                break;
    }
  }
  if(index_relop == -1) {
    if(curr_exp_type == CLK_INEQ) 
      for(i=1; i<=total_exp; i++) {
        clkconj_tail[i]->op = new_op;
      }
    else 
      for(i=1; i<=total_exp; i++) 
        boolconj_tail[i]->op = new_op;
  }
  else {
    if(clkconj_expand_start[1])
      for(i=1; i<=total_exp; i++) {
        for(c=clkconj_expand_start[i]; ; c=c->next) {
          if(c) c->op = new_op;
          if(c == clkconj_tail[i]) break;
        }
      }
    else {
      if(index_quantifier == FORALL) {
        for(i=1; i<=total_exp; i++) {
          if(boolconj_expand_start[i]) {
            for(b=boolconj_expand_start[i]; ; b=b->next) {
              if(b) b->op = new_op;
              if(b == boolconj_tail[i]) break;
            }
          }
        }
      }
      else {
        for(i=1; i<=total_exp; i++) {
          if(boolexp_expand_start[i]) {
            for(e=boolexp_expand_start[i]; ; e=e->next) {
              if(e) e->boolconj->op = new_op;
              if(e == boolexp_tail[i]) break;
            }
          }
        }
      }
    }
  }
  return 0;
}

int install_assign_op(int op) {
  int i;
  for(i=1; i<=current_ta_num; i++)
    assign_tail[i]->op = op;
  return 0;
}

/* 
   check if (op1, num1) and (op2, num2) are compatible intervals:
      (1) if they are overlapping intervals, return OVERLAPPING
      (2) if they can be reduced into 1 interval, return TRUE 
             ( the resulting interval is in (*op3, *num3) )
      (3) if they are not compatible, return FALSE
*/

int compatible(int op1, int num1, int op2, int num2, int *op3, int *num3) {

  if(op1 == LESS || op1 == LEQ) {

    if(op2 == LESS || op2 == LEQ) {
      if(num1 == num2) {
        switch(op1) {
          case LESS: *op3 =  LESS;
                     break;
          case LEQ: switch(op2) {
                      case LESS: *op3 = LESS;
                                 break;
                      case LEQ: *op3 = LEQ;
                                break;
                    }
        }
      }
      else if(op1 == LEQ && op2 == LEQ && num1 == -num2) {
        *op3 = EQ;
        *num3 = (num1 >= 0) ? num1 : num2;
        return TRUE;
      }
      else {
        *op3 = (num1 < num2) ? op1 : op2;
      }

      *num3 = (num1 <= num2) ? num1 : num2;
      return TRUE;
    }

    else
    if(op2 == EQ) {
      if(num2 < num1) {
        *op3 = EQ;
        *num3 = num2;
        return TRUE;
      }
      else if(num2 > num1) {
             op3 = NULL;
             num3 = NULL;
             return FALSE;
      }
      else {                    /* num1 == num2 */
        if(op1 == LESS) {
          op3 = NULL;
          num3 = NULL;
          return FALSE;
        }
        else {
          *op3 = EQ;
          *num3 = num2;
          return TRUE;
        }
      }
    }

    else
    if(op2 == GEQ || op2 == GREATER) {
      if(num2 < num1) {
        op3 = NULL;
        num3 = NULL;
        return OVERLAPPING;
      }
      else if(num2 > num1) {
        op3 = NULL;
        num3 = NULL;
        return FALSE;
      }
      else {                 /* num1 == num2 */
        switch(op1) {
          case LESS:
                     op3 = NULL;
                     num3 = NULL;
                     return FALSE;
                     break;
          case LEQ: switch(op2) {
                      case GREATER: 
                                    op3 = NULL;
                                    num3 = NULL;
                                    return FALSE;
                                    break;
                      case GEQ: *op3 = EQ;
                                *num3 = num1;
                                return TRUE;
                                break;
                    }
        }
      }
    }

  } /* END if(op1 == LESS || op1 == LEQ) */

  else
  if(op1 == GREATER || op1 == GEQ) {

    if(op2 == GREATER || op2 == GEQ) {
      if(num2 == num1) {
        switch(op1) {
          case GREATER: *op3 =  GREATER;
                        break;
          case GEQ: switch(op2) {
                      case GREATER: *op3 = GREATER;
                                 break;
                      case GEQ: *op3 = GEQ;
                                break;
                    }
        }
      }
      else
        *op3 = (num1 > num2) ? op1 : op2;

      *num3 = (num1 >= num2) ? num1 : num2;
      return TRUE;
    }

    else
    if(op2 == EQ) {
      if(num2 > num1) {
        *op3 = EQ;
        *num3 = num2;
        return TRUE;
      }
      else if(num2 < num1) {
             op3 = NULL;
             num3 = NULL;
             return FALSE;
      }
      else {                    /* num1 == num2 */
        if(op1 == GREATER) return FALSE;
        else {      /* op1 == GEQ */
          *op3 = EQ;
          *num3 = num2;
          return TRUE;
        }
      }
    }

    else
    if(op2 == LEQ || op2 == LESS) {
      if(num2 > num1) {
        op3 = NULL;
        num3 = NULL;
        return OVERLAPPING;
      }
      else if(num2 < num1) {
        op3 = NULL;
        num3 = NULL;
        return FALSE;
      }
      else {                 /* num1 == num2 */
        switch(op1) {
          case GREATER: op3 = NULL;
                        num3 = NULL;
                        return FALSE;
                        break;
          case GEQ: switch(op2) {
                      case LESS: op3 = NULL;
                                 num3 = NULL;
                                 return FALSE;
                                 break;
                      case LEQ: *op3 = EQ;
                                *num3 = num1;
                                return TRUE;
                                break;
                    }
        }
      }
    }

  } /* END if(op1 == GREATER || op1 == GEQ) */

  else {     /* if(op1 == EQ) { */
    if(op2 == GREATER || op2 == GEQ) {
      if(num2 > num1) {
        op3 = NULL;
        num3 = NULL;
        return FALSE;
      }
      else if(num2 < num1) {
        *op3 = EQ;
        *num3 = num1;
        return TRUE;
      }
      else {         /* num2 == num1 */
        switch(op2) {
          case GREATER: op3 = NULL;
                        num3 = NULL;
                        return FALSE; 
                        break;
          case GEQ: *op3 = EQ;
                    *num3 = num1;
                    return TRUE;
        }
      }
    }
    if(op2 == LESS || op2 == LEQ) {
      if(num2 < num1) {
        op3 = NULL;
        num3 = NULL;
        return FALSE;
      }
      else if(num2 > num1) {
        *op3 = EQ;
        *num3 = num1;
        return TRUE;
      }
      else {         /* num2 == num1 */
        switch(op2) {
          case LESS: op3 = NULL;
                     num3 = NULL;
                     return FALSE; 
                     break;
          case LEQ: *op3 = EQ;
                    *num3 = num1;
                    return TRUE;
        }
      }
    }
    if(op2 == EQ) {
      if(num1 == num2) {
        *op3 = EQ;
        *num3 = num1;
        return TRUE;
      }
      else {
        op3 = NULL;
        num3 = NULL;
        return FALSE;
      }
    }
  } /* END if(op1 == EQ) */
  return OVERLAPPING;
}

struct clkconj_type *
clkconj_and(struct clkconj_type *x1, struct clkconj_type *x2) {
  int tmp_clk, i, j;
  int nx1_count = 0;
  int nx2_count = 0;
  int comp = DIFFERENT;
  int comp_new = DIFFERENT;
  int *tmp_op = NULL, *tmp_diff = NULL;
  struct clkconj_type *nx1, *nx2, *tmp = NULL, 
                      *y1, *y2, *py1, *py2, 
                      *p_true_ptr1 = NULL, *true_ptr1 = NULL;

/*
  output("Conjunction of clkconj: ");
  print_clkconj(x1); output(" AND "); print_clkconj(x2); output(" is:\n");
*/

  if(!x1 && !x2) {
    tmp = alloc_clkconj();
    tmp->op = TRUE;
    return tmp;
  }

  nx1 = clkconj_copy(x1);
  nx2 = clkconj_copy(x2);

  if(x1 == NULL || x1->op == TRUE) {
    free_clkconj(&nx1); nx1 = NULL;
    if(x2 == NULL) {
      tmp = alloc_clkconj();
      tmp->op = TRUE;
      return tmp;
    }
    else return nx2;
  }
  else if(x1->op == FALSE) {
    free_clkconj(&nx1); nx1 = NULL;
    free_clkconj(&nx2); nx2 = NULL;
    tmp = alloc_clkconj();
    tmp->op = FALSE;
    return tmp;
  }

  if(x2 == NULL || x2->op == TRUE) {
    free_clkconj(&nx2); nx2 = NULL;
    if(x1 == NULL) {
      tmp = alloc_clkconj();
      tmp->op = TRUE;
      return tmp;
    }
    else return nx1;
  }
  else if(x2->op == FALSE) {
    free_clkconj(&nx1); nx1 = NULL;
    free_clkconj(&nx2); nx2 = NULL;
    tmp = alloc_clkconj();
    tmp->op = FALSE;
    return tmp;
  }

  for(y1=nx1->next; ; y1=y1->next) {
    nx1_count++;
    if(y1 == nx1) break;
  }

  for(y2=nx2->next; ; y2=y2->next) {
    nx2_count++;
    if(y2 == nx2) break;
  }

  for(j=1, py2=nx2, y2=nx2->next; j<=nx2_count; j++, py2 = y2, y2=y2->next) {
    for(i=1, py1=nx1, y1=nx1->next; i<=nx1_count; i++, py1 = y1, y1=y1->next) {
      if(y1->clk1 == y2->clk2 && y1->clk2 == y2->clk1) {
	/* if x11=x22 and x12==x21, then make them the same by swapping */
        if(y1->clk1) {
          tmp_clk = y2->clk1;  
          y2->clk1 = y2->clk2;
          y2->clk2 = tmp_clk;
          y2->op = reverse_op(y2->op);
          y2->diff = -(y2->diff);
        }
        else {
          tmp_clk = y1->clk1; 
          y1->clk1 = y1->clk2;
          y1->clk2 = tmp_clk;
          y1->op = reverse_op(y1->op);
          y1->diff = -(y1->diff);
        }
      }
      if(y1->clk1 == y2->clk1 && y1->clk2 == y2->clk2) {
/*        
        printf("y1: "); print_one_clkconj(y1); printf("\n");
        printf("y2: "); print_one_clkconj(y2); printf("\n");
*/
        tmp_op = (int *)malloc(sizeof(int));
        tmp_diff = (int *)malloc(sizeof(int));

        comp_new =  compatible( y1->op, y1->diff, y2->op, y2->diff, 
                                tmp_op, tmp_diff);
        switch(comp_new) { 
          case OVERLAPPING : /*printf("Got OVERLAPPING\n"); */
                             switch(comp) {
                               case DIFFERENT: comp = OVERLAPPING;
                                               break;
                               case TRUE: break;
                             }
                             break;

          case FALSE: /* printf("Got FALSE\n"); */
                      if(comp == TRUE && tmp != NULL) {
                        free_clkconj(&tmp);  tmp = NULL;
                      }
                      comp = FALSE;
                      break;

          case TRUE: /* printf("Got TRUE\n"); */
                     if(comp == TRUE && tmp != NULL) {
                       free_clkconj(&tmp);  tmp = NULL;
                     }
                     comp = TRUE; 
/*			 printf("Got TRUE for %s, %d, %s, %d\n", 
                         relop_string(y1->op), y1->diff, 
                         relop_string(y2->op), y2->diff);
*/
                     p_true_ptr1 = py1;
                     true_ptr1 = y1;
                     tmp = alloc_clkconj();
                     tmp->clk1 = y1->clk1;
                     tmp->clk2 = y1->clk2;
                     tmp->op = *tmp_op;
                     tmp->diff = *tmp_diff;
                     break;

          default: 
              printf("Error: invalid return value from compatible() = %d\n", 
                      comp_new);
              break;

        } /* END switch(comp_new) */
	free(tmp_op); tmp_op = NULL;
	free(tmp_diff); tmp_diff = NULL;

      } /* END if(y1->clk1 == y2->clk1 && y1->clk2 == y2->clk2) */

      if(comp == FALSE) break;

    } /* END for(; ; py1 = y1, y1=y1->next) */

    switch(comp) { 
      case TRUE:                            /* replace true_ptr1 with tmp */
/*
                 printf("Before replacing ");
                 print_one_clkconj(true_ptr1);
                 printf("nx1: ");
                 print_clkconj(nx1);
                 printf("\n");
*/
                 if(p_true_ptr1 == true_ptr1) {
                   tmp->next = tmp;
		   free_clkconj(&nx1);
                   nx1 = tmp;
                 }
                 else {
                   p_true_ptr1->next = tmp;
                   tmp->next = true_ptr1->next;
                   if(nx1 == true_ptr1) 
                     nx1 = tmp;
		   true_ptr1->next = true_ptr1;
		   free_clkconj(&true_ptr1); true_ptr1 = NULL;
                 }
/*
                 printf("After replacement, nx1: ");
                 print_clkconj(nx1);
                 printf("\n");
*/
                 break;
      case FALSE: free_clkconj(&nx1);
                  nx1 = NULL;
                  free_clkconj(&nx2);
                  nx2 = NULL;
                  tmp = alloc_clkconj();
                  tmp->op = FALSE;
                  return tmp;
                  break;
      case DIFFERENT:
      case OVERLAPPING:  
                        if(nx2->next == nx2)
                          nx2 = (struct clkconj_type *)NULL;
                        else { 			/* Skip y2 */
                          if(y2 == nx2) nx2 = py2;
                          py2->next = y2->next; 
                        }
                        y2->next = nx1->next;
                        nx1->next = y2;        	/* append y2 into nx1 */
                        nx1 = y2;
                        nx1_count++;
                        y2 = py2;
                        break;
    }
    if(!nx2) break;
    comp = comp_new = DIFFERENT; /* reset the comparison results */
  }    /* END for(; ; py2 = y2, y2=y2->next) */

/*
  if(true_ptr1) true_ptr1->next = true_ptr1;
  free_clkconj(&true_ptr1); true_ptr1 = NULL;
*/
  free_clkconj(&nx2);
  nx2 = NULL;
/*
  print_clkconj(nx1); output("\n");
*/
  return nx1;
}

int clkconj_compare(struct clkconj_type *x, struct clkconj_type *y) {
  struct clkconj_type *x1, *y1;

/*
    output("Comparing clkconj:\n");
    print_clkconj(x); output("\n");
    print_clkconj(y); output("\n\n");
*/
  if(x && !y) return -1;
  if(!x && y) return 1;
  if(!x && !y) return EQUAL;

  for(x1=x->next; ; x1=x1->next) {
    if(clkconj_contained(x1, y)) ;
    else {
      for(y1=y->next; ; y1=y1->next) {
        if(clkconj_contained(y1, x)) ;
        else {
          return clkconj_compare_one(x1, y1);
        }
        if(y1 == y) break;
      }
    }
    if(x1 == x) break;
  }
  return EQUAL;
}

/* Check if single clkconj c1 is contained in the cycle of clkconj c2 */

int clkconj_contained(struct clkconj_type *c1, struct clkconj_type *c2) {
  struct clkconj_type *c;
  if(c2) {
    for(c=c2->next; ; c=c->next) {
      if(c->op == c1->op && c->clk1 == c1->clk1 && 
         c->clk2 == c1->clk2 && c->diff == c1->diff) return TRUE;
      if(c->op == reverse_op(c1->op) && c->clk1 == c1->clk1 && 
         c->clk2 == c1->clk2 && c->diff == -(c1->diff)) return TRUE;
      if(c == c2) break;
    }
  }
  return FALSE;
}

/* Compare two single clkconj */
int clkconj_compare_one(struct clkconj_type *c1, struct clkconj_type *c2) {
  if(c1->op < c2->op) return -1;
  if(c1->op > c2->op) return 1;
  if(c1->clk1 < c2->clk1) return -1;
  if(c1->clk1 > c2->clk1) return 1;
  if(c1->clk2 < c2->clk2) return -1;
  if(c1->clk2 > c2->clk2) return 1;
  if(c1->diff < c2->diff) return -1;
  if(c1->diff > c2->diff) return 1;
  return EQUAL;
}

/* Create a new mode */

int new_mode(char *name){
  struct automaton *a;
  struct mode_type *x;
  int i, j, k;

  for(i=1; i<=current_ta_num; i++) {
    x = alloc_mode();
    x->index = ++mode_count;
    x->name = str_preserve(name);
/*
    x = (struct mode_type *)malloc(sizeof(struct mode_type));
    x->mode_vec = (char **)malloc(sizeof(char *)*system_ta_num);
    for(j=0; j<system_ta_num; j++) {
      x->mode_vec[j] = (char *)NULL;
    }
    x->xconj_cycle = (struct clkconj_type *)NULL;
    x->boolexp = (struct boolexp_type *)NULL;
    x->zone = NULL;
    x->xconj_length = 0;
    x->in_trans_count = x->out_trans_count = 0;
    x->in_trans_cycle = x->out_trans_cycle = (struct transition_type *)NULL;
    x->cvl = NULL;
    x->dvl = NULL;
    x->shielded_clocks = NULL;
    x->reqd_clocks = NULL;
    x->norm_perm = (struct permutation_type *)NULL;
*/
    for(k=0, j=0, a=automaton_tail->next; j < system_ta_num; j++, a=a->next) {
      if(!strcmp(current_ta->name, a->name)) k++;
      if(k == i) break;
    }
    x->mode_vec[j] = str_preserve(name);

    if(mode_tail[i]) {
      x->next = mode_tail[i]->next;
      mode_tail[i]->next = x;
    }
    else x->next = x; 

    mode_tail[i] = x;
  }
  return 0;
}

/* Install a mode's clock and discrete variable invariants,
 * namely, clkconj_tail and boolexp_tail
 */
int install_inv() {
  struct clkconj_type *x;
  int i;

  for(i=1; i<=total_exp; i++) {
    if(clkconj_tail[i]) {
      for(x = clkconj_tail[i]->next; ; x=x->next) {
        (mode_tail[i]->xconj_length)++;
        if(x == clkconj_tail[i]) break;
      }
      mode_tail[i]->xconj_cycle = x;
    }
    if(boolexp_tail[i]) 
      mode_tail[i]->boolexp = boolexp_tail[i];
  }
  return 0;
}

/* Create a new specification transition and insert it into out_trans_tail 
 * cycle 
 */
int new_transition() {
  struct spec_trans_type *x;
  int i;
  for(i=1; i<=current_ta_num; i++) {
    x = (struct spec_trans_type *)malloc(sizeof(struct spec_trans_type));
    x->index = ++trans_count;
    x->sync_cycle = (struct var_link_type *)NULL;
    x->clk_trigger_length = x->prp_trigger_length = x->assign_length = 0;
    x->clk_trigger_cycle = (struct clkconj_type *)NULL;
    x->prp_trigger_cycle = (struct boolconj_type *)NULL;
    x->assign_cycle = (struct assign_type *)NULL;
    x->trigger_bdd = (bdd)NULL;
    x->goto_mode = (char *)NULL;
    if(out_trans_tail[i]) {
      x->next = out_trans_tail[i]->next;
      out_trans_tail[i]->next = x;
    }
    else x->next = x;
    out_trans_tail[i] = x;
    (mode_tail[i]->out_trans_count)++;
  }
  return 0;
}

/* Install the trigger cycles: clk_trigger_cycle and prp_trigger_cycle
 * from clkconj_tail and boolconj_tail, respectively
 */
int install_trigger() {
  struct clkconj_type *x;
  struct boolconj_type *y;
  int i;

  for(i=1; i<=total_exp; i++) {
    if(clkconj_tail[i]) {
      for(x = clkconj_tail[i]->next; ; x=x->next) {
        (out_trans_tail[i]->clk_trigger_length)++;
        if(x == clkconj_tail[i]) break;
      }
      out_trans_tail[i]->clk_trigger_cycle = x;
    }

    if(boolconj_tail[i]) {
      for(y = boolconj_tail[i]->next; ; y=y->next) {
        (out_trans_tail[i]->prp_trigger_length)++;
        if(y == boolconj_tail[i]) break;
      }
      out_trans_tail[i]->prp_trigger_cycle = y;
    }
    free(boolexp_tail[i]);
    boolexp_tail[i] = NULL;
  }
  return 0;
}
  
/* Install the goto mode name to each of the current_ta_num TA */

int assign_goto_mode(char *name) {
  char *x;
  int i;

  for(i=1; i<=current_ta_num; i++) {
    x = (char *)malloc(strlen(name)+1);
    strcpy(x, name);
    out_trans_tail[i]->goto_mode = x;
  }
  x = NULL;
  return 0;
}

/* Create current_ta_num new assignments */
int new_assign() {
  struct assign_type *x;
  int i;
  for(i=1; i<=current_ta_num; i++) {
    x = alloc_assign();
    if(assign_tail[i]) {
      x->next = (assign_tail[i])->next;
      (assign_tail[i])->next = x;
    }
    else x->next = x;
    assign_tail[i] = x;
  }
  return 0;
}

struct assign_type *alloc_assign() {
  struct assign_type *a;

  a = (struct assign_type *)malloc(sizeof(struct assign_type));
  a->var = 0;
  a->use_var = 0;
  a->op = -1;
  a->value = 0;
  a->mod_num = 0;
  a->mod_var = 0;
  a->next = a;
  return a;
}

int install_assign_param() {
  int i;
  struct variable_type *v;

  for(i=1; i<=current_ta_num; i++) {
    assign_tail[i]->value = i;
    v = var_ptr_from_index(assign_tail[i]->var);
    if(v->par_val == FALSE) v->par_val = TRUE;
  }
  return 0;
}

int install_assign_var(char *name, char *par) {
  int i, v;

  if(par) {
    if(var_ptr(name, EXISTS_VAR) == NULL) {
     fprintf(stderr, "\nError: %s[%s] undeclared in line #%d.\n",
                     name, par, yylineno);
     exit(1);
    }
  }
  else {
    if(var_ptr(name, NON_PARAM) == NULL) {
     fprintf(stderr, "\nError: %s undeclared in line #%d.\n",
                     name, yylineno);
     exit(1);
    }
  }

  if(par) {
    if(sign == 1)
      for(i=1; i<=current_ta_num; i++) 
        assign_tail[i]->var = var_index(name, i);
    else if(sign == -1)
      for(i=1; i<=current_ta_num; i++) 
        assign_tail[i]->use_var = var_index(name, i);
  }
  else {
    v = var_index(name, NON_PARAM);
    if(sign == 1)
      for(i=1; i<=current_ta_num; i++) 
        assign_tail[i]->var = v;
    else if(sign == -1)
      for(i=1; i<=current_ta_num; i++) 
        assign_tail[i]->use_var = v;
  }
  return 0;
}

int install_assign_value() {
  int i;
  for(i=1; i<=current_ta_num; i++) 
    assign_tail[i]->value = last_int;
  return 0;
}

int install_assignment() {
  struct assign_type *x;
  int i;
  for(i=1; i<=current_ta_num; i++) {
    if(assign_tail[i]) {
      for(x = assign_tail[i]->next; ; x=x->next) {
        (out_trans_tail[i]->assign_length)++;
        if(x == assign_tail[i]) break;
      }
      out_trans_tail[i]->assign_cycle = assign_tail[i];
      assign_tail[i] = (struct assign_type *)NULL; 
    }
  }
  return 0;
}

int install_out_trans() {
  int i;
  struct transition_type *tr;
  struct spec_trans_type *str;

  for(i=1; i<=current_ta_num; i++) {
    if(out_trans_tail[i]) {
      for(str = out_trans_tail[i]->next; ; str=str->next) {
        tr = alloc_transition();
        tr->index = ++trans_count;
        tr->trn = str;
        tr->automaton = get_ta(current_ta->name, i);
        if(mode_tail[i]->out_trans_cycle) {
          tr->next = mode_tail[i]->out_trans_cycle->next;
          mode_tail[i]->out_trans_cycle->next = tr;
        }
        else tr->next = tr;
        mode_tail[i]->out_trans_cycle = tr;
        if(str == out_trans_tail[i]) break;
      }
      out_trans_tail[i] = (struct spec_trans_type *)NULL;
    }
  }
  str = NULL;
  tr = NULL;
  return 0;
}

int install_mode() {
  struct automaton *x;
  int i;
  if(current_ta) {
    for(x = current_ta, i = 1; i <= current_ta_num; x=x->next, i++) {
      x->mode_cycle = mode_tail[i];
      mode_tail[i] = (struct mode_type *)NULL;
    }
  }
  free(mode_tail[0]);
  mode_tail[0] = (struct mode_type *)NULL;
  return 0;
}

int install_in_trans() {
  struct automaton *a;
  struct mode_type *m, *x;
  struct transition_type *t, *u;

  if(automaton_tail)
  for(a = automaton_tail->next; ; a = a->next) {
    if(a->mode_cycle)
    for(m = a->mode_cycle->next; ; m = m->next) {
      if(m->out_trans_cycle)
      for(t = m->out_trans_cycle->next; ; t = t->next) {
        for(x = a->mode_cycle->next; ; x = x->next) {
          if(!(strcmp(x->name, t->trn->goto_mode))) { /* m-->x:  t(out),u(in)*/ 
            (x->in_trans_count)++; 
            u = copy_transition(t);
            u->mode = m;
/*
            u->goto_mode = (char *)malloc(1+strlen(m->name));
            strcpy(u->goto_mode, m->name);
*/
            if(x->in_trans_cycle) {
              u->next = x->in_trans_cycle->next;
              x->in_trans_cycle->next = u;
            }
            else u->next = u;
            x->in_trans_cycle = u;
            break;
          }
          if(x == a->mode_cycle) break;
        }
        if(t == m->out_trans_cycle) break;
      }
      if(m == a->mode_cycle) break;
    }
    if(a == automaton_tail) break;
  }
  return 0;
}

struct transition_type *
copy_transition(struct transition_type *t) {
  struct transition_type *u = alloc_transition();
  u->index = t->index;
  u->automaton = t->automaton;
/*
  u->clk_trigger_length = t->clk_trigger_length;
  u->clk_trigger_cycle = clkconj_copy(t->clk_trigger_cycle);
  u->prp_trigger_length = t->prp_trigger_length;
  u->prp_trigger_cycle = boolconj_copy(t->prp_trigger_cycle);
  u->trigger_bdd = t->trigger_bdd;
  bdd_unfree(bddm, t->trigger_bdd);
  u->assign_length = t->assign_length;
  u->assign_cycle = assign_copy(t->assign_cycle);
  u->goto_mode = str_preserve(t->goto_mode);
*/
  u->trn = t->trn;
  u->permutation = t->permutation;
  u->mode = t->mode;
  u->bypassed_internal_vars = t->bypassed_internal_vars;
  return u;
}

/* Install the transition goto-mode pointer:  transition->mode 
 *  from transition->goto_mode (mode name)
 */

int install_trans_mode() {
  struct automaton *a;
  struct mode_type *m, *x;
  struct transition_type *t;
  
  if(automaton_tail)
  for(a = automaton_tail->next; ; a=a->next) {
    if(a->mode_cycle)
    for(m = a->mode_cycle->next; ; m=m->next) {
      if(m->out_trans_cycle) {
        for(t = m->out_trans_cycle->next; ; t=t->next) {
          for(x = a->mode_cycle->next; ; x=x->next) {
            if(!(strcmp(x->name, t->trn->goto_mode))) { /* : t: m-->x */ 
              t->mode = x;
/*
              sprintf(tmpstring, "(OUT)Installed mode %s(%d) as the target mode of transition %d from mode %s(%d).\n", 
                t->mode->name, t->mode->index, t->index, m->name, m->index);
              output(tmpstring);
*/
              break;
            }
            if(x == a->mode_cycle) break;
          }
          if(t == m->out_trans_cycle) break;
        }
      }
/*
      if(m->in_trans_cycle) {
        for(t = m->in_trans_cycle->next; ; t=t->next) {
          for(x = a->mode_cycle->next; ; x=x->next) {
            if(!(strcmp(x->name, t->trn->goto_mode))) { 
              t->mode = x;
              sprintf(tmpstring, "(IN)Installed mode %s(%d) as the source mode of transition %d to mode %s(%d).\n", t->mode->name, t->mode->index, t->index, m->name, m->index);
              output(tmpstring);
              break;
            }
            if(x == a->mode_cycle) break;
          }
          if(t == m->in_trans_cycle) break;
        }
      }
*/
      if(m == a->mode_cycle) break;
    }
    if(a == automaton_tail) break;
  }
  return 0;
}

char *str_preserve(char *str) {
  char *p = NULL;
  if (str) {
    p = (char *)malloc(1 + strlen(str));
    strcpy(p, str);
    return (p);
  }
  else {
    return (char *)NULL;
  }
}

char *itoa(int someint){
  char *c = (char *)malloc(5);
  sprintf(c, "%d",someint);
  return c;
}

char *relop_string(int relop_type) {
  switch(relop_type) {
    case LESS:         return ("<");
    case LEQ:          return ("<=");
    case EQ:           return ("=");
    case GEQ:          return (">=");
    case GREATER:      return (">");
    case NEQ:	       return ("!=");
    default:	       fprintf(stderr, "\nError in relop_string!\n");
                       fprintf(stderr, "\nGot invalid relop: %d\n", relop_type);
  }
  return (char *)NULL;
}


void yywarningmessage(char *warningstr) {
  fprintf(stderr, "\nWarning (line %d): %s\n",yylineno,warningstr);
}


int get_max_const() {
  struct automaton *a;
  struct mode_type *m;
  struct transition_type *t;
  struct clkconj_type *c, *tc;
  int max = 0;

  if(automaton_tail) {
    for(a=automaton_tail->next; ; a=a->next) {
      if(a->mode_cycle) {
        for(m = a->mode_cycle->next; ; m=m->next) {
          if(m->xconj_cycle) {
            for(c = m->xconj_cycle->next; ; c=c->next) {
              if(c->diff > max) max = c->diff;
              if(c == m->xconj_cycle) break;
            }
          } 
          if(m->out_trans_cycle) {
            for(t = m->out_trans_cycle->next; ; t=t->next) {
              if(t->trn->clk_trigger_cycle) {
                for(tc = t->trn->clk_trigger_cycle->next; ; tc=tc->next) {
                  if(tc->diff > max) max = tc->diff;
                  if(tc == t->trn->clk_trigger_cycle) break;
                  }          
              }
              if(t == m->out_trans_cycle) break;
            }
          }
          if(m == a->mode_cycle) break;
        }
      }
      if(a == automaton_tail) break;
    }
  }
  max_const = max;
  return max;
}

/*******************************************************************
 *  Functions for parsing specification 
 *******************************************************************/

int init_specification() {
  vspec = cspec = (struct spec_type *)NULL;
  reset_exp_tails();
  return 0;
}

struct spec_type *alloc_spec() {
  struct spec_type *s = (struct spec_type *)malloc(sizeof(struct spec_type));
  s->ms = (struct mode_spec_type *)NULL;
  s->b1 = s->b2 = (struct boolexp_type *)NULL;
  s->c1 = s->c2 = (struct clkconj_type *)NULL;
  s->s1 = s->s2 = s->next = (struct spec_type *)NULL;
  s->boolcon = 0;
  s->path = 0;
  s->relop = 0;
  s->constant = 0;
  s->sign = POSITIVE;
  s->blevel = blevel;
  s->ustat = -1;
  return s;
}

int new_spec(int type) {
  
  struct spec_type *s = alloc_spec();
  switch(type) {
    case NEGATIVE: 	s->sign = NEGATIVE;
                     	break;
    case POSITIVE:	s->sign = POSITIVE;
                        break;
    case ATOM:		s->sign = ATOM;
                        break;
    default: 	sprintf(tmpstring,
                     "Error: illegal type (%d) of new specification.\n", type);
                output(tmpstring);
                exit;
  }
  if(vspec) {
    if(cspec->boolcon == AND || cspec->boolcon == OR) {
      s->ustat = cspec->ustat;
      s->next = cspec->next;
      cspec->next = s;
    }
    else if(cspec->sign == NEGATIVE || cspec->sign == POSITIVE ||
            cspec->stype == HENCEFORTH) {
      s->ustat = cspec->ustat;
      s->blevel = ++blevel;
      cspec->s1 = s;
      s->next = s;
    }
    else if(cspec->stype == EVENTUALLY) { 
      s->ustat = cspec->ustat;
      s->blevel = ++blevel;
      cspec->s2 = s;
      s->next = s;
    }
    else if(cspec->stype == UNTIL) {
      s->ustat = USECOND;
      s->blevel = ++blevel;
      cspec->s2 = s;
      s->next = s;
    }
    else {
      s->ustat = UFIRST;
      s->blevel = ++blevel;
      cspec->s1 = s;
      s->next = s;
    }
  }
  else {
    vspec = s;
    vspec->next = vspec;
  }
  cspec = s;
  return 0;
}

int init_spec_atom(void) {
  reset_exp_tails();
  total_exp = 1;
  return 0;
}

struct spec_type *set_cspec(void) {
  struct spec_type *t = (struct spec_type *)NULL, *s;
  if(cspec->blevel == blevel) return cspec;
  else {
    do {
      if(t) {
        if(cspec->ustat == USECOND && t->s2) {
          for(s=t=t->s2; ; t=t->next) if(t->next == s) break;
        }
        else {
          for(s=t=t->s1; ; t=t->next) if(t->next == s) break;
        }
      }
      else {
        for(t=vspec; ; t=t->next) if(t->next == vspec) break;
      }
    } while (t->blevel != blevel);
  }
  return t;
}

/*
struct spec_type *set_cspec(struct spec_type *u) {
  struct spec_type *t;

  if(u->boolcon == AND || u->boolcon == OR) {
    if(u->s1 == cspec) return u;
    else if(u->s2 == cspec) return NULL;
    else if(t = set_cspec(u->s1)) return t;
    else if(t = set_cspec(u->s2)) return t;
  }
  else return set_cspec(u->s1);
}
*/

int install_exp(void) {
  if(boolexp_tail[1] && clkconj_tail[1]) {
    if(cspec->b1 == NULL && cspec->c1 == NULL) {
      cspec->b1 = boolexp_tail[1];
      cspec->c1 = clkconj_tail[1];
    }
    else {
      cspec->b2 = boolexp_tail[1];
      cspec->c2 = clkconj_tail[1];
    }
  }
  else if(boolexp_tail[1]) {
    if(cspec->b1 == NULL && cspec->c1 == NULL) {
      cspec->b1 = boolexp_tail[1];
    }
    else {
      cspec->b2 = boolexp_tail[1];
    }
  }
  else if(clkconj_tail[1]) {
    if(cspec->b1 == NULL && cspec->c1 == NULL) {
      cspec->c1 = clkconj_tail[1];
    }
    else {
      cspec->c2 = clkconj_tail[1];
    }
  }
  return 0;
}

int new_mode_spec() {
  struct mode_spec_type *ms;
  ms = (struct mode_spec_type *)malloc(sizeof(struct mode_spec_type));
  ms->auto_name = ms->mode_name = (char *)NULL;
  ms->a = (struct automaton *)NULL;
  cspec->ms = ms;
  return 0;
}

int install_auto_name(int n) {
  char *tmps;
  if(n==0)
    cspec->ms->auto_name = str_preserve(string_buffer);
  else if(n==-1) {
    cspec->ms->auto_name = (char *)malloc(sizeof(string_buffer)+3);
    cspec->ms->auto_name = strcpy(cspec->ms->auto_name, string_buffer);
    cspec->ms->auto_name = strcat(cspec->ms->auto_name, "[i]");
  }
  else { /* it is an integer numbered automaton */
    cspec->ms->auto_name = (char *)malloc(sizeof(string_buffer)+3);
    cspec->ms->auto_name = strcpy(cspec->ms->auto_name, string_buffer);

    if(last_int <= 0 || last_int > current_ta_num) {
      printf("\nError: illegal automaton index (%d) found in line %d.\n", 
             last_int, yylineno);
      exit(1);
    }
    if(last_int <= 9)
      tmps = (char *)malloc(4);
    else if(last_int <= 99)
      tmps = (char *)malloc(5);
    else 
      tmps = (char *)malloc(6);
    sprintf(tmps, "[%d]", last_int);
    cspec->ms->auto_name = strcat(cspec->ms->auto_name, tmps);
    free(tmps);
  }
  return 0;
}

char *bool_string(int b) {
  char *c = NULL;
  switch(b) {
    case AND: c = str_preserve(" and ");
              break;
    case OR:  c = str_preserve(" or ");
              break;
    case NOT: c = str_preserve(" not ");
              break;
    default:  printf("Error: illegal boolean value (%d)\n", b);
  }
  return c;
}

char *automaton_fullname(struct automaton *a) {
  char *c;
  if(current_ta_num < 10) 
    c = (char *)malloc(sizeof(char) * (strlen(a->name) + 3));
  else if(current_ta_num < 100)
    c = (char *)malloc(sizeof(char) * (strlen(a->name) + 4));
  else 
    c = (char *)malloc(sizeof(char) * (strlen(a->name) + 5));
  c = strcpy(c, a->name);
  sprintf(tmpstring, "[%d]", a->index);
  c = strcat(c, tmpstring);
  return c;
}

/* Functions for manipulating state-graphs */
int init_action_list() {
  action_list = (struct action_type *)NULL;
  curr_action = (struct action_type *)NULL;
  tmp_action = (struct action_type *)NULL;
  return 0;
}

int init_graph_list() {
  graph_list = (struct thread_type *)NULL;
  return 0;
}

int exists_graph(char *name, int n) {
  int i;
  struct thread_type *t;

  if(graph_list) {
    for(t=graph_list->next; ; t=t->next) {
      if(!strcmp(t->name, name) && t->automaton_index == n)
        return TRUE;
      if(t == graph_list) break;
    }
  }
  if(thread)
  for(i=0; i<2*system_ta_num-1; i++) {
    if(!strcmp(thread[i].name, name) && thread[i].automaton_index == n) 
      return TRUE;
  }
  return FALSE;
}

/* interval_start == -1 ==>  a non-indexed graph name (string_buffer)
 * interval_start >= 0  ==>  indexed graphs:
 *                             interval_varname[interval_start..interval_end]
 */
int new_graph() {
  int i;
  struct thread_type *t=NULL;

  for(i=interval_start; i<=interval_end; i++) {
    if(exists_graph(interval_varname, i)) {
      if(SGM_MODE == GUI || SGM_MODE == PROCESS) {
        GUI_ERROR = E_DECL;
        sprintf(tmpstring,"Error: graph %s[%d] already declared before.\n",
                          interval_varname, i);
        append_string(tmpstring, &answer);
        sprintf(tmpstring,"\tIgnoring entire new declaration.\n");
        append_string(tmpstring, &answer);
      }
      else {
	if(i >= 0)
          fprintf(stderr, "Error:  graph %s[%d] already declared before.\n", 
                        interval_varname, i);
	else
          fprintf(stderr, "Error:  graph %s already declared before.\n", 
                        interval_varname);
        fprintf(stderr, "\tIgnoring entire new declaration.\n");
      }
      return FALSE;
    }
  }
  
  if(interval_start == -1) {
    t = alloc_thread();
    t->index = ++decl_graph_num;
    t->name = str_preserve(interval_varname);
    if(graph_list) {
      t->next = graph_list->next;
      graph_list->next = t;
    }
    else
      t->next = t;
    graph_list = t;
  }
  else if(interval_start >= 0) {
    for(i=interval_start; i<=interval_end; i++) {
      t = alloc_thread();
      t->index = ++decl_graph_num;
      t->name = str_preserve(interval_varname);
      t->automaton_index = i;
      if(graph_list) {
        t->next = graph_list->next;
        graph_list->next = t;
      }
      else
        t->next = t;
      graph_list = t;
      if(SGM_MODE == GUI || SGM_MODE == PROCESS) {
        append_string(print_size(t), &answer);
      }
    }
  }
  return 0;
}

int delete_graph(struct thread_type *g) {
  struct thread_type *t, *pt;
  if(graph_list) {
    for(pt=graph_list, t=pt->next; ; pt=t, t=t->next) {
      if(!strcmp(t->name, g->name) &&
         t->automaton_index == g->automaton_index) {
         pt->next = t->next;
         if(graph_list == t) {
           if(t == pt) graph_list = NULL;
           else graph_list = pt;
         }
         free(t->ta_list); free(t); t = NULL;
        return TRUE;
      }
      if(t == graph_list) break;
    }
  }
  return 0;
}

struct thread_type *alloc_thread() {
  struct thread_type *t;
  int i;

  t = (struct thread_type *)malloc(sizeof(struct thread_type));
  t->name = (char *)NULL;
  t->index = -1;
  t->subthread1 = t->subthread2 = (struct thread_type *)NULL;
  t->automaton_index = -1;
  t->num_ta = 0;
  t->mode_count = t->trans_count = t->invalid_trans_count = 0;
  t->mnpl = 0;
  t->total_sec = 0;
  t->total_usec = 0;
  t->last_sec = 0;
  t->last_usec = 0;
  t->starting_mode = (struct mode_type *)NULL;
  t->invalid_cycle = (struct invalid_mode_type *)NULL;
  t->init_boolexp = (struct boolexp_type *)NULL;
  t->init_clkconj = (struct clkconj_type *)NULL;
  t->modetree = (struct a23tree_type *)NULL;
  t->next = (struct thread_type *)NULL;
  t->ta_list = (int *)malloc(sizeof(int)*system_ta_num);;
  for(i=0; i<system_ta_num; i++) {
    t->ta_list[i] = FALSE;
  }
  return t;
}

struct action_type *alloc_action() {
  struct action_type *a;
  a = (struct action_type *)malloc(sizeof(struct action_type));
  a->type = 0;
  a->if_relop = a->if_num = 0;
  a->for_init_num = a->for_cond_relop = a->for_cond_num = 0;
  a->for_step = 0;
  a->g0 = a->g1 = a->g2 = (struct thread_type *)NULL;
  a->n0 = a->n1 = a->n2 = (char *)NULL;
  a->c0 = a->c1 = a->c2 = -1;
  a->p0 = a->p1 = a->p2 = -1;
  a->if_list = a->else_list = a->loop_list = a->parent = a->next = 
           (struct action_type *)NULL;
  return a;
}

int free_action(struct action_type *a) {
  if(a->if_list) free_action(a->if_list);
  if(a->else_list) free_action(a->else_list);
  if(a->loop_list) free_action(a->loop_list);
  if(a->parent) free_action(a->parent);
  if(a->n0) free(a->n0);
  if(a->n1) free(a->n1);
  if(a->n2) free(a->n2);
  free(a);
  return 0;
}

int new_action() {
  struct action_type *a = alloc_action();
  if(curr_action) {
    a->next = curr_action->next;
    curr_action->next = a;
  }
  else a->next = a;
  curr_action = a;
  if(action_list == NULL) action_list = curr_action;
  if(tmp_action) {
    if(tmp_action->type == IE) {
      if(tmp_action->if_list == NULL)
        tmp_action->if_list = curr_action;
      else
        tmp_action->else_list = curr_action;
    }
    else if(tmp_action->type == FL) {
      tmp_action->loop_list = curr_action;
    }
    curr_action->parent = tmp_action;
    tmp_action = (struct action_type *)NULL;
  }
  return 0;
}

int install_graph(int n) {
  switch(n) {
    case 0: curr_action->n0 = str_preserve(string_buffer);
            curr_action->c0 = last_int;
            if(p_num != 0) {
              curr_action->p0 = p_num;
              p_num = 0;
            }
            break;
    case 1: curr_action->n1 = str_preserve(string_buffer);
            curr_action->c1 = last_int;
            if(p_num != 0) {
              curr_action->p1 = p_num;
              p_num = 0;
            }
            break;
    case 2: curr_action->n2 = str_preserve(string_buffer);
            curr_action->c2 = last_int;
            if(p_num != 0) {
              curr_action->p2 = p_num;
              p_num = 0;
            }
            break;
  }
  return 0;
}


/* FIND thread pointer for name[id] */

struct thread_type *thread_ptr(char *name, int id) {
  int i;
  for(i=0; i<2*system_ta_num-1; i++) {
    if(thread[i].automaton_index == id && !strcmp(thread[i].name, name))
      return (&thread[i]);
  }
/*
  if(id > 0)
    fprintf(stderr, "\nThread %s[%d] not found.\n", name, id);
  else if(id == -1)
    fprintf(stderr, "\nThread %s[i] not found.\n", name);
  else
    fprintf(stderr, "\nThread (name = %s, id = %d) not found.\n", name, id);
*/
  return NULL;
}


/* FIND graph pointer for name[id] */

struct thread_type *graph_ptr(char *name, int id) {
  struct thread_type *t;
  if(graph_list) {
    for(t = graph_list->next; ; t=t->next) {
      if(!strcmp(name, t->name) && id == t->automaton_index) return t;
      if(t == graph_list) break;
    }
/*
    if(id > 0)
      fprintf(stderr, "\nGraph %s[%d] not found in declared graph list.\n", 
                      name, id);
    else 
      fprintf(stderr, "\nGraph %s not found in declared graph list.\n", name);
*/
  }
  return NULL;
}

int check_relop(int type1, int type2) {
  if(type1 != type2)
    fprintf(stderr, "\nWrong relation operator (%s) on line %d.\n",
                       relop_string(type1), yylineno);
  return 0;
}

int init_action_management(void) {
  list_update(action_list);
  return 0;
}

int list_update(struct action_type *alist) {
  struct action_type *a;
  for(a = alist->next; ; a=a->next) {
    if(a->type == IE) {
      if(a->if_list) list_update(a->if_list);
      if(a->else_list) list_update(a->else_list);
    }
    if(a->type == FL) list_update(a->loop_list);
    action_update(a);
    if(a == alist) break;
  }
  return 0;
}

ERROR action_update(struct action_type *a) {
  if(a->c0 != -2 && a->n0) {
    if(a->g0 == NULL)  a->g0 = thread_ptr(a->n0, a->c0);
    if(a->g0 == NULL)  a->g0 = graph_ptr(a->n0, a->c0);
    if(a->g0 == NULL) {
      if(a->c0 >= 0) {
        if(SGM_MODE != GUI && SGM_MODE != PROCESS)
          fprintf(stderr, "\nError: graph %s[%d] was not declared.\n", 
                          a->n0, a->c0);
      }
      else if(a->c0 == -1) {
        if(SGM_MODE != GUI && SGM_MODE != PROCESS)
          fprintf(stderr, "\nError: graph %s was not declared.\n", 
                          a->n0);
      }
      return E_UNDECL;
    }
  }

  if(a->c1 != -2 && a->n1) {
    if(a->g1 == NULL)  a->g1 = thread_ptr(a->n1, a->c1);
    if(a->g1 == NULL)  a->g1 = graph_ptr(a->n1, a->c1);
    if(a->g1 == NULL) {
      if(a->c1 >= 0) {
        if(SGM_MODE != GUI && SGM_MODE != PROCESS)
          fprintf(stderr, "\nError: graph %s[%d] was not declared.\n", 
                          a->n1, a->c1);
      }
      else if(a->c1 == -1) {
        if(SGM_MODE != GUI && SGM_MODE != PROCESS)
          fprintf(stderr, "\nError: graph %s was not declared.\n", 
                          a->n1);
      }
      return E_UNDECL;
    }
  }
  if(a->c2 != -2 && a->n2) {
    if(a->g2 == NULL)  a->g2 = thread_ptr(a->n2, a->c2); 
    if(a->g2 == NULL)  a->g2 = graph_ptr(a->n2, a->c2);
    if(a->g2 == NULL) {
      if(a->c2 >= 0) {
        if(SGM_MODE != GUI && SGM_MODE != PROCESS)
          fprintf(stderr, "\nError: graph %s[%d] was not declared.\n", 
                          a->n2, a->c2);
      }
      else if(a->c2 == -1) {
        if(SGM_MODE != GUI && SGM_MODE != PROCESS)
          fprintf(stderr, "\nError: graph %s was not declared.\n", 
                          a->n2);
      }
      return E_UNDECL;
    }
  }
  return E_NONE;
}

struct thread_type *get_graph(char *fullname) {
  char *realname = NULL;
  int start=-1;
  int end=-1;
  struct thread_type *t = NULL;
  if(fullname == NULL) {
    fprintf(stderr, "Error: get_graph(NULL) encountered.\n");
    return NULL;
  }
  get_gd_info(fullname, &realname, &start, &end);
  t = thread_ptr(realname, start);
  if(!t) t = graph_ptr(realname, start);
  free(realname);
  return t;
}

int remove_duplicate_trans(void) {
  struct automaton *a;
  struct mode_type *m;
  struct transition_type *t1, *t2, *st2;
  if(automaton_tail) {
    for(a=automaton_tail->next; ; a=a->next) {
      if(a->mode_cycle) {
        for(m = a->mode_cycle->next; ; m=m->next) {
          if(m->out_trans_cycle) {
            for(t1 = m->out_trans_cycle->next; ; t1=t1->next) {
              for(t2 = m->out_trans_cycle->next, st2=t2->next; ; 
                  t2 = st2, st2 = st2->next) {
                if(t1->index == t2->index) {
                  if(t2 == m->out_trans_cycle) break;
                  else continue;
                }
                if(spec_trans_comp(t1->trn, t2->trn) == EQUAL) {
                  eliminate_transition(t2, m); 
                  if(m->out_trans_cycle==NULL || st2==m->out_trans_cycle->next)
                    break;
                }
                if(t2 == m->out_trans_cycle) break;
              }
              if(t1 == m->out_trans_cycle) break;
            }
          }
          if(m == a->mode_cycle) break;
        }
      }
      if(a == automaton_tail) break;
    }
  }
  return 0;
}

int install_mod_num() {
  int i;
  struct boolconj_type *b;
  struct boolexp_type *e;

  if(curr_exp_type == CLK_INEQ) {
    fprintf(stderr, 
    "Error (line #%d): MOD not allowed in a clock expression.\n", yylineno);
    fprintf(stderr, "Exiting ...\n");
    exit(1);
  }
  if(index_relop == -1) {
    for(i=1; i<=total_exp; i++)
      boolconj_tail[i]->mod_num = last_int;
  }
  else {
    if(index_quantifier == FORALL)
      for(i=1; i<=total_exp; i++) {
        if(boolconj_expand_start[i]) {
          for(b=boolconj_expand_start[i]; ; b=b->next) {
            if(b) b->mod_num = last_int;
            if(b == boolconj_tail[i]) break;
          }
        }
      }
    else if(index_quantifier == EXISTS)
      for(i=1; i<=total_exp; i++) {
        if(boolexp_expand_start[i]) {
          for(e=boolexp_expand_start[i]; ; e=e->next) {
            if(e) e->boolconj->mod_num = last_int;
            if(e == boolexp_tail[i]) break;
          }
        }
      }
  }
  return 0;
}

/* n = 0  ==>  mod var
   n = 1  ==>  mod var[num]
   n = 2  ==>  mod var[param]
 */
int install_mod_var(int n) {
  int i, v=0;
  struct boolconj_type *b;
  struct boolexp_type *e;

  if(n == 0) v = var_index(string_buffer, NON_PARAM);
  else if(n == 1) v = var_index(interval_varname, last_int);

  if(curr_exp_type == CLK_INEQ) {
    fprintf(stderr, 
    "Error (line #%d): MOD not allowed in a clock expression.\n", yylineno);
    fprintf(stderr, "Exiting ...\n");
    exit(1);
  }
  if(index_relop == -1) {
    for(i=1; i<=total_exp; i++)
      if(n == 0 || n == 1) boolconj_tail[i]->mod_var = v;
      else if(n == 2) boolconj_tail[i]->mod_var = var_index(interval_varname,i);
  }
  else {
    if(index_quantifier == FORALL)
      for(i=1; i<=total_exp; i++) {
        if(boolconj_expand_start[i]) {
          for(b=boolconj_expand_start[i]; ; b=b->next) {
            if(b) {
              if(n == 0 || n == 1) b->mod_var = v;
              else if(n == 2) b->mod_var = var_index(interval_varname, i);
            }
            if(b == boolconj_tail[i]) break;
          }
        }
      }
    else if(index_quantifier == EXISTS)
      for(i=1; i<=total_exp; i++) {
        if(boolexp_expand_start[i]) {
          for(e=boolexp_expand_start[i]; ; e=e->next) {
            if(e) {
              if(n == 0 || n == 1) e->boolconj->mod_var = v;
              else if(n == 2) 
                e->boolconj->mod_var = var_index(interval_varname, i);
            }
            if(e == boolexp_tail[i]) break;
          }
        }
      }
  }
  return 0;
}

int install_assign_mod_num() {
  int i;

  for(i=1; i<=current_ta_num; i++) {
    assign_tail[i]->mod_num = last_int;
  }
  return 0;
}

/* n = 0  ==>  mod var
   n = 1  ==>  mod var[num]
   n = 2  ==>  mod var[param]
 */
int install_assign_mod_var(int n) {
  int i, v=0;

  if(n == 0){
    if(var_ptr(interval_varname, NON_PARAM) == NULL) {
     fprintf(stderr, "\nError: %s undeclared in line #%d.\n",
                     interval_varname, yylineno);
     exit(1);
    }
  }
  else if(n == 1){
    if(var_ptr(interval_varname, last_int) == NULL) {
     fprintf(stderr, "\nError: %s[%d] undeclared in line #%d.\n",
                     interval_varname, last_int, yylineno);
     exit(1);
    }
  }
  else if(n == 2) {
    if(var_ptr(interval_varname, EXISTS_VAR) == NULL) {
     fprintf(stderr, "\nError: %s[%s] undeclared in line #%d.\n",
                    interval_varname, parameter, yylineno);
     exit(1);
    }
  }

  if(n == 0) v = var_index(interval_varname, NON_PARAM);
  else if(n == 1) v = var_index(interval_varname, last_int);

  for(i=1; i<=current_ta_num; i++) {
    if(n == 0 || n == 1) assign_tail[i]->mod_var = v;
    else if(n == 2) assign_tail[i]->mod_var = var_index(interval_varname, i);
  }
  return 0;
}

int check_automata_defn() {
  struct automaton *a;
  if(automaton_tail == NULL) {
    fprintf(stderr, "\nNo automaton defined! Exiting ...\n");
    exit(1);
  }
  for(a=automaton_tail->next; ; a=a->next) {
    if(a->init_mode == NULL) {
      if(a->index >= 0)
      fprintf(stderr, 
            "\nError: automaton %s[%d] declared but NOT defined. Exiting ...\n",
              a->name, a->index);
      else if(a->index == -1)
        fprintf(stderr, 
               "\nError: automaton %s declared but NOT defined. Exiting ...\n",
              a->name);
      exit(1);
    }
    if(a == automaton_tail) break;
  }
  return 0;
}

/* buggy */
struct mode_type *source_mode1(struct transition_type *tr) {
  struct mode_type *m;
  struct transition_type *tt;

  if(NEW_THREAD)
  for(m=NEW_THREAD->starting_mode; m; m=m->cnext) {
    if(m->out_trans_cycle)
    for(tt=m->out_trans_cycle->next; ;tt=tt->next) {
      if(tt->index == tr->index) return m;
      if(tt == m->out_trans_cycle) break;
    }
  }
  return NULL;
}

struct mode_type *source_mode(struct transition_type *tr) {
  struct transition_type *tt;

  for(tt = tr->mode->out_trans_cycle->next; ; tt = tt->next) {
    if(tt->index == tr->index) return tt->mode;
    if(tt == tr->mode->out_trans_cycle) break;
  }
  fprintf(stderr, "Fatal error in source_mode (include.c).\n");
  return NULL;
}

/* Synchronization Functions */

int init_sync() {
  int i;

  sync_cycle = (struct var_link_type **)
               malloc((1+current_ta_num) * sizeof(struct var_link_type *));
  for(i=0; i<=current_ta_num; i++)
    sync_cycle[i] = (struct var_link_type *)NULL;
  return 0;
}

int sync_index(char *name, int id){
  struct variable_type *x;
  if(sync_tail) {
    for(x=sync_tail->next; ; x=x->next) {
      if(!strcmp(x->name, name) && x->id == id) 
        return x->index;
      if(x == sync_tail) break;
    }
  }
  return -1;  /* not found */
}

struct variable_type *sync_ptr(char *name, int id) {
  struct variable_type *x;
  if(sync_tail) {
    for(x=sync_tail->next; ; x=x->next) {
      if(!strcmp(x->name, name) && 
         (x->id == id || id==EXISTS_VAR)
        )
        return x;
      if(x == sync_tail) break;
    }
  }
  return (struct variable_type *)NULL; /* not found */
}

int insert_sync(char *sn) {
  int i;
  struct variable_type *x;
  struct var_link_type *vl;
  
  if(-1 == sync_index(sn, -1)) { /* A NEW synclab */
    sync_count++;
    x = (struct variable_type *)malloc(sizeof(struct variable_type));
    x->index = sync_count;
    x->id = NON_PARAM;   /* Not a parametric variable like x[i] */
    x->par_val = FALSE;  /* Does not have a parametric value */
    x->name = (char *)malloc(1 + strlen(sn));
    strcpy(x->name, sn);
  
    if(sync_tail) {
      x->next = sync_tail->next;
      sync_tail->next = x;
    }
    else x->next = x;
    sync_tail = x;
  }
  else {
    x = sync_ptr(sn, EXISTS_VAR);
  }
  for(i=1; i<=current_ta_num; i++) {
    vl = (struct var_link_type *)malloc(sizeof(struct var_link_type));
    vl->var = x;
    if(sync_cycle[i]) {
      vl->next = sync_cycle[i]->next;
      sync_cycle[i]->next = vl;
    }
    else vl->next = vl;
    sync_cycle[i] = vl;
  }
  return 0;
} 

int install_sync() {
  struct automaton *x;
  int i;
  if(current_ta) {
    for(x = current_ta, i = 1; i <= current_ta_num; x=x->next, i++) {
      x->sync_cycle = sync_cycle[i];
      sync_cycle[i] = (struct var_link_type *)NULL;
    }
  }
  free(sync_cycle[0]);
  sync_cycle[0] = (struct var_link_type *)NULL;
  return 0;
}

int init_tr_sync() {
  int i;
  tr_sync_cycle = (struct var_link_type **)
                  malloc((1+current_ta_num) * sizeof(struct var_link_type *));
  for(i=0; i<=current_ta_num; i++)
    tr_sync_cycle[i] = (struct var_link_type *)NULL;
  return 0;
}

int insert_trans_sync(char *name) {
  int i;
  struct var_link_type *vl;

  for(i=1; i<=current_ta_num; i++) {
    vl = (struct var_link_type *)malloc(sizeof(struct var_link_type));
    if( (vl->var = sync_ptr(name, EXISTS_VAR)) == NULL) {
     fprintf(stderr,"Error (line %d): synchronization label %s not declared.\n",
                       yylineno, name);
      free(vl);
      exit(1);
    }
    if(tr_sync_cycle[i]) {
      vl->next = tr_sync_cycle[i]->next;
      tr_sync_cycle[i]->next = vl;
    }
    else vl->next = vl;
    tr_sync_cycle[i] = vl;
  }
  return 0;
}

int install_trans_sync() {
  int i;
  for(i=1; i<=current_ta_num; i++) {
    out_trans_tail[i]->sync_cycle = tr_sync_cycle[i];
    tr_sync_cycle[i] = (struct var_link_type *)NULL;
  }
  return 0;
}

int free_assign_cycle(struct assign_type **ac) {
  struct assign_type *a;
  if(*ac) {
    a = (*ac)->next;
    (*ac)->next = NULL;
    (*ac) = a;
    a = a->next;
    while(1) {
      free(*ac);
      if(a == NULL) break;
      else {
        *ac = a;
        a = a->next;
      }
    }
  }
  return 0;
}
