#ifndef __INCLUDE_H
#define __INCLUDE_H
#define __EXTENSIONS__

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <malloc.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "memuser.h"
#include "bdduser.h"
#include "setman.h"
#include "setman.e"
#include "treeman.h"
#include "treeman.e"

#ifdef __SUNOS
extern printf();
extern fprintf();
extern fflush();
extern _filbuf();
extern fclose();
extern fwrite();
extern ungetc();
extern perror();
extern system();
#endif

#define FALSE                   0
#define TRUE                    1

#define EXISTS_VAR		-2
#define NON_PARAM		-1

#define NON_INDEXED		0
#define INDEXED			1

#define OVERLAPPING             2
#define DIFFERENT		3

#define START			6
#define END			7

#define ADD			8
#define SUB			9
#define NOOP			10

#define LESS                    11
#define LEQ                     12
#define EQ                      13
#define GEQ                     14
#define GREATER                 15
#define NEQ                     16

#define FORALL			17
#define EXISTS			18

#define AND                     21
#define OR                      22
#define NOT                     23

#define NO_PATH			 0
#define EXISTS_PATH		31
#define ALL_PATHS		32

#define NEGATIVE		41
#define POSITIVE		42
#define ATOM			43

#define UNTIL			51
#define HENCEFORTH		52
#define EVENTUALLY		53

#define UFIRST			61
#define USECOND			62

#define NUM_MODES		66
#define NUM_TRANS		67

#define MG			1	/* MERGE_GRAPH */
#define SC			2	/* SHIELD_CLOCK */
#define NR			4	/* NORMALIZE_REGION */
#define RW			8	/* READWRITE REDUCTION */
#define TB			16	/* TRANSITION_BYPASS */
#define SV			32	/* HIDE_VARIABLES */

#define TRW			69	/* TOGGLE READ-WRITE (on MERGE) */
#define XG			70	/* RENAME_GRAPH */
#define FG			71	/* FREE_GRAPH */
#define RG			72	/* RESET_GRAPH */
#define EG			73	/* EMPTY_GRAPH */
#define VG			74	/* VIEW_GRAPH */
#define KG			75	/* KEEP_GRAPHS */

#define VI			76	/* VI */
#define EMACS			77	/* EMACS */
#define LA			78	/* LOOK_AHEAD */
#define MC			79	/* MODEL_CHECK */
#define PG			80	/* PRINT_GRAPH */
#define IG			81	/* INFO_GRAPH */
#define SI			82	/* SHORT_INFO */

#define PS			83	/* PRINT_SIZE */
#define SM			84	/* SEQUENTIAL_MERGE */
#define PM			85	/* PAIRWISE_MERGE */
#define IE			86	/* IF_ELSE */
#define FL			87	/* FOR_LOOP */
#define VL			88	/* VERBOSITY LEVEL */
#define TS			89	/* TOGGLE SEMANTICS */

#define GD			90	/* Graph Declaration */
#define LOAD			91      /* Load specification input file */
#define SETENV			92      /* Set environment variables */
#define PROG_INIT		93      /* Program Initialization */

#define HELP			94      /* Help interactive user */

#define CD			95	/* Change Directory */
#define LS			96	/* List Directory Files */
#define PWD                     97      /* Present Working Directory */

#define LE	                98      /* LAST ERROR */
#define QP	                99      /* QUIT_PROGRAM */

#define CONTAINED       	-2
#define INCOMPARABLE    	-1
#define LATER			-1
#define EQUAL            	 0
#define EARLIER			 1
#define CONTAINS         	 2
#define BOOB         	         3

#define NO_OVERLAP		-1
#define OVERLAP	 	 	 1

#define CLK_INEQ                  0
#define BOOL_EXP                  1

#define MAX_CHAR                200
#define MAX_LINE 	 	 30
#define MAX_NAME_LEN             25

#define V_CLOCK	  		  0
#define V_REGISTER		  1
#define V_SYNC			  2

/*
#define WRITE_0_TRANS		  4
#define WRITE_I_TRANS		  5
#define READ_0_TRANS		  6
#define READ_I_TRANS		  7

#define FIRST_WRITE	  	  8
#define SECOND_WRITE		  9
*/

#define INF			10000

typedef enum {
  BATCH,        /* batch mode */
  INTERACTIVE,  /* interactive mode */
  GUI,          /* GUI mode */
  PROCESS       /* PROCESS mode */
} MODE;

struct variable_type {
  char 			*name;		/* x[i] => x */
  int 			index, 		/* a unique index */
			type,		/* V_CLOCK, V_REGISTER, or V_SYNC */
                        id, 		/* x[i] => id = i */
			par_val;	/* if it ever has a parametric value */
  struct variable_type 	*next;		/* next variable on cycle */
} ;

struct var_link_type {
  struct variable_type *var;
  struct var_link_type *next;
} ;

struct var_link_link_type {
  struct var_link_type *vl;
  struct var_link_link_type *next;
} ;

struct automaton {
  char 			*name;		/* name of automaton */
  char			*par;		/* parametric index: i, j, k, ... */
  int 			index;		/* g[n] ==> name = "g" & index = n */
                                        /* g    ==> name = "g" & index = -1 */
  int 			u_index;	/* a UNIQUE index: 0, 1, 2, ... */
  char 			*init_mode;	/* initial mode name */
  struct var_link_type 	*sync_cycle;	/* list of synclabs in a cycle */
  struct clkconj_type	*init_clkconj;	/* initial clock constraints */
  struct boolexp_type	*init_boolexp;  /* initial boolean expression */
  struct mode_type 	*mode_cycle;	/* list of modes in a cycle */
  struct automaton 	*next;		/* next automaton */
} ;

/* A clock conjunct (clkconj) is an inequality of form:
      'clk1 - clk2 op diff'  or  'clk1 op diff' */

struct clkconj_type	{
  int			clk1,   	/* clock 1 */
			clk2,   	/* clock 2 */
			op, 		/* LESS (11), LEQ (12), EQ= (13), 
                                 	 * GEQ (14), GREATER (15) */ 
			diff; 		/* an integer constant */
  struct clkconj_type	*next; 
} ;

/* A boolconj is a conjunctive list of boolean conditions: 
   'var1 - var2 op constant (mod mod_num)'
   'var1 - var2 op constant (mod mod_var)'
 */

struct boolconj_type    	{
  int				var1,
  				var2,
				op,        /* LESS (11), LEQ (12), EQ (13), 
                                 	    * GEQ (14), GREATER (15) */ 
                        	constant,
				mod_num,
				mod_var;
  bdd   			prop_bdd;
  struct boolconj_type  	*next;
} ;

/* A boolexp is a disjunctive list of boolean conjuncts (boolconj) */

struct boolexp_type     {
  bdd			boolexp_bdd;/* BDD for boolean expression */
  struct boolconj_type  *boolconj;  /* conjunctive list of boolean conjuncts */
  struct boolexp_type   *next;      /* pointer to next on disjunctive list */
} ;

struct mode_type		{
  char 				*name;		  /* mode name */
  char 				**mode_vec;	  /* mode vector */
                       /* mode_vec[0], ..., mode_vec[system_ta_num-1]: 
                          thread[0], ..., thread[system_ta_num-1] mode names */
  int				index;            /* a unique index */
  struct a23tree_type		*parent; 	  /* used by 23tree */
  int				count;		  /* used by 23tree */
  int 				status; 	 
/*
  int				*accessed;	*/  /* to store access status */
/*
  int				*history;	*/ /* to store history status */
  struct permutation_type	*norm_perm;	  /* permutation */
  struct mode_type 		*submode1, 	  /* component state 1 */
                                *submode2;        /* component state 2 */
  int				xconj_length;  	  /* clock conjunct length */
  struct clkconj_type		*xconj_cycle;     /* clock conjunct cycle */
  struct boolexp_type		*boolexp; 	  /* boolean expression */
  struct boolconj_type		*dvl;		  /* discrete var literal set */
  struct clkconj_type		*cvl;		  /* clock var literal set */
  int				in_trans_count,   /* in transition count */
                                out_trans_count;  /* out transition count */
  struct transition_type 	*in_trans_cycle,  /* in transition cycle */
                                *out_trans_cycle; /* out transition cycle */
  struct zone_type		*zone;		  /* zone of mode */
  bdd				dval_bdd;	  /* discrete vars valuation */
  struct var_link_type		*shielded_clocks; /* set of shielded clocks */
  struct var_link_type		*reqd_clocks;     /* set of required clocks */
  struct mode_type 		*cnext;		  /* next on mode list  */
  struct mode_type 		*next;		  /* next on mode list 
                                                     used for current stack */
} ;

struct assign_type	{               
  int			var,        /* (1) var := value mod mod_num*/
                        use_var,    /* (2) var := use_var op value mod mod_var*/
			op, 	    /* LESS (11), LEQ (12), EQ= (13), 
                                     * GEQ (14), GREATER (15) */ 
			value, 
			mod_num,
			mod_var;
  struct assign_type	*next; 
} ; 

struct spec_trans_type {
  int				index;
  char 				*goto_mode;
  int                           clk_trigger_length;
  struct clkconj_type           *clk_trigger_cycle;
  int                           prp_trigger_length;
  struct boolconj_type          *prp_trigger_cycle;
  bdd                           trigger_bdd;
  int                           assign_length;
  struct assign_type            *assign_cycle;
  struct var_link_type		*sync_cycle;
  struct spec_trans_type	*next;
} ;

struct transition_type	{
  int				index; 
  struct automaton		*automaton; 
/*struct a23tree_type           *parent; */       /* used by 23tree */
/*int                           count;   */       /* used by 23tree */
  struct spec_trans_type	*trn;
  struct mode_type		*mode; 
  struct permutation_type 	*permutation;
  struct var_link_type		*bypassed_internal_vars;
  struct transition_type 	*next;
} ;

struct trans_link_type {
  struct transition_type	*tr;
  struct trans_link_type	*next;
};

struct thread_type 		{
  char 				*name;
  int				index, 
				automaton_index,
                                num_ta,
				mnpl,
				mode_count, 
				trans_count, 
				invalid_trans_count; 
  int				*ta_list;
  long				total_sec,
  				total_usec,
  				last_sec,
  				last_usec;
  struct mode_type		*starting_mode; 
  struct invalid_mode_type 	*invalid_cycle; 
  struct var_link_type 		*sync_cycle;   /* list of synclabs in a cycle */
  struct clkconj_type		*init_clkconj;
  struct boolexp_type		*init_boolexp;
  struct a23tree_type		*modetree; 
  struct thread_type 		*subthread1, *subthread2, *next;
} ;

struct invalid_mode_type 	{
  struct mode_type 		*invalid_mode;
  struct invalid_mode_type 	*next;
} ;

struct proposition_type {
  int var1, var2, op, value;
  bdd prop_bdd;
  struct proposition_type *next;
} ;

struct permutation_type {
  struct a23tree_type *parent; 	  /* used by 23tree */
  int count;		  	  /* used by 23tree */
  int *perm;
} ;

/* Note: path = 0  => A Simple Non-Path Formulae
         path != 0 => EXISTS_PATH or ALL_PATHS
         relop = 0 => NO UNTIL
         relop != 0 => An UNTIL formulae
         b1,b2,c1,c2,s1,s2 => a full-fledged until formula
         b1,c1,s1 => a henceforth formula
         b2,c2,s2 => an eventually formula
 */
         
struct spec_type {
  struct mode_spec_type *ms;
  struct boolexp_type *b1, *b2;
  struct clkconj_type *c1, *c2;
  struct spec_type *s1, *s2, *next;
  int stype;      /* UNTIL, EVENTUALLY, HENCEFORTH */
  int boolcon;    /* AND, OR */
  int path;       /* NO_PATH, EXISTS_PATH, ALL_PATHS */
  int relop;      /* 0, LEQ, LESS, EQ, GEQ, GREATER */
  int constant;   /* an integer in N */
  int sign;       /* NEGATIVE, POSITIVE */
  int blevel;     /* my bracket level: 0, 1, 2, ... */
  int ustat;      /* UFIRST, USECOND */
} ;

struct mode_spec_type {
  char *auto_name, *mode_name;
  int relop;
  struct automaton *a;
} ;

/* 
struct graph_type {
  char *name;
  int id;
  struct graph_type *next;
} ;
*/

struct action_type {
  int type;              	/* MG, SC, NR, LA, MC, PG, IE, FL */
  int if_relop, if_num;  	/* if(i  if_relop  if_num) { 
                          	 *   if_list 
                          	 * }  
                          	 * else_list  // a list of else if(...) {...} 
                          	 */
  struct thread_type *g0, 	/* g0 = manipulator(g1, g2) */
                     *g1, 
                     *g2;
  char *n0, *n1, *n2;    	/* names for graphs or threads          */
  int c0, c1, c2;	 	/* indices for graphs or threads        
                                 * c? >=  0   ==>   indexed graphs
                                 * c? == -1   ==>   no index
                                 * c? == -2   ==>   parametric index
                                 */
  int p0, p1, p2;               /* parametric index increment/decrement */
  int  			   	/* for(i := for_init_num;   */ 
      for_init_num,            	/*     i for_cond_relop for_cond_num;   */
      for_cond_relop,          	/*     i PLUS_PLUS, i PLUS_EQ num,      */
      for_cond_num,            	/*     i := i + num, i := i - num */
                                /*    )                                 */
      for_step;                 /* each iteration step size             */
  struct action_type *if_list, *else_list, 
                     *loop_list, 
                     *parent, *next;
} ;

struct manipulator_type {
  int type;		/* SC, NR, LA */
  struct manipulator_type *next;
} ;

struct int_link_type {
  int k;
  struct int_link_type *next;
} ;

#endif
