#include "include.h"
#include "global.h"
#include "modeman.h"
#include "bddman.h"
#include "lf.h"
#include "print.h"
#include "symmetry.h"

int look_forward_reduce(struct thread_type *t, struct thread_type *nt);
int reset_status(struct thread_type *t);
int dormant_cycles(struct mode_type *m, struct transition_type *e,
                   struct thread_type *tt);
int forward_dfs_visit(struct mode_type *mm, struct mode_type *m, 
                      struct transition_type *e, struct thread_type *tt);
int backward_dfs_visit(struct mode_type *mm, struct mode_type *m, 
                       struct transition_type *e, struct thread_type *tt);
int visit_rest(struct transition_type *ee, struct mode_type *m);
int own_index_trigger(struct transition_type *t);
int delusively_valid(struct mode_type *m, struct transition_type *e,
                     struct thread_type *tt);
int eliminate_derived_transitions(struct transition_type *e, 
                                  struct mode_type *m, 
                                  struct thread_type *tt);
int no_dormant_cycles(struct mode_type *, struct transition_type *,
                      struct thread_type *);

int look_forward_reduce(struct thread_type *t, struct thread_type *tt) {
  struct mode_type *m;
  struct transition_type *e, *se;
  int changed = FALSE;

  printf("Look-Forward reducing thread %d using thread %d.\n", 
         t->index, tt->index);
  if(verbose >= 0) {
    sprintf(tmpstring,"Look-Forward reducing thread %d using thread %d:\n",
            t->index, tt->index);
    output(tmpstring);
  }

  for(m=t->starting_mode; m; m=m->cnext) {
    if(m->out_trans_cycle) {
     for(e=m->out_trans_cycle->next, se=e->next; ; e=se, se=se->next) {
      if(BDD_AND(m->dval_bdd, e->trn->trigger_bdd) == BDD_FALSE) {
        if(verbose >= 0) {
         sprintf(tmpstring, 
           "Found invalid mode-edge pair: <%s(%d), %d(%d)> to %s(%d).\n",
           m->name, m->index, e->index, e->trn->index, 
                              e->mode->name, e->mode->index);
         output(tmpstring);
        }
        if(no_dormant_cycles(m, e, tt) == FALSE){
          if(changed==FALSE) changed = TRUE;
          eliminate_derived_transitions(e, m, tt);
          eliminate_transition(e, m);
          t->trans_count--;
          if(verbose >= 0) {
            sprintf(tmpstring, 
                    "Eliminated invalid transition %d(%d): %s(%d) to %s(%d).\n",
                     e->index, e->trn->index, m->name, m->index, 
                     e->mode->name, e->mode->index);
            output(tmpstring);
          }
          if(m->out_trans_cycle == NULL || 
             se == m->out_trans_cycle->next) break;
        }
      }
      else if(EDV) {                /* A valid transition */
        if(verbose >= 0) {
         sprintf(tmpstring,
           "Found valid mode-edge pair: <%s(%d), %d(%d)> to %s(%d).\n",
           m->name, m->index, e->index, e->trn->index,
                              e->mode->name, e->mode->index);
         output(tmpstring);
        }
        if(delusively_valid(m, e, tt) == TRUE){
          if(changed==FALSE) changed = TRUE;
          eliminate_derived_transitions(e, m, tt);
          eliminate_transition(e, m);
          t->trans_count--;
          if(verbose >= 0) {
            sprintf(tmpstring,
                    "Eliminated valid transition %d(%d): %s(%d) to %s(%d).\n",
                     e->index, e->trn->index, m->name, m->index,
                     e->mode->name, e->mode->index);
            output(tmpstring);
          }
          if(m->out_trans_cycle == NULL ||
             se == m->out_trans_cycle->next) break;
        }
      }
      if(e == m->out_trans_cycle) break;
     }
    }
  }

  printf("\tRemoving irreachable modes in thread %d.\n\t", tt->index);
  remove_irreachable_mode(tt);
  print_counts(tt);
  printf("\tRemoving dormant cycles modes in thread %d.\n\t", tt->index);
  remove_dormant_cycles(tt);

  printf("\tRemoving irreachable modes in thread %d.\n\t", t->index);
  composed_ta--;
  remove_irreachable_mode(t);
  print_counts(t);
  printf("\tRemoving dormant cycles modes in thread %d.\n\t", t->index);
  remove_dormant_cycles(t);
  composed_ta++;

  return changed;
}

int eliminate_derived_transitions(struct transition_type *e, 
                                  struct mode_type *m, 
                                  struct thread_type *tt) {
  struct transition_type *ee, *ees;
  struct mode_type *mm;
  for(mm = tt->starting_mode; mm; mm = mm->cnext) {
    if(mm->submode1->index == m->index) {
      if(mm->out_trans_cycle)
      for(ee=mm->out_trans_cycle->next, ees=ee->next; ; ee=ees, ees=ees->next) {
        if(same_trans(ee, e)) {
          eliminate_transition(ee, mm);
          tt->trans_count--;
          if(verbose >= 0) {
            sprintf(tmpstring,
                    "Eliminated lf transition %d(%d): %s(%d) to %s(%d).\n",
                     ee->index, ee->trn->index, mm->name, mm->index,
                     ee->mode->name, ee->mode->index);
            output(tmpstring);
          }
        }
        if(mm->out_trans_cycle == NULL || ees == mm->out_trans_cycle->next) 
          break;
      }
    } 
  }
  return 0;
}

int reset_status(struct thread_type *t) {
  struct mode_type *m;
  for(m=t->starting_mode; m; m=m->cnext) {
    m->status = UNVISITED;
  }
  return 0;
}

int same_loc(struct mode_type *m1, struct mode_type *m2) {
  if(m1->index == m2->index) return TRUE;
  if(strcmp(m1->name, m2->name) == EQUAL) return TRUE;
  return FALSE;
}

int no_dormant_cycles(struct mode_type *m, struct transition_type *e,
                      struct thread_type *tt) {
  struct mode_type *mm;

  reset_status(tt);
  for(mm=tt->starting_mode; mm; mm=mm->cnext) {
    if(same_loc(mm->submode1, m) && mm->status == UNVISITED) {
      if(verbose >= 0) {
        sprintf(tmpstring, "  Visiting %s(%d):\n", mm->name, mm->index);
        output(tmpstring);
      }
      if(BDD_AND(mm->dval_bdd, e->trn->trigger_bdd) == BDD_FALSE) {
        if(forward_dfs_visit(mm, m, e, tt) == TRUE) return TRUE;
      }
      else {
        if(backward_dfs_visit(mm, m, e, tt) == TRUE) return TRUE;
      }
    }
  }
  return FALSE;
}

int forward_dfs_visit(struct mode_type *mm, struct mode_type *m, 
                      struct transition_type *e, struct thread_type *tt) {
  struct transition_type *ee;

  mm->status = VISITED;
  if(verbose >= 0) {
    sprintf(tmpstring, "  Forward visited mode %s(%d)\n", mm->name, mm->index);
    output(tmpstring);
  }
  if(mm->out_trans_cycle)
   for(ee=mm->out_trans_cycle->next; ; ee=ee->next) {
    if(corr_trans(ee, e) && 
       BDD_AND(ee->trn->trigger_bdd, mm->dval_bdd) == BDD_FALSE) {
      visit_rest(ee, m);
      if(ee == mm->out_trans_cycle) break;
      else continue;
    }
    if(ee->trn->index == e->trn->index) { 
      if(BDD_AND(ee->trn->trigger_bdd, mm->dval_bdd) != BDD_FALSE) {
        if(verbose >= 0) {
          sprintf(tmpstring, "  Trns %d(%d) is valid with Mode %s(%d).\n", 
                             ee->index, ee->trn->index, mm->name, mm->index);
          output(tmpstring);
          output("Back-tracking!\n");
        }
        if(backward_dfs_visit(mm,m,e,tt) == TRUE) return TRUE;
/*
        return TRUE;
*/
      }
      else if(ee == mm->out_trans_cycle) break;
      else continue;
    }
    if(same_loc(ee->mode->submode1, m) && ee->mode->status == UNVISITED)
      if(forward_dfs_visit(ee->mode, m, e, tt) == TRUE) return TRUE;
    if(ee == mm->out_trans_cycle) break;
   }
  return FALSE;
}

int backward_dfs_visit(struct mode_type *mm, struct mode_type *m, 
                       struct transition_type *e, struct thread_type *tt) {
  struct transition_type *ee;

  mm->status = VISITED;
  if(verbose >= 0) {
    sprintf(tmpstring, "  Backward visited mode %s(%d)\n", mm->name, mm->index);
    output(tmpstring);
  }
  if(mm->in_trans_cycle)
  for(ee=mm->in_trans_cycle->next; ; ee=ee->next) {
    if(corr_trans(ee, e) && 
       BDD_AND(ee->trn->trigger_bdd, ee->mode->dval_bdd) == BDD_FALSE) {
       if(verbose >= 0) output(" STOPPED.\n");
/*
      visit_rest(ee, m);
*/
      if(ee == mm->in_trans_cycle) break;
      else continue;
    }
/*
    if(ee->trn->index == e->trn->index) { 
      if(BDD_AND(ee->trn->trigger_bdd, ee->mode->dval_bdd) != BDD_FALSE) {
        if(verbose >= 0) {
          sprintf(tmpstring, "  Trns %d(%d) is valid with Mode %s(%d).\n", 
                  ee->index, ee->trn->index, ee->mode->name, ee->mode->index);
          output(tmpstring);
        }
        return TRUE;
      }
      else if(ee == mm->in_trans_cycle) break;
      else continue;
    }
*/
    if(ee->mode->index == tt->starting_mode->index) {
      if(verbose >= 0) 
        output("  Reached initial mode.\n");
      return TRUE;
    }
    if(ee->mode->status == UNVISITED) {
      if(verbose >= 0) {
        sprintf(tmpstring, "FROM %s(%d) BY tr# %d TO %s(%d).\n", 
               mm->name, mm->index, ee->index, ee->mode->name, ee->mode->index);
        output(tmpstring);
      }
      if(backward_dfs_visit(ee->mode, m, e, tt) == TRUE) return TRUE;
    }
    if(ee == mm->in_trans_cycle) break;
  }
  return FALSE;
}

int visit_rest(struct transition_type *ee, struct mode_type *m) {
  struct transition_type *ep;

  ee->mode->status = VISITED;
  if(ee->mode->out_trans_cycle) {
    for(ep=ee->mode->out_trans_cycle->next; ; ep=ep->next) {
      if(same_loc(ep->mode->submode1, m) && ep->mode->status == UNVISITED)
        visit_rest(ep, m);
      if(ep == ee->mode->out_trans_cycle) break;
    }
  }
  return 0;
}

/* Return TRUE if a transition's trigger only checks whether 
   the value of a global discrete variable is its own index 
   or not. 
   ONLY ONE variable is assumed here.
 */
int own_index_trigger(struct transition_type *t) {
  struct boolconj_type *b;

  if(t->automaton->index >= 0)
  if(t->trn->prp_trigger_cycle) {
    for(b=t->trn->prp_trigger_cycle->next; ; b=b->next) {
      if(b->op == EQ && b->constant == t->automaton->index) return TRUE;
      if(b == t->trn->prp_trigger_cycle) break;
    }
  }
  return FALSE;
}

int own_index_assign(struct transition_type *t) {
  struct assign_type *a;

  if(t->automaton->index >= 0)
  if(t->trn->assign_cycle) {
    for(a=t->trn->assign_cycle->next; ; a=a->next) {
      if(a->value == t->automaton->index) return TRUE;
      if(a == t->trn->assign_cycle) break;
    }
  }
  return FALSE;
}

int complement_read(struct transition_type *t) {
  struct boolconj_type *b;

  if(t->trn->prp_trigger_cycle) {
    for(b=t->trn->prp_trigger_cycle->next; ; b=b->next) {
      if(b->op == NEQ) return TRUE;
      if(b == t->trn->prp_trigger_cycle) break;
    }
  }
  return FALSE;
}

int read_value(struct transition_type *t) {
  struct boolconj_type *b;

  if(t->trn->prp_trigger_cycle) {
    for(b=t->trn->prp_trigger_cycle->next; ; b=b->next) {
      return b->constant;
      if(b == t->trn->prp_trigger_cycle) break;
    }
  }
  return -1;
}

int delusively_valid(struct mode_type *m, struct transition_type *e,
                     struct thread_type *tt) {
  struct transition_type *ee;

  if(own_index_trigger(e) && m->in_trans_cycle) {
    for(ee=m->in_trans_cycle->next; ; ee=ee->next) {
      if(ee->permutation) {
        if(own_index_trigger(ee) && 
            (perm_index(ee->automaton->index, ee->permutation->perm) 
             == e->automaton->index)
          ) return FALSE;
        if(own_index_assign(ee) &&
            (perm_index(ee->automaton->index, ee->permutation->perm) 
             == e->automaton->index)
          ) return FALSE;
        if(complement_read(ee) &&
           (perm_index(read_value(ee), ee->permutation->perm) 
            != e->automaton->index)
          ) return FALSE;
      }
      else {
        if(own_index_trigger(ee) && ee->automaton->index == e->automaton->index)
          return FALSE;
        if(own_index_assign(ee) && ee->automaton->index == e->automaton->index)
          return FALSE;
        if(complement_read(ee) && read_value(ee) != e->automaton->index)
          return FALSE;
      }
      if(ee == m->in_trans_cycle) break;
    }
  }
  else return FALSE;
/*
  for(mm=tt->starting_mode; mm; mm=mm->cnext) {
    if(same_loc(mm->submode1, m)) {
      if(verbose >= 0) {
        sprintf(tmpstring, "  Checking %s(%d):\n", mm->name, mm->index);
        output(tmpstring);
      }
      for(ee=mm->out_trans_cycle->next; ; ee=ee->next) {
        if(ee->trn->index == e->trn->index) {
          if(BDD_AND(mm->dval_bdd, ee->trn->trigger_bdd) != BDD_FALSE)
            return FALSE;
        }
        if(ee == mm->out_trans_cycle) break;
      }
    }
  }
*/
  return TRUE;
}
