#include "include.h"
#include "global.h"
#include "print.h"
#include "bddman.h"
#include "modeman.h"
#include "zoneman.h"
#include "symmetry.h"
#include "transman.h"
#include "readwrite.h"
#include "action.h"
#include "gui.h"
#include "shieldclk.h"
#include "lf.h"
#include "shieldvar.h"
#include "treeman.e"

extern void qsort();
extern int stdoutput;
struct thread_type 		*NEW_THREAD, *THREAD1, *THREAD2;

struct mode_type 		*TOP;           /* TOP of stack (end=NULL) */

struct mode_type		*mode_pool; 
int				mode_count, free_mode_count;

struct transition_type		*transition_pool; 
int				transition_count, free_transition_count; 

int				comp_stat = EQUAL;
struct mode_type 		*same_mode = (struct mode_type *)NULL;
struct mode_type		*last_mode =(struct mode_type *)NULL; 
struct clkconj_type *clk_literal_cycle = NULL;

int init_mode_management();
int init_thread_management();
int get_mode_count(struct thread_type *);
int get_trans_count(struct thread_type *);

int reset_thread(struct thread_type *);
struct mode_type *alloc_mode();
int free_mode_vec(struct mode_type *);

int free_mode(struct mode_type *);
int free_mode_list(struct mode_type *);
int mode_count_inc(struct mode_type   *);
int mode_parent_set(struct mode_type *, struct a23tree_type *);
struct transition_type  *alloc_transition();
int free_transition(struct transition_type *);
int free_transition_cycle(struct transition_type *trc, int cc);
int transition_candidate_comp(struct transition_type *tr, struct mode_type *s); 
struct mode_type *initial_mode(struct thread_type *, struct thread_type *);
struct boolconj_type *initial_dvl1(struct thread_type *, struct thread_type *, 
                                  int);
struct boolconj_type *append_boolconj(struct boolconj_type *, 
                                      struct boolconj_type *);
struct boolconj_type *boolconj_union(struct boolconj_type *,
                                     struct boolconj_type *);
struct boolconj_type *check_boolconj(struct boolconj_type *,
                                     struct boolconj_type *);
int insert_int(int val, int var_pos, int **int_vals, int tot_vals);
int delete_int(int val, int var_pos, int **int_vals, int tot_vals);
int print_dvl(struct boolconj_type *);

int install_dvl_bdd(struct boolconj_type *dvl);
int dvl_comp(struct mode_type *m1, struct mode_type *m2);
int mode_comp(struct mode_type *s1, struct mode_type *s2);
struct transition_type *add_transition(struct transition_type *, 
                   struct mode_type *, struct mode_type *);
struct transition_type *add_transition_fifo(struct transition_type *, 
                   struct mode_type *, struct mode_type *);

struct boolconj_type *compute_dvl(struct mode_type *m,
                                  struct transition_type *tr);
int free_int_link(struct int_link_type *);
struct boolconj_type *extract_dvl(struct boolconj_type *, 
                                  struct variable_type *);
int neg_lit_trigger(struct transition_type *tr, struct variable_type *);
int pos_lit_trigger(struct transition_type *tr, struct variable_type *);
struct boolconj_type *new_dvl_neg(struct transition_type *, 
                                  struct mode_type *, 
                                  struct variable_type *);
struct boolconj_type *new_dvl_pos(struct transition_type *tr, 
                                  struct variable_type *, struct mode_type *);
int not_written_by_others(struct boolconj_type *b);
int not_accessed_by_self(struct boolconj_type *, struct transition_type *);
int no_dvar_assign(struct transition_type *tr, struct variable_type *v);
struct boolconj_type *delete_boolconj(int, int, struct boolconj_type *);
bdd dvl_bdd(struct mode_type *m);
bdd get_bdd_from_dvl(struct boolconj_type *dvl);
int print_dvl_bdd(struct boolconj_type *);

void print_info_tr(struct transition_type *tr);
void print_info_nm(struct mode_type *, struct transition_type *,
                   struct mode_type *);
void print_info_zone(struct mode_type *);
void print_info_bdd(struct mode_type *);
void handle_comp_stat();
int *get_perm(struct mode_type *, struct transition_type *);
int create_permute_dvl(struct mode_type *, struct transition_type *,
                       struct mode_type *, int *);
int sync_create_permute_dvl(struct mode_type *, struct transition_type *,
                            struct transition_type *, struct mode_type *);
int create_permute_boolexp(struct mode_type *, struct transition_type *,
                           struct mode_type *, int *);
int sync_create_permute_boolexp(struct mode_type *, struct transition_type *,
                                struct transition_type *, struct mode_type *, 
                                int *, int *);
int handle_last(struct mode_type *, struct transition_type *);
int sync_handle_last(struct mode_type *, struct transition_type *,
                     struct transition_type *);
int create_permute_dval(struct mode_type *, struct transition_type *,
                        struct mode_type *, int *, int);
int sync_create_permute_dval(struct mode_type *, struct transition_type *,
                             struct transition_type *, struct mode_type *,
                             int *, int *);
struct mode_type *handle_otf_norm(struct mode_type *);
void push_mode(struct mode_type *, struct mode_type *);
void print_progress(struct mode_type *);
int create_permute_zone(struct mode_type *, struct transition_type *,
                        struct mode_type *);
int sync_create_permute_zone(struct mode_type *, struct transition_type *,
                             struct transition_type *, struct mode_type *);
int unsync_transition(struct mode_type *, struct transition_type *, int);
int sync_transition(struct mode_type *, struct transition_type *,
                    struct transition_type *);

int transition_m1_first(struct mode_type *);
int transition_m2_first(struct mode_type *);
int compose(struct thread_type *, struct thread_type *, struct thread_type *);
int merge_iden_red_multi(struct thread_type *t);

struct zone_type *
post_xconj(struct mode_type *,        	/* mode 	   */
           struct clkconj_type *, 	/* xconj_cycle of new submode1 */
           struct clkconj_type *, 	/* xconj_cycle of new submode2 */
           struct clkconj_type *, 	/* clk_trigger_cycle   */
           struct assign_type *, 	/* assign_cycle    */
           struct permutation_type *,	/* permutation on transition */
           struct clkconj_type **     	/* new xconj_cycle */
          );

struct clkconj_type *process_max(struct clkconj_type **);
struct clkconj_type *advance(struct clkconj_type *, int);
struct clkconj_type *conjoin(struct clkconj_type *, int);
int transitive_op(int, int);
struct clkconj_type *assign(struct clkconj_type *, struct assign_type *);
bdd bdd_assign(bdd, struct assign_type *);
struct boolconj_type *new_assign_boolconj(struct assign_type *, bdd);
struct int_link_type *get_curr_values(int, bdd);
char *get_name(char ***);
char *assign_mode_name(struct mode_type *);
char *assign_mode_name1(struct mode_type *, int *, int);

int init_clk_literals();
int append_clkcycle(struct clkconj_type *c, struct clkconj_type **lc);
int clkconj_not_in_cycle(struct clkconj_type *cp, struct clkconj_type **lc);
int append_clkconj(struct clkconj_type *cp, struct clkconj_type **lc);

struct clkconj_type *initial_literal_set();
int reset_literals(int var, struct clkconj_type **lc);
int delete_literals(int var, struct clkconj_type **lc);
int insert_literals(int var, struct clkconj_type **lc);
int update_literals(struct clkconj_type *trigger, struct clkconj_type **lc);
int incompatible_literals(struct clkconj_type *c1, struct clkconj_type *c2);

struct mode_type *normalize_mode(struct mode_type *m);
int *get_mode_perm(struct mode_type *m);

int mode_shield_clocks(struct mode_type *m);
struct var_link_type *var_link_copy(struct var_link_type *sc);
struct var_link_type *clocks_union(struct var_link_type *sc1, 
                                   struct var_link_type *sc2);
struct var_link_type *insert_var_link(struct var_link_type *c, 
                                      struct var_link_type *sc);
int print_var_links(struct var_link_type *sc);
int free_var_links(struct var_link_type *);

int shield_clocks_compare(struct var_link_type *, struct var_link_type *);
int update_permutation(struct mode_type *m, struct transition_type *tr,
                       int *new_mode_perm, struct transition_type *ntr);

int print_new_mode(struct mode_type *m);
int perm_comp(int *p1, int *p2);
int mode_perm_comp(struct mode_type *m1, struct mode_type *m2);

int permutation_comp(struct permutation_type *p1, struct permutation_type *p2);
int perm_count_inc(struct permutation_type *p);
int perm_parent_set(struct permutation_type *p, struct a23tree_type *parent);
int free_perm(struct permutation_type *p);

struct permutation_type *alloc_permutation();
int print_permutation(struct permutation_type *p);
struct var_link_type *sort_var_links(struct var_link_type *sc);

int reset_ta_list(struct thread_type *);
int update_ta_list(struct thread_type *, struct thread_type *,
                   struct thread_type *);
int clear_line();
int clean_line();


/* SYNC related functions */

struct variable_type *get_same_sync(struct transition_type *, 
                                    struct transition_type *);
struct trans_link_type *get_sync_trans(struct transition_type *,
                                       struct mode_type *);
int check_sync(struct transition_type *, struct mode_type *, int);

struct spec_trans_type *alloc_spec_trans();
int get_clk_trigger_length(struct clkconj_type *);
int get_prp_trigger_length(struct boolconj_type *);
struct transition_type *compose_sync_transition(struct transition_type *,
					        struct transition_type *);
struct assign_type *assign_union(struct assign_type *,
                                 struct assign_type *);
struct var_link_type *var_link_union(struct var_link_type *,
                                     struct var_link_type *);
struct transition_type *add_sync_transition_fifo(struct transition_type *, 
                                                 struct transition_type *, 
                                                 struct mode_type *, 
                                                 struct mode_type *);
int sync_update_permutation(struct mode_type *, struct transition_type *,
                            struct transition_type *,
                            int *new_mode_perm, struct transition_type *);

/************************ FUNCTION DEFINITIONS **********************/

int init_mode_management()
{
  mode_pool = NULL; 
  mode_count = 0;
  free_mode_count = 0;

  transition_pool = NULL; 
  transition_count = 0; 
  free_transition_count = 0; 

  init_23tree(&zone_tree); 
  init_23tree(&bdd_tree); 
  init_23tree(&trans_tree);
  return 0;
}
  /* init_mode_management() */


int init_thread_management() {
  struct automaton *x;
  struct mode_type *m;
  int i;

  thread = (struct thread_type *)
              malloc( ((2 * system_ta_num) - 1) * sizeof(struct thread_type));
  for(i=0; i<2*system_ta_num-1; i++) {
    thread[i].index = i;
    thread[i].automaton_index = i;
    thread[i].ta_list = (int *)malloc(sizeof(int)*system_ta_num);
    reset_ta_list(&thread[i]);
    thread[i].name = NULL;
    thread[i].starting_mode = NULL;
    thread[i].invalid_cycle = NULL;
    thread[i].num_ta = 0;
    thread[i].mnpl = 0;
    thread[i].mode_count = 0;
    thread[i].trans_count = 0;
    thread[i].invalid_trans_count = 0;
    thread[i].init_boolexp = NULL;
    thread[i].init_clkconj = NULL;
    thread[i].subthread1 = thread[i].subthread2 = (struct thread_type *)NULL;
    thread[i].total_sec = thread[i].total_usec = 
                          thread[i].last_sec = thread[i].last_usec = 0;
    thread[i].modetree = (struct a23tree_type *)NULL;
    thread[i].next = (struct thread_type *)NULL;
  }
  if(automaton_tail) {
    for(x = automaton_tail->next, i = 0; ; i++, x = x->next) {
      thread[i].name = str_preserve(x->name);
      thread[i].automaton_index = x->index;
      thread[i].num_ta = 1;
      thread[i].sync_cycle = x->sync_cycle;
      thread[i].init_boolexp = x->init_boolexp;
      thread[i].init_clkconj = x->init_clkconj;
      if(x->mode_cycle) {
        for(m = x->mode_cycle->next; ; m=m->next) {
          if(!strcmp(m->name, x->init_mode)) break;
          if(m == x->mode_cycle) break;
        }
        thread[i].starting_mode = m;
      }
      (thread[i].ta_list)[i] = TRUE;
      if(x==automaton_tail) break;
    }
    for(i=0; i<system_ta_num; i++) {
      for(m=thread[i].starting_mode; ; m=m->next) {
        if(m->next == thread[i].starting_mode) {
          m->cnext = NULL;
          break;
        }
        m->cnext = m->next;
      }
    }
  }
  for(i=0; i<2*system_ta_num-1; i++) {
    if(thread[i].name == NULL) 
      thread[i].name = str_preserve("_T");
/*
    if(thread[i].automaton_index == 0) 
      thread[i].automaton_index = thread[i].index;
*/
  }
  
  for(i=0; i<2*system_ta_num-1; i++) {
    thread[i].mode_count = get_mode_count(&thread[i]);
    thread[i].trans_count = get_trans_count(&thread[i]);
  }
  decl_graph_num = system_ta_num;
  return 0;
}

int get_mode_count(struct thread_type *t) {
  struct mode_type *m;
  int i;
  if(t->starting_mode) {
    for(m=t->starting_mode,i=0; m; m=m->cnext) i++;
    return i;
  }
  else return 0;
}

int get_trans_count(struct thread_type *t) {
  struct mode_type *m;
  struct transition_type *tr;
  int i=0;
  if(t->starting_mode) {
    for(m=t->starting_mode; m; m=m->cnext) {
      if(m->out_trans_cycle) {
        for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
          i++;
          if(tr == m->out_trans_cycle) break;
        }
      }
    }
    return i;
  }
  else return 0;
}

int reset_thread(struct thread_type *t) {
  char *c = get_graph_name(t);

  if(t->index < system_ta_num) {
    if(SGM_MODE == GUI || SGM_MODE == PROCESS) GUI_ERROR = E_RESET_FAIL;
    else {
      fprintf(stderr, "Error: Cannot reset a system graph %s.\n", c);
    }
    free(c);
    return FALSE;
  } 
  if(t->mode_count == 0 && t->trans_count == 0) return TRUE;
  printf("Resetting graph %s ... \n", c);
  t->mode_count = 0;
  t->trans_count = 0;
  t->invalid_trans_count = 0;
  t->num_ta = 0;
  t->mnpl = 0;
  t->total_sec = t->total_usec = t->last_sec, t->last_usec=0;

  free_clkconj(&(t->init_clkconj));
  t->init_clkconj = NULL;

  free_boolexp(t->init_boolexp);
  t->init_boolexp = NULL;

  t->subthread1 = t->subthread2 = NULL;

  reset_ta_list(t);

/*
  composed_ta--;
  printf("Reached before free_mode_list().\n");
*/
  free_mode_list(t->starting_mode);
  t->starting_mode = NULL;
/*
  printf("Reached after free_mode_list().\n");
*/
  t->invalid_cycle = 0;
/*
  printf("Reached before free_23tree().\n");
*/
  if(t->modetree) free_23tree(t->modetree);
/*
  printf("Reached after free_23tree().\n");
*/
  t->modetree = NULL;

/*
  printf("done.\n");
  composed_ta++;
*/
  free(c);
  return TRUE;
}

int free_mode_list(struct mode_type *m) {
  struct mode_type *x, *y;
  if(m) {
    for(x=m, y=m->cnext; x; x=y, y=y->cnext) {
/*
      printf("Freeing mode %d (%s).\n", x->index, x->name);
*/
      free_mode(x); x = NULL;
/*
      printf("  Freed mode %d (%s).\n", x->index, x->name);
      printf("\tNext on list is %d (%s).\n", y->index, y->name);
*/
      if(y==NULL) break;
    }
    m = NULL;
  }
  return 0;
}

struct mode_type *alloc_mode() {
  int j;
  struct mode_type *m;

  if (free_mode_count) {
    m = mode_pool;
/*
    mode_pool = (struct mode_type *) m->out_trans_cycle;
*/
    mode_pool = mode_pool->next; 	/* modified by ERIC from above */
    free_mode_count--;
/*
    printf("Reusing freed mode: free_mode_count = %d.\n", free_mode_count);
*/
  }
  else {
/*
    printf("Allocating new mode: free_mode_count = %d.\n", free_mode_count);
*/
    m = (struct mode_type *) malloc(sizeof(struct mode_type));
    m->index = mode_count++;
  }
  m->name = (char *)NULL;
  m->mode_vec = (char **)malloc(sizeof(char *)*system_ta_num);
  for(j=0;j<system_ta_num;j++) {
    m->mode_vec[j] = (char *)NULL;
  }

  m->parent = (struct a23tree_type *)NULL;
  m->status = 0;
  m->count = 0;
  m->xconj_cycle = (struct clkconj_type *)NULL;
  m->xconj_length = 0;
  m->boolexp = (struct boolexp_type *)NULL;
  m->dvl = (struct boolconj_type *)NULL;
  m->cvl = (struct clkconj_type *)NULL;
  m->norm_perm = (struct permutation_type *)NULL;

  m->out_trans_count = 0;
  m->out_trans_cycle = (struct transition_type *)NULL;
  m->in_trans_count = 0;
  m->in_trans_cycle = (struct transition_type *)NULL;

  m->submode1 = (struct mode_type *)NULL;
  m->submode2 = (struct mode_type *)NULL;
  m->zone = (struct zone_type *)NULL;
  m->dval_bdd = (bdd)NULL;
  m->shielded_clocks = (struct var_link_type *)NULL;
  m->reqd_clocks = (struct var_link_type *)NULL;
  m->next = (struct mode_type *)NULL;

  return (m);
}

int free_mode(struct mode_type *s) {
  struct transition_type *tr, *trs, *ts;
  int i;

/*
  printf("Inside free_mode(%d).\n", s->index); 
*/
  if(s->name) free(s->name);
  s->name = NULL;

  free_mode_vec(s);

  if(s->out_trans_cycle) {
/*
    printf("\t this mode has %d out-going trans.\n", s->out_trans_count);
*/
    tr = s->out_trans_cycle->next;
    trs = tr->next;
    for(i=s->out_trans_count; i; i--, tr=trs, trs=trs->next) {
      eliminate_transition(tr, s);
    }
  }
/*
  printf("Eliminated all out-going transitions.\n");
*/
  if(s->in_trans_cycle) {
/*
    printf("\t this mode has %d in-coming trans.\n", s->in_trans_count);
*/
    tr = s->in_trans_cycle->next;
    trs = tr->next;
    for(i=s->in_trans_count; i; i--, tr=trs, trs=trs->next) {
/*
      printf(
      "eliminating in-coming trans %d (from mode %d (%s) with %d out-trans).\n",
         tr->index, tr->mode->index, tr->mode->name, tr->mode->out_trans_count);
*/
      for(ts = tr->mode->out_trans_cycle->next; ; ts=ts->next) {
        if(ts->index == tr->index) break;
        if(ts==tr->mode->out_trans_cycle) break;
      }
      eliminate_transition(ts, tr->mode);
    }
  }
/*
  printf("Eliminated all in-coming transitions.\n");
*/
/*
  if (s->out_trans_cycle) {
    tr = s->out_trans_cycle->next;
    s->out_trans_cycle->next = transition_pool;
    transition_pool = tr;
  }
  if (s->in_trans_cycle) {
    tr = s->in_trans_cycle->next;
    s->in_trans_cycle->next = transition_pool;
    transition_pool = tr;
  }
  free_transition_count
    = free_transition_count + s->in_trans_count + s->out_trans_count;

  printf("Freed all transitions: free_transition_count = %d.\n", 
                    free_transition_count); 

  s->in_trans_count = s->out_trans_count = 0;
*/

  if(s->boolexp)
    free_boolexp(s->boolexp);
/*
  printf("\tFreed boolexp.\n"); 
*/

/*
  bdd_free(bddm, s->dval_bdd);
  printf("\tFreed dval_bdd.\n"); 
*/

  free_clkconj(&s->xconj_cycle);
  s->xconj_length = 0;
/*
  printf("\tFreed xconj_cycle.\n"); 
*/

  if(s->zone) {
    free_zone(s->zone);
    s->zone = NULL;
  }
/*
  printf("\tFreed zone.\n");
*/

  if(s->dvl) {
    free_boolconj(s->dvl);
    s->dvl = NULL;
  }

  if(s->cvl) {
    free_clkconj(&s->cvl);
    s->cvl = NULL;
  }

  if(s->shielded_clocks) {
    free_var_links(s->shielded_clocks);
    s->shielded_clocks = NULL;
  }

  if(s->reqd_clocks) {
    free_var_links(s->reqd_clocks);
    s->reqd_clocks = NULL;
  }

  s->next = mode_pool;
  mode_pool = s;

  free_mode_count++;
/*
  printf("\tFreed mode: free_mode_count = %d.\n", free_mode_count); 
*/
  return 0;
}

int free_mode_vec(struct mode_type *m) {
  int i;

  if(m->mode_vec == NULL) return 0;

  for(i=0; i<system_ta_num; i++) {
    free(m->mode_vec[i]);
  }
  free(m->mode_vec);
  m->mode_vec = NULL;
  return 0;
}

int mode_count_inc(struct mode_type *m) {
  (m->count)++;
  return 0;
}

int mode_parent_set(struct mode_type *m, struct a23tree_type *parent) {
  m->parent = parent;
  return 0;
}

struct transition_type  *alloc_transition() {
  struct transition_type        *tr;

  if (free_transition_count) {
    tr = transition_pool;
    transition_pool = tr->next;
    free_transition_count--;
  /*
   *  printf("Reusing transition: free_transition_count = %d.\n", 
   *                 free_transition_count); 
   *  fprintf(logfile,"$$$ Alocating old transition %x with %d tr in pool %x\n",
   *                    tr, free_transition_count, transition_pool);
   */
  }
  else {
    tr = (struct transition_type *)malloc(sizeof(struct transition_type));
    transition_count++;
    /*
     *printf("Allocating new transition: free_transition_count = %d.\n", 
     *             free_transition_count);
     *fprintf(logfile,"$$$ Alocating new transition %x with %d tr in pool %x\n",
     *             tr, free_transition_count, transition_pool);
     */
  }
  /*  fflush(logfile);
   */
  tr->automaton = NULL;
  tr->mode = (struct mode_type *)NULL;
/*
  tr->goto_mode = (char *)NULL;
  tr->clk_trigger_length = 0;
  tr->clk_trigger_cycle = NULL;
  tr->prp_trigger_length = 0;
  tr->prp_trigger_cycle = NULL;
  tr->assign_length = 0;
  tr->assign_cycle = NULL;
  tr->trigger_bdd = NULL;
  tr->valid = TRUE;
*/
  tr->trn = (struct spec_trans_type *)NULL;
  tr->permutation = (struct permutation_type *)NULL;
  tr->bypassed_internal_vars = (struct var_link_type *)NULL;
  tr->next = (struct transition_type *)NULL;

  return(tr);
}

int free_transition(struct transition_type *tr) {
/*  fprintf(logfile, "$$$ Freeing transition %x with %d tr in pool %x\n",
 *                    tr, free_transition_count, transition_pool);
 *  fflush(logfile);
 */
/*
  printf("Freeing transition (%d) with dir = %d, mode = %s (%d).\n", 
         tr->index, direction, tr->mode->name, tr->mode->index);
*/
/*
  free(tr->goto_mode);
  tr->goto_mode = NULL;
  printf("Freed goto_mode.\n");
  free_clkconj(&tr->clk_trigger_cycle);
  printf("Freed clk_trigger_cycle.\n");
  free_boolconj(tr->prp_trigger_cycle);
  printf("Freed prp_trigger_cycle.\n");
  bdd_free(bddm, tr->trigger_bdd);
  printf("Freed trigger_bdd.\n");
  free(tr->assign_cycle);
  tr->assign_cycle = NULL;
  printf("Freed assign_cycle.\n");
*/

/* a SYNC'd transition */
/*
  if(tr->automaton == NULL) 
    free_spec_trans(tr);
*/
  tr->permutation = (struct permutation_type *)NULL;
  tr->next = transition_pool;
  transition_pool = tr;
  fflush(logfile);
  free_transition_count++;
/*
  printf("Freed transition: free_transition_count = %d.\n", 
                      free_transition_count);
*/
  return 0;
}

int free_transition_cycle(struct transition_type *trc, int cc) {
  if(trc) {
    if(cc) {
/*    fprintf(logfile,
              "$$$Freeing %d transition cycle %x with and %d tr in pool %x.\n",
              cc, trc, free_transition_count, transition_pool);
      fflush(logfile);
*/
      trc->next = transition_pool;
      transition_pool = trc;
      free_transition_count = free_transition_count + cc;
    }
    else {
      fprintf(logfile, "A mismatch in the exec cycle %x with %d exec's.\n",
                        (unsigned int)trc, cc);;
      while(1);
    }
  }
  else if(cc) {
    fprintf(logfile, "A mismatch in the exec cycle %x with %d exec's.\n",
                      (unsigned int)trc, cc);;
    while(1);
  }
  return 0;
}

int transition_candidate_comp(struct transition_type *tr, struct mode_type *s) {
  int   comp;

  if (tr->mode->index < s->index)
    return(-1);
  else if (tr->mode->index > s->index)
    return(1);
  else {
    comp = mode_comp(tr->mode, s);
    return(comp);
  }
}

struct mode_type *initial_mode(struct thread_type *t1, struct thread_type *t2) {
  struct mode_type              *ns, *s1, *s2, *nns;
  struct clkconj_type 		*tmp_clkconj, *tmp_clkconj2;

  ns = alloc_mode();

  ns->submode1 = s1 = t1->starting_mode;
  ns->submode2 = s2 = t2->starting_mode;
  ns->name = assign_mode_name(ns);

  ns->boolexp = alloc_boolexp();
  ns->dval_bdd = (bdd)NULL;
  
  ns->boolexp->boolexp_bdd = BDD_AND(s1->boolexp->boolexp_bdd,
                                     s2->boolexp->boolexp_bdd);

  if(t1->init_boolexp && t2->init_boolexp) {
    ns->dval_bdd = BDD_AND( BDD_AND(t1->init_boolexp->boolexp_bdd, 
                                    t2->init_boolexp->boolexp_bdd), 
                            BDD_AND(s1->boolexp->boolexp_bdd, 
                                    s2->boolexp->boolexp_bdd)
                   );
  }
  else if(t1->init_boolexp) {
    ns->dval_bdd = BDD_AND(t1->init_boolexp->boolexp_bdd, 
                           BDD_AND(s1->boolexp->boolexp_bdd, 
                                   s2->boolexp->boolexp_bdd)
                   );
  }
  else if(t2->init_boolexp) {
    ns->dval_bdd = BDD_AND(t2->init_boolexp->boolexp_bdd, 
                           BDD_AND(s1->boolexp->boolexp_bdd, 
                                   s2->boolexp->boolexp_bdd)
                   );
  }
  else {
    ns->dval_bdd = BDD_AND(s1->boolexp->boolexp_bdd, s2->boolexp->boolexp_bdd);
  }

  if(clock_shielding)
    ns->shielded_clocks = clocks_union(ns->submode1->shielded_clocks,
                                       ns->submode2->shielded_clocks);


/*
  sprintf(tmpstring, "Thread %s[%d] has init_clkconj: ", t1->name, t1->index);
  output(tmpstring);
  print_clkconj(t1->init_clkconj);
  output("\n");
  output("Starting mode has clkconj: ");
  print_clkconj(s1->xconj_cycle);
  output("\n");
  sprintf(tmpstring, "Thread %s[%d] has init_clkconj: ", t2->name, t2->index);
  output(tmpstring);
  print_clkconj(t2->init_clkconj);
  output("\n");
  output("Starting mode has clkconj: ");
  print_clkconj(s2->xconj_cycle);
  output("\n");
*/
  
  ns->xconj_cycle = clkconj_and(t1->init_clkconj, 
                                t2->init_clkconj);
/*
  output("Mode clkconj and is: "); 
  print_clkconj(clkconj_and(s1->xconj_cycle, s2->xconj_cycle));
  output("\n");
*/
  tmp_clkconj2 = clkconj_and(s1->xconj_cycle, s2->xconj_cycle);
  tmp_clkconj = clkconj_and(ns->xconj_cycle, tmp_clkconj2);
  free_clkconj(&ns->xconj_cycle);
  free_clkconj(&tmp_clkconj2);
  tmp_clkconj2 = NULL;
  ns->xconj_cycle = tmp_clkconj;
  tmp_clkconj = NULL;

/*
  output("Before advance+conjoin: ");
  print_clkconj(ns->xconj_cycle);
  output("\n");
*/

  ns->xconj_cycle = advance(ns->xconj_cycle, FALSE);
  ns->xconj_cycle = conjoin(ns->xconj_cycle, DELTA);	

/*
  sprintf(tmpstring, "Initial mode %d has clkconj: ", ns->index);
  output(tmpstring);
  print_clkconj(ns->xconj_cycle);
  output("\n");
*/

  ns->zone = zone(ns->xconj_cycle, NEW_THREAD);
  if(clock_shielding)
    mode_shield_clocks(ns);
  if(ns->zone->clk_index)        /* not a TRUE ZONE */
    ns->zone = process_zone_max_const(ns->zone);

  free_clkconj(&(ns->xconj_cycle));
  ns->xconj_cycle = zone2clkconj(ns->zone);

  if(!lazy_bdd_merging)
    ns->dvl = initial_dvl(t1, t2, TRUE, NULL); /* TRUE => is_initial_mode */
/*
  ns->cvl = initial_literal_set();
*/

  if(verbose == 0) {
    sprintf(tmpstring, "mode %s(%d) [%s(%d), %s(%d)]: \n", 
              ns->name, ns->index, ns->submode1->name, ns->submode1->index, 
                                   ns->submode2->name, ns->submode2->index);
    output(tmpstring); 
/*
    output("\tCVL set: ");
    print_literals(ns->cvl);
    output("\tinv bdd: ");
    print_bdd(ns->boolexp->boolexp_bdd);
*/
    if(!lazy_bdd_merging) {
      output("\tDVL Set: ");
      print_dvl(ns->dvl);
    }
    else {
      output("\tdval bdd: ");
      print_bdd(ns->dval_bdd);
    }
    if(total_clocks > 0) {
      output("\tZone: ");
      print_zone(ns->zone);
      output("\n");
    }
  }

  if(verbose > 0) {
    sprintf(tmpstring, 
   "\n###  Initial mode created %s(%d) [%s(%d), %s(%d)]  ###\n", 
                 ns->name, ns->index, s1->name, s1->index, s2->name, s2->index);
    output(tmpstring);
    sprintf(tmpstring, "BDD of mode %s(%d) [%s(%d), %s(%d)] is: \n\t", 
                       ns->name, ns->index, s1->name, s1->index, 
                       s2->name, s2->index);
    output(tmpstring);
    print_bdd(ns->boolexp->boolexp_bdd);
    if(total_clocks > 0) {
      sprintf(tmpstring, "ZONE of mode %s(%d) [%s(%d), %s(%d)] is:\n\t", 
                         ns->name, ns->index, s1->name, s1->index, 
                         s2->name, s2->index);
      output(tmpstring);
      print_zone(ns->zone);
      output("\n");
    }
  }
        
  init_23tree(&(NEW_THREAD->modetree));

  nns = (struct mode_type *)
    add_23tree(ns, NEW_THREAD->modetree, mode_comp,
               free_mode, mode_count_inc, mode_parent_set, print_mode);
  return(ns);
}

/* Calculation of the initial discrete variable literal set for 
 * merging threads t1 and t2 
 * (1) get all the values (int_val) never written by processes outside 
 *     the hypersite of all currently composed timed automata (# = composed_ta)
 * (2) return the cycle of boolean conjuncts (boolconj) with each one
 *     equal to (dvar != int_val), where dvar is discrete variable.
 */

struct boolconj_type *initial_dvl1(struct thread_type *t1, 
                                  struct thread_type *t2, int is_initial_mode) {
  int i, j, tot_vars = 0, tot_vals, **int_vals = NULL;
  struct variable_type *v;
  struct boolconj_type *b, *dvl=NULL;
  struct mode_type *m;
  struct transition_type *tr;
  struct assign_type *a;
  struct proposition_type *p;
  bdd f = (bdd)NULL;

  for(v=var_tail->next; ; v=v->next) {
    if(v->type == V_REGISTER) {
      tot_vars++;
      if(int_vals)
        int_vals = (int **)realloc(int_vals, sizeof(int *)*tot_vars);
      else
        int_vals = (int **)malloc(sizeof(int *));
      int_vals[tot_vars-1] = (int *)malloc(sizeof(int));
      int_vals[tot_vars-1][0] = v->index;
    }
    if(v == var_tail) break;
  }

  for(j=0; j<tot_vars; j++) {
    tot_vals = 0;

    /* (1) get all the values written by thread[0] ... thread[composed_ta-1] 
     *     to each of the global discrete variables
     */

    for(i=0; i<composed_ta; i++) {
      for(m=thread[i].starting_mode; m; m=m->cnext) {
        if(m->out_trans_cycle) {
          for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
            if(tr->trn->assign_cycle) {
              for(a=tr->trn->assign_cycle->next; ; a=a->next) {
                if(a->var == int_vals[j][0]) {
                  if(insert_int(a->value, j, int_vals, tot_vals) == TRUE)
                    tot_vals++;
/*
                  sprintf(tmpstring, "after appending: ");
                  output(tmpstring);
                  print_vals(int_vals[j], tot_vals);
*/
                }
                if(a==tr->trn->assign_cycle) break;
              }
            }
            if(tr==m->out_trans_cycle) break;
          }
        }
      }
    }

    /* (2) delete all the values written by 
     *     thread[composed_ta] ... thread[system_ta_num] 
     *     to each of the global discrete variables
     */

    if(!lastcomp) {
      for(i=composed_ta; i<system_ta_num; i++) {
        for(m=thread[i].starting_mode; m; m=m->cnext) {
          if(m->out_trans_cycle) {
            for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
              if(tr->trn->assign_cycle) {
                for(a=tr->trn->assign_cycle->next; ; a=a->next) {
                  if(a->var == int_vals[j][0]) {
                    if(delete_int(a->value, j, int_vals, tot_vals) == TRUE)
                      tot_vals--;
  /*
                    sprintf(tmpstring, "after deleting: ");
                    output(tmpstring);
                    print_vals(int_vals[j], tot_vals);
  */
                  }
                  if(a==tr->trn->assign_cycle) break;
                }
              }
              if(tr==m->out_trans_cycle) break;
            }
          }
        }
      }
    }

    /* (3) construct the literals for int_vals[j] */

    for(i=1; i<=tot_vals; i++) {
      if(lastcomp && is_initial_mode) {
        for(p = prop_tail->next; ; p=p->next) {
          if(p->var1 == int_vals[j][0] && p->value == int_vals[j][i]) {
            f = bdd_not(bddm, p->prop_bdd);
            break;
          }
          if(p == prop_tail) break;
        }
        if(bdd_and(bddm, f, t1->init_boolexp->boolexp_bdd)
           == BDD_FALSE) {
          continue;
        }
        else if(bdd_and(bddm, f, t2->init_boolexp->boolexp_bdd)
           == BDD_FALSE) {
          continue;
        }
        else {
          b = alloc_boolconj();
          b->var1 = int_vals[j][0];
          b->op = NEQ;
          b->constant = int_vals[j][i];
          dvl = append_boolconj(b, dvl);
        }
      }
      else {
        b = alloc_boolconj();
        b->var1 = int_vals[j][0];
        b->op = NEQ;
        b->constant = int_vals[j][i];
        dvl = append_boolconj(b, dvl);
      }
    }

  }
  install_dvl_bdd(dvl);
/*
  sprintf(tmpstring, "Number of composed timed automata = %d.\n", composed_ta);
  output(tmpstring);
  sprintf(tmpstring, "Number of global discrete variables = %d.\n", tot_vars);
  output(tmpstring);
  print_dvl(dvl);
  output("DVL BDD: "); 
  print_dvl_bdd(dvl);
*/
  return dvl;
}

struct boolconj_type *append_boolconj(struct boolconj_type *b, 
                                      struct boolconj_type *dvl) {
/*
  sprintf(tmpstring, "appending %s %s %d to ", 
                     var_name(b->var1), relop_string(b->op), b->constant);
  output(tmpstring);
  print_dvl(dvl);
*/
  struct boolconj_type *bb;
  if(dvl) {
    for(bb=dvl->next; ; bb=bb->next) {
      if( (bb->var1 == b->var1) &&
          (bb->var2 == b->var2) &&
          (bb->op == b->op) &&
          (bb->constant == b->constant)
        ) {
        free_boolconj(b);
        b = NULL;
        return dvl;
      }
      if(bb==dvl) break;
    }
  }

  if(dvl) {
    b->next = dvl->next;
    dvl->next = b;
  }
  else b->next = b;
  dvl = b;
  return dvl;
}

struct boolconj_type *boolconj_union(struct boolconj_type *b1,
                                     struct boolconj_type *b2) {
  int found = FALSE;
  struct boolconj_type *b = NULL, *bp1, *bp2, *bn;

  if(b1 == NULL && b2 == NULL) return NULL;
  if(b1 == NULL) { b = boolconj_copy(b2); return b; }
  if(b2 == NULL) { b = boolconj_copy(b1); return b; }

  if(b1->op == TRUE) { b = boolconj_copy(b2); return b; }
  if(b2->op == TRUE) { b = boolconj_copy(b1); return b; }

  if(b1->op == FALSE) { b = boolconj_copy(b1); return b; }
  if(b2->op == FALSE) { b = boolconj_copy(b2); return b; }

  b = boolconj_copy(b1);
  for(bp2=b2->next; ; bp2=bp2->next) {
    for(bp1=b1->next; ; bp1=bp1->next) {
      if(bp1->var1 == bp2->var1 &&
         bp1->var2 == bp2->var2 &&
         bp1->op == bp2->op &&
         bp1->constant == bp2->constant) {
         found = TRUE;
         break;
      }
      if(bp1 == b1) break;
    }
    if(found) found = FALSE;
    else {
      bn = (struct boolconj_type *)malloc(sizeof(struct boolconj_type));
      bn->var1 = bp2->var1;
      bn->var2 = bp2->var2;
      bn->op = bp2->op;
      bn->constant  = bp2->constant;
      bn->prop_bdd = bp2->prop_bdd;
      if(b) {
        bn->next = b->next;
        b->next = bn;
      }
      else bn->next = bn;
      b = bn;
    }
    if(bp2 == b2) break;
  }
  b = check_boolconj(b, b2);
  return b;
}

struct boolconj_type *check_boolconj(struct boolconj_type *b,
                                     struct boolconj_type *b2) {
  struct boolconj_type *bp, *bps;

  if(get_bdd_from_dvl(b) != BDD_FALSE) return b;
  if(b) {
    for(bp=b->next, bps=bp->next; ; bp=bps, bps=bps->next) {
      if(bp->var1 == b2->var1 && bp->constant != b2->constant) {
        b = delete_boolconj(bp->var1, bp->constant, b);
        if(b == NULL) break;
      }
      if(bps == b) break;
    }
  }
  return b;
}

/* insert val into int_vals array in an ascending order of the values */

int insert_int(int val, int var_pos, int **int_vals, int tot_vals) {
  int i, j;

/*
  sprintf(tmpstring, "  appending value %d to ", val);
  output(tmpstring);
  print_vals(int_vals[var_pos], tot_vals);
*/

  for(i=1; i<=tot_vals; i++) {
    if(int_vals[var_pos][i] == val)  /* duplicate value exists in array */
      return FALSE;
    if(int_vals[var_pos][i] > val) break;
  }
  if(int_vals[var_pos])
    int_vals[var_pos] = (int *)realloc(int_vals[var_pos], 
                                       sizeof(int)*(tot_vals+2));
  else
    int_vals[var_pos] = (int *)malloc(sizeof(int)*(tot_vals+2));
  for(j=tot_vals+1; j>i; j--) {
    int_vals[var_pos][j] = int_vals[var_pos][j-1];
  }
  int_vals[var_pos][i] = val;
  return TRUE;
}

int delete_int(int val, int var_pos, int **int_vals, int tot_vals) {
  int i, j, found = FALSE;

/*
  sprintf(tmpstring, "  deleting value %d from ", val);
  output(tmpstring);
  print_vals(int_vals[var_pos], tot_vals);
*/

  for(i=1; i<=tot_vals; i++) {
    if(int_vals[var_pos][i] == val) {
      found = TRUE;
      break;
    }
  }
  if(found) {
    if(i < tot_vals) {
      for(j=i; j<tot_vals; j++) {
        int_vals[var_pos][j] = int_vals[var_pos][j+1];
      }
    }
    if(int_vals[var_pos])
      int_vals[var_pos] = (int *)realloc(int_vals[var_pos], 
                                         sizeof(int)*(tot_vals+1));
    else
      int_vals[var_pos] = (int *)malloc(sizeof(int)*(tot_vals+1));
    return TRUE;
  }
  else return FALSE;
}

int print_vals(int *v, int tot_vals) {
  int i;
  sprintf(tmpstring, "Var %s: {", var_name(v[0]));
  output(tmpstring);
  for(i=1; i<=tot_vals; i++) {
    sprintf(tmpstring, "%d", v[i]);
    output(tmpstring);
    if(i<tot_vals) output(", ");
  }
  output("}\n");
  return 0;
}

int print_dvl(struct boolconj_type *dvl) {
  int v;
  struct boolconj_type *b;

  output("{");
  if(dvl) {
    for(b=dvl->next; ; b=b->next) {
      v = var_id(b->var1);

      if(v >= 0) sprintf(tmpstring, "%s[%d] ", var_name(b->var1), v);
      else       sprintf(tmpstring, "%s ", var_name(b->var1));
      output(tmpstring);

      if(b->var2 > 0) {
        v = var_id(b->var2);
        if(v >= 0) sprintf(tmpstring, "- %s[%d] ", var_name(b->var2), v);
        else       sprintf(tmpstring, "- %s ", var_name(b->var2));
        output(tmpstring);
      }

      sprintf(tmpstring, "%s %d", relop_string(b->op), b->constant);
      output(tmpstring);

      if(b==dvl) break;
      else output(", ");
    }
  }
  output("}\n");
  return 0;
}

int install_dvl_bdd(struct boolconj_type *dvl) {
  struct proposition_type *p;
  struct boolconj_type *b;
  bdd f = BDD_FALSE;

  if(dvl) {
/*
    if(verbose >= 3) {
      output("For DVL: ");
      print_dvl(dvl);
    }
*/
    for(b = dvl->next; ; b=b->next) {
      for(p = prop_tail->next; ; p=p->next) {
        if(p->var1 == b->var1 && p->var2 == b->var2) {
          if(p->value == b->constant)
            b->prop_bdd = bdd_not(bddm, p->prop_bdd);
          else
            f = bdd_or(bddm, f, p->prop_bdd);
        } 
        if(p == prop_tail) break;
      }
      b->prop_bdd = bdd_and(bddm, b->prop_bdd, f);
      f = BDD_FALSE;
      if(b == dvl) break;
    }
  }
/*
  if(verbose >= 3) {
    output("... Installed DVL BDD: ");
    print_dvl_bdd(dvl);
  }
*/
  return 0;
}

int print_dvl_bdd(struct boolconj_type *dvl) {
  struct boolconj_type *b; 
  bdd dvl_bdd = BDD_TRUE;
  if(dvl) {
    for(b=dvl->next; ; b=b->next) {
      dvl_bdd = bdd_and(bddm, dvl_bdd, b->prop_bdd);
      if(b==dvl) break;
    }
  }
  print_bdd(dvl_bdd);
  return 0;
}

int dvl_comp(struct mode_type *m1, struct mode_type *m2) {
  bdd b1, b2;
  b1 = dvl_bdd(m1);
  b2 = dvl_bdd(m2);
  return bdd_comp(b1, b2);
}

/**************************************************************************
*
*       returns -2 if s1 is contained by s2
*               -1 if s1 is incomparable with s2
*                0 if s1 is identical to s2
*                2 if s1 contains s2
*
*       Assumption : the bdds and zones in s1 and s2 are all normalized and
*       have been recorded in bdd_tree and zone_tree.
#define CONTAINED       -2
#define LATER		-1
#define EQUAL            0
#define EARLIER    	 1
#define CONTAINS         2
*/

int mode_comp(struct mode_type *m1, struct mode_type *m2) {
  int comp0, comp1, comp2, comp3;

/*
sprintf(tmpstring, "Comparing modes: %d(%s) and %d(%s)\n", 
                    m1->index, m1->name, m2->index, m2->name);
output(tmpstring);
*/

  comp0 = strcmp(m1->name, m2->name);

/*
  sprintf(tmpstring, "Name comparison = %d\n", comp0);
  output(tmpstring);
*/
  if(comp0 < 0) return LATER;
  else if(comp0 > 0) return EARLIER;
  else {                              
/*
    if(compact_reduction_compare) {
      comp1 = bdd_comp_internal(m1->dval_bdd, m2->dval_bdd);
    }
    else 
*/
/*
    comp1 = bdd_comp(m1->dval_bdd, m2->dval_bdd);
    comp1 = cvl_comp(m1->cvl, m2->cvl);
*/
    if(clock_shielding)
      comp1 = shield_clocks_compare(m1->shielded_clocks, m2->shielded_clocks);
    else comp1 = EQUAL;
/*
    sprintf(tmpstring, "Dval BDD comparison = %d\n", comp1);
    output(tmpstring);
*/
    switch(comp1) {
      case EARLIER: 
        return EARLIER;
        break;
      case EQUAL:
        if(m1->dvl && m2->dvl) 
          comp2 = dvl_comp(m1, m2);
        else
          comp2 = bdd_comp(m1->dval_bdd, m2->dval_bdd);
        switch(comp2) {
          case EARLIER: return EARLIER; break;
          case LATER: return LATER; break;
          case EQUAL:
            comp3 = zone_comp(m1->zone, m2->zone);  

/*
            sprintf(tmpstring, "ZONE comparison = %d, where:\n", comp3);
            output(tmpstring);
            print_zone(m1->zone); output("\n");
            print_zone(m2->zone); output("\n");
*/

/*
            if(comp3 == CONTAINED) {
              comp_stat = CONTAINED;
              same_mode = m2;
              return EQUAL;
            }
            if(comp3 == CONTAINS) {
              comp_stat = CONTAINS;
              return EQUAL;
            }
*/
            if(comp3 == EQUAL) {
              if(boob == TRUE) {
                comp_stat = BOOB;
                boob = FALSE;
              }
              same_mode = m2;
            }
            return comp3;
            break;
        }
        break;
      case LATER:
        return LATER;
        break;
    }
  } 
  return EARLIER;
}
/*
  switch (comp1) {
  case CONTAINED :
    switch (comp2) {
    case CONTAINED :
      return(CONTAINED);
      break;
    case INCOMPARABLE :
      return(INCOMPARABLE);
      break;
    case EQUAL :
      return(CONTAINED);
      break;
    case CONTAINS :
      return(INCOMPARABLE);
      break;
    }
    break;
  case INCOMPARABLE :
    return(INCOMPARABLE);
    break;
  case EQUAL :
    switch (comp2) {
    case CONTAINED :
      return(CONTAINED);
      break;
    case INCOMPARABLE :
      return(INCOMPARABLE);
      break;
    case EQUAL :
      return(EQUAL);
      break;
    case CONTAINS :
      return(INCOMPARABLE);
      break;
    }
    break;
  case CONTAINS :
    switch (comp2) {
    case CONTAINED :
    case INCOMPARABLE :
      return(INCOMPARABLE);
      break;
    case EQUAL :
    case CONTAINS :
      return(CONTAINS);
      break;
    }
    break;
  }
*/
  /* mode_comp() */

struct transition_type *add_transition_fifo(tr, sp, ss)
     struct mode_type *sp, *ss;
     struct transition_type *tr;
{
  struct transition_type *trn;

  trn = copy_transition(tr);
  trn->mode = sp;
  trn->index = ++trans_count;
  ss->in_trans_count++;
  if(ss->in_trans_cycle) {
    trn->next = ss->in_trans_cycle->next;
    ss->in_trans_cycle->next = trn;
    ss->in_trans_cycle = trn;
  }
  else {
    ss->in_trans_cycle = trn->next = trn;
  }

  trn = copy_transition(tr);
  trn->mode = ss;
  trn->index = trans_count;
  sp->out_trans_count++;
  if(sp->out_trans_cycle) {
    trn->next = sp->out_trans_cycle->next;
    sp->out_trans_cycle->next = trn;
    sp->out_trans_cycle = trn;
  }
  else {
    sp->out_trans_cycle = trn->next = trn;
  }
  return trn;
}

struct transition_type *add_transition(tr, sp, ss)
     struct mode_type   *sp, *ss;
     struct transition_type *tr;
{
  struct transition_type        *trn=NULL, *ctr;
  int                           comp;

  trans_count++;

  if(ss->in_trans_cycle) {
    comp = transition_candidate_comp(ss->in_trans_cycle, sp);
    if (comp < 0) {
      trn = copy_transition(tr);
      trn->mode = sp;
/*
      trn->goto_mode = str_preserve(sp->name);
*/
      trn->index = trans_count;
      (ss->in_trans_count)++;
      trn->next = ss->in_trans_cycle->next;
      ss->in_trans_cycle->next = trn;
      ss->in_trans_cycle = trn;
    }
    else if (comp > 0) {
      for (ctr = ss->in_trans_cycle;
           (comp = transition_candidate_comp(ctr->next, sp)
            ) < 0;
           ctr = ctr->next) ;

      if (comp > 0) {
        trn = copy_transition(tr);
        trn->mode = sp;
/*
        trn->goto_mode = str_preserve(sp->name);
*/
        trn->index = trans_count;
        (ss->in_trans_count)++;
        trn->next = ctr->next;
        ctr->next = trn;
      }
    }
  }
  else {
    trn = copy_transition(tr);
    trn->mode = sp;
/*
    trn->goto_mode = str_preserve(sp->name);
*/
    trn->index = trans_count;
    ss->in_trans_count++;
    ss->in_trans_cycle = trn->next = trn;
  }

  if (sp->out_trans_cycle) {
    comp = transition_candidate_comp(sp->out_trans_cycle, ss);
    if (comp < 0) {
      trn = copy_transition(tr);
      trn->mode = ss;
/*
      trn->goto_mode = str_preserve(ss->name);
*/
      trn->index = trans_count;
      (sp->out_trans_count)++;
      trn->next = sp->out_trans_cycle->next;
      sp->out_trans_cycle->next = trn;
      sp->out_trans_cycle = trn;
    }
    else if (comp > 0) {
      for (ctr = sp->out_trans_cycle;
           (comp = transition_candidate_comp(ctr->next, ss)
            ) < 0;
           ctr = ctr->next) ;

      if (comp > 0) {
        trn = copy_transition(tr);
        trn->mode = ss;
/*
        trn->goto_mode = str_preserve(ss->name);
*/
        trn->index = trans_count;
        (sp->out_trans_count)++;           /* added by eric */
        trn->next = ctr->next;
        ctr->next = trn;
      }
    }
  }
  else {
    trn = copy_transition(tr);
    trn->mode = ss;
/*
    trn->goto_mode = str_preserve(ss->name);
*/
    trn->index = trans_count;
    sp->out_trans_count++;
    sp->out_trans_cycle = trn->next = trn;
  }

  return trn;
}
	  /* add_transition() */


/* DVL set computation functions */

/* Compute DVL (Discrete Variable Literal set) 
 * due to transition "tr", which is an out-going transition of mode "m"
 */
struct boolconj_type *compute_dvl(struct mode_type *m,
                                  struct transition_type *tr) {
  struct boolconj_type *dvl = NULL, *d1 = NULL, *d2 = NULL, *b = NULL;
  struct assign_type *a;
  struct variable_type *v;
  int val=-1;
  struct int_link_type *vallist = NULL, *vl;

  if(var_tail)
  for(v=var_tail->next; ; v=v->next) {
    if(v->type == V_REGISTER) {
      if(no_dvar_assign(tr, v))
    /* CASE (1): NO assignment to discrete-global-variables on transition tr */
      {
        if(neg_lit_trigger(tr, v)) {
          d1 = new_dvl_neg(tr, m, v);
          d2 = boolconj_union(dvl, d1);
          free_boolconj(dvl);
          free_boolconj(d1); d1 = NULL;
          dvl = d2;
          d2 = NULL;
          if(verbose >= 2) {
            sprintf(tmpstring, 
               "  Trans %d has NO assignment & NEGATIVE trigger (var = %s):\n",
                    tr->index, v->name);
            output(tmpstring);
            output("    DVL: ");
            print_dvl(dvl);
          }
        }
        else if(pos_lit_trigger(tr, v)) {
          d1 = new_dvl_pos(tr, v, m);

/*
output("after new_dvl_pos(): d1 = "); print_dvl(d1);
output("dvl = "); print_dvl(dvl);
*/

          d2 = boolconj_union(dvl, d1);

/*
output("after boolconj_union(): d2 = "); print_dvl(d2);
*/

          free_boolconj(dvl);
          free_boolconj(d1); d1 = NULL;
          dvl = d2;
          d2 = NULL;
          if(verbose >= 2) {
            if(v->id >= 0)
              sprintf(tmpstring,
            "  Trans %d has NO assignment & POSITIVE trigger (var = %s[%d]):\n",
                    tr->index, v->name, v->id);
            else
              sprintf(tmpstring,
                "  Trans %d has NO assignment & POSITIVE trigger (var = %s):\n",
                    tr->index, v->name);
            output(tmpstring);
            output("    DVL: ");
            print_dvl(dvl);
          }
        }
        else {
          d1 = extract_dvl(m->dvl, v);
/*
          output("after extract parent dvl: "); print_dvl(d1);
*/
          if(d1) {
            for(d2=d1->next, b=d2->next; ; d2=b, b=b->next) {
              if(not_accessed_by_self(d2, tr) == FALSE) {
                d1 = delete_boolconj(d2->var1, d2->constant, d1);
                if(d1 == NULL) break;
              }
              if(b == d1) break;
            }
          }
/*
          output("after deleting self: "); print_dvl(d1);
*/
          d2 = boolconj_union(dvl, d1);
          free_boolconj(dvl);
          free_boolconj(d1); d1 = NULL;
          dvl = d2;
          d2 = NULL;
          if(verbose >= 2) {
            if(v->id >= 0)
              sprintf(tmpstring, 
                "  Trans %d has NO assignment & NO trigger (var = %s[%d]):\n",
                    tr->index, v->name, v->id);
            else
              sprintf(tmpstring, 
                "  Trans %d has NO assignment & NO trigger (var = %s):\n",
                    tr->index, v->name);
            output(tmpstring);
            output("    DVL: ");
            print_dvl(dvl);
          }
        }
      }
      else
    /* CASE (2) : Assignment to discrete-global-variables on transition tr */
      {  
        for(a=tr->trn->assign_cycle->next; ; a=a->next) {
          if(a->var == v->index) {
/*
sprintf(tmpstring, "d1 = "); output(tmpstring); print_dvl(d1);
sprintf(tmpstring, "dvl = "); output(tmpstring); print_dvl(dvl);
*/
            d1 = initial_dvl(THREAD1, THREAD2, FALSE, v);
/*
sprintf(tmpstring, "after initial_dvl(): d1 = "); output(tmpstring); print_dvl(d1);
*/
            if(d1) {
              for(d2=d1->next, b=d2->next; ; d2=b, b=b->next) {
                if(not_accessed_by_self(d2, tr) == FALSE) {
                  d1 = delete_boolconj(d2->var1, d2->constant, d1);
                  if(d1 == NULL) break;
                }
                if(b == d1) break;
              }
            }
/*
sprintf(tmpstring, "after deleting self: d1 = "); output(tmpstring); print_dvl(d1);
*/
            d2 = NULL;
            if(a->use_var == 0) {
              if(a->mod_var) 
                fprintf(stderr, "MOD var not implemented.\n");
              else if(a->mod_num)
                val = (int)fmod(a->value, a->mod_num);
              else
                val = a->value;
              d1 = delete_boolconj(a->var, val, d1);
            }
            else if(a->use_var) {
              vallist = get_curr_values(a->use_var, m->dval_bdd);
              if(vallist)
              for(vl=vallist->next; ; vl=vl->next) {
                if(a->op == ADD) {
                  if(a->mod_num)
                    val = (int)fmod(vl->k + a->value, a->mod_num);
                  else if(a->mod_var)
                    fprintf(stderr,
                    "Error: assignment with MOD var not yet implemented.\n");
                  else val = vl->k + a->value;
                } 
                else if(a->op == SUB) {
                  if(a->mod_num)
                    val = (int)fmod(vl->k - a->value, a->mod_num);
                  else if(a->mod_var)
                    fprintf(stderr,
                      "Error: assignment with MOD var not yet implemented.\n");
                  else val = vl->k - a->value;
                }
                else {
                  if(a->mod_num)
                    val = (int)fmod(vl->k, a->mod_num);
                  else if(a->mod_var)
                    fprintf(stderr,
                      "Error: assignment with MOD var not yet implemented.\n");
                  else val = vl->k;
                }
                d1 = delete_boolconj(a->var, val, d1);
                if(vl == vallist) break; 
              }
              free_int_link(vallist); vallist = NULL;
            }
/*
sprintf(tmpstring, "after delete_boolconj(): d1 = "); output(tmpstring); print_dvl(d1);
*/
            d2 = boolconj_union(dvl, d1);
/*
sprintf(tmpstring, "after boolconj_union(): d2 = "); output(tmpstring); print_dvl(d2);
*/
            free_boolconj(dvl);
            free_boolconj(d1); d1 = NULL;
            dvl = d2;
/*
sprintf(tmpstring, "finally: dvl = "); output(tmpstring); print_dvl(dvl);
*/
            d2 = NULL;
          }
          if(a == tr->trn->assign_cycle) break;
        }
        if(verbose >= 2) {
          if(v->id >= 0)
            sprintf(tmpstring, "  Trans %d has ASSIGNMENT (var = %s[%d]):\n",
                    tr->index, v->name, v->id);
          else
            sprintf(tmpstring, "  Trans %d has ASSIGNMENT (var = %s):\n",
                    tr->index, v->name);
          output(tmpstring);
          output("    DVL: ");
          print_dvl(dvl);
        }
      }
    }
    if(v == var_tail) break;
  }
  return dvl;
}

int free_int_link(struct int_link_type *i) {
  struct int_link_type *il;
  if(i) {
    il = i->next;
    i->next = NULL;
    i = il;
    while(i) {
      il = i;
      i = i->next;
      free(il);
      il = NULL;
    }
  }
  return 0;
}

/* Extract all literals in "d" that are related to "v"
 */
struct boolconj_type *extract_dvl(struct boolconj_type *d, 
                                  struct variable_type *v) {
  struct boolconj_type *b;
  struct boolconj_type *ans = NULL, *copy;

  if(d) {
    for(b=d->next; ; b=b->next) {
      if(b->var1 == v->index) {
        copy = (struct boolconj_type *)malloc(sizeof(struct boolconj_type));
        copy->var1 = b->var1;
        copy->var2 = b->var2;
        copy->op = b->op;
        copy->constant  = b->constant;
        copy->prop_bdd = b->prop_bdd;
        bdd_unfree(bddm, b->prop_bdd);
        if(ans) {
          copy->next = ans->next;
          ans->next = copy;
        }
        else copy->next = copy;
        ans = copy;
      }
      if(b == d) break;
    }
  }
  return ans;
}

/* Returns TRUE if there exists a trigger "b->var1 != b->constant" of tr
 */
int neg_lit_trigger(struct transition_type *tr, struct variable_type *v) {
  struct boolconj_type *b;
  if(tr->trn->prp_trigger_cycle) {
    for(b=tr->trn->prp_trigger_cycle->next; ; b=b->next) {
      if(b->var1 == v->index && b->op == NEQ) 
        return TRUE;
      if(b == tr->trn->prp_trigger_cycle) break;
    }
  }
  return FALSE;
}

/* Returns TRUE if there exists a trigger "b->var1 == b->constant" of tr
 */
int pos_lit_trigger(struct transition_type *tr, struct variable_type *v) {
  struct boolconj_type *b;
  if(tr->trn->prp_trigger_cycle) {
    for(b=tr->trn->prp_trigger_cycle->next; ; b=b->next) {
      if(b->var1 == v->index && b->op == EQ)
        return TRUE;
      if(b == tr->trn->prp_trigger_cycle) break;
    }
  }
  return FALSE;
}

/* (1) Extract all literals in "dvl" related to "v" into "ndvl",
 * (2) Append new literals "b->var1 != b->constant" into "ndvl",
 * (3) Return "ndvl".
 */
struct boolconj_type *new_dvl_neg(struct transition_type *tr, 
                                  struct mode_type *m,
                                  struct variable_type *v) {
  struct boolconj_type *b, *nb, *ndvl, *dvl=m->dvl;
  struct int_link_type *vallist, *vl;
  bdd f = m->dval_bdd;
  int val = 0;

  ndvl = extract_dvl(dvl, v);
/*
  output("Parent DVL: ");
  print_dvl(dvl);
  output("Extracted DVL: ");
  print_dvl(ndvl);
*/
  if(ndvl) {
    for(nb=ndvl->next, b=nb->next; ; nb=b, b=b->next) {
      if(not_accessed_by_self(nb, tr) == FALSE) {
        ndvl = delete_boolconj(nb->var1, nb->constant, ndvl);
        if(ndvl == NULL) break;
      }
      if(b == ndvl) break;
    }
  }
  if(tr->trn->prp_trigger_cycle) {
    for(b=tr->trn->prp_trigger_cycle->next; ; b=b->next) {
      if( (b->var1 == v->index) && (b->op == NEQ) && 
          not_written_by_others(b) && not_accessed_by_self(b, tr) ){
        if(b->var2) {
          vallist = get_curr_values(b->var2, f);
          if(vallist)       
            for(vl=vallist->next; ; vl=vl->next) {
              if(b->mod_num)
                  val = (int)fmod(vl->k + b->constant, b->mod_num);
              else if(b->mod_var)
                  fprintf(stderr,
                    "Error: assignment with MOD var not yet implemented.\n");
              else val = vl->k + b->constant;
              nb = alloc_boolconj();
              nb->var1 = b->var1;
              nb->op = b->op;
              nb->constant = val;
              ndvl = append_boolconj(nb, ndvl);
              if(vl == vallist) break;
            }
          free_int_link(vallist); vallist = NULL;
        }
        else {
          if(b->mod_num)
            val = (int)fmod(b->constant, b->mod_num);
          else if(b->mod_var)
            fprintf(stderr,
                   "Error: assignment with MOD var not yet implemented.\n");
          else val = b->constant;
          nb = alloc_boolconj();
          nb->var1 = b->var1;
          nb->op = b->op;
          nb->constant = val;
          ndvl = append_boolconj(nb, ndvl);
        }
/*
        nb = alloc_boolconj();
        nb->var1 = b->var1;
        nb->op = b->op;
        nb->var2 = b->var2;
        nb->constant = b->constant;
        ndvl = append_boolconj(nb, ndvl);
*/
  /*
        if(tr->permutation) {
          tmp_bool = permute_boolconj(nb, tr->permutation->perm);
          free_boolconj(nb);
          nb = tmp_bool;
        }
  */
      }
      if(b == tr->trn->prp_trigger_cycle) break;
    }
  }
  install_dvl_bdd(ndvl);
  return ndvl;
}

/* (1) Construct an initial DVL set with literals "v != x," where "x" is an 
 *     internal value of "THREAD1" and "THREAD2" (i.e., "v" is not assigned "x"
 *     by any external TA).
 * (2) Delete from initial DVL set all literals "v != y," where "y" is a value 
 *     that "x" is supposed to have in the trigger of "tr" 
 *     (i.e., "x == y" is a trigger of "tr")
 * (3) Return the resultant DVL set.
 */
struct boolconj_type *new_dvl_pos(struct transition_type *tr, 
                                  struct variable_type *v, struct mode_type *m) {
  struct boolconj_type *b, *d2, *dvl;
  struct int_link_type *vallist = NULL, *vl = NULL;
  int val=-1;
  bdd f = m->dval_bdd;

  dvl = initial_dvl(THREAD1, THREAD2, FALSE, v);

/*
output("after initial_dvl(): dvl = "); print_dvl(dvl);
*/

  if(dvl) {
    for(d2=dvl->next, b=d2->next; ; d2=b, b=b->next) {
      if(not_accessed_by_self(d2, tr) == FALSE) {
        if(d2 == dvl) {
          dvl = delete_boolconj(d2->var1, d2->constant, dvl);
          break;
        }
        else {
          dvl = delete_boolconj(d2->var1, d2->constant, dvl);
          continue;
        }
        if(dvl == NULL) break;
      }
      if(d2 == dvl) break;
    }
  }
/*
output("after deleting self: dvl = "); print_dvl(dvl);
*/
  if(m->dvl) {
    for(d2=m->dvl->next; ; d2=d2->next) {
      if(not_accessed_by_self(d2, tr) == FALSE) {
        dvl = delete_boolconj(d2->var1, d2->constant, dvl);
      }
      if(d2 == m->dvl) break;
    }
  }
/*
output("parent dvl = "); print_dvl(m->dvl);
output("after deleting m->dvl: dvl = "); print_dvl(dvl);
*/
  if(tr->trn->prp_trigger_cycle) {
    for(b = tr->trn->prp_trigger_cycle->next; ; b = b->next) {
      if(b->var1 == v->index && b->op == EQ) {
        if(b->var2 == 0) {
          if(b->mod_var)
            fprintf(stderr, "MOD var not implemented.\n");
          else if(b->mod_num)
            dvl = delete_boolconj(b->var1, (int)fmod(b->constant, b->mod_num), 
                                  dvl);
          else
            dvl = delete_boolconj(b->var1, b->constant, dvl);
        }
        else { /* b->var1 - b->var2 = b->constant {MOD mod_num} */
          vallist = get_curr_values(b->var2, f);
          if(vallist)
            for(vl=vallist->next; ; vl=vl->next) {
              if(b->mod_num)
                  val = (int)fmod(vl->k + b->constant, b->mod_num);
              else if(b->mod_var)
                  fprintf(stderr, 
                    "Error: assignment with MOD var not yet implemented.\n");
              else val = vl->k + b->constant;
              dvl = delete_boolconj(b->var1, val, dvl);
              if(vl == vallist) break;
            }
          free_int_link(vallist); vallist = NULL;
        }
      }
      if(b == tr->trn->prp_trigger_cycle) break;
    }
  }
  return dvl;
}

/* Returns FALSE if variable "b->var1" is not assigned value "b->constant"
 * by any external TA
 */
int not_written_by_others(struct boolconj_type *b) {
  int i;
/*
  struct mode_type *m;
  struct transition_type *tr;
  struct assign_type *a;
*/

  for(i=0; i<system_ta_num; i++) {
    if(!component_graph(i, NEW_THREAD))
      if(varval_written(b->var1, b->constant, &thread[i])) 
        return FALSE;
  }
/*
    for(m=thread[i].starting_mode; m; m=m->cnext) {
      if(m->out_trans_cycle) {
        for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
          if(tr->trn->assign_cycle) {
            for(a=tr->trn->assign_cycle->next; ; a=a->next) {
              if( (a->var1 == b->var1) && 
                  (a->var2 == b->var2) && 
                  (a->value == b->constant) )
                return FALSE;
              if(a == tr->trn->assign_cycle) break;
            }
          }
          if(tr==m->out_trans_cycle) break;
        }
      }
    }
*/
  return TRUE;
}

int not_accessed_by_self1(struct boolconj_type *b, struct transition_type *t) {
  return TRUE;
}
/* Returns FALSE when variable "b->var1" is written/read value "b->constant"
 * by any other in-coming transition of mode "t->mode" 
 */
int not_accessed_by_self(struct boolconj_type *b1, struct transition_type *t) {
  struct transition_type *tr;
  struct assign_type *a;
  struct int_link_type *vallist=NULL, *vl=NULL;
  struct boolconj_type *bc, *b=NULL;
  int val=-1, *tmp_perm1 = NULL, *tmp_perm2 = NULL, found_pos_trigger=FALSE;


  for(tr=t->mode->in_trans_cycle->next; ; tr=tr->next) {
    if(tr->index == t->index) {
      if(tr == t->mode->in_trans_cycle) break;
      else continue;
    }

    if(tr->permutation)
      if(tr->permutation->perm)
        tmp_perm1 = inverse_permutation(tr->permutation->perm);
    if(t->permutation) {
      if(t->permutation->perm)
        tmp_perm2 = compose_permutation(t->permutation->perm, tmp_perm1);
    }
    else if(tmp_perm1)
      tmp_perm2 = tmp_perm1;
    if(tmp_perm2)
      b = permute_boolconj1(b1, tmp_perm2);
    else b = boolconj_copy1(b1);
    if(tmp_perm1 == tmp_perm2) {
      free(tmp_perm1); tmp_perm1 = NULL; tmp_perm2 = NULL;
    }
    else {
      free(tmp_perm1); tmp_perm1 = NULL;
      free(tmp_perm2); tmp_perm2 = NULL;
    }

    if(tr->trn->assign_cycle)
    for(a=tr->trn->assign_cycle->next; ; a=a->next) {
      if(a->var == b->var1) {
        if(a->use_var) {
          vallist = get_curr_values(a->use_var, t->mode->dval_bdd);
          if(vallist) {
            for(vl=vallist->next; ; vl=vl->next) {
              if(a->op == ADD) {
                if(a->mod_num)
                  val = (int)fmod(vl->k + a->value, a->mod_num);
                else if(a->mod_var)
                  fprintf(stderr,
                  "Error: assignment with MOD var not yet implemented.\n");
                else val = vl->k + a->value;
              }
              else if(a->op == SUB) {
                if(a->mod_num)
                  val = (int)fmod(vl->k - a->value, a->mod_num);
                else if(a->mod_var)
                  fprintf(stderr,
                    "Error: assignment with MOD var not yet implemented.\n");
                else val = vl->k - a->value;
              }
              else {
                if(a->mod_num)
                  val = (int)fmod(vl->k, a->mod_num);
                else if(a->mod_var)
                  fprintf(stderr,
                    "Error: assignment with MOD var not yet implemented.\n");
                else val = vl->k;
              }
              if(val == b->constant) {
                free_int_link(vallist); vallist = NULL;
                if(b) free_boolconj(b);
                return FALSE;
              }
              if(vl == vallist) break;
            }
            free_int_link(vallist); vallist = NULL;
          }
        }
        else {
          if(a->mod_num)
            val = (int)fmod(a->value, a->mod_num);
          else if(a->mod_var)
                fprintf(stderr,
                  "Error: assignment with MOD var not yet implemented.\n");
          else val = a->value;
          if(val == b->constant) {
            if(b) free_boolconj(b);
            return FALSE;
          }
        }
        if(b) free_boolconj(b);
        return TRUE;
      }
      if(a == tr->trn->assign_cycle) break;
    }
    if(tr->trn->prp_trigger_cycle)
    for(bc = tr->trn->prp_trigger_cycle->next; ; bc = bc->next) {
      if(bc->var1 == b->var1 && bc->op == EQ) {
        if(bc->var2 == 0) {
          if(bc->mod_var)
            fprintf(stderr, "MOD var not implemented.\n");
          else if(bc->mod_num)
            val = (int)fmod(bc->constant, bc->mod_num);
          else
            val = bc->constant;
          if(b->constant == val) {
            if(b) free_boolconj(b);
            return FALSE;
          }
        }
        else { /* bc->var1 - bc->var2 = bc->constant {MOD mod_num} */
          vallist = get_curr_values(bc->var2, t->mode->dval_bdd);
          if(vallist) {
            for(vl=vallist->next; ; vl=vl->next) {
              if(bc->mod_num)
                  val = (int)fmod(vl->k + bc->constant, bc->mod_num);
              else if(bc->mod_var)
                  fprintf(stderr,
                    "Error: assignment with MOD var not yet implemented.\n");
              else val = vl->k + bc->constant;
              if(val == b->constant) {
                free_int_link(vallist); vallist = NULL;
                if(b) free_boolconj(b);
                return FALSE;
              }
              if(vl == vallist) break;
            }
            free_int_link(vallist); vallist = NULL;
          }
        }
        found_pos_trigger = TRUE;
      }
      if(bc == tr->trn->prp_trigger_cycle) break;
    }
    if(found_pos_trigger == FALSE) {
      if(BDD_AND(get_bdd_from_dvl(b), 
                 get_bdd_from_dvl(source_mode(tr)->dvl)) == BDD_FALSE)
      if(b) { free_boolconj(b); b = NULL; }
      return FALSE;
    }
    if(b) { free_boolconj(b); b = NULL; }
    if(tr == t->mode->in_trans_cycle) break;
  }
  return TRUE;
}

int no_dvar_assign(struct transition_type *tr, struct variable_type *v) {
  struct assign_type *a;
  if(tr->trn->assign_cycle) {
    for(a=tr->trn->assign_cycle->next; ; a=a->next) {
      if(a->var == v->index) return FALSE;
      if(a == tr->trn->assign_cycle) break;
    }
  }
  return TRUE;
}

struct boolconj_type *delete_boolconj(int var, int val,
                                      struct boolconj_type *dvl) {
  struct boolconj_type *bp, *b;

  if(dvl == NULL) return dvl;
  if(dvl->next == dvl) {
    if(dvl->var1 == var && dvl->constant == val) {
      free_boolconj(dvl);
      return NULL;
    }
    else return dvl;
  }
  for(bp=dvl, b=bp->next; ; bp=b, b=b->next) {
    if(var == b->var1 && val == b->constant) {
      bp->next = b->next;
      if(b == dvl) dvl = bp;
      b->next = b;
      free_boolconj(b);
      break;
    }
    if(b==dvl) break;
  }
  return dvl;
}

bdd dvl_bdd(struct mode_type *m) {
  struct boolconj_type *b;
  bdd f = BDD_TRUE;
  if(m->dvl) {
    for(b=m->dvl->next; ; b=b->next) {
      f = BDD_AND(f, b->prop_bdd);
      if(b==m->dvl) break;
    }
  }
  return f;
}

bdd get_bdd_from_dvl(struct boolconj_type *dvl) {
  struct boolconj_type *b;
  bdd f = BDD_TRUE;
  if(dvl) {
    for(b = dvl->next; ; b=b->next) {
      f = BDD_AND(f, b->prop_bdd);
      if(b == dvl) break;
    }
  }
  return f;
}


/* 
   Following are functions used in transition_m1_first() and 
                                   transition_m2_first() 
*/

void print_info_tr(struct transition_type *tr) {
  if(verbose >= 0) {
   if(tr->automaton) {
     if(tr->automaton->index >= 0) {
      sprintf(tmpstring,"\n* considering transition %d (of %s[%d] to %s)",
      tr->index, tr->automaton->name, tr->automaton->index, tr->mode->name);
      output(tmpstring);
     }
     else {
      sprintf(tmpstring,"\n* considering transition %d (of %s to %s)",
      tr->index, tr->automaton->name, tr->mode->name);
      output(tmpstring);
     }
   }
   else {
      sprintf(tmpstring, "\n* considering sync transition %d", tr->index);
      output(tmpstring);
   }
   if(region_normalization) {
     output(" with ");
     print_permutation(tr->permutation);
   }
   else output("\n");
  }
}

/* Used in transition_m1_first() and transition_m2_first() */

void print_info_tr2(struct transition_type *tr) {
  if(verbose >= 0) {
   if(tr->automaton) {
     if(tr->automaton->index >= 0) {
      sprintf(tmpstring,"\n  * syncing with transition %d (of %s[%d] to %s)",
      tr->index, tr->automaton->name, tr->automaton->index, tr->mode->name);
      output(tmpstring);
     }
     else {
      sprintf(tmpstring,"\n  * syncing with transition %d (of %s to %s)",
      tr->index, tr->automaton->name, tr->mode->name);
      output(tmpstring);
     }
   }
   else {
      sprintf(tmpstring,"\n  * syncing with sync transition %d", tr->index);
      output(tmpstring);
   }
   if(region_normalization) {
     output(" with ");
     print_permutation(tr->permutation);
   }
   else output("\n");
  }
}

/* Used in transition_m1_first() and transition_m2_first() */

void print_info_nm(struct mode_type *m, struct transition_type *tr,
                   struct mode_type *nm) {
  if(verbose == 0) {
    sprintf(tmpstring, "mode %s(%d) [%s(%d), %s(%d)]: ", 
     nm->name, nm->index, nm->submode1->name, nm->submode1->index, 
                                 nm->submode2->name, nm->submode2->index);
    output(tmpstring);
    if(clock_shielding) {
      output("shielded clocks = ");
      print_var_links(nm->shielded_clocks);
    }
    else output("\n");
  }

  if(verbose > 0) {
    sprintf(tmpstring, 
    "\n## New mode created %s[%d(%d,%d)] from parent mode %s[%d(%d,%d)] ##\n",
    nm->name, nm->index, nm->submode1->index, nm->submode2->index,
    m->name, m->index, m->submode1->index, m->submode2->index);
    output(tmpstring);
  }

  if(verbose >= 2) {
    sprintf(tmpstring, 
     	"Parent mode %s[%d(%d,%d)] has invariant BDD:\n\t",
	   m->name, m->index, m->submode1->index, m->submode2->index);
    output(tmpstring);
    print_bdd(m->boolexp->boolexp_bdd);
    sprintf(tmpstring, 
    	"Parent mode %s[%d(%d,%d)] has discrete valuation BDD:\n\t",
	   m->name, m->index, m->submode1->index, m->submode2->index);
    output(tmpstring);
    print_bdd(m->dval_bdd);
    sprintf(tmpstring, "Transition (%d) has trigger BDD:\n\t", tr->index);
    output(tmpstring);
    if(tr->trn->trigger_bdd) print_bdd(tr->trn->trigger_bdd);
    else output("NULL BDD\n");
  }
}

/* Used in transition_m1_first() and transition_m2_first() */

void print_info_zone(struct mode_type *nm) {
  if(total_clocks > 0) {
    if(verbose == 0) {
      output("\tZone: ");
      print_zone(nm->zone);
      output("\n");
    }
    if(verbose > 0) {
      sprintf(tmpstring, "ZONE of mode (%s) [%d(%d, %d)] is:\n\t", 
              nm->name, nm->index, nm->submode1->index, nm->submode2->index);
      output(tmpstring);
      print_zone(nm->zone);
      output("\n");
    }
  }
}

/* Used in transition_m1_first() and transition_m2_first() */

void print_info_bdd(struct mode_type *nm) {
  if(verbose == 0) {
/*
    output("\tinv bdd: "); 
    print_bdd(nm->boolexp->boolexp_bdd);
*/
    if(lazy_bdd_merging) {
      output("\tdval bdd: "); 
      print_bdd(nm->dval_bdd);
    }
  }
  if(verbose > 0) {
/*
    sprintf(tmpstring, "Invariant BDD of mode (%s) [%d(%d, %d)] is:\n\t", 
	 nm->name, nm->index, nm->submode1->index, nm->submode2->index);
    output(tmpstring);
    print_bdd(nm->boolexp->boolexp_bdd);
*/
    if(lazy_bdd_merging) {
      sprintf(tmpstring, 
	"Discrete variable valuation BDD of mode (%s) [%d(%d, %d)] is:\n\t",
	 nm->name, nm->index, nm->submode1->index, nm->submode2->index);
      output(tmpstring);
      print_bdd(nm->dval_bdd);
    }
  }
}

/* Used in transition_m1_first() and transition_m2_first() */

void handle_comp_stat() {
  if(comp_stat == EQUAL) {
/*
    if(on_the_fly_normalization) {
      update_same_mode_trans_perm(new_mode_perm, ntr1, nnm);
      free(new_mode_perm);
    }
*/
    if(verbose == 0) {
      sprintf(tmpstring,  
              "\tNew mode similar to existing mode #%d.\n", 
                                                   same_mode->index);
      output(tmpstring);
    }
    if(verbose > 0) {
      sprintf(tmpstring, 
              "Found similar mode #%d in the composed modes\n", 
                               same_mode->index);
      output(tmpstring);
    }
  }
  if(comp_stat == BOOB) {
    comp_stat = EQUAL;
    if(verbose == 0) {
      sprintf(tmpstring,  
              "\tNew mode out of bounds with existing mode #%d.\n", 
                                                   same_mode->index);
      output(tmpstring);
    }
    if(verbose > 0) {
      sprintf(tmpstring, 
           "Found out of bounds mode #%d in the composed modes\n", 
                               same_mode->index);
      output(tmpstring);
    }
  }
/*
  else if(comp_stat == CONTAINS) {
    comp_stat = EQUAL;
    if(verbose == 0) {
      sprintf(tmpstring,
       "\tNew mode contains some existing mode %d.\n", same_mode->index);
      output(tmpstring);
    }
    if(verbose > 0) 
      output("New mode contains some existing composed mode.\n");
  }
*/
  else if(comp_stat == CONTAINED) {
    comp_stat = EQUAL;
    if(verbose == 0) { 
      sprintf(tmpstring, 
       "\tNew mode contained in existing mode #%d.\n", same_mode->index);
      output(tmpstring);
    }
    if(verbose > 0) {
      sprintf(tmpstring, 
       "New mode contained in existing mode #%d.\n", same_mode->index);
      output(tmpstring);
    }
  }
}

/* Used in transition_m1_first() and transition_m2_first() */

int *get_perm(struct mode_type *m, struct transition_type *tr) {
  int *tmp_perm1 = (int *)NULL, 
      *tmp_perm2 = (int *)NULL;

  if(on_the_fly_normalization && m->norm_perm)
    tmp_perm1 = inverse_permutation(m->norm_perm->perm);
  if(tr->permutation) {
    tmp_perm2 = compose_permutation(tmp_perm1, tr->permutation->perm);
    free(tmp_perm1);
    tmp_perm1 = (int *)NULL;
  }
  else {
    tmp_perm2 = tmp_perm1;
    tmp_perm1 = (int *)NULL;
  }
  return tmp_perm2;
}

int create_permute_dvl(struct mode_type *m, struct transition_type *tr,
                       struct mode_type *nm, int *tmp_perm2) {
  struct boolconj_type *tmp_bc = NULL;

  if(!lazy_bdd_merging) {
    if(internal_trans_bypass) {
      if(!lastcomp && 
        (BDD_AND(dvl_bdd(m), 
               get_bdd_from_dvl(tmp_bc=hide_vars(tr->bypassed_internal_vars, 
                                            tr->trn->prp_trigger_cycle)))
         == BDD_FALSE))
      {
        free_boolconj(tmp_bc); tmp_bc=NULL;
        if(verbose >= 0) 
          output("Invalid transition (BDD: DVL & trigger = false).\n");
        return 1;
      }
      free_boolconj(tmp_bc); tmp_bc=NULL;
    }
    else
    if(!lastcomp && (BDD_AND(dvl_bdd(m), tr->trn->trigger_bdd) == BDD_FALSE))
    {
      if(verbose >= 0) 
        output("Invalid transition (BDD: DVL & trigger = false).\n");
      return 1;
    }
    nm->dvl = compute_dvl(m, tr);
    if(tmp_perm2) {
      tmp_bc = permute_boolconj(nm->dvl, tmp_perm2);
      free_boolconj(nm->dvl);
      nm->dvl = tmp_bc;
      tmp_bc = NULL;
    }
    if(verbose >= 0) {
      output("\tDVL Set: ");
      print_dvl(nm->dvl);
    }
  }
  return 0;
}

int sync_create_permute_dvl(struct mode_type *m, struct transition_type *tr1,
                            struct transition_type *tr2, struct mode_type *nm) {
  struct boolconj_type *tmp_bc = NULL, *dvl1, *dvl2;

  if(!lazy_bdd_merging) {
    /* CHECK if new DVL becomes false */
    if(internal_trans_bypass) {
      if(!lastcomp && 
        (BDD_AND(dvl_bdd(m), 
               get_bdd_from_dvl(tmp_bc=hide_vars(tr1->bypassed_internal_vars, 
                                            tr1->trn->prp_trigger_cycle)))
         == BDD_FALSE))
      {
        free_boolconj(tmp_bc); tmp_bc=NULL;
        if(verbose >= 0) 
          output("Invalid transition (BDD: DVL & trigger = false).\n");
        return 1;
      }
      free_boolconj(tmp_bc); tmp_bc=NULL;
      if(!lastcomp && 
        (BDD_AND(dvl_bdd(m), 
               get_bdd_from_dvl(tmp_bc=hide_vars(tr2->bypassed_internal_vars, 
                                            tr2->trn->prp_trigger_cycle)))
         == BDD_FALSE))
      {
        free_boolconj(tmp_bc); tmp_bc=NULL;
        if(verbose >= 0) 
          output("Invalid transition (BDD: DVL & trigger = false).\n");
        return 1;
      }
      free_boolconj(tmp_bc); tmp_bc=NULL;
    }
    else
    if(!lastcomp && (BDD_AND(dvl_bdd(m), tr1->trn->trigger_bdd) == BDD_FALSE))
    {
      if(verbose >= 0) 
        output("Invalid transition (BDD: DVL & trigger = false).\n");
      return 1;
    }
    if(!lastcomp && (BDD_AND(dvl_bdd(m), tr2->trn->trigger_bdd) == BDD_FALSE))
    {
      if(verbose >= 0) 
        output("Invalid transition (BDD: DVL & trigger = false).\n");
      return 1;
    }
    dvl1 = get_new_dvl(m, tr1); /* compute and permute new DVL */
    dvl2 = get_new_dvl(m, tr2);
    nm->dvl = boolconj_union(dvl1, dvl2);
    free_boolconj(dvl1);
    free_boolconj(dvl2);
    if(verbose >= 0) {
      output("\tDVL Set: ");
      print_dvl(nm->dvl);
    }
  }
  return 0;
}

/* Used in transition_m1_first() and transition_m2_first() */

int create_permute_boolexp(struct mode_type *m, struct transition_type *tr,
                           struct mode_type *nm, int *tmp_perm2) {
  bdd tmp_bdd;

  if( (BDD_AND(m->boolexp->boolexp_bdd, 
               tr->trn->trigger_bdd)) == BDD_FALSE ) {
    if(verbose >= 0)
      output("Invalid transition (BDD: invariant & trigger = false).\n");
    return 1;
  }
  nm->boolexp = alloc_boolexp();
  nm->dval_bdd = (bdd)NULL;
  nm->boolexp->boolexp_bdd = BDD_AND(nm->submode1->boolexp->boolexp_bdd,
		                     nm->submode2->boolexp->boolexp_bdd); 

  if(region_normalization && tr->permutation) {
    tmp_bdd = permute_bdd(nm->boolexp->boolexp_bdd, tr->permutation->perm);
    nm->boolexp->boolexp_bdd = tmp_bdd;
  }
  if(on_the_fly_normalization) 
    nm->boolexp->boolexp_bdd = permute_bdd(nm->boolexp->boolexp_bdd, 
                                           tmp_perm2);
  return 0;
}

/* Used in transition_m1_first() and transition_m2_first() */

int sync_create_permute_boolexp(struct mode_type *m, 
                                struct transition_type *tr1,
                                struct transition_type *tr2,
                                struct mode_type *nm, 
                                int *tmp_perm1, int *tmp_perm2) {
  bdd tmp_bdd1, tmp_bdd2;

  if( BDD_AND(m->boolexp->boolexp_bdd, tr1->trn->trigger_bdd) == BDD_FALSE
     || 
      BDD_AND(m->boolexp->boolexp_bdd, tr2->trn->trigger_bdd) == BDD_FALSE ) {
    if(verbose >= 0)
      output("Invalid transition (BDD: invariant & trigger = false).\n");
    return 1;
  }
  nm->boolexp = alloc_boolexp();
  nm->dval_bdd = (bdd)NULL;
  if(region_normalization) {
    if(tr1->permutation)
      tmp_bdd1 = permute_bdd(nm->submode1->boolexp->boolexp_bdd, 
                             tr1->permutation->perm);
    if(tr2->permutation)
      tmp_bdd2 = permute_bdd(nm->submode2->boolexp->boolexp_bdd, 
                             tr2->permutation->perm);
    nm->boolexp->boolexp_bdd = BDD_AND(tmp_bdd1, tmp_bdd2);
  }
  else
    nm->boolexp->boolexp_bdd = BDD_AND(nm->submode1->boolexp->boolexp_bdd,
		                       nm->submode2->boolexp->boolexp_bdd); 

  if(on_the_fly_normalization) 
    nm->boolexp->boolexp_bdd = 
        BDD_AND(permute_bdd(nm->submode1->boolexp->boolexp_bdd, tmp_perm1),
                permute_bdd(nm->submode2->boolexp->boolexp_bdd, tmp_perm2));
  return 0;
}

/* Used in transition_m1_first() and transition_m2_first() */

int handle_last(struct mode_type *m, struct transition_type *tr) {
  struct boolconj_type *tmp_bc = NULL;
  bdd tmp_bdd;

  if( lastcomp ) {
/*
    tmp_bdd = BDD_AND(m->dval_bdd, tr->trn->trigger_bdd);
*/
    tmp_bdd = BDD_AND(m->dval_bdd, 
                  get_bdd_from_dvl(tmp_bc=hide_vars(tr->bypassed_internal_vars,
                                             tr->trn->prp_trigger_cycle)));
    free_boolconj(tmp_bc); tmp_bc = NULL;
    if(tmp_bdd == BDD_FALSE) {
	if(verbose >= 0) 
	  output("Invalid mode (BDD: discrete variable valuation = false).\n");
      return 1;
    }
  }
  return 0;
}

/* Used in transition_m1_first() and transition_m2_first() */

int sync_handle_last(struct mode_type *m, struct transition_type *tr1,
                     struct transition_type *tr2) {
  struct boolconj_type *tmp_bc = NULL;
  bdd tmp_bdd;

  if( lastcomp ) {
/*
    tmp_bdd = BDD_AND(m->dval_bdd, tr->trn->trigger_bdd);
*/
    tmp_bdd = BDD_AND(m->dval_bdd, 
                  get_bdd_from_dvl(tmp_bc=hide_vars(tr1->bypassed_internal_vars,
                                             tr1->trn->prp_trigger_cycle)));
    free_boolconj(tmp_bc); tmp_bc = NULL;
    if(tmp_bdd == BDD_FALSE) {
	if(verbose >= 0) 
	  output("Invalid mode (BDD: discrete variable valuation = false).\n");
      return 1;
    }
    tmp_bdd = BDD_AND(m->dval_bdd, 
                  get_bdd_from_dvl(tmp_bc=hide_vars(tr2->bypassed_internal_vars,
                                             tr2->trn->prp_trigger_cycle)));
    free_boolconj(tmp_bc); tmp_bc = NULL;
    if(tmp_bdd == BDD_FALSE) {
	if(verbose >= 0) 
	  output("Invalid mode (BDD: discrete variable valuation = false).\n");
      return 1;
    }
  }
  return 0;
}

/* Used in transition_m1_first() and transition_m2_first() */

int create_permute_dval(struct mode_type *m, struct transition_type *tr,
                        struct mode_type *nm, int *tmp_perm2, int type) {
  bdd tmp_bdd;

  if( BDD_AND((nm->dval_bdd = bdd_assign(m->dval_bdd, 
                                         tr->trn->assign_cycle)),
              nm->boolexp->boolexp_bdd
             ) == BDD_FALSE) {
    if(verbose >= 0) 
	  output("Invalid mode (BDD: discrete variable valuation = false).\n");
      return 1;
  }

  if(type==1 && EIV && composed_ta > 2) {
    if(BDD_AND(m->submode1->dval_bdd, tr->trn->trigger_bdd) != BDD_FALSE) {
      if(BDD_AND(m->dval_bdd, tr->trn->trigger_bdd) == BDD_FALSE) {
        if(verbose >= 0)
          output("Valid run became invalid after composition.\n");
        return 1;
      }
    }
  }

  if(type==2 && EOI && composed_ta > 2) {
    if(BDD_AND(m->dval_bdd, tr->trn->trigger_bdd) == BDD_FALSE) {
      if(own_index_trigger(tr)) {
        if(verbose >= 0)
          output("False own index trigger.\n");
        return 1;
      }
    }
  }

  if(region_normalization && tr->permutation) {
    tmp_bdd = permute_bdd(nm->dval_bdd, tr->permutation->perm);
    nm->dval_bdd = tmp_bdd;
  }
  if(on_the_fly_normalization) 
    nm->dval_bdd = permute_bdd(nm->dval_bdd, tmp_perm2);

  return 0;
}

/* Used in transition_m1_first() and transition_m2_first() */

int sync_create_permute_dval(struct mode_type *m, struct transition_type *tr1,
                             struct transition_type *tr2, struct mode_type *nm,
                             int *tmp_perm1, int *tmp_perm2) {
  bdd tmp_bdd;

  if(EIV && composed_ta > 2) {
    if(BDD_AND(m->submode1->dval_bdd, tr1->trn->trigger_bdd) != BDD_FALSE) {
      if(BDD_AND(m->dval_bdd, tr1->trn->trigger_bdd) == BDD_FALSE) {
        if(verbose >= 0)
          output("Valid run became invalid after composition.\n");
        return 1;
      }
    }
  }

  if(EOI && composed_ta > 2) {
    if(BDD_AND(m->dval_bdd, tr2->trn->trigger_bdd) == BDD_FALSE) {
      if(own_index_trigger(tr2)) {
        if(verbose >= 0)
          output("False own index trigger.\n");
        return 1;
      }
    }
  }

  if( BDD_AND((nm->dval_bdd = bdd_assign( bdd_assign(m->dval_bdd, 
                                          tr1->trn->assign_cycle), 
                                          tr2->trn->assign_cycle)),
              nm->boolexp->boolexp_bdd
             ) == BDD_FALSE) {
    if(verbose >= 0) 
	  output("Invalid mode (BDD: discrete variable valuation = false).\n");
      return 1;
  }

  if(region_normalization) {
    if (tr1->permutation) {
      tmp_bdd = permute_bdd(nm->dval_bdd, tr1->permutation->perm);
      nm->dval_bdd = tmp_bdd;
    }
    if (tr2->permutation) {
      tmp_bdd = permute_bdd(nm->dval_bdd, tr2->permutation->perm);
      nm->dval_bdd = tmp_bdd;
    }
  }
  if(on_the_fly_normalization) 
    nm->dval_bdd = permute_bdd(permute_bdd(nm->dval_bdd, tmp_perm1), tmp_perm2);

  return 0;
}

/* Used in transition_m1_first() and transition_m2_first() */

struct mode_type *handle_otf_norm(struct mode_type *nm) {
  struct mode_type *nnm;
  nm->norm_perm = alloc_permutation();
  nm->norm_perm->perm = get_mode_perm(nm);
  nm->norm_perm = (struct permutation_type *)
                  add_23tree(nm->norm_perm, perm_tree, 
                  permutation_comp, free_perm, perm_count_inc, 
                  perm_parent_set, print_permutation);
  if(verbose >= 0) {
    output("\tMode ");
    print_permutation(nm->norm_perm);
  }
  if(is_normal_perm(nm->norm_perm->perm) == FALSE) {
    nnm = normalize_mode(nm);
    free_mode(nm);
    nm = nnm;
    if(verbose >= 0)
      print_new_mode(nm);
    nnm = NULL;
  }
  return nm;
}

/* Used in transition_m1_first() and transition_m2_first() */

void push_mode(struct mode_type *nm, struct mode_type *m) {
  last_mode->cnext = nm;
  last_mode = nm;
  (NEW_THREAD->mode_count)++;
                      /* Push the new mode into the stack */
  nm->next = TOP;
  TOP = nm;
  if(verbose > 0) {
    sprintf(tmpstring, 
     "Pushed mode (%s) [%d(%d, %d)] {parent=(%s)[%d(%d, %d)]} into stack\n",
        nm->name, nm->index, nm->submode1->index, nm->submode2->index,
         m->name, m->index, m->submode1->index, m->submode2->index);
      output(tmpstring);
  }
}

/* Used in transition_m1_first() and transition_m2_first() */

void print_progress(struct mode_type *nm) {
  if(stdoutput == FALSE && SGM_MODE != GUI) {
    if(nm->index < 100) {
      if(fmod(nm->index, 10)==0) 
        fprintf(stderr, ".");
    }
    else if(fmod(nm->index, 100)==0) 
      fprintf(stderr, ".");
    fflush(stdout);
  }
}

/* Used in transition_m1_first() and transition_m2_first() */

int create_permute_zone(struct mode_type *m, struct transition_type *tr,
                        struct mode_type *nm) {
  int *new_mode_perm = NULL;
  struct mode_type *nnm = NULL;
  struct transition_type *ntr;

  if(total_clocks > 0)
    nm->zone = post_xconj(m, nm->submode1->xconj_cycle, 
                          nm->submode2->xconj_cycle, 
			  tr->trn->clk_trigger_cycle, tr->trn->assign_cycle,
                          tr->permutation, &(nm->xconj_cycle));
  else {
    nm->zone = alloc_zone(NEW_THREAD);
    nm->zone->relop[0][0] = TRUE;
  }
  /* NOTE: the zone is already permuted according to tr->permutation->perm */

  if(clock_shielding) mode_shield_clocks(nm);

  free_clkconj(&(nm->xconj_cycle));
  nm->xconj_cycle = zone2clkconj(nm->zone);

  if(nm->zone->relop[0][0] == FALSE) {
    if(verbose >= 0) output("\tFalse ZONE.\n");
    return 1;
  }
  else {                   /* ZONE is not false */
    print_info_zone(nm);
    if(on_the_fly_normalization) {
      nm = handle_otf_norm(nm);
      new_mode_perm = copy_permutation(new_mode_perm, nm->norm_perm->perm);
    }

    nnm = (struct mode_type *)
 	     add_23tree(nm, NEW_THREAD->modetree, mode_comp, 
		     free_mode, mode_count_inc, mode_parent_set, print_mode);

    ntr = add_transition_fifo(tr, m, nnm);
    if(on_the_fly_normalization) {
      update_permutation(m, tr, new_mode_perm, ntr);
    }
    (NEW_THREAD->trans_count)++;

    if (nnm == nm) {		/* A NEW MODE, PUSH IT! */
      push_mode(nm, m);
/*    print_progress(nm); */
    }
    else {                          /* nnm != nm */
      handle_comp_stat();
    } /* end else nnm!=nm */
  } /* end else ZONE is not false */
  return 0;
}

/* Used in transition_m1_first() and transition_m2_first() */

int sync_create_permute_zone(struct mode_type *m, struct transition_type *tr1,
                             struct transition_type *tr2, struct mode_type *nm){
  int *new_mode_perm = NULL;
  struct mode_type *nnm = NULL;
  struct transition_type *ntr;
  struct clkconj_type *c;
  struct assign_type *a;
  struct permutation_type *p;

  if(total_clocks > 0) {
    c = clkconj_and(tr1->trn->clk_trigger_cycle, tr2->trn->clk_trigger_cycle);
    if(c->op == FALSE) {
      if(verbose >= 0)
        output("  New sync trans has FALSE clock trigger.\n");
      free_clkconj(&c);
      return 1;
    }
    a = assign_union(tr1->trn->assign_cycle, tr2->trn->assign_cycle);
    p = alloc_permutation();

    if(tr1->permutation) {
       if(tr2->permutation)
          p->perm = compose_permutation(tr1->permutation->perm,
                                                      tr2->permutation->perm);
       else p->perm = copy_permutation(p->perm, tr1->permutation->perm);
    }
    else if(tr2->permutation)
           p->perm = copy_permutation(p->perm, tr2->permutation->perm);

    nm->zone = post_xconj(m, nm->submode1->xconj_cycle, 
                          nm->submode2->xconj_cycle, c, a, p, 
                          &(nm->xconj_cycle));
    free_clkconj(&c);
    free_assign_cycle(&a);
    free_perm(p);
/*
    c1 = zone2clkconj(z1);
    free_zone(z1);
    z2 = post_xconj(m, nm->submode1->xconj_cycle, nm->submode2->xconj_cycle, 
                    tr2->trn->clk_trigger_cycle, tr2->trn->assign_cycle,
                    tr2->permutation, &(nm->xconj_cycle));
    c2 = zone2clkconj(z2);
    free_zone(z2);
    c = clkconj_and(c1, c2);
    free_clkconj(&c1);
    free_clkconj(&c2);
    nm->zone = zone(c, NEW_THREAD);
*/
    if(nm->zone->clk_index) nm->zone = process_zone_max_const(nm->zone);
  }
  else {
    nm->zone = alloc_zone(NEW_THREAD);
    nm->zone->relop[0][0] = TRUE;
  }
  /* NOTE: the zone is already permuted according to tr->permutation->perm */

  if(clock_shielding) mode_shield_clocks(nm);

  free_clkconj(&(nm->xconj_cycle));
  nm->xconj_cycle = zone2clkconj(nm->zone);

  if(nm->zone->relop[0][0] == FALSE) {
    if(verbose >= 0) output("\tFalse ZONE.\n");
    return 1;
  }
  else {                   /* ZONE is not false */
    print_info_zone(nm);
    if(on_the_fly_normalization) {
      nm = handle_otf_norm(nm);
      new_mode_perm = copy_permutation(new_mode_perm, nm->norm_perm->perm);
    }

    nnm = (struct mode_type *)
 	     add_23tree(nm, NEW_THREAD->modetree, mode_comp, 
		     free_mode, mode_count_inc, mode_parent_set, print_mode);

    ntr = add_sync_transition_fifo(tr1, tr2, m, nnm);
    if(on_the_fly_normalization) {
      sync_update_permutation(m, tr1, tr2, new_mode_perm, ntr);
    }
    (NEW_THREAD->trans_count)++;

    if (nnm == nm) {		/* A NEW MODE, PUSH IT! */
      push_mode(nm, m);
/*    print_progress(nm); */
    }
    else {                          /* nnm != nm */
      handle_comp_stat();
    } /* end else nnm!=nm */
  } /* end else ZONE is not false */
  return 0;
}

/* Handle an unsynchronized transition (no sync label)
   type = 1 ==> m1_first
          2 ==> m2_first
 */

int unsync_transition(struct mode_type *m, struct transition_type *tr, 
                      int type) {
  struct mode_type *nm;
  int *tmp_perm2 = NULL;

  /* CREATE NEW DESTINATION MODE */
  nm = alloc_mode();
  switch(type) {
    case 1: nm->submode1 = tr->mode;
            nm->submode2 = m->submode2;
            break;
    case 2: nm->submode1 = m->submode1;
            nm->submode2 = tr->mode;
  }
  nm->name = assign_mode_name(nm);

  if(clock_shielding) 
    nm->shielded_clocks = clocks_union(nm->submode1->shielded_clocks,
                                       nm->submode2->shielded_clocks);

  print_info_nm(m, tr, nm);

  if(tmp_perm2) free(tmp_perm2);
  tmp_perm2 = get_perm(m, tr);

  if(create_permute_dvl(m, tr, nm, tmp_perm2)) {
    free_mode(nm);
    return 1;
  }
  if(create_permute_boolexp(m, tr, nm, tmp_perm2)) {
    free_mode(nm);
    return 1;
  }
  if(handle_last(m, tr)) {
    free_mode(nm);
    return 1;
  }
  if(create_permute_dval(m, tr, nm, tmp_perm2, type)) {
    free_mode(nm);
    return 1;
  }
  print_info_bdd(nm);
  if(create_permute_zone(m, tr, nm)) {
    free_mode(nm);
    return 1;
  }
  free(tmp_perm2);
  return 0;
}

int free_trans_link(struct trans_link_type *tl) {
  struct trans_link_type *tlp;

  if(tl == NULL) return 1;

  /* break tl cycle */
  tlp = tl;
  tl = tl->next;
  tlp->next = NULL;

  /* pop head (tlp) of list (tl) and delete it */
  for(tlp = tl, tl=tl->next; ; tlp=tl, tl=tl->next) {
    free(tlp);
    if(tl == NULL) break;
  }
  return 0;
}

/* Synchronize a pair of transitions */

int sync_transition(struct mode_type *m, struct transition_type *tr1, 
                    struct transition_type *tr2) {
  struct mode_type *nm;
  int *tmp_perm1 = NULL, *tmp_perm2 = NULL;

  /* CREATE NEW DESTINATION MODE */
  nm = alloc_mode();
  nm->submode1 = tr1->mode;
  nm->submode2 = tr2->mode;
  nm->name = assign_mode_name(nm);

  if(clock_shielding) 
    nm->shielded_clocks = clocks_union(nm->submode1->shielded_clocks,
                                       nm->submode2->shielded_clocks);

  print_info_nm(m, tr1, nm);
  print_info_tr2(tr2);

  if(tmp_perm1) free(tmp_perm1);
  tmp_perm1 = get_perm(m, tr1);

  if(tmp_perm2) free(tmp_perm2);
  tmp_perm2 = get_perm(m, tr2);

  if(sync_create_permute_dvl(m, tr1, tr2, nm)) {
    free_mode(nm);
    return 1;
  }
  if(sync_create_permute_boolexp(m, tr1, tr2, nm, tmp_perm1, tmp_perm2)) {
    free_mode(nm);
    return 1;
  }
  if(sync_handle_last(m, tr1, tr2)) {
    free_mode(nm);
    return 1;
  }
  if(sync_create_permute_dval(m, tr1, tr2, nm, tmp_perm1, tmp_perm2)) {
    free_mode(nm);
    return 1;
  }
  print_info_bdd(nm);
  if(sync_create_permute_zone(m, tr1, tr2, nm)) {
    free_mode(nm);
    return 1;
  }
  free(tmp_perm1);
  free(tmp_perm2);
  return 0;
}

/* TRANSITION_M1_FIRST */

int transition_m1_first(struct mode_type *m) {
  int                           i;
  struct transition_type        *rtr1, *tr1;
  struct trans_link_type	*tl = NULL, *tlp = NULL;

  for (i = 0, rtr1 = m->submode1->out_trans_cycle->next;
       i < m->submode1->out_trans_count;
       i++, rtr1 = rtr1->next) {

    if(on_the_fly_normalization) {
      tr1 = copy_transition(rtr1);
      if(m->norm_perm != NULL) {
        if(is_normal_perm(m->norm_perm->perm) == FALSE)
          permute_transition(tr1, m->norm_perm->perm);
      }
    }
    else tr1 = rtr1;

    print_info_tr(tr1);

    if(tr1->trn->sync_cycle == NULL) {
      unsync_transition(m, tr1, 1);
    }
    else
    /* HANDLE SYNC */
    switch(check_sync(tr1, m->submode2, 2)) {
      case -1: return 1; 
      case  0: if(lastcomp && tr1->automaton != NULL)
                              /* not a sync'd transition */
                 fprintf(stderr, 
                 "Warning: sync on trans %d could not be sync'd with any other trans.\n", tr1->index);
               else
                 unsync_transition(m, tr1, 1);   /* type = 1 ==> m1_first */
               break;		/* no same sync in THREAD2, single trans */
      case  1: 			/* sync now: once for each pair (tr1, tl->tr)*/
               tl = get_sync_trans(tr1, m->submode2);
               for(tlp = tl->next; ; tlp = tlp->next) {
                 sync_transition(m, tr1, tlp->tr);
                 if(tlp == tl) break;
               }
               free_trans_link(tl);
               tl = (struct trans_link_type *)NULL;
               break;
      case  2: continue;	/* sync later in THREAD2, jump to next trans */
    }
  } /* endfor */
  return 0;
}
  /* transition_m1_first() */



int transition_m2_first(struct mode_type *m) {
  int                           j;
  struct transition_type        *rtr2, *tr2;

  for (j = 0, rtr2 = m->submode2->out_trans_cycle->next;
       j < m->submode2->out_trans_count;
       j++, rtr2 = rtr2->next) {
    if(on_the_fly_normalization) {
      tr2 = copy_transition(rtr2);
      if(m->norm_perm != NULL) {
        if(is_normal_perm(m->norm_perm->perm) == FALSE) {
          permute_transition(tr2, m->norm_perm->perm);
        }
      }
    }
    else tr2 = rtr2;
 
    print_info_tr(tr2);

    /* HANDLE SYNC */

    if(tr2->trn->sync_cycle == NULL) {
      unsync_transition(m, tr2, 2); 	/* type = 2 ==> m2_first */
    }
    else
    /* HANDLE SYNC */
    switch(check_sync(tr2, m->submode1, 1)) {
      case -1: return 1; 
      case  0: if(lastcomp && tr2->automaton != NULL)
                              /* not a sync'd transition */
                 fprintf(stderr, 
                 "Warning: sync on trans %d could not be sync'd with any other trans.\n", tr2->index);
               else
                 unsync_transition(m, tr2, 2);   /* type = 2 ==> m2_first */
               break;		/* no same sync in THREAD2, single trans */
      case  1: 
      case  2: continue;	/* sync later in THREAD2, jump to next trans */
    }

/*
    if(tr2->trn->sync_cycle) {
      if(verbose >= 0) {
        sprintf(tmpstring, "Sync on transition %d. Ignoring it.\n", tr2->index);
        output(tmpstring);
      }
      continue;
    }
*/

/*
    if(partial_order_reduction) {
      if(po_reduce(tr2, m)) continue;
    }
    if(partial_order_with_symmetry_reduction) {
      if(pos_reduce(tr2, m)) {
        sprintf(tmpstring, "\tREDUCED transition %d.\n", tr2->index);
        output(tmpstring);
        continue;
      }
    }
*/
  } /* end for */
  return 0;
}
  /* transition_m2_first() */


int merge_iden_red_multi(struct thread_type *t) {
  char *c = get_graph_name(t);;
  /* MERGE IDENTICAL MODES */
  printf("\tMerging identical modes in graph %s.\n\t", c);
  if(verbose >= 0) {
    sprintf(tmpstring, "\nMerging identical modes in graph %s.\n", c);
    output(tmpstring);
  }
  merge_identical_modes(t);
  print_counts(t);

  /* REDUCE MULTI-GRAPH */
  if(sibling_trans_multi_reduction) {
    printf("\tReducing multigraph of graph %s.\n", c);
    if(verbose >= 0) {
      sprintf(tmpstring, "\nReducing multigraph of graph %s.\n", c);
      output(tmpstring);
    }
    reduce_multigraph(t);
    if(SGM_MODE != GUI && SGM_MODE != PROCESS) {
      fprintf(stderr, "\t");
      print_counts(t);
    }
  }
  free(c);
  return 0;
}


int compose(struct thread_type *t1, struct thread_type *t2, 
            struct thread_type *t3) {
  int i, nc1, nc2;
  struct mode_type *x;
  char *n1, *n2, *n3;

  if(t1->num_ta == 0 || t2->num_ta == 0) {
    fprintf(stderr, "Error merging with an empty state-graph!\n");
    return FALSE;
  }
  composed_ta = t3->num_ta = t1->num_ta + t2->num_ta;

  if(composed_ta == system_ta_num) lastcomp = TRUE;
  else lastcomp = FALSE;

  n1 = get_graph_name(t1);
  n2 = get_graph_name(t2);
  n3 = get_graph_name(t3);

  printf("Merging graphs %s & %s into %s [%d automata] ...\n", n1, n2, n3,
         composed_ta
        );

  if(verbose >= 0) {
    output("\n************************************************************\n");
    sprintf(tmpstring, "Merging automata %s & %s into %s [%d automata]... \n",
            n1, n2, n3, composed_ta);
    output(tmpstring);
  }
  NEW_THREAD = t3;
  THREAD1 = t1;
  THREAD2 = t2;
  t3->subthread1 = t1;
  t3->subthread2 = t2;
  t3->mnpl |= MG;
  update_ta_list(t1, t2, t3);

  t3->init_clkconj = clkconj_and(t1->init_clkconj, t2->init_clkconj);
  t3->init_boolexp = alloc_boolexp();
  if(t1->init_boolexp && t2->init_boolexp)
    t3->init_boolexp->boolexp_bdd = BDD_AND(t1->init_boolexp->boolexp_bdd, 
             		                    t2->init_boolexp->boolexp_bdd);
  else if(t1->init_boolexp && !t2->init_boolexp)
    t3->init_boolexp->boolexp_bdd = t1->init_boolexp->boolexp_bdd; 
  else if(!t1->init_boolexp && t2->init_boolexp)
    t3->init_boolexp->boolexp_bdd = t2->init_boolexp->boolexp_bdd; 
  else
    t3->init_boolexp->boolexp_bdd = BDD_TRUE;
  t3->starting_mode = initial_mode(t1, t2);
  last_mode = t3->starting_mode;
  t3->mode_count = 1;
  t3->trans_count = 0;
  t3->invalid_trans_count = 0;
  t3->invalid_cycle = 0;

  if(t3->starting_mode) {
    TOP = t3->starting_mode;
    TOP->next = NULL;

    for( ; TOP; ) {
      if(verbose > 0) {
        x = TOP;
        output("\nCurrent stack status: ");
        for(i=0; x; x=x->next, i++) { 
          sprintf(tmpstring, "%d(%d, %d)", x->index, x->submode1->index, 
                                                     x->submode2->index); 
          output(tmpstring);
          if(x->next) output(", ");
          if(i%4==0 && x->next) output("\n");
        }
        output("\n");
      }
      x = TOP;
      TOP = TOP->next;
      x->next = (struct mode_type *)NULL;
      if(verbose >= 2) {
        sprintf(tmpstring, "Processing mode %s [%d(%d, %d)]\n", 
                     x->name, x->index, x->submode1->index, x->submode2->index);
        output(tmpstring);
      }
      if(verbose == 0) {
        sprintf(tmpstring, "--- Predecessor mode %s(%d) [%s(%d), %s(%d)]\n",
                     x->name, x->index, x->submode1->name, x->submode1->index, 
                                        x->submode2->name, x->submode2->index);
        output(tmpstring);
      }
      nc1 = x->submode1->out_trans_count;
      nc2 = x->submode2->out_trans_count;

/*
      if(partial_order_reduction || partial_order_with_symmetry_reduction) 
        reset_partial_order();
*/

      if(nc1) {
        if(verbose >= 2) {
          sprintf(tmpstring,"Submode1: No. of outgoing transitions = %d\n",nc1);
          output(tmpstring);
        }
        transition_m1_first(x);
        
        if(stats == TRUE && stdoutput == FALSE && 
           SGM_MODE != GUI && SGM_MODE != PROCESS) {
          clear_line();
          fprintf(stderr, "\t%d modes, %d transitions ", 
                          t3->mode_count, t3->trans_count);
        }

        if(verbose >= 2) 
          output("---- Finished submode1 transitions ...\n");
      }
      if(nc2) {
        if(verbose >= 2) {
          sprintf(tmpstring,
                  "\nSubmode2: No. of outgoing transitions = %d\n",nc2);
          output(tmpstring);
        }
        transition_m2_first(x);
        
        if(stats == TRUE && stdoutput == FALSE && 
           SGM_MODE != GUI && SGM_MODE != PROCESS) {
          clear_line();
          fprintf(stderr, "\t%d modes, %d transitions ", 
                        t3->mode_count, t3->trans_count);
        }

        if(verbose >= 2)
          output("---- Finished submode2 transitions ...\n");
      }
/* Removed to save memory 
      if(partial_order_with_symmetry_reduction) {
        if(x->accessed) free(x->accessed);
        if(x->history) free(x->history);
      }
*/
    }
  }
  else {
    fprintf(stderr, "A conflict in initial mode merging.\n");
    free(n1);
    free(n2);
    free(n3);
    return (-1);
  }
  last_mode->cnext = (struct mode_type *)NULL;

  if(verbose >= 0)
    output("\n");

  clean_line();
  print_counts(t3);
/*
  print_thread(t3);
  if(on_the_fly_normalization)
    merge_iden_red_multi(t3);
*/
  free(n1);
  free(n2);
  free(n3);
  return TRUE;
}

int clear_line() {
  fprintf(stderr, "\r");
  return 0;
}

int clean_line() {
  int i;
  fprintf(stderr, "\r");
  for(i=0;i<80;i++)
    fprintf(stderr, " "); 
  fprintf(stderr, "\r");
  return 0;
}

struct zone_type *
post_xconj(struct mode_type *m,  	/* mode 	   */
           struct clkconj_type *x, 	/* xconj_cycle of new submode1 */
           struct clkconj_type *y, 	/* xconj_cycle of new submode2 */
           struct clkconj_type *t,	/* clk_trigger_cycle   */
           struct assign_type *a, 	/* assign_cycle    */
           struct permutation_type *p,  /* permutation on transition */
           struct clkconj_type **nx	/* new xconj_cycle */
          ) 
{
  struct clkconj_type *c, *tmp_clkconj, *d, *px, *py;
  struct zone_type *z;
  int *tmp_perm1, *tmp_perm2;

  if(verbose >= 2) {
    sprintf(tmpstring, "Parent mode %d(%d, %d) has clock conjuncts: \n\t", 
                         m->index, m->submode1->index, m->submode2->index);
    output(tmpstring);
    print_clkconj(m->xconj_cycle);
    output("\n");
  }

  if(m->xconj_cycle) {
    c = clkconj_copy(m->xconj_cycle);

/* Step (0) ADVANCE(xconj_cycle) */
    
    c = advance(c, TRUE);

/* Step (1) CONJOIN(new xconj_cycle, DELTA) */

    *nx = conjoin(c, DELTA);	
    if(verbose >= 3) {
      output("(1) After advance + conjoin, *nx = ");
      print_clkconj(*nx);
      output("\n");
    }
  }

/* Step (2) AND(new xconj_cycle, clk_trigger) */

  if(t) {
    d = clkconj_and(*nx, t);
    free_clkconj(nx);
    *nx = d;

    if(verbose >= 3) {
      output("(2.1) After clkconj_and with clk_trigger (");
      print_clkconj(t);
      output("), *nx = ");  
      print_clkconj(*nx); 
      output("\n");
    }
  }
  else if(verbose >= 3) 
    output("(2.1) Clock trigger is NULL\n");

  if((*nx)->op == FALSE) {
    z = alloc_zone(NEW_THREAD);
    z->relop[0][0] = FALSE;
    return z;
  }

  d = clkconj_and(*nx, m->xconj_cycle);
  free_clkconj(nx);
  *nx = d;

  if(verbose >= 3) {
    output("(2.2) After clkconj_and with old cklconj (");
    print_clkconj(m->xconj_cycle);
    output("), *nx = ");  
    print_clkconj(*nx); 
    output("\n");
  }

  if((*nx)->op == FALSE) {
    z = alloc_zone(NEW_THREAD);
    z->relop[0][0] = FALSE;
    return z;
  }

/* Step (3) ASSIGN(new xconj_cycle, assign) */

  if(a) {
    *nx = assign(*nx, a);          
    if(verbose >= 3) {
      output("(3) After assign with (");
      print_assignment(a);
      output("), *nx = ");  
      print_clkconj(*nx); 
      output("\n");
    }
  }

  if((*nx)->op == FALSE) {
    z = alloc_zone(NEW_THREAD);
    z->relop[0][0] = FALSE;
    return z;
  }

/* Step (4) ADVANCE and CONJOIN the new xconj_cycle */

  *nx = advance(*nx, FALSE);
  *nx = conjoin(*nx, DELTA);	
  if(verbose >= 3) {
    sprintf(tmpstring, "(4) After advance + conjoin, *nx = ");
    output(tmpstring);
    print_clkconj(*nx);
    output("\n");
  }

/* Step (5) AND(new xconj_cycle, submode1 xconj_cycle) *
 *          AND(new xconj_cycle, submode2 xconj_cycle) */
          
  if(region_normalization && p) {
    tmp_clkconj = permute_clkconj((*nx), p->perm);
    free_clkconj(nx);
    (*nx) = tmp_clkconj;
    if(verbose >= 3) {
      output("... permuted xconj_cycle using transition ");
      print_permutation(p);
      output("\n *nx = ");
      print_clkconj(*nx);
      output("\n");
    }
  }
  if(on_the_fly_normalization) {
    if(m->norm_perm)
      tmp_perm1 = inverse_permutation(m->norm_perm->perm);
    else
      tmp_perm1 = NULL;
    if(p) {
      tmp_perm2 = compose_permutation(tmp_perm1, p->perm);
      free(tmp_perm1);
      tmp_perm1 = NULL;
    }
    else {
      tmp_perm2 = tmp_perm1;
      tmp_perm1 = NULL;
    }
    tmp_clkconj = permute_clkconj((*nx), tmp_perm2);
    free_clkconj(nx);
    (*nx) = tmp_clkconj;
  }

/*
  if(on_the_fly_normalization && m->norm_perm) {
    px = permute_clkconj(x, m->norm_perm->perm);
    py = permute_clkconj(y, m->norm_perm->perm);
  }
  else {
*/
    px = x; 
    py = y;
/*
  }
*/

  if(verbose >= 3) {
    output("... taking AND with submode1 xconj_cycle: ");
    print_clkconj(px);
    output(";\n *nx = ");
    print_clkconj(*nx);
    output("\n");
  }
  
  d = clkconj_and(*nx, px);
  free_clkconj(nx);
  *nx = d;
  
  if((*nx)->op == FALSE) { 
    z = alloc_zone(NEW_THREAD);
    z->relop[0][0] = FALSE;
    return z;
  }

  if(verbose >= 3) {
    output("... taking AND with submode2 xconj_cycle: ");
    print_clkconj(py);
    output(";\n *nx = ");
    print_clkconj(*nx);
    output("\n");
  }

  d = clkconj_and(*nx, py);
  free_clkconj(nx);
  *nx = d;

  if((*nx)->op == FALSE) {
    z = alloc_zone(NEW_THREAD);
    z->relop[0][0] = FALSE;
    return z;
  }

  if(verbose >= 3) {
    output("(5) After clkconj_and with submodes' xconj_cycles: (");
    print_clkconj(px);
    output(", ");
    print_clkconj(py);
    output("), *nx = ");
    print_clkconj(*nx);
    output("\n");
  }

  *nx = process_max(nx);
  if(verbose >= 3) {
    output("(5') After processing maximum constant: (");
    print_clkconj(*nx);
    output(")\n");
  }

/* Step (6) ZONE(new xconj_cycle) */

  z = zone(*nx, NEW_THREAD);
  if(verbose >= 3) {
    output("(6) After zone, zone(*nx) =[");
    print_zone(z);
    output("]\n");
  }
  if(z->clk_index)
    z = process_zone_max_const(z);
  return z;
}

struct clkconj_type *process_max(struct clkconj_type **c) {
  struct clkconj_type *p;

  if(*c) {
    for(p=(*c)->next; ; p=p->next) {
      if(p->diff > max_const) 
        p->diff = INF;
      if(p->diff < -max_const) 
        p->diff = -INF;
      if(p==(*c)) break;
    }
  }
  return *c;
}

struct clkconj_type *advance(struct clkconj_type *c, int adv_type) {
  struct clkconj_type *x, *tmp;

  if(c) {
    for(x = c->next; ; x = x->next) {
      if(x->clk2 == 0) 
        x->clk2 = DELTA;
      if(x->clk1 == 0) 
        x->clk1 = DELTA;
      if(x == c) break;
    }
    x = alloc_clkconj();
    x->clk1 = DELTA;       /* DELTA GEQ 0 */
    if(strict_semantics || adv_type == FALSE)
      x->op = GEQ;
    else
      x->op = GREATER;
    tmp = clkconj_and(c, x);
    free_clkconj(&x);
    free_clkconj(&c);
    c = tmp;
    tmp = NULL;
  }
  return c;
}


/* Note c is altered by conjoin() */

struct clkconj_type *conjoin(struct clkconj_type *c, int someclk) {
  struct clkconj_type *L1, *L2, *L3, *t, *st, *tmp, *tmp2 = NULL;
  int must_append=1;

  L1 = (struct clkconj_type *)NULL;  /* a clkconj cycle: L1 --> tail */
  L2 = (struct clkconj_type *)NULL;  /* a clkconj list:  L2 --> head */
  L3 = (struct clkconj_type *)NULL;  /* a clkconj list:  L3 --> head */

  if(c) {
/*
    output("	conjoining clkconj: ");  
    print_clkconj(c);
    output(" with respect to: ");
    sprintf(tmpstring, "(%d)", someclk);
    output(tmpstring);
    output("\n");
*/

    if(c->op == TRUE || c->op == FALSE) return c;

    for(t = c->next, st = t->next; ; t = st, st = st->next) {
      if(t->clk2 == someclk) {
                                        /* change "clk1 - clk2 op diff" into 
                                         * "clk2 - clk1 reverse_op(op) -diff"*/
        t->clk2 = t->clk1; 		/* swap clk1 and clk2 */
        t->clk1 = someclk;
        t->op = reverse_op(t->op);	/* reverse op */
        t->diff = -(t->diff); 		/* negate diff */
      }

      if(t->clk1 == someclk) {
        if(t->op == LESS || t->op == LEQ) {
          t->next = L2;
          L2 = t;
        }
        else if(t->op == GREATER || t->op == GEQ) {
          t->next = L3;
          L3 = t;
        }
        else if(t->op == EQ) {
          tmp = alloc_clkconj();
          tmp->clk1 = t->clk1;
          tmp->clk2 = t->clk2;
          tmp->diff = t->diff;

          t->op = LEQ;
          t->next = L2;
          L2 = t;

          tmp->op = GEQ;
          tmp->next = L3;
          L3 = tmp;
          tmp = (struct clkconj_type *)NULL;
        }
        else {		/* t->op == NEQ */
          fprintf(stderr, "Error: ignoring NEQ (!=) encountered in clkconj\n");
        }
      }
      else {
        if(L1) {
          t->next = L1->next;
          L1->next = t;
        }
        else {
          t->next = t;
        }
        L1 = t;
      }
      if(t == c) break;
    }  /* endfor */

    if(L2 && L3) {

/*
      output("Current status of L3 is:\n");
      for(st=L3; st; st=st->next) {
        print_one_clkconj(st);
      }
      output("Current status of L2 is:\n");
      for(st=L2; st; st=st->next) {
        print_one_clkconj(st);
      }
*/

      for(st=L3; st; st=st->next) {
        for(t=L2; t; t=t->next) {
          tmp = alloc_clkconj();
          tmp->clk1 = st->clk2;
          tmp->clk2 = t->clk2;
          tmp->diff = t->diff - st->diff;
          tmp->op = transitive_op(reverse_op(st->op), t->op);

/*
          printf("Processing clkconj: ");
          print_one_clkconj(tmp);
*/

          if(tmp->clk1 == tmp->clk2) {
            if(   (tmp->diff == 0)
               && (tmp->op == LEQ || tmp->op == EQ || tmp->op == GEQ)
              ) 
              must_append = FALSE;
            else if(   (tmp->diff >0)
                    && (tmp->op == LESS || tmp->op == LEQ)
                   )
              must_append = FALSE;
            else if(   (tmp->diff < 0)
                    && (tmp->op == GREATER || tmp->op == GEQ)
                   )
              must_append = FALSE;
          }
          if(must_append == TRUE) {
            if(tmp->clk1 == 0) {
              tmp->clk1 = tmp->clk2;
              tmp->clk2 = 0;
              tmp->op = reverse_op(tmp->op);
              tmp->diff = -(tmp->diff);
            }
            if(L1) {               /* conjunct tmp with L1 cycle */
/*
              printf("Conjunct'ing clkconj to L1:\n");
              print_one_clkconj(tmp);
              printf("\n");
*/

              if(tmp2) { free_clkconj(&tmp2); tmp2 = NULL; }
              tmp2 = clkconj_and(L1, tmp);
	      tmp->next = tmp;
	      free_clkconj(&tmp); tmp = NULL;
/*
              output("Before freeing L1: ");
              print_clkconj(L1);
*/
              free_clkconj(&L1);
              L1 = tmp2;
              tmp2 = NULL;
            }
            else {
/*
              printf("Conjunct'ing clkconj to L1:\n");
              print_one_clkconj(tmp);
              printf("\n");
*/
              tmp->next = tmp;
              L1 = tmp;
            }
/*
            printf("Current status of L1 is: ");
            print_clkconj(L1);
            printf("\n");
*/
          }
          else {
	    free_clkconj(&tmp); tmp = NULL;
            must_append = TRUE;
          }
        }  /* endfor  (L2) */
      }  /* endfor  (L3) */
    }
  } /* end if(c) */

  if(L2) {				/* free L2 */
    for(t=L2; t->next; t=t->next);
    t->next = L2;
    free_clkconj(&t);
  }

  if(L3) {				/* free L3 */
    for(st=L3; st->next; st=st->next);
    st->next = L3;
    free_clkconj(&st);
  }

  if(!L1) {
    L1 = alloc_clkconj();
    L1->op = TRUE;
  }
  return L1;
}

/* L3: {>, >=}    and    L2: {<, <=} */
/* op1 \in L3,   op2 \in L2 */

int transitive_op(int op1, int op2) {
  switch(op1) {
    case LESS: return LESS;
               break;
    case LEQ: switch(op2) {
                case LESS: return LESS;
                           break;
                case EQ:
                case LEQ: return LEQ;
                          break;
              }
              break;
    case EQ: return(op2);
             break;
    default: fprintf(stderr, "Error in transitive_op() [%s]\n", 
                                                          relop_string(op1));
             exit(1);
             break;
  }
  return 1;
}

struct clkconj_type *assign(struct clkconj_type *c, struct assign_type *a) {
  struct assign_type *x;
  struct clkconj_type *na, *tmp;

  if(a) {
    for(x = a->next; ; x=x->next) {
      if(get_var_type(x->var) == V_CLOCK) {
        c = conjoin(c, x->var);		/* eliminate x->var from c */
/*
	output("	After conjoin in assign: ");
	print_clkconj(c);
	output("\n");
*/
      }
      if(x == a) break;
    }

    for(x = a->next; ; x=x->next) {
      if(get_var_type(x->var) == V_CLOCK) {
	na = alloc_clkconj();
	na->clk1 = x->var;
	na->op = EQ;
	na->diff = x->value;
	na->next = na;
	tmp = clkconj_and(c, na);		/* c = c & x->var = x->value */
        free_clkconj(&na);
        free_clkconj(&c);
        c = tmp;
        tmp = NULL;
      }
      if(x == a) break;
    }
    return c;
  }
  return c;
}

bdd bdd_assign(bdd f, struct assign_type *a) {
  struct assign_type *x;
  struct proposition_type *p;
  struct boolconj_type *na, *b, *bb;
  int i=0;
  bdd assoc[bdd_vars(bddm)+1];
  bdd t = f;

  bdd_unfree(bddm, t);
  if(verbose >= 2) {
    output("Assigning ("); 
    print_assignment(a);
    output(") TO bdd: \n\t");
    print_bdd(f);
  }
  if(a) {
    for(x = a->next; ; x=x->next) {
      if(get_var_type(x->var) == V_REGISTER) {
        if(prop_tail) {
          for(p = prop_tail->next; ; p = p->next) {
            if(x->var == p->var1)
              assoc[i++] = p->prop_bdd;
            if(p == prop_tail) break;
          }
        }
      }
      if(x == a) break;
    }
/*
    sprintf(tmpstring, "No. of vars in temp assoc = %d\n", i);
*/
    assoc[i] = (bdd)NULL;
/*
    output(tmpstring);
*/
    bdd_temp_assoc(bddm, assoc, 0);
    t = bdd_exists(bddm, t);
/*
    output("After bdd_exists, t =\n"); print_bdd(t);
*/
    for(x = a->next; ; x=x->next) {
      if(get_var_type(x->var) == V_REGISTER) {
        na = new_assign_boolconj(x, f);
        if(na) 
          for(b=na->next; ; b=b->next) {
            install_bdd(b);
            t = bdd_and(bddm, t, b->prop_bdd);
            if(b == na) break;
          }
/*
        output("After install_bdd and bdd_and, t =\n"); 
        print_bdd(t);
*/
        bb = na;
        na = na->next;
        bb->next = NULL;
        for(b=na; ; b=bb) {
          bdd_free(bddm, b->prop_bdd);
          bb = b->next;
          b->next = b;
	  free_boolconj(b);
          if(bb == NULL) break;
        }
      }
      if(x == a) break;
    }
    return t;
  }
  return f;
}

struct boolconj_type *new_assign_boolconj(struct assign_type *x, bdd f) {
  struct int_link_type *vallist = NULL, *vl;
  struct boolconj_type *na = NULL, *ans = NULL;
  int val;

  if(verbose >= 3) {
    sprintf(tmpstring, "New boolconj for assignment: ");
    output(tmpstring);
    print_assignment1(x);
    output("\n");
    sprintf(tmpstring, "  with mode dval bdd: ");
    output(tmpstring);
    print_bdd(f);
  }
  if(x->use_var == 0) {   /* x->var := x->value {MOD mod_num} */
    na = alloc_boolconj();
    na->var1 = x->var;
    na->op = EQ;
    na->next = na;
    if(x->mod_num)
      na->constant = (int)fmod(x->value, x->mod_num);
    else if(x->mod_var)
      fprintf(stderr, "Error: assignment with MOD var not yet implemented.\n");
    else
      na->constant = x->value;
    ans = na;
  }
  else { /* x->var := x->use_var {+|-} x->value {MOD mod_num} */
    vallist = get_curr_values(x->use_var, f);
    if(vallist)
      for(vl=vallist->next; ; vl=vl->next) {
        val = vl->k;
        na = alloc_boolconj();
        na->var1 = x->var;
        na->op = EQ;
        if(x->op == ADD) {
          if(x->mod_num)
            na->constant = (int)fmod(val + x->value, x->mod_num);
          else if(x->mod_var)
            fprintf(stderr, 
                    "Error: assignment with MOD var not yet implemented.\n");
          else na->constant = val + x->value;
        }
        else if(x->op == SUB) {
          if(x->mod_num)
            na->constant = (int)fmod(val - x->value, x->mod_num);
          else if(x->mod_var)
            fprintf(stderr, 
                    "Error: assignment with MOD var not yet implemented.\n");
          else na->constant = val - x->value;
        }
        else {
          if(x->mod_num)
            na->constant = (int)fmod(val, x->mod_num);
          else if(x->mod_var)
            fprintf(stderr, 
                    "Error: assignment with MOD var not yet implemented.\n");
          else na->constant = val;
        }
        if(ans) {
          na->next = ans->next;
          ans->next = na;
        }
        else na->next = na;
        ans = na;
        if(vl == vallist) break;
      }
    free_int_link(vallist); vallist = NULL;
  }
  if(verbose >= 3) {
    output("  IS ");
    print_boolconj(ans);
    output("\n");
  }
  return ans;
}

struct int_link_type *get_curr_values(int v, bdd f) {
/*
  bdd assoc[bdd_vars(bddm)+1];
  bdd t = f;
  int i;
*/
  struct int_link_type *ans = NULL, *il = NULL;
  struct proposition_type *p;
  int vid;

  if(prop_tail) {
    for(p = prop_tail->next; ; p = p->next) {
      if(p->var1 == v && BDD_AND(f, p->prop_bdd) != BDD_FALSE) {
        il = (struct int_link_type *)malloc(sizeof(struct int_link_type));
        il->k = p->value;
        if(ans) {
          il->next = ans->next;
          ans->next = il;
        }
        else il->next = il;
        ans = il;
      }
      if(p == prop_tail) break;
    }
  }
  if(verbose >= 3) {
    vid = var_id(v);
    if(vid >= 0)
      sprintf(tmpstring, "Current Values of %s[%d]: ", var_name(v), vid);
    else
      sprintf(tmpstring, "Current Values of %s: ", var_name(v));
    output(tmpstring);
    if(ans)
      for(il=ans->next; ; il=il->next) {
        sprintf(tmpstring, "%d", il->k);
        output(tmpstring);
        if(il == ans) break;
        else output(", ");
      }
    output("\n");
  }
  return ans;
/*
  if(prop_tail) {
    for(p = prop_tail->next; ; p = p->next) {
      if(p->var1 != v)
        assoc[i++] = p->prop_bdd;
      if(p == prop_tail) break;
    }
  }
  assoc[i] = (bdd)NULL;
  bdd_temp_assoc(bddm, assoc, 0);
  t = bdd_exists(bddm, t);

  for(i=0; i<bdd_vars(bddm)+1; i++) 
    bdd_free(bddm, assoc[i]);
*/
}

char *get_name(char ***mv) {
  int i;
  char *c = NULL;
  for(i=0; i<system_ta_num; i++) {
    if((*mv)[i]) {
      if(c) {
        c = (char *)realloc(c, strlen(c) + strlen((*mv)[i]) + 2);
        c = strcat(c, ".");
      }
      else {
        c = (char *)malloc(strlen((*mv)[i]) + 1);
        c = strcpy(c, "");
      }
      c = strcat(c, (*mv)[i]);
    }
  }
  return c;
}

char *assign_mode_name(struct mode_type *m) {
  int i;
  for(i=0; i<system_ta_num; i++) {
    if(m->submode1->mode_vec[i] && !(m->submode2->mode_vec[i])) 
      m->mode_vec[i] = str_preserve(m->submode1->mode_vec[i]);
    else
    if(m->submode2->mode_vec[i] && !(m->submode1->mode_vec[i])) 
      m->mode_vec[i] = str_preserve(m->submode2->mode_vec[i]);
    else
    if(m->submode1->mode_vec[i] && m->submode2->mode_vec[i]) {
      if(!strcmp(m->submode1->mode_vec[i], m->submode2->mode_vec[i])) 
        m->mode_vec[i] = str_preserve(m->submode1->mode_vec[i]);
      else {
        fprintf(stderr, "Fatal Error in assign_mode_name ...\n");
        fprintf(stderr, "\tSubmode1: %s(%d)->mode_vec[%d] = %s\n", 
                        m->submode1->name, m->submode1->index, i,
                        m->submode1->mode_vec[i]); 
        fprintf(stderr, "\tSubmode2: %s(%d)->mode_vec[%d] = %s\n", 
                        m->submode2->name, m->submode2->index, i,
                        m->submode2->mode_vec[i]); 
      }
    }
  }
  m->name = get_name(&(m->mode_vec));
  return m->name;
}

char *assign_mode_name1(struct mode_type *m, int *perm, int perm_submode) {
  char *tmp_name = (char *)malloc( strlen(m->submode1->name) +
                                   strlen(m->submode2->name) + 2 );
  char *tc;
/*
  sprintf(tmpstring, "Assigning mode name (%d): \n", m->index);
  output(tmpstring);
  sprintf(tmpstring, "submode1 index = %d\n", m->submode1->index);
  output(tmpstring);
  sprintf(tmpstring, "submode2 index = %d\n", m->submode2->index);
  output(tmpstring);

  sprintf(tmpstring, "submode1 name = %s\n", m->submode1->name);
  output(tmpstring);
  sprintf(tmpstring, "submode2 name = %s\n", m->submode2->name);
  output(tmpstring);
*/
  if(perm_submode == 1) 
    tc = permute_mode_name(m->submode1, perm, NULL);
  else tc = m->submode1->name;

  tmp_name = strcpy(tmp_name, tc);
  tmp_name = strcat(tmp_name, ".");

  if(perm_submode == 2) 
    tc = permute_mode_name(m->submode2, perm, NULL);
  else tc = m->submode2->name;

  tmp_name = strcat(tmp_name, tc);
  free(m->name);
  m->name = tmp_name;
  return m->name;
}

/*
char *composed_mode_name(struct mode_type *m, int t1, int t2) {
  struct name_type {
    char n[25];
  };
  struct name_type *name[system_ta_num];
  struct mode_type *x = m, *y = m;
  char *mn = (char *)malloc(27*system_ta_num + 1);

  mn = strcat(mn, "(");

  while(TRUE) {
    if(t1 < system_ta_num && t2 < system_ta_num) {
      mn = strcat(mn, x->submode1->name);
      mn = strcat(mn, ", ");
      mn = strcat(mn, x->submode2->name);
    }
    if(t1 < system_ta_num && t2 >= system_ta_num) {
      mn = strcat(mn, x->submode2
      
  }

  while(x) {
    if(x->submode1->name)
      mn = strcat(mn, x->submode1->name);
    y = x;
    x = x->submode1;
  }
  while(!x->submode1->name) {
    mn = strcat(mn, x->submode1->name);
    mn = strcat(mn, " ");
    while(!y->submode2->name) {
      y = y->submode1;
    }
    x = x->submode1;
  }

  x = m;

  while(!x->submode2->name) 
    x = x->submode2;
  do {
    mn = strcat(mn, x->submode2->name);
    mn = strcat(mn, " ");
    x = x->submode2; }
  while(!x->submode2->name);
  mn = strcat(mn, ")");
  return mn;
}
*/

/************ Clock literal set updating functions ***********/

int init_clk_literals() {
  int i;
  struct mode_type *m;
  struct transition_type *t;
  struct clkconj_type *c;

  for(i = 0; i<BASIC_THREAD_COUNT; i++) {
    for(m = thread[i].starting_mode; m; m = m->cnext) {
      c = clkconj_copy(m->xconj_cycle);
      append_clkcycle(c, &clk_literal_cycle);
      if(m->out_trans_cycle) {
        for(t = m->out_trans_cycle->next; ; t=t->next) {
          c = clkconj_copy(t->trn->clk_trigger_cycle);
          append_clkcycle(c, &clk_literal_cycle);
          if(t == m->out_trans_cycle) break;
        }
      }
    }
  }
  if(verbose >= 0) {
    output("Literal List:\n");
    print_literals(clk_literal_cycle);
  }
  return 0;
}

int append_clkcycle(struct clkconj_type *c, struct clkconj_type **lc) {
  struct clkconj_type *cp, *cps;
  if(c) {
    if(c->op == TRUE || c->op == FALSE) return 0;
    cp = c->next;
    c->next = NULL;
    cps = cp->next;
    while(cp) {
      if(clkconj_not_in_cycle(cp, lc)) {
        append_clkconj(cp, lc);
      }
      if(cps == NULL) break;
      else {
        cp = cps;
        cps = cps->next;
      }
    }
  }
  return 0;
}

int clkconj_not_in_cycle(struct clkconj_type *cp, struct clkconj_type **lc) {
  struct clkconj_type *x;
  if(*lc) {
    for(x = (*lc)->next; ; x=x->next) {
      if( (x->clk1 == cp->clk1) && (x->clk2 == cp->clk2) &&
          (x->op == cp->op) && (x->diff == cp->diff) ) return FALSE;
      if(x == (*lc)) break;
    }
  }
  return TRUE;
}

int append_clkconj(struct clkconj_type *cp, struct clkconj_type **lc) {
  if(*lc) {
    cp->next = (*lc)->next;
    (*lc)->next = cp;
  }
  else cp->next = cp;
  (*lc) = cp;
  return 0;
}

struct clkconj_type *initial_literal_set() {
  struct clkconj_type *c, *set = NULL, *cc;
  for(c=clk_literal_cycle->next; ; c=c->next) {
    if( (c->clk1 <= composed_ta) && (c->clk2 <= composed_ta) ) {
      cc = clkconj_copy1(c);
      append_clkconj(cc, &set);
    }
    if(c == clk_literal_cycle) break;
  }
  return set;
}

int reset_literals(int var, struct clkconj_type **lc) {
  delete_literals(var, lc);
  insert_literals(var, lc);
  return 0;
}

int delete_literals(int var, struct clkconj_type **lc) {
  struct clkconj_type *pc, *c;

  if( (*lc)->next == (*lc) && 
      ((*lc)->clk1 == var || (*lc)->clk2 == var) 
    ) {
    free_clkconj(lc);
    (*lc) = NULL;
    return 0;
  }
  for(pc = (*lc), c = pc->next; ;) {
    if(c->clk1 == var || c->clk2 == var) {
      pc->next = c->next;
      if(c == (*lc)) {
        c->next = c;
        free_clkconj(&c);
        (*lc) = pc;
        break;
      }
      else {
        c->next = c;
        free_clkconj(&c);
        c = pc->next;
        continue;
      }
    }
    if(c == (*lc)) break;
    else {
      pc = c; 
      c = c->next;
    }
  } 
  return 0;
}

int insert_literals(int var, struct clkconj_type **lc) {
  struct clkconj_type *pc = NULL, *c, *cc, *set = NULL;

  for(c=clk_literal_cycle->next; ; c=c->next) {
    if(c->clk1 == var || c->clk2 == var) {
      cc = clkconj_copy1(c);
      append_clkconj(cc, &set);
    }
    if(c == clk_literal_cycle) break;
  }
  if(*lc) {
    for(pc = (*lc), c=pc->next; ; pc=c, c=c->next) {
      if(c->clk1 > var) {
        pc->next = set->next;
        set->next = c;
        break;
      }
      if(c == (*lc)) {
        append_clkcycle(set, lc);
        break;
      }
    }
  }
  else {
    (*lc) = set;
  }
  return 0;
}

int update_literals(struct clkconj_type *trigger, struct clkconj_type **lc) {
  struct clkconj_type *c, *pcc, *cc;

  if(trigger) {
    for(c = trigger->next; ; c=c->next) {
      if(*lc) {
        for(pcc = (*lc), cc = pcc->next; ;) {
          if(c->clk1 == cc->clk1) {
            if(incompatible_literals(c, cc)) {
              pcc->next = cc->next;
              if(cc == (*lc)) {
                cc->next = cc;
                free_clkconj(&cc);
                *lc = pcc;
                break;
              }
              else {
                cc->next = cc;
                free_clkconj(&cc);
                cc = pcc->next;
                continue;
              }
            }
          }
          if(cc == (*lc)) break;
          else {
            pcc = cc;
            cc = cc->next;
          }
        }
      }
      if(c == trigger) break;
    }
  }
  return 0;
}

/* c1 = literal from transition trigger */
/* c2 = literal from source mode clock literals set */
int incompatible_literals(struct clkconj_type *c1, struct clkconj_type *c2) {
  switch(c1->op) {
    case GEQ:
      switch(c2->op) {
        case LESS:
          if(c1->diff >= c2->diff) return TRUE;
          break;
      }
      break;
    case GREATER:
      switch(c2->op) {
        case LEQ:
        case LESS:
          if(c1->diff >= c2->diff) return TRUE;
      }
      break;
  }
  return FALSE;
}

int clk_is_reset(struct transition_type *tr) {
  struct assign_type *a;
  if(tr->trn->assign_cycle) {
    for(a = tr->trn->assign_cycle->next; ; a = a->next) {
      if(get_var_type(a->var) == V_CLOCK) return TRUE; 
      if(a == tr->trn->assign_cycle) break;
    }
  }
  return FALSE;
}

int get_clk_reset(struct transition_type *tr) {
  struct assign_type *a;
  if(tr->trn->assign_cycle) {
    for(a = tr->trn->assign_cycle->next; ; a = a->next) {
      if(get_var_type(a->var) == V_CLOCK) return a->var; 
      if(a == tr->trn->assign_cycle) break;
    }
  }
  return -1;
}

int cvl_comp(struct clkconj_type *cvl1, struct clkconj_type *cvl2) {
  struct clkconj_type *c1, *c2;

  if(!cvl1 && cvl2) return -1;
  if(cvl1 && !cvl2) return 1;
  if(cvl1 && cvl2) {
    for(c1 = cvl1->next, c2 = cvl2->next; ; c1=c1->next, c2=c2->next) {
      if(c1->clk1 != c2->clk1) return (c1->clk1 - c2->clk1);
      if(c1->op != c2->op) return (c1->op - c2->op);
      if(c1->diff != c2->diff) return (c1->diff - c2->diff);
      if(c1 == cvl1 && c2 != cvl2) return -1;
      if(c1 != cvl1 && c2 == cvl2) return 1;
      if(c1 == cvl1 && c2 == cvl2) break;
    }
  }
  return EQUAL;
}

struct mode_type *normalize_mode(struct mode_type *m) {
  struct mode_type *nm;

  if(verbose >= 0) {
    sprintf(tmpstring, "\tNormalizing mode %s(%d).\n", m->name, m->index);
    output(tmpstring);
  }
  nm = alloc_mode();
  nm->name = permute_mode_name(m, m->norm_perm->perm, &(nm->mode_vec));
  nm->submode1 = m->submode1;
  nm->submode2 = m->submode2;
/*
  nm->norm_perm = alloc_permutation();
  nm->norm_perm->perm = 
        copy_permutation(nm->norm_perm->perm, m->norm_perm->perm);
*/
  nm->norm_perm = m->norm_perm;
  nm->dvl = permute_boolconj(m->dvl, m->norm_perm->perm);
/*
  nm->cvl = permute_clkconj(m->cvl, m->norm_perm->perm);
*/
  nm->zone = permute_zone(m->zone, m->norm_perm->perm, NEW_THREAD);
  nm->dval_bdd = permute_bdd(m->dval_bdd, m->norm_perm->perm);
  nm->boolexp = alloc_boolexp();
  nm->boolexp->boolexp_bdd = permute_bdd(m->boolexp->boolexp_bdd, 
                                         m->norm_perm->perm);
  nm->xconj_cycle = zone2clkconj(nm->zone);

  return nm;
}

int *get_mode_perm(struct mode_type *m) {
  int *perm = (int *)malloc(sizeof(int) * (total_clocks+1));
  int i;
  for(i=0; i<=total_clocks; i++) {
    perm[i] = i;
  }
  CURR_MODE = m;
  qsort(&perm[1], composed_ta, sizeof(int), normal_index_compare); 
  return perm;
}


/* Clock shielding functions during compose() */

int mode_shield_clocks(struct mode_type *m) {
  struct var_link_type *vl;
  if(m->shielded_clocks) {
    for(vl = m->shielded_clocks->next; ; vl=vl->next) {
      m->xconj_cycle = conjoin(m->xconj_cycle, vl->var->index);
      free_zone(m->zone);
      m->zone = zone(m->xconj_cycle, NEW_THREAD);
      m->zone = process_zone_max_const(m->zone);
      if(vl == m->shielded_clocks) break;
    }
  }
  return 0;
}

struct var_link_type *var_link_copy(struct var_link_type *sc) {
  struct var_link_type *scp, *vl, *ans = NULL;
  if(sc) {
    for(scp = sc->next; ; scp = scp->next) {
      vl = (struct var_link_type *)malloc(sizeof(struct var_link_type));
      vl->var = scp->var;
      if(ans) {
        vl->next = ans->next;
        ans->next = vl;
      }
      else
        vl->next = vl;
      ans = vl;
      if(scp == sc) break;
    }
  }
  return ans;
}

struct var_link_type *clocks_union(struct var_link_type *sc1, 
                                   struct var_link_type *sc2) {
  struct var_link_type *scc1, *scc2, *vl2, *c;

  if(sc1 == NULL && sc2 == NULL) return NULL;
  if(sc1 == NULL) {
    scc2 = var_link_copy(sc2);
    return scc2;
  }
  if(sc2 == NULL) {
    scc1 = var_link_copy(sc1);
    return scc1;
  }
  scc1 = var_link_copy(sc1);

  for(vl2 = sc2->next; ; vl2=vl2->next) {
    c = (struct var_link_type *)malloc(sizeof(struct var_link_type));
    c->var = vl2->var;
    scc1 = insert_var_link(c, scc1);
    if(vl2 == sc2) break;
  }

  return scc1;  
}

struct var_link_type *insert_var_link(struct var_link_type *c, 
                                      struct var_link_type *sc) {
  struct var_link_type *vlp, *vl;
  if(sc) {
    for(vlp = sc, vl = vlp->next; ; vlp=vl, vl=vl->next) {
      if(vl->var->index == c->var->index) {
        free(c);
        c = NULL;
        break;
      }
      if(vl->var->index > c->var->index) {
        vlp->next = c;
        c->next = vl;
        break;
      }
      if(vl == sc) {
        c->next = vl->next;
        vl->next = c;
        sc = c;
        break;
      }
    }
  }
  else {
    sc = c->next = c;
  }
  return sc;
}

int print_var_links(struct var_link_type *sc) {
  struct var_link_type *vl;
  output("{");
  if(sc) {
    for(vl = sc->next; ; vl = vl->next) {
      if(vl->var->id >= 0) 
        sprintf(tmpstring, "%s[%d]", vl->var->name, vl->var->id);
      else
        sprintf(tmpstring, "%s", vl->var->name);
      output(tmpstring);
      if(vl == sc) break;
      else output(", ");
    }    
  }
  output("}\n");
  return 0;
}

int free_var_links(struct var_link_type *sc) {
  struct var_link_type *vl;
  if(sc) {
    vl = sc->next;
    sc->next = NULL;
    sc = vl;
    while(sc) {
      vl = sc;
      sc = sc->next;
      free(vl);
      vl = NULL;
    }
  }
  return 0;
}

int shield_clocks_compare(struct var_link_type *sc1, 
                          struct var_link_type *sc2) {
  struct var_link_type *vl1, *vl2;
  if(!sc1 && !sc2) return EQUAL;
  if(!sc1 && sc2) return -1;
  if(sc1 && !sc2) return 1;
  for(vl1=sc1->next, vl2=sc2->next; ; vl1=vl1->next, vl2=vl2->next) {
    if((vl1->var->index) < (vl2->var->index)) 
      return -1;
    if((vl1->var->index) > (vl2->var->index)) 
      return 1;
    if(vl1 == sc1 && vl2 != sc2) return -1;
    if(vl1 != sc1 && vl2 == sc2) return -1;
    if(vl1 == sc1 && vl2 == sc2) return EQUAL;
  }
}

/* update the permutation on transition for on-the-fly region normalization
 * (using indice permutation) during merging
 */
int update_permutation(struct mode_type *m, struct transition_type *tr,
                       int *new_mode_perm, struct transition_type *ntr) {
  int *tmp1, *tmp2;
  struct permutation_type *p;

  if(verbose >= 3) {
    sprintf(tmpstring, "    Installing permutation for transition %d.\n", 
            ntr->index);
    output(tmpstring);
    sprintf(tmpstring, "\tSource mode %s(%d) has ", m->name, m->index);
    output(tmpstring);
    print_permutation(m->norm_perm);
  }
  
  if(m->norm_perm)
    tmp1 = inverse_permutation(m->norm_perm->perm);
  else tmp1 = (int *)NULL;

  if(verbose >= 3) {
    output("\tSource mode inverse ");
    print_perm(tmp1);
    sprintf(tmpstring, "\tTransition %d has ", tr->index);
    output(tmpstring);
    print_permutation(tr->permutation);
  }
  
  if(tr->permutation) {
    tmp2 = compose_permutation(tmp1, tr->permutation->perm);
    free(tmp1);
    tmp1 = NULL;
  }
  else {
    tmp2 = tmp1;
    tmp1 = (int *)NULL;
  }

  if(verbose >= 3) {
    output("\tComposed inverse source_mode and trans permutations is ");
    print_perm(tmp2);
  }

/*
  if(verbose >= 3) {
    sprintf(tmpstring, "\tDestination mode %s(%d) has ", nm->name, nm->index);
    output(tmpstring);
    print_permutation(nm->norm_perm);
  }
*/
  if(new_mode_perm) {
    tmp1 = compose_permutation(tmp2, new_mode_perm);
    free(tmp2);
    tmp2 = NULL;
  }
  else {
    tmp1 = tmp2;
    tmp2 = (int *)NULL;
  }

  
  if(verbose >= 3) {
    output("Final ");
    print_perm(tmp1);
  }

  p = alloc_permutation();
  p->perm = tmp1;
  ntr->permutation = (struct permutation_type *)
                     add_23tree(p, perm_tree, permutation_comp,
                     free_perm, perm_count_inc, perm_parent_set,
                     print_permutation);

  if(verbose >= 0) {
    sprintf(tmpstring,
            "\tInstalled permutation on transition (%d), ", ntr->index);
    output(tmpstring);
    print_trans_permutation(ntr);
  }
  return 0;
}

int print_new_mode(struct mode_type *m) {
  sprintf(tmpstring,
          "Normalized mode %s(%d) [%s(%d), %s(%d)], ",
          m->name, m->index, m->submode1->name, m->submode1->index, 
                             m->submode2->name, m->submode2->index);
  output(tmpstring);
  print_permutation(m->norm_perm);
  if(clock_shielding) {
    output("    Shielded Clocks: ");
    print_var_links(m->shielded_clocks);
  }
  if(!lazy_bdd_merging) {
    output("    Literals Set: ");
    print_dvl(m->dvl);
  }
/*
  output("    CVL SET: ");
  print_literals(m->cvl);
  output("    Inv BDD: ");
  print_bdd(m->boolexp->boolexp_bdd);
*/
  else {
    output("    Dval BDD: ");
    print_bdd(m->dval_bdd);
  }
  if(total_clocks > 0) {
    output("    Zone: ");
    print_zone(m->zone);
    output("\n");
  }
  return 0;
}

int perm_comp(int *p1, int *p2) {
  int i;
  if(p1 == NULL && p2 == NULL) return EQUAL;
  if(p1 == NULL && p2 != NULL) return -1;
  if(p1 != NULL && p2 == NULL) return 1;
  for(i=1; i<=composed_ta; i++) {
    if(p1[i] < p2[i]) return -1;
    if(p1[i] > p2[i]) return 1;
  }
  return EQUAL;
}

int mode_perm_comp(struct mode_type *m1, struct mode_type *m2) {
  int comp = mode_comp(m1, m2);
  if(comp == EQUAL) 
    return permutation_comp(m1->norm_perm, m2->norm_perm);
  else 
    return comp;
}


/* Functions for storing permutation structs in 23tree database */

int permutation_comp(struct permutation_type *p1, struct permutation_type *p2) {
  if(p1 == NULL && p2 != NULL) return 1;
  if(p1 != NULL && p2 == NULL) return -1;
  if(p1 == NULL && p2 == NULL) return EQUAL;
  return perm_comp(p1->perm, p2->perm);
}

/*
int full_permutation_comp(struct permutation_type *p1, 
                          struct permutation_type *p2) {
  int i;
  if(p1 == NULL && p2 == NULL) return EQUAL;
  if(p1 == NULL && p2 != NULL) return -1;
  if(p1 != NULL && p2 == NULL) return 1;
  for(i=1; i<=system_ta_num; i++) {
    if(p1->perm[i] < p2->perm[i]) return -1;
    if(p1->perm[i] > p2->perm[i]) return 1;
  }
  return EQUAL;
}
*/


int perm_count_inc(struct permutation_type *p) {
  (p->count)++;
  return 0;
}

int perm_parent_set(struct permutation_type *p, struct a23tree_type *parent) {
  p->parent = parent;
  return 0;
}

int free_perm(struct permutation_type *p) {
  free(p->perm);
  free(p);
  return 0;
}

struct permutation_type *alloc_permutation() {
  struct permutation_type *p = (struct permutation_type *)
                               malloc(sizeof(struct permutation_type));
  p->parent = (struct a23tree_type *)NULL;
  p->count = 0;
  p->perm = (int *)NULL;
  return p;
}

int print_permutation(struct permutation_type *p) {
  if(p)
    print_perm(p->perm);
  else print_perm(NULL);
  return 0;
}

struct var_link_type *sort_var_links(struct var_link_type *sc) {
  struct var_link_type *vl, *ans = NULL;
  if(sc) {
    vl=sc->next;
    sc->next = NULL;
    sc=vl;
    while(sc) {
      sc = sc->next;
      ans = insert_var_link(vl, ans);
      vl = sc;
    }
  }
  return ans;
}


/* for on-the-fly permutation during merging:
 * used to update the transition permutation when an existing mode
 * is used for the new mode 
 * compose(compose(tr_perm, inv_perm(new_mode_perm)), exist_mode_perm);
 */

/*
int update_same_mode_trans_perm(int *new_mode_perm,
                                struct transition_type *tr, 
                                struct mode_type *nm) {
  int *tmp_perm1 = NULL, 
      *tmp_perm2 = NULL;
  struct permutation_type *p;

  if(new_mode_perm) {
    tmp_perm1 = inverse_permutation(new_mode_perm);
    if(tr->permutation) {
      tmp_perm2 = compose_permutation(tr->permutation->perm, tmp_perm1);
      free(tmp_perm1);
      tmp_perm1 = NULL;
    }
    else tmp_perm2 = tmp_perm1;
  }
  else if(tr->permutation)
    tmp_perm2 = copy_permutation(tmp_perm2, tr->permutation->perm);
  if(nm->norm_perm) {
    tmp_perm1 = compose_permutation(tmp_perm2, nm->norm_perm->perm);
    free(tmp_perm2);
    tmp_perm2 = NULL;
  }
  else tmp_perm1 = tmp_perm2;

  p = alloc_permutation();
  p->perm = tmp_perm1;
  tr->permutation = (struct permutation_type *)
                     add_23tree(p, perm_tree, permutation_comp,
                     free_perm, perm_count_inc, perm_parent_set,
                     print_permutation);
}
*/


/* Must record the permutation in the merge hierarchy for consistency */

/*
int update_new_mode_perm(struct mode_type *nm, struct mode_type *m) {
  struct permutation_type *p;

  if(m->norm_perm) {
    p = alloc_permutation();
    if(nm->norm_perm) {
      p->perm = compose_permutation(m->norm_perm->perm, nm->norm_perm->perm);
    }
    else p->perm = copy_permutation(p->perm, m->norm_perm->perm);
    nm->norm_perm = (struct permutation_type *)
                    add_23tree(p, perm_tree, permutation_comp,
                    free_perm, perm_count_inc, perm_parent_set,
                    print_permutation);
  }
  if(verbose >= 0) { 
    sprintf(tmpstring, 
            "Updated permutation on mode %s(%d) [%s(%d), %s(%d)] to ", 
            nm->name, nm->index, nm->submode1->name, nm->submode1->index,
            nm->submode2->name, nm->submode2->index);
    output(tmpstring);
    print_permutation(nm->norm_perm);
  }
  return;
}
*/

int reset_ta_list(struct thread_type *t) {
  int i;
  for(i=0; i<system_ta_num; i++) {
    t->ta_list[i] = FALSE;
  }
  return 0;
}

int update_ta_list(struct thread_type *t1,
                   struct thread_type *t2,
                   struct thread_type *t3) {
  int i;
  for(i=0; i<system_ta_num; i++) {
    if(t1->ta_list[i] == TRUE || t2->ta_list[i] == TRUE)
      t3->ta_list[i] = TRUE;
    else
      t3->ta_list[i] = FALSE;
  }
  return 0;
}


/* SYNC related functions */

struct variable_type *get_same_sync(struct transition_type *tr1, 
                                    struct transition_type *tr2) {
  struct var_link_type *vl1, *vl2;

/*
  sprintf(tmpstring, "checking same sync of trans: %d and %d.\n",
                     tr1->index, tr2->index);
  output(tmpstring);
*/

  if(tr1->trn->sync_cycle && tr2->trn->sync_cycle) {
    for(vl1=tr1->trn->sync_cycle->next; ; vl1=vl1->next) {
      for(vl2=tr2->trn->sync_cycle->next; ; vl2=vl2->next) {

/*
        sprintf(tmpstring, "  syncs: %d and %d.\n", 
                           vl1->var->index, vl2->var->index);
        output(tmpstring);
*/

        if(vl1->var->index == vl2->var->index) return vl1->var;
        if(vl2 == tr2->trn->sync_cycle) break;
      }
      if(vl1 == tr1->trn->sync_cycle) break;
    }
  }
  return (struct variable_type *)NULL;
}

/* Get a cycle of all out-going transitions of m2 
   that has a sync label same as one of tr1
 */

struct trans_link_type *get_sync_trans(struct transition_type *tr1,
                                       struct mode_type *m2) {
  struct trans_link_type *tl = NULL,    /* Cycle of trans_link */
                         *tlp = NULL;   /* Pointer to a new trans_link */
  struct transition_type *tr;

  if(m2->out_trans_cycle) {
    for(tr=m2->out_trans_cycle->next; ; tr=tr->next) {
      if(get_same_sync(tr1, tr) != NULL) {
        tlp = (struct trans_link_type *)malloc(sizeof(struct trans_link_type));
        tlp->tr = tr;
        if(tl) {
          tlp->next = tl->next;
          tl->next = tlp;
        }
        else tlp->next = tlp;
        tl = tlp;
      }
      if(tr == m2->out_trans_cycle) break;
    }
  }
  return tl;
}

/* check_sync() return values:
   -1 : No THREAD2			   => Error
    0 : No same sync in THREAD2            => Sync Later with some external TA!
    1 : Same sync in out_trans_cycle of m2 => Sync Now!
    2 : Same sync in THREAD2 Later       => Sync Later with THREAD2, Ignore Now!
 */

int check_sync(struct transition_type *tr1, struct mode_type *m2, int th) {

/* th == 1  ==>  search for sync in THREAD1
   th == 2  ==>  search for sync in THREAD2 */
   
  struct mode_type *m;
  struct transition_type *tr;

  if(verbose >= 3) {
    sprintf(tmpstring, "  Checking sync of transition %d with mode %s(%d).\n", 
                       tr1->index, m2->name, m2->index);
    output(tmpstring);
  }
  if(th == 2 && THREAD2->starting_mode == NULL) return -1;
  if(th == 1 && THREAD1->starting_mode == NULL) return -1;

  if(get_sync_trans(tr1, m2) != NULL) {
    if(verbose >= 3) 
      output("    Found same sync transitions. Sync NOW!\n");
    return 1;
  }
  for(m=(th==1)?THREAD1->starting_mode:THREAD2->starting_mode; m; m=m->cnext) {
    if(m->out_trans_cycle) {
      for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
        if(get_same_sync(tr1, tr) != NULL) {
          if(verbose >= 3) 
            output("    Found later same sync transition. Sync LATER!\n");
          return 2;
        }
        if(tr == m->out_trans_cycle) break;
      } /* end for(tr=m...) */
    } /* end if(m->...) */
  } /* end for(m=THREAD2->starting_mode ...) */
  if(verbose >= 3)
    output("    Sync later with external TA! Continue unsync composition.\n");
  return 0;
}

/* Allocate a new struct spec_trans_type */

struct spec_trans_type *alloc_spec_trans() {
  struct spec_trans_type *x;

  x = (struct spec_trans_type *)malloc(sizeof(struct spec_trans_type));
  x->index = ++trans_count;
  x->goto_mode = (char *)NULL;
  x->clk_trigger_length = x->prp_trigger_length = x->assign_length = 0;
  x->clk_trigger_cycle = (struct clkconj_type *)NULL;
  x->prp_trigger_cycle = (struct boolconj_type *)NULL;
  x->assign_cycle = (struct assign_type *)NULL;
  x->trigger_bdd = (bdd)NULL;
  x->sync_cycle = (struct var_link_type *)NULL;
  return x;
}

/* Get the number of clock conjuncts in c */

int get_clk_trigger_length(struct clkconj_type *c) {
  struct clkconj_type *x;
  int n = 0;

  if(c)
    for(x = c->next; ; x=x->next) {
      n++;
      if(x == c) break;
    }
  return n;
}

/* Get the number of proposition conjuncts in b */

int get_prp_trigger_length(struct boolconj_type *b) {
  struct boolconj_type *x;
  int n = 0;

  if(b)
    for(x = b->next; ; x=x->next) {
      n++;
      if(x == b) break;
    }
  return n;
}

/* Get the number of assignments in b */

int get_assign_length(struct assign_type *b) {
  struct assign_type *x;
  int n = 0;

  if(b)
    for(x = b->next; ; x=x->next) {
      n++;
      if(x == b) break;
    }
  return n;
}

/* Synchronize two transitions into a new single transition */

struct transition_type *compose_sync_transition(struct transition_type *t1,
					        struct transition_type *t2) {
  struct spec_trans_type *str;
  struct transition_type *tr;

  str = alloc_spec_trans();
  tr = alloc_transition();
  tr->index = ++trans_count;

/* how to assign tr->automaton ? */
/* tr->automaton == NULL ==> a SYNC'ed transition */

  if(verbose >= 0) {
    sprintf(tmpstring, "Syncing transitions %d and %d into %d.\n", 
                       t1->index, t2->index, tr->index);
    output(tmpstring);
  }

  /* Compose tr->trn */
  str->goto_mode = (char *)malloc(sizeof(char)*5);
  strcpy(str->goto_mode, "SYNC");

  str->clk_trigger_cycle = clkconj_and(t1->trn->clk_trigger_cycle, 
                                       t2->trn->clk_trigger_cycle);
  if(str->clk_trigger_cycle->op == FALSE) {
    if(verbose >= 0) {
      sprintf(tmpstring, "  New sync trans %d has FALSE clock trigger.\n", 
                         tr->index);
      output(tmpstring);
    }
    free(str);
    free_transition(tr);
    return (struct transition_type *)NULL;
  }
  else str->clk_trigger_length = get_clk_trigger_length(str->clk_trigger_cycle);

  str->prp_trigger_cycle = boolconj_union(t1->trn->prp_trigger_cycle,
                                          t2->trn->prp_trigger_cycle);
  str->trigger_bdd = BDD_AND(t1->trn->trigger_bdd, t2->trn->trigger_bdd);
  if(str->trigger_bdd == BDD_FALSE) {
    if(verbose >= 0) {
      sprintf(tmpstring, "  New sync trans %d has FALSE proposition trigger.\n",
                         tr->index);
      output(tmpstring);
    }
    free(str);
    free_transition(tr);
    return (struct transition_type *)NULL;
  }
  else str->prp_trigger_length = get_prp_trigger_length(str->prp_trigger_cycle);

  str->assign_cycle = assign_union(t1->trn->assign_cycle,t2->trn->assign_cycle);
  str->assign_length = get_assign_length(str->assign_cycle);
  str->sync_cycle = var_link_union(t1->trn->sync_cycle, t2->trn->sync_cycle);

  tr->trn = str;

  /* Compose tr->permutation */
  tr->permutation = alloc_permutation();
  if(t1->permutation) {
     if(t2->permutation) 
        tr->permutation->perm = compose_permutation(t1->permutation->perm,
                                                    t2->permutation->perm);
     else tr->permutation->perm = 
              copy_permutation(tr->permutation->perm, t1->permutation->perm);
  }
  else if(t2->permutation)
         tr->permutation->perm = 
            copy_permutation(tr->permutation->perm, t2->permutation->perm);

  /* how to handle tr->mode */
  /* tr->mode assigned in add_sync_transition_fifo() */

  /* how to compose tr->bypassed_internal_vars */
  /* Take union of two sets of bivs and hash it into a unique pointer */
  tr->bypassed_internal_vars = hash_bypassed_internal_var_list(
                                  var_link_union(t1->bypassed_internal_vars, 
                                                 t2->bypassed_internal_vars));
  return tr;
}

/* Take union of two assignment cycles */

struct assign_type *assign_union(struct assign_type *a1,
                                 struct assign_type *a2) {
  int found = FALSE;
  struct assign_type *a = NULL, *ap1, *ap2, *an;

  if(a1 == NULL && a2 == NULL) return NULL;
  if(a1 == NULL) { a = assign_copy(a2); return a; }
  if(a2 == NULL) { a = assign_copy(a1); return a; }
  a = assign_copy(a1);
  for(ap2=a2->next; ; ap2=ap2->next) {
    for(ap1=a1->next; ; ap1=ap1->next) {
      if(ap1->var == ap2->var &&
         ap1->use_var == ap2->use_var &&
         ap1->op == ap2->op &&
         ap1->mod_num == ap2->mod_num &&
         ap1->mod_var == ap2->mod_var &&
         ap1->value == ap2->value) {
         found = TRUE;
         break;
      }
      if(ap1 == a1) break;
    }
    if(found) found = FALSE;
    else {
      an = alloc_assign();
      an->var = ap2->var;
      an->use_var = ap2->use_var;
      an->op = ap2->op;
      an->value  = ap2->value;
      an->mod_num = ap2->mod_num;
      an->mod_var = ap2->mod_var;
      if(a) {
        an->next = a->next;
        a->next = an;
      }
      else an->next = an;
      a = an;
    }
    if(ap2 == a2) break;
  }
  return a;
}

/* Take union of two sync cycles */

struct var_link_type *var_link_union(struct var_link_type *a1,
                                     struct var_link_type *a2) {
  int found = FALSE;
  struct var_link_type *a = NULL, *ap1, *ap2, *an;

  if(a1 == NULL && a2 == NULL) return NULL;
  if(a1 == NULL) { a = var_link_copy(a2); return a; }
  if(a2 == NULL) { a = var_link_copy(a1); return a; }
  a = var_link_copy(a1);
  for(ap2=a2->next; ; ap2=ap2->next) {
    for(ap1=a1->next; ; ap1=ap1->next) {
      if(ap1->var->index == ap2->var->index) {
         found = TRUE;
         break;
      }
      if(ap1 == a1) break;
    }
    if(found) found = FALSE;
    else {
      an = (struct var_link_type *)malloc(sizeof(struct var_link_type));
      an->var = ap2->var;
      if(a) {
        an->next = a->next;
        a->next = an;
      }
      else an->next = an;
      a = an;
    }
    if(ap2 == a2) break;
  }
  return a;
}

struct transition_type *add_sync_transition_fifo(struct transition_type *tr1, 
                                                 struct transition_type *tr2, 
                                                 struct mode_type *sp, 
                                                 struct mode_type *ss) {
  struct transition_type *trn;
  trn = compose_sync_transition(tr1, tr2);
  trn->mode = sp;
  ss->in_trans_count++;
  if(ss->in_trans_cycle) {
    trn->next = ss->in_trans_cycle->next;
    ss->in_trans_cycle->next = trn;
    ss->in_trans_cycle = trn;
  }
  else {
    ss->in_trans_cycle = trn->next = trn;
  }

  trn = copy_transition(trn); /* make another copy for out_trans_cycle */
  trn->mode = ss;
  trn->index = trans_count;
  sp->out_trans_count++;
  if(sp->out_trans_cycle) {
    trn->next = sp->out_trans_cycle->next;
    sp->out_trans_cycle->next = trn;
    sp->out_trans_cycle = trn;
  }
  else {
    sp->out_trans_cycle = trn->next = trn;
  }
  return trn;
}

/* update the permutation on sync transition for on-the-fly region normalization
 * (using indice permutation) during merging
 */
int sync_update_permutation(struct mode_type *m, struct transition_type *tr1,
                            struct transition_type *tr2,
                            int *new_mode_perm, struct transition_type *ntr) {
  int *tmp1, *tmp2;
  struct permutation_type *p;

  if(verbose >= 3) {
    sprintf(tmpstring, "    Installing permutation for transition %d.\n", 
            ntr->index);
    output(tmpstring);
    sprintf(tmpstring, "\tSource mode %s(%d) has ", m->name, m->index);
    output(tmpstring);
    print_permutation(m->norm_perm);
  }
  
  if(m->norm_perm)
    tmp1 = inverse_permutation(m->norm_perm->perm);
  else tmp1 = (int *)NULL;

  if(verbose >= 3) {
    output("\tSource mode inverse ");
    print_perm(tmp1);
    sprintf(tmpstring, "\tTransition %d has ", tr1->index);
    output(tmpstring);
    print_permutation(tr1->permutation);
    sprintf(tmpstring, "and transition %d has ", tr2->index);
    output(tmpstring);
    print_permutation(tr2->permutation);
  }
  
  if(tr1->permutation) {
    tmp2 = compose_permutation(tmp1, tr1->permutation->perm);
    free(tmp1);
    tmp1 = NULL;
  }
  else {
    tmp2 = tmp1;
    tmp1 = (int *)NULL;
  }
  if(tr2->permutation) {
    tmp1 = compose_permutation(tmp2, tr2->permutation->perm);
    free(tmp2);
    tmp2 = tmp1;
    tmp1 = NULL;
  }

  if(verbose >= 3) {
    output("\tComposed inverse source_mode and trans permutations is ");
    print_perm(tmp2);
  }

/*
  if(verbose >= 3) {
    sprintf(tmpstring, "\tDestination mode %s(%d) has ", nm->name, nm->index);
    output(tmpstring);
    print_permutation(nm->norm_perm);
  }
*/
  if(new_mode_perm) {
    tmp1 = compose_permutation(tmp2, new_mode_perm);
    free(tmp2);
    tmp2 = NULL;
  }
  else {
    tmp1 = tmp2;
    tmp2 = (int *)NULL;
  }

  
  if(verbose >= 3) {
    output("Final ");
    print_perm(tmp1);
  }

  p = alloc_permutation();
  p->perm = tmp1;
  ntr->permutation = (struct permutation_type *)
                     add_23tree(p, perm_tree, permutation_comp,
                     free_perm, perm_count_inc, perm_parent_set,
                     print_permutation);

  if(verbose >= 0) {
    sprintf(tmpstring,
            "\tInstalled permutation on transition (%d), ", ntr->index);
    output(tmpstring);
    print_trans_permutation(ntr);
  }
  return 0;
}

