#include "global.h"
#include "include.h"
#include "modeman.h"
#include "symmetry.h"
#include "bddman.h"
#include "zoneman.h"
#include "print.h"
#include "shieldvar.h"

int total_dvar = 0;
int **val_count = NULL;   /* val_count[i][j] => i : automaton u_index
                           *                    j : discrete variable ordering
                           *  i.e., automaton (u_index=i) writes 
                           *  "val_count[i][j]" values
                           *  to the discrete variable with 
                           *  index "varval[i][j][0]"
                           */
int ***varval = NULL;
  /* varval[i][j][k] => i : automaton u_index
   *                    j : discrete variable index (varval[i][j][0])
   *                    k : 0     --> variable index
   *                        1 ... --> values written by thread[i] to 
   *                                  variable varval[i][j][0]
   */

int **vv = NULL, *vv_count = NULL;;

int init_varval(void);
int get_varval(struct automaton *t);
int get_dvar_id(int, struct automaton *);

int varval_written(int v, int, struct thread_type *t);
int insert_val(int val, int **int_array, int size);
int delete_val(int val, int **int_array, int size);
int print_varval(void);
struct boolconj_type *varval_literals(int ***v, 
                                      struct thread_type *t1, 
                                      struct thread_type *t2, 
                                      int is_initial_mode, 
                                      struct variable_type *);
struct boolconj_type *initial_dvl(struct thread_type *t1, 
                                  struct thread_type *t2, 
                                  int is_initial_mode, struct variable_type *);
int append_vals(int ***v, struct thread_type *t);
int append_varval(int ***v, int ti);
int remove_varval(int ***v, int ti);
int component_graph(int id, struct thread_type *t);
int remove_vals(int ***v, struct thread_type *t1, struct thread_type *t2);
int readwrite_reduce(struct thread_type *t);
struct boolconj_type *get_new_dvl(struct mode_type *m, 
                                  struct transition_type *tr);

int get_all_dvar_values(int ***, int **);
int get_vv_explicit(int ***, int *);
int **get_local_bounds(int v, struct transition_type *);
void free_bounds(int **);
int get_vv_self(int ***, int *);
int get_var_init_values(int, int **, int *);
int calc_possible_values(struct assign_type *, int **, int *, int, 
                         struct transition_type *);
int get_vv_implicit(int ***, int *);
int *get_vals_array(int, int);
int get_vals_count(int, int);
int print_val_array(int **, int *);


int init_varval() {
  int i, j;
  struct variable_type *v;
  struct automaton *a;

  varval = (int ***)malloc(sizeof(int **)*system_ta_num);
  for(i = 0; i < system_ta_num; i++) {
    varval[i] = (int **)NULL;
  }

  for(v=var_tail->next; ; v=v->next) {
    if(v->type == V_REGISTER) {
      total_dvar++;
      for(i = 0; i < system_ta_num; i++) {
        if(varval[i])
          varval[i] = (int **)realloc(varval[i], sizeof(int *)*total_dvar);
        else
          varval[i] = (int **)malloc(sizeof(int *));
        varval[i][total_dvar-1] = (int *)malloc(sizeof(int));
        varval[i][total_dvar-1][0] = v->index;
      }
    }
    if(v == var_tail) break;
  }

  val_count = (int **)malloc(sizeof(int *)*system_ta_num);
  for(i = 0; i < system_ta_num; i++) {
    val_count[i] = (int *)malloc(sizeof(int)*total_dvar);
  }
  for(i = 0; i < system_ta_num; i++) {
    for(j = 0; j < total_dvar; j++) {
      val_count[i][j] = 0;
    }
  }
  get_all_dvar_values(&vv, &vv_count);
  if(automaton_tail) {
    for(a = automaton_tail->next; ; a=a->next) {
      get_varval(a);
      if(a == automaton_tail) break;
    }
  }
  if(verbose >= 0)
    print_varval();
  return 0;
}

int get_varval(struct automaton *t) {
  struct mode_type *m;
  struct transition_type *tr;
  struct assign_type *a;
  int j, k;

  if(t->mode_cycle) {
    for(m=t->mode_cycle->next; ; m=m->next) {
      if(m->out_trans_cycle) {
        for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
          if(tr->trn->assign_cycle) {
            for(a=tr->trn->assign_cycle->next; ; a=a->next) {
              for(j=0; j<total_dvar; j++) {
                if(a->var == varval[t->u_index][j][0]) {
                  if(a->use_var == 0) {
                    if(insert_val(a->value, &varval[t->u_index][j], 
                                            val_count[t->u_index][j]) == TRUE)
                      val_count[t->u_index][j]++;
                  }
                  else {
                    for(k=1; k<=vv_count[j]; k++) {
                      if(insert_val(vv[j][k], &varval[t->u_index][j], 
                                              val_count[t->u_index][j]) == TRUE)
                        val_count[t->u_index][j]++;
                    }
                  }
                  break;
                }
              }
              if(a==tr->trn->assign_cycle) break;
            }
          }
          if(tr==m->out_trans_cycle) break;
        }
      }
      if(m == t->mode_cycle) break;
    }
  }
  return 0;
}

int varval_written(int v, int val, struct thread_type *t) {
  int i, j, found = FALSE;
  for(i=0; i<total_dvar; i++) {
    if(varval[t->index][i][0] == v) {
      found = TRUE;
      break;
    }
  }
  if(found) {
    for(j=1; j<val_count[t->index][i]; j++) {
      if(varval[t->index][i][j] == val)
        return TRUE;
    }
  }
  else {
    fprintf(stderr, "Variable %s[%d] not found in varval[][][].\n",
                    var_name(v), var_id(v));
    exit(1);
  }
  return FALSE;
}

int varval_written_auto(int v, int val, struct automaton *t) {
  int i, j, found = FALSE;

  for(i=0; i<total_dvar; i++) {
    if(varval[t->u_index][i][0] == v) {
      found = TRUE;
      break;
    }
  }
  if(found) {
    for(j=1; j<=val_count[t->u_index][i]; j++) {
      if(varval[t->u_index][i][j] == val)
        return TRUE;
    }
  }
  return FALSE;
}

int get_dvar_id(int v, struct automaton *t) {
  int i;
  for(i=0; i<total_dvar; i++) {
    if(varval[t->u_index][i][0] == v) {
      return i;
    }
  }
  return -1;  /* not found */
}

int insert_val(int val, int **int_array, int size) {
  int i, j;

  for(i=1; i<=size; i++) {
    if((*int_array)[i] == val)  /* duplicate value exists in array */
      return FALSE;
    if((*int_array)[i] > val) break;
  }
  if(*int_array)
    (*int_array) = (int *)realloc(*int_array, sizeof(int)*(size+2));
  else
    (*int_array) = (int *)malloc(sizeof(int)*(size+2));
  for(j = size + 1; j > i; j--) {
    (*int_array)[j] = (*int_array)[j-1];
  }
  (*int_array)[i] = val;
  return TRUE;
}

int delete_val(int val, int **int_array, int size) {
  int i, j, found = FALSE;

  for(i=1; i<=size; i++) {
    if((*int_array)[i] == val) {
      found = TRUE;
      break;
    }
  }
  if(found) {
    if(i < size) {
      for(j=i; j<size; j++) {
        (*int_array)[j] = (*int_array)[j+1];
      }
    }
    if(*int_array) 
      (*int_array) = (int *)realloc(*int_array, sizeof(int)*size);
    else
      (*int_array) = (int *)malloc(sizeof(int)*size);
    return TRUE;
  }
  else return FALSE;
}

int print_varval() {
  int i, j, k;
  struct automaton *a;

  output("#  Graph\tVariable\tValues\n");
 
  for(i=0, a=automaton_tail->next; i<system_ta_num; i++, a=a->next) {
    for(j=0; j<total_dvar; j++) {
      if(var_id(varval[i][j][0]) != -1) {
        if(a->index >= 0)
          sprintf(tmpstring, 
              "%d: %s[%d]\t%s[%d]\t\t", a->u_index, a->name, a->index,
                          var_name(varval[i][j][0]), var_id(varval[i][j][0]));
        else
          sprintf(tmpstring, 
              "%d: %s\t\t%s[%d]\t\t", a->u_index, a->name,
                          var_name(varval[i][j][0]), var_id(varval[i][j][0]));
        output(tmpstring);
      }
      else {
        if(a->index >= 0)
          sprintf(tmpstring, 
           "%d: %s[%d]\t%s\t\t", a->u_index, a->name, a->index,
                                 var_name(varval[i][j][0]));
        else 
          sprintf(tmpstring, 
           "%d: %s\t\t%s\t\t", a->u_index, a->name,
                                 var_name(varval[i][j][0]));
        output(tmpstring);
      }
      if(val_count[i][j])
      for(k=1; k<=val_count[i][j]; k++) {
        sprintf(tmpstring, "%d", varval[i][j][k]);
        output(tmpstring);
        if(k<val_count[i][j]) output(", ");
        else output("\n");
      }
      else output("\n");
    }
  }
  return 0;
}

struct boolconj_type *varval_literals(int ***v, 
                                      struct thread_type *t1, 
                                      struct thread_type *t2, 
                                      int is_initial_mode, 
                                      struct variable_type *vr) {
  int i, j;
  struct boolconj_type *b, *dvl=NULL;
  struct proposition_type *p;
  bdd f = NULL;

  for(j=0; j<total_dvar; j++) {
   if(vr == NULL || (vr && varval[0][j][0] == vr->index)) {
    for(i=1; i<=(*v)[j][0]; i++) {
      if(is_initial_mode) {
        for(p = prop_tail->next; ; p=p->next) {
          if(p->var1 == varval[0][j][0] && p->value == (*v)[j][i]) {
            f = bdd_not(bddm, p->prop_bdd);
            break;
          }
          if(p == prop_tail) break;
        }
        if(bdd_and(bddm, f, t1->init_boolexp->boolexp_bdd)
           == BDD_FALSE) {
          continue;
        }
        else if(bdd_and(bddm, f, t2->init_boolexp->boolexp_bdd)
           == BDD_FALSE) {
          continue;
        }
        else {
          b = alloc_boolconj();
          b->var1 = varval[0][j][0];
          b->op = NEQ;
          b->constant = (*v)[j][i];
          dvl = append_boolconj(b, dvl);
        }
      }
      else {
        b = alloc_boolconj();
        b->var1 = varval[0][j][0];
        b->op = NEQ;
        b->constant = (*v)[j][i];
        dvl = append_boolconj(b, dvl);
      }
    }
   if(vr) break;
   }
  }
  install_dvl_bdd(dvl);
  return dvl;
}


/* "val" contains all possible values assigned to all discrete variables 
 * by all TA in "t1" or by all TA in "t2", 
 * except those values that are also assigned by any TA not in "t1" or "t2" 
 */

struct boolconj_type *initial_dvl(struct thread_type *t1, 
                                  struct thread_type *t2, 
                                  int is_initial_mode, 
                                  struct variable_type *v) {
  int i;
  struct boolconj_type *b;
  int **val = (int **)malloc(sizeof(int *)*total_dvar);

/*
  sprintf(tmpstring, "Getting Initial DVL from threads %d and %d.\n", 
                     t1->index, t2->index);
  output(tmpstring);

  sprintf(tmpstring, "Total discrete vars = %d.\n", total_dvar);
  output(tmpstring);
*/

  for(i=0; i<total_dvar; i++) {
    val[i] = (int *)malloc(sizeof(int));
    val[i][0] = 0;
  }
  if(t1)
    append_vals(&val, t1);
  if(t2)
    append_vals(&val, t2);
  if(t1 && t2)
    remove_vals(&val, t1, t2);

  b = varval_literals(&val, t1, t2, is_initial_mode, v);

  for(i=0; i<total_dvar; i++) {
    free(val[i]);
  }
  free(val);
  install_dvl_bdd(b);
  return b;
}

int append_vals(int ***v, struct thread_type *t) {
  int i;
  for(i=0; i<system_ta_num; i++) {
    if(component_graph(i, t)) {
      append_varval(v, i);
    }
  }
/* The following does not work with graph resetting */
/*
  if(t->subthread1 == NULL && t->subthread2 == NULL) {
    append_varval(v, t->index);
  }
  else {
    append_vals(v, t->subthread1);
    append_vals(v, t->subthread2);
  }
*/
  return 0;
}

int append_varval(int ***v, int ti) {
  int i, j;
  for(i=0; i<total_dvar; i++) {
    for(j=1; j<=val_count[ti][i]; j++) {
      if(insert_val(varval[ti][i][j], &((*v)[i]), (*v)[i][0]) == TRUE)
        (*v)[i][0]++;
    }
  }
  return 0;
}

int remove_varval(int ***v, int ti) {
  int i, j;
  for(i=0; i<total_dvar; i++) {
    for(j=1; j<=val_count[ti][i]; j++) {
      if(delete_val(varval[ti][i][j], &((*v)[i]), (*v)[i][0]) == TRUE)
        (*v)[i][0]--;
    }
  }
  return 0;
}

int component_graph(int id, struct thread_type *t) {
  if(t) return(t->ta_list[id]);
/* The following does not work with graph resetting */
/*
  if(t->subthread1 == NULL && t->subthread2 == NULL) {
    if(t->index == id) return TRUE;
  }
  else {
    if(component_graph(id, t->subthread1)) return TRUE;
    if(component_graph(id, t->subthread2)) return TRUE;
  }
*/
  return FALSE;
}

int remove_vals(int ***v, struct thread_type *t1, struct thread_type *t2) {
  int i;
  for(i=0; i<system_ta_num; i++) {
    if( !(component_graph(i, t1) || component_graph(i, t2)) ) {
      remove_varval(v, i);
    }
  }
  return 0;
}

int readwrite_reduce(struct thread_type *t) {
  int i;
  struct mode_type *m;
  struct transition_type *tr, *str;
  struct boolconj_type *b1=NULL, *b2=NULL;
  int rw = lazy_bdd_merging, last = FALSE;
  char *c = get_graph_name(t);

  if(t->mode_count == 0) {
    fprintf(stderr, "Thread %s contains no mode.\n", c);
    return FALSE;
  }

  if(t->subthread1 == NULL && t->subthread2 == NULL) {
    fprintf(stderr, "Input graph %s cannot be readwrite reduced.\n", 
                    c);
    return FALSE;
  }

  NEW_THREAD = t;
  t->mnpl |= RW;
  lazy_bdd_merging = FALSE;

  if(verbose >= 0) {
    sprintf(tmpstring, "Read-Write reducing graph %s...\n", c);
    output(tmpstring);
  }
  printf("    Read-Write reducing graph %s ...\n", c);
  t->starting_mode->dvl = initial_dvl(t->subthread1, t->subthread2, TRUE, NULL);

/*
  sprintf(tmpstring, "initial dvl of starting mode %s(%d) of thread %s is: ", 
                     t->starting_mode->name, t->starting_mode->index, c);
  output(tmpstring);
  print_dvl(t->starting_mode->dvl);
  output("BDD: ");
  print_bdd(dvl_bdd(t->starting_mode));
*/
  
  for(m=t->starting_mode; m; m=m->cnext) {
    if(m->out_trans_cycle) {
      for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
/*
        if(tr->mode->dvl == NULL) get_new_dvl(m, tr);
*/
	b1 = tr->mode->dvl;
	b2 = get_new_dvl(m, tr);
	tr->mode->dvl = boolconj_union(b1, b2);
	free_boolconj(b1); b1 = NULL;
	free_boolconj(b2); b2 = NULL;
        if(tr == m->out_trans_cycle) break;
      }
    }
  }
  for(m=t->starting_mode,i=1; m; m=m->cnext,i++) {
    if(verbose >= 3) {
      sprintf(tmpstring, "Mode %s(%d) has DVL BDD: ", m->name, m->index);
      output(tmpstring);
      print_bdd(dvl_bdd(m));
      output("\n");
      output("DVL: ");
      print_dvl(m->dvl);
      output("\n");
    }
    if(m->out_trans_cycle) {
      last = FALSE;
      for(tr=m->out_trans_cycle->next, str=tr->next; ; tr=str, str=str->next) {
        if(verbose >= 3) {
	  sprintf(tmpstring, "Trans %d has TRIGGER BDD: ", tr->index);
	  output(tmpstring);
	  print_bdd(tr->trn->trigger_bdd);
        }
        if(!lastcomp && 
/*
           (BDD_AND(dvl_bdd(m), tr->trn->trigger_bdd) == BDD_FALSE))
*/
           (BDD_AND(dvl_bdd(m), 
                    get_bdd_from_dvl(b1=hide_vars(tr->bypassed_internal_vars, 
                                     tr->trn->prp_trigger_cycle))) 
                    == BDD_FALSE))
        {
          free_boolconj(b1); b1 = NULL;
          if(tr == m->out_trans_cycle) last = TRUE;
          if(verbose >= 0) {
            sprintf(tmpstring,
                   "\tEliminating transition (%d) from %s (%d) to %s (%d).\n",
             tr->index, m->name, m->index, tr->mode->name, tr->mode->index);
            output(tmpstring);
          }
          eliminate_transition(tr, m);
          t->trans_count--;
          if(m->out_trans_cycle == NULL || last == TRUE) break;
          else continue;
        }
        free_boolconj(b1); b1 = NULL;
        if(tr == m->out_trans_cycle) break;
      }
    }
    if(stats == TRUE && SGM_MODE != GUI && SGM_MODE != PROCESS) {
      clear_line();
      fprintf(stderr, "    RW-Checked %d modes, Remaining %d modes (%.1f%%)\r",
                      i, t->mode_count-i,
              ((float)(t->mode_count-i) / (float)t->mode_count)*100.0);
    }
  }
  clean_line();
  printf("    "); print_counts(t);
  merge_iden_red_multi(t);

  printf("\tRemoving irreachable modes in graph %s.\n\t", c);
  if(verbose >= 0) {
    outputline();
    sprintf(tmpstring,
    "Removing irreachable modes in graph %s (%d modes, %d transitions).\n",
       c, t->mode_count, t->trans_count);
    output(tmpstring);
  }
  remove_irreachable_mode(t);
  if(verbose >= 0) output("\t");
  print_counts(t);

  lazy_bdd_merging = rw;
  free(c);
  return 0;
}

struct boolconj_type *get_new_dvl(struct mode_type *m, 
                                  struct transition_type *tr) {
  int *tmp_perm1 = NULL, *tmp_perm2 = NULL;
  struct boolconj_type *tmp_bc = NULL;
  struct boolconj_type *ans = NULL;

  if(verbose >= 2) {
    sprintf(tmpstring,
             "Getting DVL from mode %s(%d) through trans %d for mode %s(%d):\n",
             m->name, m->index, tr->index, tr->mode->name, tr->mode->index);
    output(tmpstring);
    print_trans(tr);
  }
  ans = compute_dvl(m, tr);

  if(verbose >= 2) {
    sprintf(tmpstring, 
            "  After compute_dvl(tr=%d), mode %s(%d) will have DVL:\n",
               tr->index, tr->mode->name, tr->mode->index);
    output(tmpstring);
    print_dvl(ans);
    print_trans(tr);
  }

  if(on_the_fly_normalization && m->norm_perm)
    tmp_perm1 = inverse_permutation(m->norm_perm->perm);
  else
    tmp_perm1 = (int *)NULL;
  if(tr->permutation) {
    tmp_perm2 = compose_permutation(tmp_perm1, tr->permutation->perm);
    free(tmp_perm1);
  }
  else {
    tmp_perm2 = tmp_perm1;
    tmp_perm1 = (int *)NULL;
  }
  if(tmp_perm2) {
    tmp_bc = permute_boolconj(ans, tmp_perm2);
    free_boolconj(ans);
    ans = tmp_bc;
  }

  if(verbose >= 2) {
    sprintf(tmpstring, "  After permutation, mode %s(%d) will have DVL:\n",
                       tr->mode->name, tr->mode->index);
    output(tmpstring);
    print_dvl(ans);
    output("DVL BDD: ");
    print_dvl_bdd(ans);
    output("\n");
  }
  free(tmp_perm2);
  return ans;
}


/* Return the set of all possible values assigned to all discrete variables
 * by all TA with all sorts of assignments: explicit, self, and implicit.
 * (*vv)[i] => a list of all possible values assigned to variable with index
 *             varval[X][i][0], where 0 <= X < system_ta_num */

int get_all_dvar_values(int ***vv, int **vv_count) {
  int i;

  *vv = (int **)malloc(sizeof(int *) * total_dvar);
  *vv_count = (int *)malloc(sizeof(int) * total_dvar);
  for(i=0; i<total_dvar; i++) {
    (*vv)[i] = (int *)NULL;
    (*vv_count)[i] = 0;
  }
  get_vv_explicit(vv, *vv_count);
  get_vv_self(vv, *vv_count);
  get_vv_implicit(vv, *vv_count);
  if(verbose >= 3)
    print_val_array(*vv, *vv_count);

  return 0;
}


/* explicit => NO use_var in assignments, such as   var := val; */
/* Return arrays of possible values assigned to all discrete vars by all TA
 * Assignments considered: v1 := n, 
 *                   where v1 is a var and n is a positive integer */

int get_vv_explicit(int ***vv, int *vv_count) {
  struct automaton *ta;
  struct mode_type *m;
  struct transition_type *tr;
  struct assign_type *a;
  int j;

  if(automaton_tail) {
    for(ta = automaton_tail->next; ; ta = ta->next) {
      if(ta->mode_cycle) {
        for(m=ta->mode_cycle->next; ; m=m->next) {
          if(m->out_trans_cycle) {
            for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
              if(tr->trn->assign_cycle) {
                for(a=tr->trn->assign_cycle->next; ; a=a->next) {
                  if(a->use_var) {
                    if(a==tr->trn->assign_cycle) break;
                    else continue;
                  }
                  for(j=0; j<total_dvar; j++) {
                    if(a->var == varval[ta->u_index][j][0]) {
                      if(insert_val(a->value, &((*vv)[j]), vv_count[j]) == TRUE)
                        vv_count[j]++;
                      break;
                    }
                  }
                  if(a==tr->trn->assign_cycle) break;
                }
              }
              if(tr==m->out_trans_cycle) break;
            }
          }
          if(m == ta->mode_cycle) break;
        }
      }
      if(ta == automaton_tail) break;
    }
  }
  return 0;
}

/* Get the local bounds on the values of a discrete variable v 
 *  due to triggers on transition tr
 *    bound[0] and bound[1] define LOWER BOUND (operator and bound value)
 *    bound[2] and bound[3] define UPPER BOUND (operator and bound value)
 *    operator = GEQ or GREATER for lower bound
 *    operator = LEQ or LESS for upper bound
 */

int **get_local_bounds(int v, struct transition_type *tr) {
  struct boolconj_type *b;
  int **bounds = (int **)malloc(4*sizeof(int *));
  int i;

  for(i = 0; i < 4; i++) {
    bounds[i] = (int *)NULL;
  }

  if(tr->trn->prp_trigger_cycle) {
    for(b=tr->trn->prp_trigger_cycle->next; ; b=b->next) {
      if(b->var1 == v) {
        switch(b->op) {
          case GEQ:
          case GREATER:
                    bounds[0] = (int *)malloc(sizeof(int));
                    *bounds[0] = b->op;
                    bounds[1] = (int *)malloc(sizeof(int));
                    *bounds[1] = b->constant;
               break;
          case LEQ:
          case LESS:
                    bounds[2] = (int *)malloc(sizeof(int));
                    *bounds[2] = b->op;
                    bounds[3] = (int *)malloc(sizeof(int));
                    *bounds[3] = b->constant;
               break;
          case EQ:
                    bounds[0] = (int *)malloc(sizeof(int));
                    *bounds[0] = b->op;
                    bounds[1] = (int *)malloc(sizeof(int));
                    *bounds[1] = b->constant;
                    bounds[2] = (int *)malloc(sizeof(int));
                    *bounds[2] = b->op;
                    bounds[3] = (int *)malloc(sizeof(int));
                    *bounds[3] = b->constant;
               break;
        }
      }
      if(b == tr->trn->prp_trigger_cycle) break;
    }
  }
  return bounds;
}

/* Free the 4-element array of bounds */
void free_bounds(int **bound) {
  int i;
  for(i = 0; i < 4; i++) {
    if(bound[i]) free(bound[i]);
  }
  free(bound);
}

/* Return arrays of possible values assigned to all discrete vars by all TA
 * Assignments considered: v1 := v2 +/- n MOD m,  OR
 *                         v1 := v2 +/- n (with a bound on values of v2), 
 *                   where v1 == v2 are vars, n and m are positive integers */

int get_vv_self(int ***vv, int *vv_count) {
  struct automaton *ta;
  struct mode_type *m;
  struct transition_type *tr;
  struct assign_type *a;
  int j, i, count = 0, *list = NULL, valcount = 0, *vals = NULL;

  if(automaton_tail) {
    for(ta = automaton_tail->next; ; ta = ta->next) {
      if(ta->mode_cycle) {
        for(m=ta->mode_cycle->next; ; m=m->next) {
          if(m->out_trans_cycle) {
            for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
              if(tr->trn->assign_cycle) {
                for(a=tr->trn->assign_cycle->next; ; a=a->next) {
                  if(a->use_var == 0 || a->use_var != a->var) {
                    if(a==tr->trn->assign_cycle) break;
                    else continue;
                  }
                  get_var_init_values(a->use_var, &list, &count);
                  valcount = calc_possible_values(a, &vals, list, count, tr);
                  free(list); list = NULL; count = 0;
                  for(j=0; j<total_dvar; j++) {
                    if(a->var == varval[ta->u_index][j][0]) {
                      for(i=1; i<=valcount; i++) {
                       if(insert_val(vals[i], &((*vv)[j]), vv_count[j]) == TRUE)
                         vv_count[j]++;
                      }
                      free(vals); vals = NULL; valcount = 0;
                      break;
                    }
                  }
                  if(a==tr->trn->assign_cycle) break;
                }
              }
              if(tr==m->out_trans_cycle) break;
            }
          }
          if(m == ta->mode_cycle) break;
        }
      }
      if(ta == automaton_tail) break;
    }
  }
  return 0;
}


/* Returns the "list" of all initial values that variable "v" can take,
   total number of values is "count" */

int get_var_init_values(int v, int **list, int *count) {
  struct automaton *ta;
  struct boolexp_type *be;
  struct boolconj_type *bc;
  
  *count = 0;
  *list = NULL;
  if(automaton_tail) {
    for(ta = automaton_tail->next; ; ta = ta->next) {
      if(ta->init_boolexp) {
        for(be=ta->init_boolexp->next; ; be=be->next) {
          if(be->boolconj) {
            for(bc=be->boolconj->next; ; bc=bc->next) {
              if(bc->var1 == v) {
                if(bc->var2) 
                  fprintf(stderr, 
                  "Warning: Initial condition contains diff of 2 variables.\n");
                else if(bc->op != EQ)
                  fprintf(stderr, 
                  "Warning: Initial condition is not an equality (EQ).\n");
                else {
                  if(insert_val(bc->constant, list, *count) == TRUE)
                    (*count)++;
/*
                  if(*count == 0) {
                    *list = (int *)malloc(sizeof(int));
                    (*list)[0] = bc->constant;
                  }
                  else {
                    for(i=0; i<(*count); i++) {
                      if((*list)[i] == bc->constant) {
                        if(bc == be->boolconj) break;
                        else continue;
                      }
                    }
                    *list = (int *)realloc(*list, sizeof(int)*((*count)+1));
                    (*list)[*count] = bc->constant;
                  }
                  (*count)++;
*/
                }  
              }
              if(bc == be->boolconj) break;
            }
          }
          if(be == ta->init_boolexp) break;
        }
      }
      if(ta == automaton_tail) break;
    }
  }
  return 0;
}

/* Calculates list "vals" of all the values that variable "a->var" can take 
   due to the self-assignment "a" with totally "count" initial values in "list"
   here, self-assignment is : a->var := a->use_var a->op a->value MOD a->mod_num
     where a->use == a->var, a->op == ADD or SUB, 
           a->value and a->mod_num are positive integers
                      AND (semantically the same)
   Calculates list "vals" of all the values that variable "a->var" can take
   due to the assignment "a" with totally "count" values of "a->use_var" in
   "list"
 */

int calc_possible_values(struct assign_type *a, int **vals, 
                         int *list, int count, struct transition_type *tr){
  int i, total=0, val=-1, oldval;
  int **bounds;

  for(i=1; i<=count; i++) {
    oldval = list[i];
    if(a->mod_num) {
      while(1) {
        if(a->op == ADD)
          val = (int)fmod(oldval + a->value, a->mod_num);
        else if(a->op == SUB)
          val = (int)fmod(oldval - a->value, a->mod_num);
        if(insert_val(val, vals, total) == TRUE) {
          total++;
          oldval = val;
        }
        else break;
      }
    }
    else if(a->mod_var) {
      fprintf(stderr,"Error: assignment with MOD var not yet implemented.\n");
    }
    else if(a->var == a->use_var) {
      bounds = get_local_bounds(a->var, tr);
      if(a->op == ADD) {
        if(bounds[2] == NULL)
          fprintf(stderr, 
          "Error: upper unbounded self assignment without MOD not allowed (Trans %d).\n", tr->index);
        else
        switch(*bounds[2]) {
          case EQ:
            if(oldval == *bounds[3]) {
              val = oldval + a->value;
              if(insert_val(val, vals, total) == TRUE) total++;
            }
            break;
          case LEQ:
            while(oldval <= *bounds[3]) {
              val = oldval + a->value;
              if(insert_val(val, vals, total) == TRUE) total++;
              oldval = val;
            }
            break;
          case LESS:
            while(oldval < *bounds[3]) {
              val = oldval + a->value;
              if(insert_val(val, vals, total) == TRUE) total++;
              oldval = val;
            }
            break;
        }
      }
      else if(a->op == SUB) {
        if(bounds[0] == NULL)
          fprintf(stderr, 
               "Error: lower unbounded self assignment without MOD not allowed (Trans %d).\n", tr->index);
        else
        switch(*bounds[0]) {
          case EQ:
            if(oldval == *bounds[1]) {
              val = oldval - a->value;
              if(insert_val(val, vals, total) == TRUE) total++;
            }
            break;
          case GEQ:
            while(oldval >= *bounds[1]) {
              val = oldval - a->value;
              if(insert_val(val, vals, total) == TRUE) total++;
              oldval = val;
            }
            break;
          case GREATER:
            while(oldval > *bounds[1]) {
              val = oldval - a->value;
              if(insert_val(val, vals, total) == TRUE) total++;
              oldval = val;
            }
            break;
        }
      }
    } /* end else if(a->var == a->use_var) */
    else if(a->value) {
      fprintf(stderr, 
      "Error: implicit assignment with add/sub and without MOD not allowed.\n");
    }
    else {       /* a->var := a->use_var; */
      if(insert_val(list[i], vals, total) == TRUE)
        total++;
    }
  }
  return total;
}


/* Return arrays of possible values assigned to all discrete vars by all TA
 * Assignments considered: v1 := v2 +/- n MOD m, 
 *                   where v1 != v2 are vars, n and m are positive integers */

int get_vv_implicit(int ***vv, int *vv_count) {
  struct automaton *ta;
  struct mode_type *m;
  struct transition_type *tr;
  struct assign_type *a;
  int j, i, valcount, *vals = NULL, *list = NULL, count = 0;

  if(automaton_tail) {
    for(ta = automaton_tail->next; ; ta = ta->next) {
      if(ta->mode_cycle) {
        for(m=ta->mode_cycle->next; ; m=m->next) {
          if(m->out_trans_cycle) {
            for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
              if(tr->trn->assign_cycle) {
                for(a=tr->trn->assign_cycle->next; ; a=a->next) {
                  if(a->use_var == 0 || a->use_var == a->var) {
                    if(a==tr->trn->assign_cycle) break;
                    else continue;
                  }
                  for(j=0; j<total_dvar; j++) {
                    if(a->var == varval[ta->u_index][j][0]) {
                      valcount = calc_possible_values(a, &vals, 
                                      get_vals_array(a->use_var, ta->u_index),
                                      get_vals_count(a->use_var, ta->u_index),
                                      (struct transition_type *)NULL);
                      for(i=1; i<=valcount; i++) {
                       if(insert_val(vals[i], &((*vv)[j]), vv_count[j]) == TRUE)
                         vv_count[j]++;
                      }
                      free(vals); vals = NULL;
                      if(valcount == 0) {
                        get_var_init_values(a->use_var, &list, &count);
                        for(i=1; i<=count; i++) {
                         if(insert_val(list[i], &((*vv)[j]), vv_count[j])==TRUE)
                           vv_count[j]++;
                        }
                        free(list); list = NULL; count = 0;
                      }
                      else valcount = 0;
                      break;
                    }
                  }
                  if(a==tr->trn->assign_cycle) break;
                }
              }
              if(tr==m->out_trans_cycle) break;
            }
          }
          if(m == ta->mode_cycle) break;
        }
      }
      if(ta == automaton_tail) break;
    }
  }
  return 0;
}


/* Return the array of possible values assigned to variable 'var' by all TA */

int *get_vals_array(int var, int auto_index) {
  int j;

  for(j=0; j<total_dvar; j++) {
    if(var == varval[auto_index][j][0]) 
      return vv[j];
  }
  return NULL;
}


/* Return the number of possible values assigned to variable 'var' by all TA */

int get_vals_count(int var, int auto_index) {
  int j;

  for(j=0; j<total_dvar; j++) {
    if(var == varval[auto_index][j][0]) 
      return vv_count[j];
  }
  return -1;
}

/* Array "vv" contains all possible values assigned to discrete vars by all TA 
 * vv[j] contains values of variable with index varval[X][j][0], 
 *                                        where 0 <= X < system_ta_num
 */
int print_val_array(int **vv, int *vv_count) {
  int j, k;
 
  output("Lists of Possible Values assigned to Discrete Vars By All TA.\n");
  for(j=0; j<total_dvar; j++) {
    if(var_id(varval[0][j][0]) != -1) {
      sprintf(tmpstring, "%s[%d]\t\t", 
                         var_name(varval[0][j][0]), var_id(varval[0][j][0]));
      output(tmpstring);
    }
    else {
      sprintf(tmpstring, "%s\t\t", var_name(varval[0][j][0]));
      output(tmpstring);
    }
    if(vv_count[j])
    for(k=1; k<=vv_count[j]; k++) {
      sprintf(tmpstring, "%d", vv[j][k]);
      output(tmpstring);
      if(k<vv_count[j]) output(", ");
      else output("\n");
    }
    else output("\n");
  }
  output("\n");
  return 0; 
}
