#include "include.h"
#include "global.h"
#include "print.h"
#include "bddman.h"
#include "modeman.h"
#include "zoneman.h"
#include "symmetry.h"
#include "readwrite.h"

struct a23tree_type *trans_tree;
struct spec_trans_type *new_trns = NULL;
struct var_link_link_type *bypassed_internal_var_list = NULL;

int trans_count_inc(struct transition_type *);
int trans_parent_set(struct transition_type *t, struct a23tree_type *parent);
int trans_comp(struct transition_type *, struct transition_type *);
int spec_trans_comp(struct spec_trans_type *t1, struct spec_trans_type *t2);
int boolconj_compare(struct boolconj_type *b1, struct boolconj_type *b2);
int internal_transition_bypass(struct thread_type *);
int common_var(int, int, int);
int is_internal_clock(struct variable_type *, struct thread_type *);
int contains_synclab(struct thread_type *, struct variable_type *);
int common_sync(struct variable_type *, struct thread_type *, 
                struct thread_type *);
int is_internal_sync(struct variable_type *, struct thread_type *);

int is_spec_related_mode(struct mode_type *);
int is_internal_transition(struct transition_type *, struct thread_type *t);

bdd dvl_bdd_ext(struct mode_type *m);
struct spec_trans_type *copy_spec_trans(struct spec_trans_type *trn);
struct spec_trans_type *new_trn_on_bypass(struct transition_type *, 
                                          struct boolconj_type *);
int system_spec_trans(struct transition_type *);

int is_internal_var(struct variable_type *, struct thread_type *);

struct var_link_type *get_internal_var_list(struct thread_type *t);
struct var_link_type *hash_bypassed_internal_var_list(struct var_link_type *);
int var_link_comp(struct var_link_type *, struct var_link_type *);
int update_all_dvl(struct thread_type *);

/*
int trans_count_inc(struct transition_type *t) {
  (t->count)++;
  return 0;
}

int trans_parent_set(struct transition_type *t, struct a23tree_type *parent) {
  t->parent = parent;
  return 0;
}
*/

int trans_comp(struct transition_type *t1, struct transition_type *t2) {
  int comp;

  if(t1->automaton && t2->automaton)
    if(t1->automaton->u_index < t2->automaton->u_index) return -1;
  if(t1->automaton && t2->automaton)
    if(t1->automaton->u_index > t2->automaton->u_index) return 1;
  if(t1->trn->index < t2->trn->index) return -1;
  if(t1->trn->index > t2->trn->index) return 1;
  comp = mode_comp(t1->mode, t2->mode);
  if(comp != EQUAL) return comp;
  return (permutation_comp(t1->permutation, t2->permutation));
}

int spec_trans_comp(struct spec_trans_type *t1, struct spec_trans_type *t2) {
  int result;
  result = var_link_comp(t1->sync_cycle, t2->sync_cycle);
  if(result != EQUAL) return result;
  result = strcmp(t1->goto_mode, t2->goto_mode);
  if(result != EQUAL) return result;
  result = clkconj_compare(t1->clk_trigger_cycle, t2->clk_trigger_cycle);
  if(result != EQUAL) return result;
  result = boolconj_compare(t1->prp_trigger_cycle, t2->prp_trigger_cycle);
  if(result != EQUAL) return result;
  result = assignment_compare(t1->assign_cycle, t2->assign_cycle); 
  if(result != EQUAL) return result;
  return EQUAL;
}

int boolconj_compare(struct boolconj_type *b1, struct boolconj_type *b2) {
  return dvl_compare(b1, b2);
}

int internal_transition_bypass(struct thread_type *t) {
  struct mode_type *m, *trm;
  struct transition_type *ltr, *tr, *str, *tr2, *ntr;
  struct permutation_type *p;
  int i, j, cmp, tcount, *p1 = NULL, *p2 = NULL; /*, exists_assign = FALSE; */
  char *c;

  internal_trans_bypass = TRUE;
  t->mnpl |= TB;
  NEW_THREAD = t;

/*
  if(lastcomp) return E_NOTIME;
*/
  if(!t) return E_NOTIME;
  if(!(t->starting_mode)) return E_NOTIME;
  
  c = get_graph_name(t);
  printf("    Internal Transition Bypass Reducing graph %s.\n", c);
  fflush(stdout);
  if(verbose >= 0) {
    sprintf(tmpstring, 
            "\nInternal transition bypass reducing graph %s.\n", c);
    output(tmpstring);
  }
  for(m=t->starting_mode,j=0; m; m=m->cnext,j++) {
    tcount = m->out_trans_count;
/*
    sprintf(tmpstring, "Mode %s(%d): tcount = %d\n", m->name, m->index, tcount);
    output(tmpstring);
*/
    if(!(m->out_trans_cycle)) continue;
    if(verbose >= 0) {
      sprintf(tmpstring, "Mode %s(%d) has:\n  (1) Invariance: ", 
              m->name, m->index);
      output(tmpstring);
      print_bdd(m->boolexp->boolexp_bdd);
      output("  (2) EXT DVL BDD: ");
      print_bdd(dvl_bdd_ext(m));
      if(total_clocks > 0) { 
        print_zone(m->zone); 
        output("\n");
      }
    }
    ltr = m->out_trans_cycle;
    for(tr=m->out_trans_cycle->next, str=tr->next, i=0; 
        i < tcount; 
        tr=str, str=str->next, i++) {
/*
      sprintf(tmpstring, "  #%d --- Tr %d\n", i, tr->index);
      output(tmpstring);
*/
      if(!system_spec_trans(tr)) {
        if(tr == ltr) break;
        else continue;
      }
      if(verbose >= 0) {
        sprintf(tmpstring, "  Transition %d: ", tr->index);
        output(tmpstring);
        print_trans(tr);
        sprintf(tmpstring, 
                "    Destination Mode %s(%d) has:\n    (1) Invariance: ", 
                           tr->mode->name, tr->mode->index);
        output(tmpstring);
        print_bdd(tr->mode->boolexp->boolexp_bdd);
        output("    (2) EXT DVL BDD: ");
        print_bdd(dvl_bdd_ext(tr->mode));
        if(total_clocks > 0) {
          output("    ");
          print_zone(tr->mode->zone); 
          output("\n");
        }
      }
      if(tr->mode->index == m->index) {
        if(verbose >= 0) {
          sprintf(tmpstring, "  Loop Transition (%d).\n", tr->index);
          output(tmpstring);
        }
        if(tr == ltr) break;
        else continue;
      }
/*
      if(is_internal_transition(tr, NEW_THREAD->subthread1) ||
         is_internal_transition(tr, NEW_THREAD->subthread2)) {
        if(verbose >= 0) {
          sprintf(tmpstring, 
                  "  An Old Internal Transition (%d).\n", tr->index);
          output(tmpstring);
        }
        if(tr == ltr) break;
        else continue;
      }
*/
      if(is_internal_transition(tr, NEW_THREAD) == FALSE) {
        if(verbose >= 0) {
          sprintf(tmpstring, 
                  "  Not An Internal Transition (%d).\n", tr->index);
          output(tmpstring);
        }
        if(tr == ltr) break;
        else continue;
      }
      /* tr is internal now, check other things */

/* ERIC: newly added 2000/07/20 */
      if(tr->trn->assign_cycle != NULL) {
        cmp = zone_comp(tr->mode->zone, m->zone);
        if(cmp == CONTAINS || cmp == INCOMPARABLE) {
          if(verbose >= 0)
            output("  Local assign: not contained or equal zones.\n");
          if(tr == ltr) break;
          else continue;
        }
      }
/* Till here */

      if(is_spec_related_mode(tr->mode)) {
        if(verbose >= 0) {
          sprintf(tmpstring, 
              "  Dest. mode %s(%d) of trans (%d) is specificiation RELATED.\n", 
                 tr->mode->name, tr->mode->index, tr->index);
          output(tmpstring);
        }
        if(tr == ltr) break;
        else continue;
      }
/*
      if(zone_comp(tr->mode->zone, m->zone) != EQUAL) {
        if(tr == m->out_trans_cycle) break;
        else continue;
      }
*/
      if(bdd_or(bddm, bdd_not(bddm, m->boolexp->boolexp_bdd), 
                              tr->mode->boolexp->boolexp_bdd) == BDD_FALSE) {
/*
      if(bdd_implies(bddm, m->boolexp->boolexp_bdd, 
                           tr->mode->boolexp->boolexp_bdd)) {
*/
        if(tr == ltr) break;
        else continue;
      }
      if(verbose >= 0) {
        sprintf(tmpstring, "    Boolean invariant implications OK!\n");
        output(tmpstring);
      }

      if(bdd_or(bddm, bdd_not(bddm, dvl_bdd_ext(m)), 
                              dvl_bdd_ext(tr->mode)) == BDD_FALSE){
/*
      if(bdd_implies(bddm, dvl_bdd_ext(m), dvl_bdd_ext(tr->mode))){
*/
        if(tr == ltr) break;
        else continue;
      }
      if(verbose >= 0) {
        sprintf(tmpstring, "    DVL implications OK!\n");
        output(tmpstring);
      }
      trm = tr->mode;
      if(tr == ltr) ltr = NULL;
      if(tr->permutation)
        p1 = tr->permutation->perm;

/*
      if(tr->trn->assign_cycle) exists_assign = TRUE;
*/

      if(verbose >= 0) {
        sprintf(tmpstring,
                "    Eliminating transition (%d) from %s(%d) to %s(%d).\n",
                  tr->index, m->name, m->index,tr->mode->name,tr->mode->index);
        output(tmpstring);
      }
      eliminate_transition(tr, m);
      t->trans_count--;

      if(trm->out_trans_cycle) {
        for(tr2=trm->out_trans_cycle->next; ; tr2=tr2->next) {
          if(tr2->permutation)
            p2 = tr2->permutation->perm;
          ntr = add_transition_fifo(tr2, m, tr2->mode);
          if(p1 || p2) {
            ntr->permutation = alloc_permutation();
            ntr->permutation->perm = compose_permutation(p1, p2);
            p = (struct permutation_type *)
                   add_23tree(ntr->permutation, perm_tree, permutation_comp,
                   free_perm, perm_count_inc, perm_parent_set,
                   print_permutation);
            ntr->permutation = p;
            if(verbose >= 0) {
              sprintf(tmpstring,
                   "\tInstalled permutation on transition (%d): ", ntr->index);
              output(tmpstring);
              print_trans_permutation(ntr);
            }
          }

          t->trans_count++;

/*
          if(verbose >= 3) {
            sprintf(tmpstring, "Trans %d's Old Prop Trigger: ", ntr->index);
            output(tmpstring);
            print_trigger(ntr);
            output("\n");
          }

          ntr->trn = new_trn_on_bypass(ntr, trm->dvl);

          if(verbose >= 3) {
            sprintf(tmpstring, "Trans %d's New Prop Trigger: ", ntr->index);
            output(tmpstring);
            print_trigger(ntr);
            output("\n");
            output("and DVL BDD: ");
            if(ntr->trn->trigger_bdd)
              print_bdd(ntr->trn->trigger_bdd);
            else output("BDD_TRUE\n");
          }
*/

/*
	  if(exists_assign)
            ntr->bypassed_internal_vars = get_internal_var_list(NEW_THREAD);
*/

          if(verbose >= 0) {
            sprintf(tmpstring,
                "\tAdded transition (%d) between %s(%d) and %s(%d).\n",
              ntr->index, m->name, m->index, tr2->mode->name, tr2->mode->index);
            output(tmpstring);
            output("\t");
            print_trans(ntr);
          }
          if(tr2 == trm->out_trans_cycle) break;
        }
      }
      if(ltr == NULL) break;
/*
      else if(m->out_trans_cycle == NULL || str==ltr->next) break;
      else if(tr == ltr) break;
*/
    }
/*
    sprintf(tmpstring, "    new tcount = %d\n", m->out_trans_count);
    output(tmpstring);
*/
    if(stats == TRUE && SGM_MODE != GUI && SGM_MODE != PROCESS) {
      clear_line();
      fprintf(stderr, "    BIT-Checked %d modes, Remaining %d modes (%.1f%%)\r",
                      j, t->mode_count-j,
              ((float)(t->mode_count-j) / (float)t->mode_count)*100.0);
    }
  }
  clean_line();
  printf("\tRemoving irreachable modes in graph %s.\n\t", c);
  if(verbose >= 0) {
    outputline();
    sprintf(tmpstring,
    "Removing irreachable modes in graph %s (%d modes, %d transitions).\n",
       c, t->mode_count, t->trans_count);
    output(tmpstring);
  }
  remove_irreachable_mode(t);
  if(verbose >= 0) output("\t");
  print_counts(t);

  if(sibling_trans_multi_reduction) {
    printf("\tReducing transitions in (multi)graph %s[%d].\n",
           t->name, t->automaton_index);
    fflush(stdout);
    reduce_multigraph(t);
    if(verbose >= 0) {
      sprintf(tmpstring, "DONE reducing transitions in (multi)graph %s.\n",c);
      output(tmpstring);
    }
    if(SGM_MODE != GUI && SGM_MODE != PROCESS) fprintf(stderr, "\t");
    print_counts(t);
  }
  update_all_dvl(t);
  merge_iden_red_multi(t);
  free(c);
  return 0;
}


/* Check if the variable "v" is written by more than one automaton */

int is_global_var(struct variable_type *v) {
  int i, n, found = FALSE;

  if(v->type == V_CLOCK) return FALSE; 
                        /* currently all clocks are considered as local */

  n = get_dvar_id(v->index, automaton_tail);

  for(i=0; i<system_ta_num; i++) {
    if(val_count[i][n] > 0) {
      if(found) return TRUE;
      else found = TRUE;
    }
  }
  return FALSE;
}

/* Check if variable "v" is an internal variable of graph NEW_THREAD */

/*
int is_internal_var(struct variable_type *v) {
  int n, i, j, found = FALSE;

  if(v == NULL) return TRUE;
  if(v->type == V_CLOCK) return is_internal_clock(v); 
  n = get_dvar_id(v->index, automaton_tail);

  for(i=0; i<system_ta_num; i++) {
    if(component_graph(i, NEW_THREAD) && val_count[i][n] > 0) {
      found = TRUE;
      break; 
    }
  }
  if(found == FALSE) return FALSE;

  for(i=0; i<system_ta_num; i++) {
    for(j=0; j<system_ta_num; j++) {
      if(component_graph(i, NEW_THREAD) && component_graph(j, NEW_THREAD)) 
        continue;
      if(common_var(n, i, j)) return FALSE;
    }
  }
  if(verbose >= 3) {
    if(v->id >= 0)
    sprintf(tmpstring, "    Variable %s[%d] is internal w.r.t graph ", 
            v->name, v->id);
    else
    sprintf(tmpstring, "    Variable %s is internal w.r.t graph ", 
            v->name);
    output(tmpstring);
    if(NEW_THREAD->automaton_index >= 0)
      sprintf(tmpstring, "%s[%d].\n", 
              NEW_THREAD->name, NEW_THREAD->automaton_index);
    else
      sprintf(tmpstring, "%s.\n", NEW_THREAD->name);
    output(tmpstring);
  }
  return TRUE;
}
*/

/* Check if variable "v" is an internal variable of graph t */

int is_internal_var(struct variable_type *v, struct thread_type *t) {
  int n, i, j, found = FALSE;

  if(v == NULL) return TRUE;
  if(v->type == V_CLOCK) return is_internal_clock(v, t); 
                        /* currently all clocks are considered as internal */
  n = get_dvar_id(v->index, automaton_tail);

  for(i=0; i<system_ta_num; i++) {
    if(component_graph(i, t) && val_count[i][n] > 0) {
      found = TRUE;
      break; 
    }
  }
  if(found == FALSE) return FALSE;

  for(i=0; i<system_ta_num; i++) {
    for(j=i+1; j<system_ta_num; j++) {
      if(component_graph(i, t) && component_graph(j, t)) 
        continue;
      if(common_var(n, i, j)) return FALSE;
    }
  }
  if(verbose >= 3) {
    if(v->id >= 0)
    sprintf(tmpstring, "    Variable %s[%d] is internal w.r.t graph ", 
            v->name, v->id);
    else
    sprintf(tmpstring, "    Variable %s is internal w.r.t graph ", 
            v->name);
    output(tmpstring);
    if(t->automaton_index >= 0)
      sprintf(tmpstring, "%s[%d].\n", 
              t->name, t->automaton_index);
    else
      sprintf(tmpstring, "%s.\n", t->name);
    output(tmpstring);
  }
  return TRUE;
}

int is_internal_clock(struct variable_type *v, struct thread_type *t) {
  return TRUE;
}

int contains_synclab(struct thread_type *t, struct variable_type *v) {
  struct var_link_type *vl;

  if(t->sync_cycle == NULL) return FALSE;
  for(vl=t->sync_cycle->next; ; vl=vl->next) {
    if(vl->var->index == v->index) return TRUE;
    if(vl==t->sync_cycle) break;
  }
  return FALSE;
}

int common_sync(struct variable_type *v, struct thread_type *t1, 
                struct thread_type *t2) {
  if(contains_synclab(t1, v) && contains_synclab(t2, v)) return TRUE;
  else return FALSE;
}

int is_internal_sync(struct variable_type *v, struct thread_type *t) {
  int i, j, found = FALSE;

  if(v == NULL) return TRUE;

  for(i=0; i<system_ta_num; i++) {
    if(component_graph(i, t) && contains_synclab(&thread[i], v)) {
      found = TRUE;
      break; 
    }
  }
  if(found == FALSE) return FALSE;
  for(i=0; i<system_ta_num; i++) {
    for(j=i+1; j<system_ta_num; j++) {
      if(component_graph(i, t) && component_graph(j, t)) continue;
      if(common_sync(v, &thread[i], &thread[j])) return FALSE;
    }
  }
  if(verbose >= 3) {
    sprintf(tmpstring, "Synclab %s is internal w.r.t. graph ", v->name);
    output(tmpstring);
    if(t->automaton_index >= 0)
      sprintf(tmpstring, "%s[%d].\n", 
              t->name, t->automaton_index);
    else
      sprintf(tmpstring, "%s.\n", t->name);
    output(tmpstring);
  }
  return TRUE;
}


/* Check if thread[t1] and thread[t2] both access
   variable with index varval[][n][0] */

int common_var(int n, int t1, int t2) {
/*
  int i, j;
*/
  if(val_count[t1][n] > 0 && val_count[t2][n] > 0) return TRUE;
/*
  for(i=1; i<=val_count[t1][n]; i++) {
    for(j=1; j<=val_count[t2][n]; j++) {
      if(varval[t1][n][i] == varval[t2][n][j]) return TRUE;
    }
  }
*/
  return FALSE;
}

int is_internal_boolconj(struct boolconj_type *b, struct thread_type *t) {
  struct boolconj_type *bb;

  if(!b) return TRUE;
  for(bb = b->next; ; bb = bb->next) {
    if(is_internal_var(var_ptr_from_index(bb->var1), t) == FALSE) return FALSE;
    if(bb == b) break;
  }
  return TRUE;
}

int is_internal_clkconj(struct clkconj_type *c, struct thread_type *t) {
  struct clkconj_type *cc;

  if(!c) return TRUE;
  for(cc = c->next; ; cc = cc->next) {
    if(is_internal_var(var_ptr_from_index(cc->clk1), t) == FALSE) return FALSE;
    if(is_internal_var(var_ptr_from_index(cc->clk2), t) == FALSE) return FALSE;
    if(cc == c) break;
  }
  return TRUE;
}

/* all clocks are assumed to be local */
/* internal or not is decided for clocks by further checking of zone
   containment in BIT */

int is_internal_assign(struct assign_type *a, struct thread_type *t) {
  struct assign_type *aa;
  if(!a) return FALSE;
  for(aa=a->next; ; aa=aa->next) {
/*
    print_assignment(aa);
    output("\n");
*/
    if(aa->var) {
/*
      if(get_var_type(aa->var) == V_CLOCK) 
        return FALSE;              
*/
      /* Former assumption: all clock resets are not internal */
      /* New assumption: all clocks are assumed local, further checking of
                  zone containment is required, this is an extension of BIT */
      if(is_internal_var(var_ptr_from_index(aa->var), t) == FALSE) 
        return FALSE;
    }
    if(aa->use_var) {
/*
      if(get_var_type(aa->use_var) == V_CLOCK) 
        return FALSE;
*/
      if(is_internal_var(var_ptr_from_index(aa->use_var), t) == FALSE) 
        return FALSE;
    }
    if(aa==a) break;
  }
  return TRUE;
}

int is_internal_transition(struct transition_type *tr, struct thread_type *t) {
  struct var_link_type *vl;

  if(tr->trn->sync_cycle)
    for(vl=tr->trn->sync_cycle->next; ; vl=vl->next) {
      if(is_internal_sync(vl->var, t) == FALSE) {
        if(verbose >= 3) {
          sprintf(tmpstring, "  Synclab %s is not internal.\n", vl->var->name);
          output(tmpstring);
        }
        return FALSE;
      }
      if(vl==tr->trn->sync_cycle) break;
    }

  if(tr->trn->prp_trigger_length > 0)
    if(is_internal_boolconj(tr->trn->prp_trigger_cycle, t) == FALSE) {
      if(verbose >= 3) {
        output("  Not an internal boolean proposition trigger.\n");
      }
      return FALSE;
    }

  if(tr->trn->clk_trigger_length > 0) 
    if(is_internal_clkconj(tr->trn->clk_trigger_cycle, t) == FALSE) {
      if(verbose >= 3) {
        output("  Not an internal clock trigger.\n");
      }
      return FALSE;
    }

  if(tr->trn->assign_length > 0)
    if(is_internal_assign(tr->trn->assign_cycle, t) == FALSE)  {
      if(verbose >= 3) {
        output("  Not an internal assignment.\n");
      }
      return FALSE;
    }

  return TRUE;
}


/* Check if the mode "m" is related to the specification 
 * (1) m has a mode-vector component that appears in the specification
 * (2) m has a discrete variable literal that appears in the specification
 * (3) m has an invariant condition that appears in the specification
 */
                 /* NOT YET IMPLEMENTED */
int is_spec_related_mode(struct mode_type *m) {
  return FALSE;
}

int read_only_variable(struct variable_type *v) {
  int i, n;

  if(v->type == V_CLOCK) return FALSE;
                        /* currently all clocks are as writable */

  n = get_dvar_id(v->index, automaton_tail);

  for(i=0; i<system_ta_num; i++) {
    if(val_count[i][n] > 0) {
      return FALSE;
    }
  }
  return TRUE;
}

bdd dvl_bdd_ext(struct mode_type *m) {
  struct boolconj_type *b;
  bdd f = BDD_TRUE;
  if(m->dvl) {
    for(b=m->dvl->next; ; b=b->next) {
      if( (b->var1==0 || !is_internal_var(var_ptr_from_index(b->var1), NEW_THREAD)) &&
          (b->var2==0 || !is_internal_var(var_ptr_from_index(b->var2), NEW_THREAD)) )
        f = BDD_AND(f, b->prop_bdd);
      if(b==m->dvl) break;
    }
  }
  return f;
}

struct spec_trans_type *copy_spec_trans(struct spec_trans_type *trn) {
  struct spec_trans_type *ntrn;

  ntrn = (struct spec_trans_type *)malloc(sizeof(struct spec_trans_type));
  ntrn->index = trn->index;
  ntrn->clk_trigger_length = trn->clk_trigger_length;
  ntrn->prp_trigger_length = trn->prp_trigger_length;
  ntrn->assign_length = trn->assign_length;
  ntrn->prp_trigger_cycle = trn->prp_trigger_cycle;
  ntrn->clk_trigger_cycle = trn->clk_trigger_cycle;
  ntrn->assign_cycle = trn->assign_cycle;
  ntrn->trigger_bdd = trn->trigger_bdd;
  ntrn->goto_mode = trn->goto_mode;
  ntrn->next = (struct spec_trans_type *)NULL;
  return ntrn;
}

struct spec_trans_type *new_trn_on_bypass(struct transition_type *ntr, 
                                          struct boolconj_type *dvl) {
  struct boolconj_type *b, *c, *nc, *ans = NULL;
  struct spec_trans_type *ntrn = NULL;
  bdd f = BDD_TRUE;
  int ans_len=0, implied = FALSE;

  if(verbose >= 3) {
    output("    Bypassed DVL: ");
    print_dvl(dvl);
    output("    Bypassed DVL BDD: ");
    print_dvl_bdd(dvl);
  }
  
  if(dvl == NULL) return ntr->trn;
  if(ntr->trn->prp_trigger_cycle == NULL) return ntr->trn;

  ntrn = copy_spec_trans(ntr->trn);
  ntrn->index = ++trans_count;

  if(ntr->trn->prp_trigger_cycle) {
    for(c=ntr->trn->prp_trigger_cycle->next; ; c=c->next) {
      nc = boolconj_copy1(c);
      if(nc) install_bdd(nc);
      if(verbose >= 3) {
        output("    Copied Boolean Conjunct: "); 
        print_boolconj(nc); 
        output("\n");
        output("    with BDD: "); 
        print_bdd(nc->prop_bdd);
      }
      for(b=dvl->next; ; b=b->next) {
        if(c->var1 == b->var1 && c->var2 == b->var2) {
          if(bdd_and(bddm, nc->prop_bdd, b->prop_bdd) == BDD_FALSE) {
            nc->var1 = nc->var2 = 0;
            nc->op = EQ;
            nc->constant = 1;
            nc->prop_bdd = BDD_FALSE;
            free_boolconj(ans);
            ans = nc;
/*
            ans = append_boolconj(nc, ans);
            bdd_and(bddm, f, BDD_FALSE);
            ans_len++;
*/
            ntrn->prp_trigger_length = 1;
            ntrn->prp_trigger_cycle = ans;
            ntrn->trigger_bdd = BDD_FALSE;
            return ntrn;
          }
          else {
            implied = TRUE;
            break;
          }
        }
        if(b == dvl) break;
      }
      if(nc->prop_bdd == BDD_FALSE) {
        output("Found BDD_FALSE prop_bdd outside dvl loop!\nans: ");
        if(ans) print_boolconj(ans); 
        else output("True");
        output("\n");
        ntrn->prp_trigger_length = ans_len;
        ntrn->prp_trigger_cycle = ans;
        ntrn->trigger_bdd = BDD_FALSE;
        return ntrn;
      }
      if(!implied) {
        if(verbose >= 3) {
          output("    Conjunct Not Relevant!\n");
          output("    Current New Boolean Trigger: "); 
          if(ans) print_boolconj(ans); 
          else output("True");
          output("\n");
        }
        ans = append_boolconj(nc, ans);
        if(verbose >= 3) {
          output("      After appending conjunct ("); 
          print_boolconj(nc);
          output(")\n     with bdd ("); 
          print_bdd(nc->prop_bdd);
          output("    Current New Boolean Trigger: "); 
          if(ans) print_boolconj(ans);
          else output("True");
          output("\n");
        }
        f = bdd_and(bddm, f, nc->prop_bdd);
        if(verbose >= 3) {
          output("    Current New Boolean Trigger BDD: "); 
          print_bdd(f);
        }
        ans_len++;
      }
      else { /* implied */
        if(verbose >= 3) {
          output("    Conjunct Relevant! Current New Boolean Trigger Not Changed\n");
        }
        implied = FALSE;
        free_boolconj(nc);
        nc = NULL;
      }
      if(c == ntr->trn->prp_trigger_cycle) break;
    }
    ntrn->prp_trigger_length = ans_len;
    ntrn->prp_trigger_cycle = ans;
    ntrn->trigger_bdd = f;

    if(ntrn->trigger_bdd != BDD_FALSE &&
       ntrn->prp_trigger_length == ntr->trn->prp_trigger_length) {
      free_boolconj(ans);
      free(ntrn);
      return ntr->trn;
    }
    if(new_trns) {
      ntrn->next = new_trns->next;
      new_trns->next = ntrn;
    }
    else ntrn->next = ntrn;
    new_trns = ntrn;
    return ntrn;
  }
  else return ntr->trn;
}

int system_spec_trans(struct transition_type *tr) {
  struct automaton *a;
  struct mode_type *m;
  struct transition_type *t;
  for(a=automaton_tail->next; ; a=a->next) {
    if(a->mode_cycle)
    for(m=a->mode_cycle->next; ; m=m->next) {
      if(m->out_trans_cycle)
      for(t=m->out_trans_cycle->next; ; t=t->next) {
        if(tr->trn->index == t->trn->index) return TRUE;
        if(t == m->out_trans_cycle) break;
      }
      if(m == a->mode_cycle) break;
    }
    if(a==automaton_tail) break;
  }
  return FALSE;
}

struct var_link_type *get_internal_var_list(struct thread_type *t) {
  struct variable_type *v;
  struct var_link_type *vl, *ans=NULL, *new_ans;

  for(v=var_tail->next; ; v=v->next) {
    if(v->type == V_REGISTER && is_internal_var(v, t)) {
      vl = (struct var_link_type *)malloc(sizeof(struct var_link_type));
      vl->var = v;
      if(ans) {
        vl->next = ans->next;
        ans->next = vl;
      }
      else vl->next = vl;
      ans = vl;
    }
    if(v == var_tail) break;
  }
  new_ans = hash_bypassed_internal_var_list(ans);
  if(new_ans != ans) free_var_links(ans);
  return new_ans;
}


/* Keep unique list and only allocate pointers */

struct var_link_type *hash_bypassed_internal_var_list(struct var_link_type *a) {
  struct var_link_link_type *vllp;

  if(bypassed_internal_var_list) {
    for(vllp=bypassed_internal_var_list->next; ; vllp=vllp->next) {
      if(var_link_comp(vllp->vl, a) == EQUAL) return vllp->vl;
      if(vllp == bypassed_internal_var_list) break;
    }
  }
  /* not found or empty list */
  vllp=(struct var_link_link_type *)malloc(sizeof(struct var_link_link_type));
  vllp->vl = a;
  vllp->next = vllp;
  bypassed_internal_var_list = vllp;
  return a;
}

/* Assume that the var_links are ordered in the two cycles */

int var_link_comp(struct var_link_type *vl1, struct var_link_type *vl2) {
  struct var_link_type *v1, *v2;

  if(vl1 == NULL && vl2 != NULL) return -1;
  if(vl1 != NULL && vl2 == NULL) return 1;
  if(vl1 == NULL && vl2 == NULL) return EQUAL;

  for(v1=vl1->next, v2=vl2->next; ; v1=v1->next, v2=v2->next) {
    if(v1->var->index < v2->var->index)  return -1;
    if(v1->var->index > v2->var->index)  return 1;
    if(v1==vl1 && v2 != vl2) return -1;
    if(v1!=vl1 && v2 == vl2) return 1;
    if(v1==vl1 && v2 == vl2) break;
  }
  return EQUAL;
}

int update_all_dvl(struct thread_type *t) {
  struct mode_type *m;
  struct transition_type *tr;
  struct boolconj_type *b1 = NULL, *b2 = NULL;

  for(m=t->starting_mode; m; m=m->cnext) {
    if(m->out_trans_cycle) {
      for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
        b1 = tr->mode->dvl;
        b2 = get_new_dvl(m, tr);
        tr->mode->dvl = boolconj_union(b1, b2);
        free_boolconj(b1); b1 = NULL;
        free_boolconj(b2); b2 = NULL;
        if(tr == m->out_trans_cycle) break;
      }
    }
  }
  return 0;
}
